/*
 * PDF export routines
 *
 * Copyright (c) 2004-2006    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <zlib.h>
#include "export/export.h"
#include "kernel/kernel.h"
#include "font/font.h"
#include "lib/garr.h"
#include "lib/clists.h"
#include "import/import.h"

/*
 * TODO:
 * - Vypisovat widths jednotlivych znaku, nyni se nevypisuji, cimz se mirne
 *   porusuje norma. Cele to zase totiz pada na nejakem bugu ve freetype. (grrr)
 * - Dodat nektere optional atributy, dodatecne informace do ruznych dictionaries, atd.
 */


/*
 * The compresion level of pdf streams.
 * The posibilities are: Z_NO_COMPRESSION, Z_BEST_SPEED, Z_BEST_COMPRESSION, Z_DEFAULT_COMPRESSION.
 */
static int compress_level;


/*
 * White margin around the image measured in mm.
 */
#define PDF_MARGIN_X 5
#define PDF_MARGIN_Y 5

#define BBOX_MARGIN 0.2

static FILE *pdf_file;
static struct obj_doc *pdf_doc;
static struct obj_tlo *cur_tlo;

struct font {
	uns used;	// Set if this font is used in actual tlo so it must be mentioned in a resource dictionary.
	uns saved;	// Set to one in the first use of this font. See font_set.
	uns label;	// Label of indirect object font simple.
};

/*
 * Array indexed by font_id.
 * Font resource identified is same as font_id.
 * Chapter: 5.1.1
 */
static struct font *pdf_font;
static uns pdf_max_fntid;

static int embed_fonts;

/*
 * Each indirect object has one instance of this structure.
 * Chapter: 3.2.9
 */
struct indirect {
	cnode n;
	uns offset, label;
	void (*func)(struct indirect*);
	union {
		struct {
			struct obj_tlo* tlo;
			struct indirect *parent;
		} page;

		struct {
			struct obj_tlo* tlo;
		} contents;

		struct {
			uns id;
		} font;
	} u;
};

static uns cur_label;

/*
 * Array used for exporting streams.
 * Chapter: 3.2.7
 */
static garr stream_buf;
static uns stream_buf_size;

static clist indirect_list;

static void cleanup(void)
{
	struct indirect *i;

	while (i = clist_head(&indirect_list)) {
		clist_remove(&i->n);
		xfree(i);
	}
	garr_destroy(stream_buf);
	xfree(pdf_font);
	if(cur_tlo)
	{
		expand_unref(cur_tlo);
		cur_tlo = NULL;
	}
}

/*
 * The PostScript names of 14 Type 1 fonts, known as the standard fonts, are as follows:
 * Chapter: 5.5.1
 */
static const char *pdf_standard_type1_fonts[] = {
	"Times-Roman",
	"Times-Bold",
	"Times-Italic",
	"Times-BoldItalic",
	"Helvetica",
	"Helvetica-Bold",
	"Helvetica-Oblique",
	"Helvetica-BoldOblique",
	"Courier",
	"Courier-Bold",
	"Courier-Oblique",
	"Courier-BoldOblique",
	"Symbol",
	"ZapfDingbats"
};

static int font_is_standard(const char *name)
{
	for (int i = 0; i < 14; i++)
		if (strcmp(name, pdf_standard_type1_fonts[i]) == 0)
			return 1;
	return 0;
}

static int buffer_output;

static FORMAT(1,2) void out(const char *s, ...)
{
	va_list ap;
	va_start(ap, s);

	if (buffer_output) {
		char *buf = (char *)garr_get(stream_buf, stream_buf_size + 4096) - 4096;
		int wr = vsnprintf(buf, 4096, s, ap);
		if (wr < 0)
			trans_fail("Buffer printf failure");
		if (wr > 4096)
			wr = 4096;
		stream_buf_size += wr;
	}
	else
		if (vfprintf(pdf_file, s, ap) < 0)
			trans_fail("Error writing to ouput file");

	va_end(ap);
}

static void outc(int c)
{
	if (buffer_output)
		*garr_get(stream_buf, stream_buf_size++) = c;
	else
		fputc(c, pdf_file);
}

/*
 * Chapter: 3.2.3
 */
static void dump_string(const char* s)
{
	outc('(');
	while (*s) {
		switch (*s) {
		case '(':
			out("\\(");
			break;
		case ')':
			out("\\)");
			break;
		case '\\':
			out("\\\\");
			break;
		default:
			outc(*s);
		}
		s++;
	}
	outc(')');
}

static void stream_start(void)
{
	stream_buf_size = 0;
	buffer_output = 1;
}

static void stream_end(void)
{
	buffer_output = 0;
	if (compress_level != Z_NO_COMPRESSION) {
		uLongf len = compressBound(stream_buf_size);
		byte buf[len];

		if (compress2(buf, &len, garr_get(stream_buf,0), stream_buf_size, compress_level) != Z_OK)
			trans_fail("Error occured during stream compression");
		out("<<\n");
		out("/Length %lu\n", len);
		out("/Filter /FlateDecode\n");
		out(">>\n");
		out("stream\n");
		if (fwrite(buf, 1, len, pdf_file) != len)
			trans_fail("Error writing to output file");
		out("\nendstream\n");
	}
	else {
		out("<<\n/Length %u\n>>\nstream\n", stream_buf_size);
		if (fwrite(garr_get(stream_buf, 0), 1, stream_buf_size, pdf_file) != stream_buf_size)
			trans_fail("Error writing to output file");
		out("\nendstream\n");
	}
}

static struct indirect* indirect_object(void (*func)(struct indirect*))
{
	struct indirect *i = xmalloc(sizeof(struct indirect));

	i->label = ++cur_label;
	clist_add_tail(&indirect_list, &i->n);
	i->func = func;
	return i;
}

/*
 * Set the stroking color space.
 * Synopsis: r g b RG
 * Chapter 4.5.7
 */
static int set_color(struct go *go)
{
	real r, g, b, a;
	uns clr = prop_lookup_uns(go2o(go), string_stroke_color, 0);

	uns2rgba_real(clr, &r, &g, &b, &a);
	if (a >= 0.5) {
		out("%f %f %f RG\n", r, g, b);
		return 1;
	}
	return 0;
}

static int set_fillcolor(struct go *go)
{
	real r, g, b, a;
	uns clr = prop_lookup_uns(go2o(go), string_fill_color, 0);

	uns2rgba_real(clr, &r, &g, &b, &a);
	if (a >= 0.5) {
		out("%f %f %f rg\n", r, g, b);
		return 1;
	}
	return 0;
}

static void set_linewidth(struct go *go)
{
	real lw = prop_lookup_real(go2o(go), string_stroke_width, -1);

	if (lw >= 0)
		out("%f w\n", lw);
}

static void set_linecap(struct go *go)
{
	int c = prop_lookup_uns(go2o(go), string_stroke_cap, -1);

	if (c < 0)
		return;
	switch (c) {
	case PSC_BUTT:
		c = 0;
		break;
	case PSC_ROUND:
		c = 1;
		break;
	case PSC_PROJECTING:
		c = 2;
		break;
	default:
		return;
	}
	out("%d J\n", c);
}

static void set_linejoin(struct go *go)
{
	int c = prop_lookup_uns(go2o(go), string_stroke_join, -1);

	if (c < 0)
		return;
	switch (c) {
	case PSJ_MITER:
		c = 0;
		break;
	case PSJ_ROUND:
		c = 1;
		break;
	case PSJ_BEVEL:
		c = 2;
		break;
	default:
		return;
	}
	out("%d j\n", c);
}

static void dump_font_type1(struct indirect *i)
{
	int id = i->u.font.id;
	int size;
	const char *fontfile;
	FILE *f;

	fontfile = font_filename(id);
	f = fopen(fontfile, "r");
	if (!f)
		trans_fail("Cannot open fontfile '%s' for reading", fontfile);

	struct stat statbuf;
	if (fstat(fileno(f), &statbuf)<0)
		trans_fail("Cannot fstat() fontfile '%s'", fontfile);

	size = statbuf.st_size;
	byte buf[size];
	uns length, txt, encr, fixed;
	byte outbuf[size+20];

	if (fread(buf, 1, size, f) != (uns)size)
		trans_fail("Error read()ing fontfile '%s'", fontfile);

	if ((size = font_pfa_to_pfb_buf(buf, size, outbuf)) < 0)
		trans_fail("Font file '%s' is not in exportable format", fontfile);

	if (font_pfb_sections(outbuf, size, &length, &txt, &encr, &fixed) < 0)
		trans_fail("Error occured during PFB file '%s' scan", fontfile);

	if (font_pfb_data_store(outbuf, size, outbuf)<0)
		trans_fail("Error storing font file '%s'", fontfile);

	if (compress_level != Z_NO_COMPRESSION) {
		uLongf len = compressBound(length);
		byte buf[len];

		if (compress2(buf, &len, outbuf, length, compress_level) != Z_OK)
			trans_fail("Error occured during stream compression");

		out("<<\n");
		out("/Length %lu\n", len);
		out("/Filter /FlateDecode\n");
		out("/Length1 %d\n", txt);
		out("/Length2 %d\n", encr);
		out("/Length3 %d\n", fixed);
		out(">>\n");
		out("stream\n");
		if (fwrite(buf, 1, len, pdf_file) != len)
			trans_fail("Error writing data to output file");
		out("\nendstream\n");
	}
	else {
		out("<<\n");
		out("/Length %d\n", length);
		out("/Length1 %d\n", txt);
		out("/Length2 %d\n", encr);
		out("/Length3 %d\n", fixed);
		out(">>\n");
		out("stream\n");
		if (fwrite(outbuf, 1, length, pdf_file) != length)
			trans_fail("Error writing data to output file");
		out("\nendstream\n");
	}
	fclose(f);
}

static void dump_font_tt(struct indirect *i)
{
	int id = i->u.font.id;
	FILE *f;
	uns fontsize;
	struct stat statbuf;
	const char *filename = font_filename(id);

	f = fopen(filename, "r");
	if (!f)
		trans_fail("Cannot open fontfile '%s'", filename);
	if (fstat(fileno(f), &statbuf)<0)
		trans_fail("fstat() on '%s' failed", filename);
	fontsize = statbuf.st_size;

	byte fbuf[fontsize];
	if (fread(fbuf, 1, fontsize, f) != fontsize)
		trans_fail("Fontfile '%s' read() error", filename);

	if (compress_level != Z_NO_COMPRESSION) {
		uLongf len = compressBound(fontsize);
		byte buf[len];

		if (compress2(buf, &len, fbuf, fontsize, compress_level) != Z_OK)
			trans_fail("Error occured during stream compression");
		out("<<\n");
		out("/Length %lu\n", len);
		out("/Filter /FlateDecode\n");
		out(">>\n");
		out("stream\n");
		if (fwrite(buf, 1, len, pdf_file) != len)
			trans_fail("Error writing to output file");
		out("\nendstream\n");
	}
	else {
		out("<<\n");
		out("/Length %d\n", fontsize);
		out("/Length1 %d\n", fontsize);
		out(">>\n");
		out("stream\n");
		if (fwrite(fbuf, 1, fontsize, pdf_file) != fontsize)
			trans_fail("Error writing to output file");
		out("\nendstream\n");
	}
	fclose(f);
}

static void dump_font_descriptor(struct indirect *i)
{
	uns flags = 0, id = i->u.font.id;
	struct font_info fi;

	if (!font_info(id, &fi))
		trans_fail("Error getting font %d info", id);

	out("<< /Type /FontDescriptor\n");
	out("/FontName /%s\n", fi.name);
	out("/FontBBox [%f %f %f %f]\n", fi.bbox.ll.x, fi.bbox.ll.y, fi.bbox.ur.x, fi.bbox.ur.y);

	if (fi.fixed_width)
		flags |= 1;
	flags |= 4;	/* Font contains glyphs outside the Adobe standard Latin character set. */
	if (fi.italic)
		flags |= 64;

	/*
	 * We are unable to detect the following flags.
	 * Let's hope it doesn't matter too much...
	 *
	 * Script flags, mask 8.
	 * Serif flags, mask 2.
	 * Symbolic xor nonsymbolic flags, mask 4 and 32.
	 * AllCap flags, mask 1<<16.
	 * SmallCap flags, mask 1<<17.
	 * ForceBold flags, mask 1<<18.
	 */

	out("/Flags %d\n", flags);
	out("/Ascent %d\n", fi.ascender);
	out("/Descent %d\n", fi.descender);

	if (font_get_type(id) == FONT_TYPE_TYPE1) {
		struct font_t1_info t1;

		if (!font_t1_info(id, &t1))
			trans_fail("Cannot get font %d Type1 info", id);

		out("/StemV %u\n", t1.underline_thickness);
		out("/ItalicAngle %ld\n", t1.italic_angle);

		if (embed_fonts) {
			struct indirect *fontfile = indirect_object(dump_font_type1);
			fontfile->u.font.id = id;
			out("/FontFile %d 0 R\n", fontfile->label);
		}
	}
	else if (embed_fonts) {
		struct indirect *fontfile = indirect_object(dump_font_tt);

		fontfile->u.font.id = id;
		out("/FontFile2 %d 0 R\n", fontfile->label);
	}

	/*
	 * And now we are missing:
	 *
	 * CapHeight:	(Required for fonts that have Latin characters, except
	 * 		for Type 3 fonts) The vertical coordinate of the top of flat
	 * 		capital letters, measured from the baseline.
	 * ItalicAngle: [for TT fonts]
	 * StemV:	[for TT fonts]
	 *
	 * Some of them are TODO/FIXME, some are undeterminable.
	 */
	out(">>\n");
}

UNUSED static void dump_font_widths(struct indirect *i)
{
	garr width = garr_new(512, sizeof(real));
	int n;

	n = font_get_glyph_advances(i->u.font.id, width);
	if (n < 0)
		trans_fail("Computing widths for font %d failed.", i->u.font.id);

	out("[\n");
	for (int i = 0; i < n; i++)
		out("%f\n", *(real *)garr_get(width, i));
	out("]\n");
	garr_destroy(width);
}

static void dump_font_dict(struct indirect *i)
{
	uns id = i->u.font.id;
	struct font_info fi;

	if (!font_info(id, &fi))
		trans_fail("Cannot get font %d info.", id);

	out("<< /Type /Font\n");

	switch (font_get_type(id)) {
	case FONT_TYPE_TYPE1:
		out("/Subtype /Type1\n");
		out("/BaseFont /%s\n", fi.name);
		break;

	case FONT_TYPE_TRUETYPE:
		out("/Subtype /TrueType\n");
		out("/BaseFont /%s\n", fi.name);
		break;

	case FONT_TYPE_INVALID:
		trans_fail("Invalid font %d.", id);
		break;

	case FONT_TYPE_UNKNOWN:
		trans_fail("Unknown font %d type.", id);
		break;

	default:
		BUG();
	}

	if (!font_is_standard(fi.name)) {
		struct indirect *font_des = indirect_object(dump_font_descriptor);
		font_des->u.font.id = id;
		out("/FirstChar 0\n");
		out("/LastChar %d\n", fi.glyphs-1);

		// FIXME!
		// struct indirect *widths = indirect_object(dump_font_widths);
		// widths->u.font.id = id;
		// out("/Widths %d 0 R\n", widths->label);

		out("/FontDescriptor %d 0 R\n", font_des->label);
	}
	out(">>\n");
}

static void dump_segment(struct go_segment *s)
{
	out("q\n");
	if (!set_color(&s->go))
		return;
	set_linewidth(&s->go);
	set_linecap(&s->go);
	set_linejoin(&s->go);
	out("%f %f m %f %f l S\n", s->geom.a.x, s->geom.a.y, s->geom.b.x, s->geom.b.y);
	out("Q\n");
}

static void dump_point(struct go_point *p)
{
	/* hack, circle would be probably better */
	out("q\n");
	if (!set_color(&p->go))
		return;
	set_linewidth(&p->go);
	out("%f %f m %f %f l S\n", p->geom.point.x, p->geom.point.y, p->geom.point.x, p->geom.point.y);
	out("Q\n");
}

static void write_approx(struct geom_fpath *approx)
{
	struct geom_point lpt, fpt;
	void *curve;

	fpt.x = fpt.y = lpt.x = lpt.y = 1/0.0;	// just some invalid point
	GEOM_GROUP_WALK(approx, curve) {
		if (geom_instance_class(curve) == GEOM_CLASS(bezier)) {
			struct geom_bezier *b = curve;

			if (isinf(fpt.x)) {
				fpt.x = b->pt[0].x;
				fpt.y = b->pt[0].y;
			}
			if (!IS_EQUAL_XY(lpt, b->pt[0]))
				out("%f %f m\n", b->pt[0].x, b->pt[0].y);
			out("%f %f %f %f %f %f c\n", b->pt[1].x, b->pt[1].y, b->pt[2].x, b->pt[2].y, b->pt[3].x, b->pt[3].y);
			lpt.x = b->pt[3].x;
			lpt.y = b->pt[3].y;
		}
		else if (geom_instance_class(curve) == GEOM_CLASS(segment)) {
			struct geom_segment *s = curve;
			
			if (isinf(fpt.x))
				fpt = s->a;
			if (!IS_EQUAL_XY(lpt, s->a))
				out("%f %f m\n", s->a.x, s->a.y);
			out("%f %f l\n", s->b.x, s->b.y);
			lpt = s->b;
		}
		else
			BUG();
	}
	if (IS_EQUAL_XY(fpt, lpt))
		out("h\n");
}


static void dump_approx(struct go *g, void *geom_item)
{
	struct geom_fpath approx;
	int stroke, fill;

	if (!GEOM_ETEST(geom_item_convert(geom_item, GEOM_APPROX_CUBIC_NONRATIONAL | GEOM_APPROX_SEGMENT, &approx)))
		trans_fail("Error expanding object");

	out("q\n");
	stroke = set_color(g);
	fill = set_fillcolor(g);
	set_linewidth(g);
	set_linecap(g);
	set_linejoin(g);

	write_approx(&approx);

	if (stroke && fill)
		out("B\n");
	else if (stroke)
		out("S\n");
	else if (fill)
		out("f\n");
	out("Q\n");

	geom_fpath_destroy(&approx);
}

static void use_font(int id)
{
	if (!pdf_font[id].saved) {
		struct indirect *i = indirect_object(dump_font_dict);

		i->u.font.id = id;
		pdf_font[id].saved = 1;
		pdf_font[id].label = i->label;
	}
	pdf_font[id].used = 1;
}

static void dump_text(struct go_text *t)
{
	int id = t->font_id;
	struct geom_transform tr;
	struct geom_point pt;

	go_text_get_transformation(t, &tr);
	go_text_get_visualisation_point(t, &pt);
	out("q\n");
	set_fillcolor(&t->go);
	use_font(id);
	out("BT /F%d %f Tf ", id, t->font_size);
	if (tr.flags & GEOM_TRANSFORM_IDENTITY)
		out("%f %f Td\n", pt.x, pt.y);
	else out("%f %f %f %f %f %f Tm\n", tr.coef[0][0], tr.coef[1][0], tr.coef[0][1], tr.coef[1][1], pt.x, pt.y);
	dump_string(t->text->text);
	out(" Tj ET\n");
	out("Q\n");
}

#define HEX(c) ((c) < 10 ? (c) + '0' : (c) - 10 + 'a')

static void dump_tex_text(struct go_tex_text *t)
{
	uns i;
	uns n = t->process->glyphs_cnt;
	struct geom_transform tr;
	struct geom_point pt;

	go_tex_text_get_transformation(t, &tr);
	go_tex_text_get_visualisation_point(t, &pt);
	out("q\n");
	set_fillcolor(&t->go);
	for (i = 0; i < n; i++) {
		struct tex_glyph *glyph = (struct tex_glyph *)garr_get(t->process->glyphs, i);

		if (glyph->is_rule) {
			struct geom_point pos, p1, p2, p3, p4;
			real x, y;

			pos.x = glyph->xpos;
			pos.y = glyph->ypos;
			geom_point_transform(&pos, &tr, &pos);
			x = pt.x + pos.x;
			y = pt.y + pos.y;
			p1 = glyph->bbox.ll;
			p2.x = glyph->bbox.ur.x;
			p2.y = glyph->bbox.ll.y;
			p3 = glyph->bbox.ur;
			p4.x = glyph->bbox.ll.x;
			p4.y = glyph->bbox.ur.y;
			geom_point_transform(&p1, &tr, &p1);
			geom_point_transform(&p2, &tr, &p2);
			geom_point_transform(&p3, &tr, &p3);
			geom_point_transform(&p4, &tr, &p4);

			out("%f %f m %f %f l %f %f l %f %f l h f\n", x + p1.x, y + p1.y,
					x + p2.x, y + p2.y, x + p3.x, y + p3.y, x + p4.x, y + p4.y);
			continue;
		}

		use_font(glyph->font_id);
		out("BT /F%d %f Tf ", glyph->font_id, glyph->font_size);
		if (tr.flags & GEOM_TRANSFORM_IDENTITY)
			out("%f %f Td ", pt.x + glyph->xpos, pt.y + glyph->ypos);
		else {
			struct geom_point pos;

			pos.x = glyph->xpos;
			pos.y = glyph->ypos;
			tr.coef[0][2] = tr.coef[1][2] = 0;
			geom_point_transform(&pos, &tr, &pos);

			out("%f %f %f %f %f %f Tm\n", tr.coef[0][0], tr.coef[1][0], tr.coef[0][1], tr.coef[1][1],
					pt.x + pos.x, pt.y + pos.y);
		}
		out("<%c%c> Tj ET\n", HEX((glyph->code & 0xff) >> 4), HEX(glyph->code & 0xf));
	}
	out("Q\n");
}

static int go_exportable(struct go *gobj)
{
	/* Skip invisible objects */
	if (prop_lookup(go2o(gobj), string_invisible, PT_UNS, (prop_value){.u=0}).u)
		return 0;
	
	/* Ignore path members */
	if (go_get_group(gobj)->go.subtype == GOST_GROUP_PATH)
		return 0;
	
	return 1;
}

static void page_bbox(struct obj_tlo *tlo, struct geom_rectangle *bbox)
{
	struct go *go;
	struct geom_rectangle tmp;
	
	geom_rectangle_set_empty(bbox);
	WALK_GO_RECURSIVE_WITHOUT_ROOT_BACKWARDS(&tlo->tlg->group.go, go) {
		if (go_exportable(go)) {
			if (go_get_exact_bbox_after_trans(go, &geom_transform_identity, &tmp))
				tmp = *go_get_rtree_bbox(go);
			geom_rectangle_merge(&tmp, bbox, bbox);
		}
	}
	if (geom_rectangle_is_empty(bbox))
		bbox->ll.x = bbox->ll.y = bbox->ur.x = bbox->ur.y = 0;
}

/*
 * Chapter: 3.7.1
 */
static void dump_contents(struct indirect *i)
{
	struct obj_tlo* tlo = i->u.contents.tlo;
	struct go *g;
	struct geom_rectangle bbox;
	real movex, movey;

	cur_tlo = tlo;
	expand_ref(tlo);
	page_bbox(tlo, &bbox);
	movex = PDF_MARGIN_X - bbox.ll.x;
	movey = PDF_MARGIN_Y - bbox.ll.y;

	stream_start();

	/* Transform all coordinates from MM into default PDF units and translate the picture */
	out("%f %f %f %f %f %f cm\n", MM2POINT, 0.0, 0.0, MM2POINT, MM2POINT * movex, MM2POINT * movey);
	out("0.2 w\n");		// default linewidth 0.2mm

	WALK_GO_RECURSIVE_WITHOUT_ROOT_BACKWARDS(&tlo->tlg->group.go, g) {

		if (!go_exportable(g))
			continue;

		/* Take care of path */
		if (g->type == GOT_GROUP && g->subtype == GOST_GROUP_PATH) {
			dump_approx(g, &go2group(g)->geom);
			continue;
		}

		switch (g->type) {
		case GOT_SEGMENT:
			dump_segment((struct go_segment *)g);
			break;
		case GOT_POINT:
			dump_point((struct go_point *)g);
			break;
		case GOT_BEZIER:
			dump_approx(g, &((struct go_bezier *)g)->geom);
			break;
		case GOT_ELARC:
			dump_approx(g, &((struct go_elarc *)g)->geom);
			break;
		case GOT_DECORATOR_POINT:
			dump_approx(g, &((struct go_decorator_point *)g)->geom);
			break;
		case GOT_ARROW:
			dump_approx(g, &((struct go_arrow *)g)->geom);
			break;
		case GOT_TEXT:
			dump_text((struct go_text *)g);
			break;
		case GOT_TEX_TEXT:
			dump_tex_text((struct go_tex_text *)g);
			break;
		}
	}
	stream_end();
}

/*
 * Chapter: 3.6.2
 */
static void dump_page(struct indirect *i)
{
	struct obj_tlo* tlo = i->u.page.tlo;
	struct geom_rectangle bbox;
	real medx, medy;

	page_bbox(tlo, &bbox);
	medx = bbox.ur.x - bbox.ll.x + 2 * PDF_MARGIN_X;
	medy = bbox.ur.y - bbox.ll.y + 2 * PDF_MARGIN_Y;

	out("<< /Type /Page\n");
	out("/Parent %d 0 R\n", i->u.page.parent->label);
	out("/MediaBox [0 0 %f %f]\n", medx*MM2POINT, medy*MM2POINT);

	if (group_get_first_go(&tlo->tlg->group))
		out("/Contents %d 0 R\n", i->label - 1);

	out("/Resources <<\n");

	out("/Font <<\n");
	for (uns i = 0; i <= pdf_max_fntid; i++)
		if (pdf_font[i].used) {
			pdf_font[i].used = 0;
			out("/F%d %d 0 R\n", i, pdf_font[i].label);
		}
	out(">>\n>>\n>>\n");
	expand_unref(tlo);
	cur_tlo = NULL;
}

/*
 * Chapter: 3.6.2
 */
static void dump_pages(struct indirect *i)
{
	struct indirect *ind;
	struct obj_tlo* tlo;
	uns count = 0;
	out("<< /Type /Pages\n");
	out("/Kids [\n");
	WALK_DOC_TLO(pdf_doc, tlo) {
		count++;
		if (group_get_first_go(&tlo->tlg->group))
			indirect_object(dump_contents)->u.contents.tlo = tlo;
		ind = indirect_object(dump_page);
		ind->u.page.tlo = tlo;
		ind->u.page.parent = i;
		out("%d 0 R\n", ind->label);
	}
	out("]\n/Count %d\n>>\n", count);
}

/*
 * Chapter: 3.6.1
 */
static void dump_catalog(struct indirect *i UNUSED)
{
	struct indirect *n = indirect_object(dump_pages);
	out("<< /Type /Catalog\n");
	out("/Pages %d 0 R\n", n->label);
	out("/PageLabels << /Nums [ 0 << /S /D /P (Page ) >> ] >>\n");
	out(">>\n");
}

/*
 * Chapter: 10.2.1
 */
static void dump_document_information(struct indirect *i UNUSED)
{
	time_t tm = time(NULL);
	char tmbuf[50];

	out("<<\n");
	out("/Title ");
	dump_string(obj_lookup_name(&pdf_doc->obj));
	out("\n/Creator (VRR  (http://vrr.ucw.cz/))\n");
	out("/Producer (VRR  (http://vrr.ucw.cz/))\n");

	strftime(tmbuf, sizeof(tmbuf), "D:%Y%m%d%H%M%SZ00'00'", gmtime(&tm));
	out("/CreationDate (%s)\n", tmbuf);
	out(">>\n");
}

void export_pdf(struct obj_doc* doc, FILE *f, int without_fonts, int compression)
{
	struct indirect *i;
	uns crt_offset;
	char err_buf[TRANS_ERR_SIZE];

	ASSERT(trans_present);

	stream_buf = garr_new(65536, 1);
	buffer_output = 0;
	embed_fonts = !without_fonts;
	clist_init(&indirect_list);
	cur_label = 0;
	pdf_max_fntid = font_get_max_fontid();
	pdf_font = xmalloc_zero((font_get_max_fontid()+1) * sizeof(*pdf_font));
	pdf_file = f;
	pdf_doc = doc;

	switch (compression) {
	case 0:
		compress_level = Z_NO_COMPRESSION;
		break;
	case 1:
		compress_level = Z_BEST_SPEED;
		break;
	case 2:
		compress_level = Z_BEST_COMPRESSION;
		break;
	default:
		compress_level = Z_NO_COMPRESSION;
	}

	if (!obj_get_first_son(&doc->obj))
		trans_fail("Document should have at least one page");

	uns root_lab = indirect_object(dump_catalog)->label;
	uns info_lab = indirect_object(dump_document_information)->label;

	out("%%PDF-1.5 %%%c%c%c%c%c\n", 130, 131, 132, 133, 134);

	/*
	 * Body
	 * Chapter: 3.4.2
	 */
	TRANS_BEGIN_ANONYMOUS(err_buf)
		CLIST_WALK(i, indirect_list) {
			i->offset = ftell(pdf_file);
			out("%d 0 obj\n", i->label);
			i->func(i);
			out("endobj\n");
		}
	TRANS_FAILED
		cleanup();
		trans_fail("%s", err_buf);
	TRANS_END

	/*
	 * Cross-reference table
	 * Chapter: 3.4.3
	 */
	crt_offset = ftell(pdf_file);
	out("xref\n");
	out("0 %d\n", cur_label + 1);

	out("0000000000 65535 f \n");

	CLIST_WALK(i, indirect_list)
		out("%.10d 00000 n \n", i->offset);

	/*
	 * Trailer
	 * Chapter: 3.4.4
	 */
	out("trailer\n");

	out("<<\n");
	out("/Size %d\n", cur_label + 1);
	out("/Root %d 0 R\n", root_lab);
	out("/Info %d 0 R\n", info_lab);
	out(">>\n");

	out("startxref\n");
	out("%d\n", crt_offset);
	out("%%%%EOF\n");

	cleanup();
}
