/*
 * SVG export routines
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <stdio.h>
#include <libxml/xmlwriter.h>
#include "geomlib/geomlib.h"
#include "kernel/kernel.h"
#include "font/font.h"
#include "export/export.h"
#include "import/import.h"

#define SUCCESS		0
#define ERROR		1

static real xoffset, yoffset, width, height;
static xmlTextWriterPtr ptr;

static real transform_x (real x)
{
	return (x - xoffset) * MM2POINT;
}

static real transform_y (real y)
{
	return (height - y + yoffset) * MM2POINT;
}

static real transform_len (real len)
{
	return len * MM2POINT;
}

static double get_opacity_from_alpha (uns alpha)
{
	return ((double) alpha) / 255.0;
}

static int export_style (struct go * go, int fill_is_stroke)
{
	char fill_color[64], stroke_color[64];
	char *stroke_linecap, *stroke_join;
	int ret = SUCCESS;

	uns stroke = prop_lookup_uns (go2o(go), string_stroke_color, 0);
	uns fill = prop_lookup_uns (go2o(go), string_fill_color, 0);
	uns sr, sg, sb, sa, fr, fg, fb, fa;
	uns2rgba_uns (stroke, &sr, &sg, &sb, &sa);
	uns2rgba_uns (fill, &fr, &fg, &fb, &fa);

	sprintf (stroke_color, "rgb(%u, %u, %u)", sr, sg, sb);
	sprintf (fill_color, "rgb(%u, %u, %u)", fr, fg, fb);
	
	uns cap = prop_lookup_uns (go2o (go), string_stroke_cap, 0);
	switch (cap)
	{
		case 0:
			stroke_linecap = "butt";
			break;
		case 1:
			stroke_linecap = "round";
			break;
		case 2:
			stroke_linecap = "square";
			break;
		default:
			DBG ("Bad stroke cap, setting the default value.");
			stroke_linecap = "butt";
	}
	uns join = prop_lookup_uns (go2o (go), string_stroke_join, 0);
	switch (join)
	{
		case 0:
			stroke_join = "miter";
			break;
		case 1:
			stroke_join = "round";
			break;
		case 2:
			stroke_join = "bevel";
			break;
		default:
			DBG ("Bad stroke join, setting the default value.");
			stroke_join = "miter";
	}
	
	ret = SUCCESS;
	real len = prop_lookup_real (go2o (go), string_stroke_width, 0);
	
	if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke", "%s", stroke_color) < 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke-width", "%.2f", len ? transform_len ((double) prop_lookup_real (go2o (go), string_stroke_width, 1.0)) : 1 ) < 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke-opacity", "%.2f", get_opacity_from_alpha (sa)) ||
			(!fill_is_stroke && xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "fill", "%s", fill_color) < 0) ||
			(!fill_is_stroke && xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "fill-opacity", "%.2f", get_opacity_from_alpha (fa)) < 0) ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "visibility", "%s", prop_lookup_uns (go2o (go), string_invisible, 0) ? "hidden" : "visible") < 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke-linecap", "%s", stroke_linecap) < 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke-join", "%s", stroke_join) < 0)
		ret = ERROR;

	return ret;
}

static int export_svg_bezier_path (struct go *go)
{
	struct geom_approx approx;
	struct geom_approx_item *item;
	struct geom_bezier *b;
	char *str, tmp[1024];
	int ret = SUCCESS;

	geom_approx_init (&approx);

	if (go_approx (&approx, go, GEOM_APPROX_SHAPE | GEOM_APPROX_CUBIC_NONRATIONAL))
	{
	   	str = alloca (approx.count * 72 + 128);
		sprintf(str, "M%f %f",
				transform_x (approx.items[0].data.bezier.bezier.pt[0].x),
				transform_y (approx.items[0].data.bezier.bezier.pt[0].y));
		for (uns i = 0; i < approx.count; i++)
		{
			item = approx.items + i;
			b = &(item->data.bezier.bezier);
			sprintf(tmp, " C%10f,%10f %10f,%10f %10f,%10f",
					transform_x (b->pt[1].x), transform_y (b->pt[1].y),
					transform_x (b->pt[2].x), transform_y (b->pt[2].y),
					transform_x (b->pt[3].x), transform_y (b->pt[3].y));
			DBG("data: %s\n", tmp);
			str = strcat (str, tmp);
		}
		if(xmlTextWriterStartElement (ptr, (xmlChar *) "path") < 0) 
			return ERROR;
		else
		{
			if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "d", "%s", str) >= 0)
			{
				if (export_style(go, 0) == ERROR || xmlTextWriterEndElement(ptr) < 0)
					ret = ERROR; 
			}
			else
				ret = ERROR;
		}
	}
	geom_approx_destroy(&approx);
	return ret;
}

static int export_point (struct go_point *go)
{
	int ret = SUCCESS;
	
	if(xmlTextWriterStartElement (ptr, (xmlChar *) "circle") < 0) 
		return ERROR;
	else
	{
		if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "cx", "%f", transform_x (go->geom.point.x)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "cy", "%f", transform_y (go->geom.point.y)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "r",  "%d", 1) >= 0)
		{
			if (export_style(&go->go, 1) == ERROR || xmlTextWriterEndElement(ptr) < 0)
				ret = ERROR;; 
		}
		else
			ret = ERROR;
	}
	ret = ERROR;
	return ret;
}

static int export_segment (struct go_segment *go)
{
	int ret = SUCCESS;
   	real x1, y1, x2, y2;
   
	x1 = go->geom.a.x;
   	y1 = go->geom.a.y;
   	x2 = go->geom.b.x;
   	y2 = go->geom.b.y;

	if(xmlTextWriterStartElement (ptr, (xmlChar *) "line") < 0)
		return ERROR;
	else
	{
		if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x1", "%f", transform_x (x1)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y1", "%f", transform_y (y1)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x2", "%f", transform_x (x2)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y2", "%f", transform_y (y2)) >= 0)
		{
			if (export_style((struct go *) go, 1) == ERROR || xmlTextWriterEndElement (ptr) < 0)
				ret = ERROR;
		}
		else
			ret = ERROR;
	}
	return ret;
}

static int export_elliptic_arc (struct go *go)
{
	return export_svg_bezier_path (go);
}

static int export_rectangle (real x, real y, real width, real height, struct go_tex_text *t)
{
	int ret = SUCCESS;
	char color[1024];

	uns fill = prop_lookup_uns (go2o((struct go *) t), string_fill_color, 0);
	uns fr, fg, fb, fa;
	uns2rgba_uns (fill, &fr, &fg, &fb, &fa);
	sprintf (color, "rgb(%u, %u, %u)", fr, fg, fb);
	
	if(xmlTextWriterStartElement (ptr, (xmlChar *) "rect") < 0)
		ret = ERROR;
	else if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x", "%f", transform_x (x)) < 0 ||
			 xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y", "%f", transform_y (y)) < 0 ||
			 xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "width", "%f", transform_len (width)) < 0 ||
			 xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "height", "%f", transform_len (height)) < 0 ||
	 		 xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke", "%s", color) < 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "fill", "%s", color)< 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "visibility", "%s", prop_lookup_uns (go2o ((struct go *) t), string_invisible, 0) ? "hidden" : "visible") < 0 ||
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "opacity", "%f", get_opacity_from_alpha (fa)) < 0 ||
			xmlTextWriterEndElement (ptr) < 0)
				ret = ERROR;

	return ret;
}

static int export_text (struct go_text *t)
{
	const char *fontname;
	char transf[1024];
	struct geom_transform tr;
	struct geom_rectangle box;
	int ret = SUCCESS;
	char color[1024];

	DBG("Exporting text\n");

	uns fill = prop_lookup_uns (go2o((struct go *) t), string_fill_color, 0);
	uns fr, fg, fb, fa;
	uns2rgba_uns (fill, &fr, &fg, &fb, &fa);
	sprintf (color, "rgb(%u, %u, %u)", fr, fg, fb);

	if(xmlTextWriterStartElement (ptr, (xmlChar *) "text") < 0)
		return ERROR;
	
	fontname = font_get_ps_name (t->font_id);
	go_text_get_transformation (t, &tr);
	go_text_get_text_bbox (t, &box);
	sprintf(transf, "%f %f %f %f %f %f",
			tr.coef[0][0], -tr.coef[1][0], 
			-tr.coef[0][1], tr.coef[1][1],
			transform_x (tr.coef[0][2]), transform_y (tr.coef[1][2]));

	if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x", "0") >= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y", "0") >= 0 &&
	    xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "font-size", "%f", transform_len (t->font_size)) >= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "transform", "matrix(%s)", transf) >= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke", "%s", color) >= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke-opacity", "%f", get_opacity_from_alpha (fa)) >= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "fill", "%s", color)>= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "visibility", "%s", prop_lookup_uns (go2o ((struct go *) t), string_invisible, 0) ? "hidden" : "visible") >= 0 &&
		xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "opacity", "%f", get_opacity_from_alpha (fa)) >= 0)
	{
		if ((fontname && xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "font-family", "%s", fontname) < 0) ||
			xmlTextWriterWriteString (ptr, (xmlChar *) t->text->text) < 0)
				ret = ERROR;
	}
	else 
		ret = ERROR;
	
	if (xmlTextWriterEndElement (ptr) < 0)
		ret = ERROR;

	return ret;
}

static int export_tex_text (struct go_tex_text *t)
{
	uns i;
	char transf[1024];
	int ret = SUCCESS;
	char color[1024];
	if (!t->process->glyphs_cnt)
		return SUCCESS;
	struct tex_glyph *glyph = (struct tex_glyph*) garr_get (t->process->glyphs, 0);
	struct geom_point pos;
	pos.x = glyph->xpos;
	pos.y = glyph->ypos;


	uns fill = prop_lookup_uns (go2o((struct go *) t), string_fill_color, 0);
	uns fr, fg, fb, fa;
	uns2rgba_uns (fill, &fr, &fg, &fb, &fa);
	sprintf (color, "rgb(%u, %u, %u)", fr, fg, fb);

	for (i = 0; i < t->process->glyphs_cnt; i++)
	{
		glyph = (struct tex_glyph*) garr_get (t->process->glyphs, i);
		struct geom_transform tr;
		struct geom_rectangle box;
	
		pos.x = glyph->xpos;
		pos.y = glyph->ypos;
		
		go_tex_text_get_text_bbox (t, &box);	
		go_tex_text_get_transformation (t, &tr);
		if (geom_point_transform(&pos, &tr, &pos) == GEOM_ERR_NUMERIC)
			continue;	// numeric error
		sprintf(transf, "%f %f %f %f %f %f",
			tr.coef[0][0], -tr.coef[1][0], 
			-tr.coef[0][1], tr.coef[1][1],
			transform_x (pos.x), transform_y (pos.y));

		if (glyph->is_rule)
		{
			struct geom_point p1, p2;

			p1 = glyph->bbox.ll;
			p2 = glyph->bbox.ur;
			DBG("hbox: %f, %f - %f, %f\n", glyph->bbox.ll.x, glyph->bbox.ll.y, glyph->bbox.ur.x, glyph->bbox.ur.y);

			geom_point_transform (&p1, &tr, &p1);
			geom_point_transform (&p2, &tr, &p2);
			export_rectangle (pos.x, pos.y, p2.x - p1.x, p2.y - p1.y, t);
		}
		else	// character
		{
			char str[2];
			struct font_info fi;

			if(xmlTextWriterStartElement (ptr, (xmlChar *) "text") < 0)
			{
				ret = ERROR;
				continue;
			}
			else
			{
				if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x", "0") >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y", "0") >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "font-size", "%f", transform_len (glyph->font_size)) >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke", "%s", color) >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "stroke-opacity", "%f", get_opacity_from_alpha (fa)) >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "fill", "%s", color) >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "transform", "matrix(%s)", transf) >= 0 &&
					xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "visibility", "%s", prop_lookup_uns (go2o ((struct go *) t), string_invisible, 0) ? "hidden" : "visible") >= 0)
				{
					if (font_info (glyph->font_id, &fi) &&
						xmlTextWriterWriteFormatAttribute(ptr, (xmlChar *) "font-family", "%s", fi.family) < 0)
						ret = ERROR;
					else
					{
						if (glyph->code >= 32 && glyph->code < 128)
						{
							str[0] = glyph->code;
							str[1] = 0;

							if (xmlTextWriterWriteString (ptr, (xmlChar *) str) < 0)
								ret = ERROR;
						}
					}
				}
				else
					ret = ERROR;
				
				if (xmlTextWriterEndElement (ptr) < 0)
						ret = ERROR;
			}
		}
		if (ret == ERROR)
			return ERROR;
	}
	return ret;
}

static int export_bezier (struct go *go)
{
	DBG("Exporting bezier\n");
	return export_svg_bezier_path (go);
}

static int write_approx (struct geom_fpath *approx, struct go *go)
{
	void *curve;
	int ret = SUCCESS;
	
	GEOM_GROUP_WALK (approx, curve) 
	{
		if (geom_instance_class (curve) == GEOM_CLASS (bezier))
		{
			char str[256];
			struct geom_bezier *b = curve;
			
			sprintf(str, "M%f %f",
				transform_x (b->pt[0].x),
				transform_y (b->pt[0].y));
			sprintf(str + strlen (str), " C%10f,%10f %10f,%10f %10f,%10f",
					transform_x (b->pt[1].x), transform_y (b->pt[1].y),
					transform_x (b->pt[2].x), transform_y (b->pt[2].y),
					transform_x (b->pt[3].x), transform_y (b->pt[3].y));
			if(xmlTextWriterStartElement (ptr, (xmlChar *) "path") < 0) 
			{
				ret = ERROR;
				continue;
			}
			else
			{
				if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "d", "%s", str) >= 0)
				{
					if (export_style(go, 0) == ERROR || xmlTextWriterEndElement(ptr) < 0)
						ret = ERROR;
				}	
				else
					ret = ERROR;
			}
		}
		else if (geom_instance_class (curve) == GEOM_CLASS (segment))
		{
			struct geom_segment *s = curve;
			real x1 = s->a.x;
			real y1 = s->a.y;
			real x2 = s->b.x;
			real y2 = s->b.y;
			if (xmlTextWriterStartElement (ptr, (xmlChar *) "line") < 0)
			{
				ret = ERROR;
				continue;
			}
			else
			{
				if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x1", "%f", transform_x (x1)) >= 0 &&
						xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y1", "%f", transform_y (y1)) >= 0 &&
						xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "x2", "%f", transform_x (x2)) >= 0 &&
						xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "y2", "%f", transform_y (y2)) >= 0)
				{
					if (export_style(go, 1) == ERROR || xmlTextWriterEndElement (ptr) < 0)
						ret = ERROR;
				}
				else
					ret = ERROR;
			}
		}	
		else
			DBG ("Bad curve type in approximation.");
	}

	return ret;
}

static int export_decorator (struct go_decorator_point *dp)
{
	struct geom_fpath approx;
	int ret = SUCCESS;

	if (!GEOM_ETEST (geom_item_convert (&dp->geom, GEOM_APPROX_CUBIC_NONRATIONAL | GEOM_APPROX_SEGMENT, &approx)))
		trans_fail ("Error while expanding decorator");

	ret = write_approx (&approx, (struct go*) dp);	
	
	geom_fpath_destroy (&approx);
	
	return ret;
}

static int  export_arrow (struct go_arrow *arrow)
{
	struct geom_fpath approx;
	int ret = SUCCESS;

	if (!GEOM_ETEST (geom_item_convert (&arrow->geom, GEOM_APPROX_CUBIC_NONRATIONAL | GEOM_APPROX_SEGMENT, &approx)))
		trans_fail ("Error expanding arrow");
	
	write_approx (&approx, (struct go *) arrow);
	
	geom_fpath_destroy (&approx);

	return ret;
}

static void export_objects (struct obj_tlo *tlo)
{
	struct go *gobj;
	int ret = SUCCESS;

	WALK_GO_RECURSIVE_WITHOUT_ROOT_BACKWARDS(&tlo->tlg->group.go, gobj)
	{
		switch (gobj->type)
		{
			case GOT_POINT:
				ret = export_point ((struct go_point *) gobj);
				break;
			case GOT_SEGMENT:
				ret = export_segment ((struct go_segment *) gobj);
				break;
			case GOT_ELARC:
				ret = export_elliptic_arc (gobj);
				break;
			case GOT_TEXT:
				ret = export_text ((struct go_text*) gobj);
				break;
			case GOT_TEX_TEXT:
				ret = export_tex_text ((struct go_tex_text*) gobj);
				break;
			case GOT_BEZIER:
				ret = export_bezier (gobj);
				break;
			case GOT_DECORATOR_POINT:
				ret = export_decorator ((struct go_decorator_point *) gobj);
				break;
			case GOT_ARROW:
				ret = export_arrow ((struct go_arrow *) gobj);
// FIXME: Not implemented yet.
#if 0
			case GOT_PATH:
				export_path ((struct go_path *) gobj);
				break;
			case GOT_PARAMETRIC_POINT:
			case GOT_INTERSECTION:
				break; 	// do not export
			case GOT_INSTANCE:
			case GOT_SCRIPT:
			case GOT_GROUP:
				break;	//NIY
#endif
			default:
				DBG("Bad go type!!!\n");
		}
		if (ret == ERROR)
		{
			DBG("Error while exporting svg.");
			return;
		}
	}
}

static void export_tlo(struct obj_tlo *tlo, const char *filename)
{
	ptr = xmlNewTextWriterFilename(filename, 0);
	if(!ptr)
	{
		DBG("Couldn't create text writer.\n");
		return;
	}
	if ((xmlTextWriterStartDocument(ptr, NULL, "UTF-8", NULL) < 0) ||
		(xmlTextWriterSetIndent(ptr, 1) < 0) ||
		(xmlTextWriterSetIndentString(ptr, (xmlChar *) "\t") < 0))
	{
		xmlFreeTextWriter (ptr);
		return;
	}
	if(xmlTextWriterWriteDTD(ptr, (xmlChar *) "svg", (xmlChar *) "-//W3C//DTD SVG 1.1//EN", (xmlChar *) "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd", NULL) < 0) 
	{ 
		DBG("Error while exporting DTD"); 
		xmlFreeTextWriter (ptr);
		return; 
	}
	if(xmlTextWriterStartElement (ptr, (xmlChar *) "svg") >= 0)
	{
		if (xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "width", "%d", (int) (width * MM2POINT)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "height", "%d", (int) (height * MM2POINT)) >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "version", "%s", "1.1") >= 0 &&
			xmlTextWriterWriteFormatAttribute (ptr, (xmlChar *) "xmlns", "%s", "http://www.w3.org/2000/svg") >= 0 &&
			xmlTextWriterStartElement (ptr, (xmlChar *) "desc") >= 0 &&
			xmlTextWriterWriteString (ptr, (xmlChar *) "SVG file exported by VRR") >= 0 &&
			xmlTextWriterEndElement (ptr) >= 0)
		{
			export_objects(tlo);
			if (xmlTextWriterEndElement (ptr) < 0 ||
				xmlTextWriterEndDocument(ptr) < 0 ||
				xmlTextWriterFlush(ptr) < 0)
					DBG("Couldn't end the document and flush the text\n");
		}
		else
			DBG("Error while exporting svg file.\n");
	}
	xmlFreeTextWriter(ptr);
}

void export_svg (struct obj_tlo *tlo, const char *filename)
{
	const struct geom_rectangle *bbox;

	expand_ref(tlo);
	bbox = tlo_get_rtree_bbox (tlo);

	width = bbox->ur.x - bbox->ll.x;
	height = bbox->ur.y - bbox->ll.y;

	xoffset = bbox->ll.x;
	yoffset = bbox->ll.y;

	export_tlo(tlo, filename);
	expand_unref(tlo);
}
