/** @file
 *
 * 	VRR FONTLIB.
 *
 * 	Routines for rendering font bitmaps and computing bboxes.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#ifndef _KERNEL_FONT_H_
#define _KERNEL_FONT_H_

#include "geomlib/base.h"
#include "geomlib/transform.h"
#include <stdio.h>

#define FAMILY_LENGTH	50

enum font_type
{
	FONT_TYPE_TYPE1,
	FONT_TYPE_TRUETYPE,
	FONT_TYPE_UNKNOWN,
	FONT_TYPE_INVALID,
};

/**
 * Structure with information about font.
 */
struct font_info
{
	char family[FAMILY_LENGTH];		///< font family name
	char style[50];			///< font style name
	char name[50];			//< Postscript name
	int italic, bold, fixed_width;		///< flags
	int glyphs, charmaps;		///< number of objects
	int fontid;			///< font descriptor
	int type;			///< type as specified in enum font_type
	struct geom_rectangle bbox; 	//< The box is large enough to contain any glyph from the font.
	int ascender;
	int descender;
};

/**
 * Structure with information about Type1 font,
 * including some Type1 specific fields.
 */
struct font_t1_info
{
	/* Standard information */
	char family[50];		///< font family name
	char style[50];			///< font style name
	int italic, bold;		///< flags
	int glyphs, charmaps;		///< number of objects

	/* Things specific to Type1 fonts follows */
	char full_name[50];
	char family_name[50];
	char version[50];
	char notice[50];
	char weight[50];
	long italic_angle;
	struct geom_rectangle bbox;
	uns units_per_em;
	int ascender;
	int descender;
	int is_fixed_pitch;
	int underline_position;
	uns underline_thickness;
};

struct font_truetype_info
{
	long italic_angle;
};

enum font_charmap {
	FONT_CHARMAP_NONE = 0,
	FONT_CHARMAP_STD,
	FONT_CHARMAP_TEX,
	FONT_CHARMAP_UNICODE,
	FONT_CHARMAP_LATIN1,
};

/**
 * Control structure for font rendering.
 * You should call font_ctl_init to initialize the structure before first use
 * and font_ctl_cleanup after use.
 */
struct font_ctl {
	int font_id;			///< font descriptor
	real font_size;			///< font size, in font units
	uns charmap;			///< charmap conversions
	struct geom_transform transform;///< transformation to be applied
	byte *bitmap;			///< bitmap buffer
	uns bsize;			///< allocated bitmap buffer size
	int width, height;		///< resulting bitmap size in pixels
	int xoff, yoff;			///< when drawing to screen add these offsets to refpoint
};

/**
 * Prepare font_ctl structure for use.
 * Zeroes all fields and sets the default transformation to identity.
 * You should call it before the first use of any font_ctl structure.
 *
 * @a	ctl	structure to initialize
 * @r	always ctl
 */
static inline struct font_ctl *font_ctl_init(struct font_ctl *ctl)
{
	bzero(ctl, sizeof(*ctl));
	ctl->transform = geom_transform_identity;
	return ctl;
}

/**
 * Discard the guts of struct font_ctl.
 * You should call it after use of any font_ctl structure.
 *
 * @a ctl	structure to discard
 */
static inline void font_ctl_cleanup(struct font_ctl *ctl)
{
	xfree(ctl->bitmap);
}

/**
 * Initialize the font server.
 */
void font_init(const char* vrr_name);

/**
 * Cleanup the font server.
 */
void font_finish(void);

/**
 * Test for valid font descriptor.
 *
 * @ai font_id	font descriptor
 * @rv 1	valid font descriptor
 * @rv 0	invalid font descriptor
 */
int font_id_valid(int font_id);

/**
 * Return font filename corresponding to fontdescriptor.
 *
 * @ai font_id	font descriptor
 * @rv NULL	filename not found
 * @rv name	font filename
 */
const char *font_filename(int font_id);

/**
 * Load font file into fontserver.
 * When called second time with the same filename,
 * returns the same fontdescriptor as at the first time.
 * Function checks if the fontfile is valid font usable by Freetype.
 *
 * @ai fname	filename to load
 * @r 0 if error occured (fontfile is probably invalid), resulting font descriptor otherwise
 */
int font_load_file(const char *fname);

/**
 * Compute bounding box of a given glyph.
 * Does not (re)allocate memory for bitmap in ctl, thus you don't have to
 * deallocate it.
 *
 * @a ctl	render controlling structure (uses only fields font_id, font_size, charmap_conv, transform)
 * @a char_code	character code
 * @a bbox	pointer to resulting bbox, in milimeters
 * @rv -1	error occured
 * @rv 0	otherwise
 */
int font_get_glyph_bbox(struct font_ctl *ctl, uns char_code, struct geom_rectangle *bbox);

/**
 * Compute bounding box of a given string.
 * Does not (re)allocate memory for bitmap in ctl, thus you don't have to
 * deallocate it.
 *
 * @a ctl	render controlling structure (uses only fields font_id, font_size, charmap_conv, transform)
 * @a str	input string
 * @a bbox	pointer to resulting bbox, in milimeters
 * @rv -1	error occured
 * @rv 0	otherwise
 */
int font_get_string_bbox(struct font_ctl *ctl, const char *str, struct geom_rectangle *bbox);

/**
 * Render one given glyph.
 * Warning: function looks at font_ctl.bitmap and font_ctl.bsize,
 * allocates new bitmap if fields are zero or automatically reallocates bitmap
 * if it is too small.
 *
 * @a g		render controlling structure
 * @a char_code	character to be rendered
 * @rv -1	error occured
 * @rv 0	otherwise
 */
int font_render_glyph(struct font_ctl *g, uns char_code);

/**
 * Render given string.
 * Warning: function looks at font_ctl.bitmap and font_ctl.bsize,
 * allocates new bitmap if fields are zero or automatically reallocates bitmap
 * if it is too small.
 *
 * @a g		render controlling structure
 * @a str	string to be rendered
 * @rv -1	error occured
 * @rv 0	otherwise
 */
int font_render_string(struct font_ctl *g, const char *str);

/**
 * Detect font type.
 *
 * @ai	font_id	font descriptor
 * @rv	one of FONT_TYPE_xxx constants
 */
int font_get_type(int font_id);

/**
 * Return various information about given font file.
 *
 * @ai font_id	font descriptor
 * @ao fi	pointer on struct where to put informations
 * @r pointer to fi, NULL on error
 */
struct font_info *font_info(int font_id, struct font_info *fi);

/**
 * Return various information about given font file.
 * Does not load file into fontserver.
 *
 * @a fontpath	font filename
 * @ao fi	pointer on struct where to put informations
 * @r pointer to fi, NULL on error
 */
struct font_info *font_info_temporary(const char *fontpath, struct font_info *fi);

/**
 * Return various information about given Type1 font
 *
 * @a font_id	fontdescriptor
 * @ao fi	pointer on struct where to put informations
 * @r pointer to fi, NULL on error
 */
struct font_t1_info *font_t1_info(int font_id, struct font_t1_info *fi);

/**
 * Return various information about given TrueType font
 *
 * @a font_id	fontdescriptor
 * @ao tt	pointer on struct where to put informations
 * @r pointer to fi, NULL on error
 */
struct font_truetype_info *font_truetype_info(int font_id, struct font_truetype_info *tt);

struct garr;

/**
 * Return widths of all glyphs
 *
 * @a font_id	fontdescriptor
 * @ao width	growing array of reals
 * @rv -1	error occured
 * @rv num	number of glyph widths returned in width
 */
int font_get_glyph_advances(int font_id, struct garr *width);

/**
 * Dump various debugging informations about font on stdout.
 *
 * @a font_id	font descriptor corresponding the font to be dumped
 */
void font_info_dump(int font_id);

/**
 * Return maximal fontdescriptor currently known to the fontserver.
 *
 * @r integer	maximal fontdescriptor
 */
uns font_get_max_fontid(void);

/**
 * Converts Type1 PFB files into PFA.
 * Just rewrites if the original file is PFA.
 *
 * @ai font_id	fontdescriptor
 * @ao out	output file opened for writing
 * @rv -1	error occured
 * @rv 0	otherwise
 */
int font_pfb_to_pfa(int font_id, FILE *out);

/**
 * Get PostScript fontname as specified by FontName entry in font dictionary.
 *
 * @ai font_id	font descriptor
 * @rv pointer	to name (read only), NULL on error
 */
const char *font_get_ps_name(int font_id);

/**
 * Save TrueType font loaded into fontserver into external Type42 font file.
 *
 * @ai	fontid		font descriptor
 * @ai	fontname	manually set the FontName dictionary entry, autodetect if NULL
 * @a	out		output file opened for writing
 *
 * @rv	-1 on error, 0 if all OK
 */
int font_tt_to_type42(int fontid, const char *fontname, FILE *out);

struct geom_pathset;
/**
 * Decompose glyph into a curve representation.
 *
 * @ai	ctl		font control structure, set font_id, font_size, transform and charmap
 * @ai	charcode	character to decompose
 * @ao	pathset		initialized (and possibly partially filled) output geom_pathset
 * @rv	-1 on error, 0 if all OK
 */
int font_char_decompose(struct font_ctl *ctl, int charcode, struct geom_pathset *pathset);

/**
 * Try to compute section layout of PFB font files.
 * Note that this can fail on some very nonstandard PFB font files.
 * Does not modify any of the output arguments in case of error.
 *
 * @ai buf		buffer with loaded PFB font file
 * @ai size		size of the buf buffer in bytes
 * @ao length		total data size of all sections in bytes, can be NULL
 * @ao clear_text	size of the initial ASCII section in bytes, can be NULL
 * @ao encrypted	size of the encrypted body in bytes, can be NULL
 * @ao fixed		size of the ASCII trailer in bytes, can be NULL
 * @r -1 on error (invalid PFB format), 0 otherwise
 */
int font_pfb_sections(byte *buf, uns size, uns *length, uns *clear_text, uns *encrypted, uns *fixed);

/**
 * Store the PFB font file data, without segment headers.
 * Especially useful for PDF export.
 * Note that this can fail on some very nonstandard PFB font files.
 * outbuf must be at least the same size as buf. buf == outbuf is allowed.
 *
 * @ai buf	buffer with loaded PFB font file
 * @ai size	size of the buf buffer in bytes
 * @ao outbuf	output buffer where to store the data, buf == outbuf is allowed
 * @r -1 on error (invalid PFB format), number of output bytes otherwise
 */
int font_pfb_data_store(byte *buf, uns size, byte *outbuf);

/**
 * Convert the PFA font data into the binary PFB font data.
 * Output buffer must be large enough, that means at least size+20.
 * If the input data is not PFA, just copies the input buffer.
 * Note that some nonstandard PFA cannot be converted succesfully.
 *
 * @ai buf	buffer with loaded PFA font file
 * @ai size	size of the buf buffer in bytes
 * @ao outbuf	output buffer where to store the result, must be of size at least size+20
 * 		and buf and outbuf must not overlap
 * @r -1 on error (bad PFA file structure), number of output bytes otherwise
 */
int font_pfa_to_pfb_buf(byte *buf, uns size, byte *outbuf);

/**
 * Convert the PFA font file into the binary PFB font file.
 * If the input file is not PFA, just copies the file.
 * Note that some nonstandard PFA cannot be converted succesfully.
 *
 * @ai fntid	valid font descriptor
 * @ao out	output file opened for writing
 * @r -1 on error (bad PFA file structure), 0 otherwise
 */
int font_pfa_to_pfb(int fntid, FILE *out);

/**
 * Search for the best matching font filename according to fontconfig database.
 * In case outbuf is too small, the filename is truncated to fit into the buffer.
 *
 * @ai family family name, can be NULL
 * @ai style style name, can be NULL
 * @ao outbuf output buffer where to store the filename, should be large enough
 * @ai bufsize output buffer size
 * @rv NULL something wrong happened
 * @rv outbuf otherwise
 */
char *font_search(const char *family, const char *style, char *outbuf, uns bufsize);

#endif
