/*
 * GEOMLIB - Basic routines
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"


int geom_dump_depth = 0;
int geom_dump_on = 0;
uns geom_dump_flags = 0;
int geom_dump_separator = 0;
int geom_func_depth = 1;

static char *geom_err_str_tab[] = {
	[0] = "Success.",
	[-GEOM_ERR_NUMERIC] = "Numeric error.",
	[-GEOM_ERR_MATRIX_NO_SOLUTION] = "System of linear equations has no solution.",
	[-GEOM_ERR_MATRIX_SINGULAR] = "Matrix is singular.",
	[-GEOM_ERR_TRANSFORM_SINGULAR] = "Transformation matrix is singular.",
	[-GEOM_ERR_PATH_EMPTY] = "Path is empty.",
	[-GEOM_ERR_PATH_CLOSED] = "Path is closed."
};

static char *geom_err_str_invalid = "Invalid error code.";

int geom_errno = 0;


/* geom_err_str */

char *geom_err_str(int err)
{
	if (err >= 0)
		return geom_err_str_tab[0];
	err = -err;
	if ((uns)err >= ARRAY_SIZE(geom_err_str_tab))
		return geom_err_str_invalid;
	if (geom_err_str_tab[err])
		return geom_err_str_tab[err];
	else
		return geom_err_str_invalid;
}


/* geom_dump_group */

void geom_dump_group(uns flags)
{
	geom_dump_depth++;
	if (geom_dump_on) {
		if (geom_dump_on < geom_dump_depth)
			msg("(");
		else
			geom_dump_flags = flags;
	}
	geom_dump_separator = 0;
}


/* geom_dump_sep */

void geom_dump_sep(const char *c)
{
	if (geom_dump_on && geom_dump_separator)
		msg("%s", c);
	geom_dump_separator = 1;
}


/* geom_dump_leave */

void geom_dump_leave(void)
{
	if (geom_dump_on) {
		if (geom_dump_on < geom_dump_depth)
			msg(")");
		else {
			if (!(geom_dump_flags & 2))
				msg("\n");
			geom_dump_on = 0;
		}
		geom_dump_separator = 1;
	}
	geom_dump_depth--;
}


/* geom_b_assert */

int geom_b_assert(GEOM_ASSERT_DEF, int value)
{
	if (!TEST(value))
		die(GEOM_ASSERT_FORMAT ": boolean test is false\n", GEOM_ASSERT_COPY);
	return value;
}


/* geom_p_assert */

const void *geom_p_assert(GEOM_ASSERT_DEF, const void *value)
{
	if (!PTEST(value))
		die(GEOM_ASSERT_FORMAT ": pointer is NULL\n", GEOM_ASSERT_COPY);
	return value;
}


/* geom_r_assert */

real geom_r_assert(GEOM_ASSERT_DEF, real value)
{
	if (!RTEST(value))
		die(GEOM_ASSERT_FORMAT ": real %f is not finite\n", GEOM_ASSERT_COPY, value);
	return value;
}


/* geom_f_assert */

double geom_f_assert(GEOM_ASSERT_DEF, double value)
{
	if (!FTEST(value))
		die(GEOM_ASSERT_FORMAT ": double %f is not finite\n", GEOM_ASSERT_COPY, value);
	return value;
}


/* geom_e_assert */

int geom_e_assert(GEOM_ASSERT_DEF, int value)
{
	if (!ETEST(value))
		die(GEOM_ASSERT_FORMAT ": error with code %d and message `%s'\n", GEOM_ASSERT_COPY, value, geom_err_str(value));
	return value;
}


/* geom_r_minmax_assert */

real geom_r_minmax_assert(GEOM_ASSERT_DEF, real value, real min, real max)
{
	if (!RTEST_MINMAX(value, min, max))
		die(GEOM_ASSERT_FORMAT ": real %f is not in range [%f,%f]\n", GEOM_ASSERT_COPY, value, min, max);
	return value;
}


/* geom_f_minmax_assert */

double geom_f_minmax_assert(GEOM_ASSERT_DEF, double value, double min, double max)
{
	if (!FTEST_MINMAX(value, min, max))
		die(GEOM_ASSERT_FORMAT ": double %f is not in range [%f,%f]\n", GEOM_ASSERT_COPY, value, min, max);
	return value;
}


#ifdef GEOM_DEBUG

/* Dumping routines */

void geom_func(GEOM_POS_DEF, int max_depth)
{
	if (geom_func_depth++ < max_depth) {
		msg("Entering " GEOM_POS_FORMAT, GEOM_POS_COPY);
		if (geom_func_depth > 1) msg(" [%d]", geom_func_depth);
		msg("\n");
	}
}

void geom_ret(GEOM_POS_DEF, int max_depth)
{
	if (--geom_func_depth < max_depth) {
		msg("Leaving " GEOM_POS_FORMAT, GEOM_POS_COPY);
		if (geom_func_depth >= 1) msg(" [%d]", geom_func_depth + 1);
		msg("\n");
	}
}

int geom_eret(GEOM_POS_DEF, int max_depth, int max_err_depth, int value)
{
	if (--geom_func_depth < (RTEST(value) ? max_depth : max_err_depth)) {
		if (ETEST(value)) msg("Entering"); else msg("Failing");
		msg(" " GEOM_POS_FORMAT, GEOM_POS_COPY);
		if (geom_func_depth >= 1) msg(" [%d]", geom_func_depth + 1);
		msg(" with error code %d", value);
		if (!ETEST(value)) msg(" (msg=`%s')", geom_err_str(value));
		msg("\n");
	}
	return value;
}

real geom_rret(GEOM_POS_DEF, int max_depth, int max_err_depth, real value)
{
	if (--geom_func_depth < (RTEST(value) ? max_depth : max_err_depth)) {
		if (RTEST(value)) msg("Entering"); else msg("Failing");
		msg(" " GEOM_POS_FORMAT, GEOM_POS_COPY);
		if (geom_func_depth >= 1) msg(" [%d]", geom_func_depth + 1);
		msg(" with real result %f\n", value);
	}
	return value;
}

double geom_fret(GEOM_POS_DEF, int max_depth, int max_err_depth, double value)
{
	if (--geom_func_depth < (FTEST(value) ? max_depth : max_err_depth)) {
		if (FTEST(value)) msg("Entering"); else msg("Failing");
		msg(" " GEOM_POS_FORMAT, GEOM_POS_COPY);
		if (geom_func_depth >= 1) msg(" [%d]", geom_func_depth + 1);
		msg(" with double result %f\n", value);
	}
	return value;
}

void *geom_pret(GEOM_POS_DEF, int max_depth, int max_err_depth, void *value)
{
	if (--geom_func_depth < (PTEST(value) ? max_depth : max_err_depth)) {
		if (PTEST(value)) msg("Entering"); else msg("Failing");
		msg(" " GEOM_POS_FORMAT, GEOM_POS_COPY);
		if (geom_func_depth >= 1) msg(" [%d]", geom_func_depth + 1);
		msg(" with result %p\n", value);
	}
	return value;
}

#endif
