/** @file
 *
 * 	GEOMLIB - Bernstein polynomials.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#ifndef _VRR_GEOMLIB_BERNSTEIN_H_
#define _VRR_GEOMLIB_BERNSTEIN_H_

#include "geomlib/base.h"

int geom_bernstein_valid(uns degree, double coef[]);


/**
    \brief Converts polynomial in bernstein form to power form.
    \param degree integer degree of polynomial
    \param coef array of polynomial coefficients (bernstein form)
    \param result array of polynomial coefficient (power form)
    \retval degree success
    \retval -1 numeric error
*/
int geom_bernstein_to_power(uns degree, double coef[], double result[]);

/**
    \brief Converts polynomial in power form to bernstein form.
    \param degree integer degree of polynomial
    \param coef array of polynomial coefficients (power form)
    \param result array of polynomial coefficient (bernstein form)
    \retval degree success
    \retval -1 numeric error
*/
int geom_power_to_bernstein(uns degree, double coef[], double result[]);

/**
    \brief Function evaluates a real polynomial in bernstein form at given parameter t.
    \param degree integer degree of polynomial
    \param coef array of polynomial coefficients (in bernstein form)
    \param t a real number
    \retval 0 success
    \retval -1 numeric overflow
*/
double geom_bernstein_evaluate(uns degree, double coef[], double t);

/**
    \brief Function adds two polynomials in bernstein form.
    \param degree integer degree of polynomials
    \param coef1 array of first polynomial coefficients (in bernstein form)
    \param coef1 array of second polynomial coefficients
    \param result array of result polynomial coefficients (size degree)
    \retval degree success
    \retval -1 numeric error
*/
int geom_bernstein_add(uns degree, double coef1[], double coef2[], double result[]);

/**
    \brief Function subtracts two polynomials in bernstein form.
    \param degree integer degree of polynomials
    \param coef1 array of first polynomial coefficients (in bernstein form)
    \param coef1 array of second polynomial coefficients
    \param result array of result polynomial coefficients (size degree)
    \retval degree success
    \retval -1 numeric error
*/
int geom_bernstein_sub(uns degree, double coef1[], double coef2[], double result[]);

/**
    \brief Function returns product of two polynomials in bernstein form.
    \param degree1 integer degree of first polynomial
    \param coef1 array of first polynomial coefficients (in bernstein form)
    \param degree2 integer degree of second polynomial
    \param coef1 array of second polynomial coefficients
    \param result array of result polynomial coefficients (size degree1+degree2+1)
    \retval >=0 integer degree of result polynomial
    \retval -1 numeric error
*/
int geom_bernstein_mul(uns degree1, double coef1[], uns degree2, double coef2[], double result[]);

/**
    \brief Calculates first derivation of a polynomial in bernstein form.
    \param degree integer degree of polynomial
    \param coef array of bernstein polynomial coefficients
    \param result array of derivation coefficients (in bernstein form, size degree-1)
    \retval >=0 degree of result polynomial
    \retval -1 numeric error
*/
int geom_bernstein_derivate(uns degree, double coef[], double result[]);


/**
	\brief Function finds all real roots of a real polynomial in bernstein form.
		If GEOM_SOLVE_UNIT_INTERVAL is set, it always finish without numeric error.
		It is developement version only. At first, function geom_polynomial_solve is called
		and if numeric error occured, slower algorithm is used instead
		(but it cant deal with GEOM_SOLVE_MULTIPLICITY - multiple roots are returned only once).


	\param degree integer degree of polynomial
	\param coef array of bernstein polynomial coefficients
	\param flags OR combinations of
		GEOM_SOLVE_MULTIPLICITY (returns multiple roots as multiple array entries),
		GEOM_SOLVE_LEFT_ONLY (returns one root only),
		GEOM_SOLVE_UNIT_INTERVAL (returns only roots in closed interval <0,1>)
	\param result array of found roots (degree entries needed, one entry with GEOM_SOLVE_LEFT_ONLY)

	\retval >=0 number of found roots
	\retval -1 numeric error
*/
int geom_bernstein_solve(uns degree, double coef[], uns flags, double result[]);


/*int geom_bcom(int degree, double coef[]);
void geom_bcom_test(void);*/

#endif
