/*
 * GEOMLIB - Computing bezier derivations
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/numeric.h"
#include "geomlib/polynomial.h"
#include "geomlib/bernstein.h"
#include "geomlib/bezier.h"


/* geom_bezier_derivation_start */

int geom_bezier_derivation_start(struct geom_bezier *bezier, struct geom_vector *result)
{
	ITEM_ASSERT(bezier, bezier);
	PASSERT(result);

	result->dx = BEZIER_DEGREE(bezier) * (bezier->pt[1].x - bezier->pt[0].x);
	result->dy = BEZIER_DEGREE(bezier) * (bezier->pt[1].y - bezier->pt[0].y);
	if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL)) {
		real t = bezier->pt[1].w / bezier->pt[0].w;
		result->dx *= t;
		result->dy *= t;
	}
	return geom_vector_valid(result) ? 0 : GEOM_ERR_NUMERIC;
}


/* geom_bezier_derivation_end */

int geom_bezier_derivation_end(struct geom_bezier *bezier, struct geom_vector *result)
{
	uns d;
	
	ITEM_ASSERT(bezier, bezier);
	PASSERT(result);

	d = BEZIER_DEGREE(bezier);
	result->dx = d * (bezier->pt[d].x - bezier->pt[d - 1].x);
	result->dy = d * (bezier->pt[d].y - bezier->pt[d - 1].y);
	if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL)) {
		real t = bezier->pt[d - 1].w / bezier->pt[d].w;
		result->dx *= t;
		result->dy *= t;
	}
	return geom_vector_valid(result) ? 0 : GEOM_ERR_NUMERIC;
}


/* geom_bezier_derivation_at_time */

int geom_bezier_derivation_at_time(struct geom_bezier *bezier, real time, struct geom_vector *result)
{
	ITEM_ASSERT(bezier, bezier);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		return geom_bezier_derivation_start(bezier, result);
	else if (time >= 1)
		return geom_bezier_derivation_end(bezier, result);
	else {
		struct geom_bezier left, right;
		real s;
		GEOM_INSTANCE_INIT(&left, bezier);
		GEOM_INSTANCE_INIT(&right, bezier);
		geom_bezier_split_at(bezier, time, &left, &right);
		if (time <= 0.5) {
			ETEST_RET(geom_bezier_derivation_start(&right, result));
			s = 1 / (1 - time);
		} else {
			ETEST_RET(geom_bezier_derivation_end(&left, result));
			s = 1 / time;
		}
		GEOM_INSTANCE_DESTROY2(&left, bezier);
		GEOM_INSTANCE_DESTROY2(&right, bezier);
		result->dx *= s;
		result->dy *= s;
		return geom_vector_valid(result) ? 0 : GEOM_ERR_NUMERIC;
	}
}


/* geom_bezier_direction_times
 * Function returns parameters, where is the curve parallel to given vector.
 *
 * TODO:
 * - zamyslet se, zda u racionalnich bezierek nestaci pocitat polynomy mensiho stupne...
 *   treba se mi zda, ze kvadraticke krivky maj max. 1 reseni, tak proc by mel mit polynom stupen 3
 * - predpocitat koeficienty misto rady volani pro derivace, nasobeni a odcitani polynomu */

int geom_bezier_direction_times(struct geom_bezier *b, struct geom_point *v, uns flags, real result[])
{
	int i, d, n/*, m*/;
	double s, v_x, v_y, pt_x[4], pt_y[4], coef[6], w[4], dw[3], wa[4], dwa[3], w_dwa[6], wa_dw[6], temp[5];
	/*struct geom_point p;
	struct geom_bezier c;*/

	ITEM_ASSERT(bezier, b);
	GASSERT(point, v);
	PASSERT(result);

	/* linear curve should always return 0 */
	if (BEZIER_DEGREE(b) == GEOM_BEZIER_LINEAR)
		return 0;

	/* scale vector v */
	s = MAX(fabs(v->x), fabs(v->y));
	if (s == 0)
		return 0;
	v_x = v->x / s;
	v_y = v->y / s;

	/* scale bezier curve coordinates */
	s = 0;
	for (i = 0; i <= BEZIER_DEGREE(b); i++) {
		if (s < fabs(b->pt[i].x))
			s = fabs(b->pt[i].x);
		if (s < fabs(b->pt[i].y))
			s = fabs(b->pt[i].y);
	}
	if (s == 0)
		return 0;
	for (i = 0; i <= BEZIER_DEGREE(b); i++) {
		pt_x[i] = b->pt[i].x / s;
		pt_y[i] = b->pt[i].y / s;
	}

	/* rotate around v */
	for (i = 0; i <= BEZIER_DEGREE(b); i++)
		coef[i] = pt_x[i] * v_y + pt_y[i] * v_x;

	/* nonrational curve */
	if (INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL) {
		if  ((d = geom_bernstein_derivate(BEZIER_DEGREE(b), coef, coef)) < 0)
			return -1;
	}

	/* rational curve */
	else {
		/* scale weigths */
		s = 0;
		for (i = 0; i <= BEZIER_DEGREE(b); i++) {
			if (s < b->pt[i].w)
				s = b->pt[i].w;
		}

		/* prepare bernstein_polynomial */
		for (i = 0; i <= BEZIER_DEGREE(b); i++) {
			w[i] = b->pt[i].w / s;
			wa[i] = coef[i] * w[i];
		}

		d = geom_bernstein_derivate(BEZIER_DEGREE(b), w, dw);
		geom_bernstein_derivate(BEZIER_DEGREE(b), wa, dwa);
		geom_bernstein_mul(BEZIER_DEGREE(b), w, d, dwa, w_dwa);
		geom_bernstein_mul(BEZIER_DEGREE(b), wa, d, dw, wa_dw);
		d += BEZIER_DEGREE(b);
		geom_bernstein_sub(d, w_dwa, wa_dw, coef);
	}

	/* solve polynomial and filter results if necessary */
	if (!(flags & GEOM_BEZIER_DIRECTION_NOBACK)) {
		n = geom_bernstein_solve(d, coef, GEOM_SOLVE_UNIT_INTERVAL | (flags & GEOM_SOLVE_LEFT_ONLY), temp);
		for (i = 0; i < n; i++)
			RTEST_NUM(result[i] = temp[i]);
		return n;
	}
	else
	{
		BUG(); /* not implemented */
#if 0
		n = geom_bernstein_solve(d, coef, GEOM_SOLVE_UNIT_INTERVAL, temp);

		/* create scaled bezier structure */
		c.degree = b->degree;
		c.flags = b->flags & GEOM_BEZIER_NONRATIONAL;
		for (i = 0; i <= b->degree; i++) {
			c.pt[i].x = pt_x[i];
			c.pt[i].y = pt_y[i];
			c.pt[i].w = w[i];
		}
		/* filter */
		m = 0;
		for (i = 0; i < n; i++) {
			geom_bezier_derivation_at_time(&c, temp[i], (struct geom_vector *)(void*)(&p));
			if (p.x * v_x - p.y * v_y >= 0)
			{
				result[m++] = temp[i];
				if (b->flags & GEOM_SOLVE_LEFT_ONLY)
					return 1;
			}
		}
		return m;
#endif		
	}
}


