/*
 * GEOMLIB - Handling conic curves and their sections
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/vector.h"
#include "geomlib/matrix.h"
#include "geomlib/bezier.h"
#include "geomlib/path.h"
#include "geomlib/circle.h"
#include "geomlib/ellipse.h"
#include "geomlib/conic.h"


/* normalize_angle */

static void normalize_angle(real *rotation)
{
	*rotation -= (2 * M_PI) * floor(*rotation / (2 * M_PI));
	if (!isfinite(*rotation) || *rotation < 0)
		*rotation = 0;
	else if (*rotation > 2 * M_PI)
		*rotation = 2 * M_PI;
}


/* normalize_angle_interval */

/*static int normalize_angle_interval(real *start, real *end)
{
	real s, e, dif;

	dif = *end - *start;
	s = *start - (2 * M_PI) * floor(*start / (2 * M_PI));
	if (!isfinite(s) || s < 0)
		s = 0;
	else if (s > 2 * M_PI)
		s = 2 * M_PI;
	e = s + dif;
	if (!isfinite(e))
		return GEOM_ERR_NUMERIC;
	*start = s;
	*end = e;

	return 0;
}*/


/*---------------------------------------------------------------------------------------------*/


/* geom_parabolic_arc_valid */

int geom_parabolic_arc_valid(const struct geom_parabolic_arc *arc)
{
	return arc && geom_point_valid(&arc->center) && isfinite(arc->rotation) && isfinite(arc->distance) && arc->distance >= 0 && isfinite(arc->start) && isfinite(arc->end);
}


/* geom_parabolic_arc_dump */

void geom_parabolic_arc_dump(const struct geom_parabolic_arc *arc)
{
	msg("Dumping geom_parabolic_arc structure at %p", (void*)arc);
	if (!arc) {
		msg(".\n");
		return;
	}
	msg(": center=(%f, %f), rotation=%f, distance=%f, start=%f, end=%f", arc->center.x, arc->center.y, arc->rotation, arc->distance, arc->start, arc->end);
	if (!geom_parabolic_arc_valid(arc))
		msg(" (invalid)");
	msg("\n");
	return;
}


/* geom_parabolic_arc_set */

void geom_parabolic_arc_set(struct geom_parabolic_arc *arc, const struct geom_point *center, const real rotation, const real distance, const real start, const real end)
{
	BASSERT(arc != NULL);

	arc->center = *center;
	arc->rotation = rotation;
	arc->distance = distance;
	arc->start = start;
	arc->end = end;

}


/* geom_parabolic_arc_normalize */

int geom_parabolic_arc_normalize(struct geom_parabolic_arc *arc)
{

	normalize_angle(&arc->rotation);

	return 0;
}


/* geom_parabolic_arc_set_endpoints */

int geom_parabolic_arc_set_endpoints(struct geom_parabolic_arc *arc, const struct geom_point *start, const struct geom_point *end UNUSED)
{
	real cos1, sin1, t1, t2;

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);

	t1 = cos1 * (start->x - arc->center.x) + sin1 * (start->y - arc->center.y);
	t2 = cos1 * (start->x - arc->center.x) + sin1 * (start->y - arc->center.y);

	if (!isfinite(t1) || !isfinite(t2))
		return GEOM_ERR_NUMERIC;

	arc->start = t1;
	arc->end = t2;
	return 0;
}


/* geom_parabolic_arc_point_at */

int geom_parabolic_arc_point_at(const struct geom_parabolic_arc *arc, const real time, struct geom_point *point)
{
	real cos1, sin1, y;

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	y = SQR(time) / (4 * arc->distance);
	point->x = cos1 * time - sin1 * y + arc->center.x;
	point->y = sin1 * time + cos1 * y + arc->center.y;

	return isfinite(point->x) && isfinite(point->y) ? 0 : GEOM_ERR_NUMERIC;
}


/* geom_parabolic_arc_focus */

int geom_parabolic_arc_focus(const struct geom_parabolic_arc *arc, struct geom_point *focus)
{
	focus->x = arc->center.x - arc->distance * sin(arc->rotation);
	focus->y = arc->center.y + arc->distance * cos(arc->rotation);
	return isfinite(focus->x) && isfinite(focus->y) ? 0 : GEOM_ERR_NUMERIC;
}


/* geom_parabolic_arc_directrix */

int geom_parabolic_arc_directrix(const struct geom_parabolic_arc *arc, struct geom_point *point, struct geom_vector *vector)
{
	real cos1, sin1;

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	if (vector) {
		vector->dx = cos1;
		vector->dy = sin1;
	}
	if (point) {
		point->x = arc->center.x + arc->distance * sin1;
		point->y = arc->center.y - arc->distance * cos1;
		if (!isfinite(point->x) || !isfinite(point->y))
			return GEOM_ERR_NUMERIC;
	}
	return 0;
}


/* geom_parabolic_arc_to_poly */

int geom_parabolic_arc_to_poly(const struct geom_parabolic_arc *arc, real *poly)
{
	uns i;
	real cos1, sin1, cos2, sin2, cosd, sind, cosx, siny;

	BASSERT(poly != NULL);

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	cos2 = SQR(cos1);
	sin2 = SQR(sin1);
	cosx = cos1 * arc->center.x;
	siny = sin1 * arc->center.y;
	cosd = 4 * arc->distance * cos1;
	sind = 4 * arc->distance * sin1;

	poly[0] = cos2;
	poly[1] = 2 * sin1 * cos1;
	poly[2] = sin2;
	poly[3] = -2 * cos1 * (cosx + siny) + sind;
	poly[4] = -2 * sin1 * (cosx + siny) - cosd;
	poly[5] = SQR(cosx) + SQR(siny) + 2 * cosx * siny - sind * arc->center.x + cosd * arc->center.y;

	for (i = 3; i <= 5; i++)
		if (!isfinite(poly[i]))
			return GEOM_ERR_NUMERIC;

	return 0;
}


/* geom_parabolic_arc_to_bezier */

#if 0
int geom_parabolic_arc_to_bezier(const struct geom_parabolic_arc *arc, struct geom_bezier *bezier)
{
	uns i;
	real cos1, sin1, a, x, y;

	BASSERT(bezier != NULL);

	bezier->degree = GEOM_BEZIER_QUADRATIC;
	bezier->flags = GEOM_BEZIER_NONRATIONAL;
	bezier->pt[0].w = bezier->pt[1].w = bezier->pt[2].w = 1;

	a = 0.25 / arc->distance;
	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);

	y = a * SQR(arc->start);
	bezier->pt[0].x = cos1 * arc->start - sin1 * y + arc->center.x;
	bezier->pt[0].y = sin1 * arc->start + cos1 * y + arc->center.y;

	x = 0.5 * (arc->start + arc->end);
	y += a * arc->start * (arc->end - arc->start);
	bezier->pt[1].x = cos1 * x - sin1 * y + arc->center.x;
	bezier->pt[1].y = sin1 * x + cos1 * y + arc->center.y;

	y = a * SQR(arc->end);
	bezier->pt[2].x = cos1 * arc->end - sin1 * y + arc->center.x;
	bezier->pt[2].y = sin1 * arc->end + cos1 * y + arc->center.y;

	for (i = 0; i < 2; i++)
		if (!isfinite(bezier->pt[i].x) || !isfinite(bezier->pt[i].y))
			return GEOM_ERR_NUMERIC;

	return 0;
}


/* geom_parabolic_arc_to_path */

int geom_parabolic_arc_to_path(const struct geom_parabolic_arc *arc, struct geom_path *path)
{
	struct geom_bezier bezier;

	if (geom_parabolic_arc_to_bezier(arc, &bezier) < 0)
		return GEOM_ERR_NUMERIC;
	geom_path_insert_tail(path, GEOM_CURVE_BEZIER, &bezier);
	return 1;
}

#endif


/*---------------------------------------------------------------------------------------------*/


/* geom_hyperbolic_arc_valid */

int geom_hyperbolic_arc_valid(const struct geom_hyperbolic_arc *arc)
{
	return arc && geom_point_valid(&arc->center) && isfinite(arc->rotation) && isfinite(arc->major) && (arc->major >= 0) && isfinite(arc->minor) && (arc->minor >= 0) && isfinite(arc->start) && isfinite(arc->end);
}


/* geom_hyperbolic_arc_dump */

void geom_hyperbolic_arc_dump(const struct geom_hyperbolic_arc *arc)
{
	msg("Dumping geom_hyperbolic_arc structure at %p", (void*)arc);
	if (!arc) {
		msg(".\n");
		return;
	}
	msg(": center=(%f, %f), rotation=%f, major=%f, minor=%f, start=%f, end=%f", arc->center.x, arc->center.y, arc->rotation, arc->major, arc->minor, arc->start, arc->end);
	if (!geom_hyperbolic_arc_valid(arc))
		msg(" (invalid)");
	msg("\n");
	return;
}


/* geom_hyperbolic_arc_set */

void geom_hyperbolic_arc_set(struct geom_hyperbolic_arc *arc, const struct geom_point *center, const real rotation, const real major, const real minor, const real start, const real end)
{
	BASSERT(arc != NULL);

	arc->center = *center;
	arc->rotation = rotation;
	arc->major = major;
	arc->minor = minor;
	arc->start = start;
	arc->end = end;

}


/* geom_hyperbolic_arc_normalize */

int geom_hyperbolic_arc_normalize(struct geom_hyperbolic_arc *arc UNUSED)
{

	/* TODO */

	return 0;
}


/* geom_hyperbolic_arc_set_endpoints */

int geom_hyperbolic_arc_set_endpoints(struct geom_hyperbolic_arc *arc UNUSED, const struct geom_point *start UNUSED, const struct geom_point *end UNUSED)
{

	/* TODO */
	BUG();
}


/* geom_hyperbolic_arc_point_at */

int geom_hyperbolic_arc_point_at(const struct geom_hyperbolic_arc *arc UNUSED, const real time UNUSED, struct geom_point *point UNUSED)
{

	/* TODO */
	BUG();
}


/* geom_hyperbolic_arc_to_poly */

int geom_hyperbolic_arc_to_poly(const struct geom_hyperbolic_arc *arc, real *poly)
{
	uns i;
	real a2, b2, cos1, sin1, cos2, sin2, rotx, roty;
	
	BASSERT(poly != NULL);

	a2 = SQR(arc->major);
	b2 = -SQR(arc->minor);
	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	cos2 = SQR(cos1);
	sin2 = SQR(sin1);
	rotx = cos1 * arc->center.x + sin1 * arc->center.y;
	roty = sin1 * arc->center.x - cos1 * arc->center.y;

	poly[0] = b2 * cos2 + a2 * sin2;
	poly[1] = 2 * sin1 * cos1 * (b2 - a2);
	poly[2] = b2 * sin2 + a2 * cos2;
	poly[3] = -2 * (b2 * cos1 * rotx + a2 * sin1 * roty);
	poly[4] = -2 * (b2 * sin1 * rotx - a2 * cos1 * roty);
	poly[5] = b2 * (SQR(rotx) - a2) + a2 * SQR(roty);

	for (i = 0; i <= 5; i++)
		if (!isfinite(poly[i]))
			return GEOM_ERR_NUMERIC;

	return 0;
}

#if 0

/* geom_hyperbolic_arc_to_bezier */

int geom_hyperbolic_arc_to_bezier(const struct geom_hyperbolic_arc *arc UNUSED, struct geom_bezier *bezier UNUSED)
{
	BASSERT(bezier != NULL);

	/* TODO */
	BUG();
}


/* geom_hyperbolic_arc_to_path */

int geom_hyperbolic_arc_to_path(const struct geom_hyperbolic_arc *arc, struct geom_path *path)
{
	struct geom_bezier bezier;

	if (geom_hyperbolic_arc_to_bezier(arc, &bezier) < 0)
		return GEOM_ERR_NUMERIC;
	geom_path_insert_tail(path, GEOM_CURVE_BEZIER, &bezier);
	return 1;
}

#endif


/*---------------------------------------------------------------------------------------------*/


#if 0

/* geom_conic_valid */

int geom_conic_valid(struct geom_conic *conic)
{
	/*if (!conic)
		return 0;
	switch (conic->type) {
		case GEOM_CONIC_PARABOLIC_ARC:
			return geom_parabolic_arc_valid(&conic->data.parabolic_arc);

		case GEOM_CONIC_ARC:
			return geom_arc_valid(&conic->data.arc);

		case GEOM_CONIC_ELLIPTIC_ARC:
			return geom_elliptic_arc_valid(&conic->data.elliptic_arc);

		case GEOM_CONIC_HYPERBOLIC_ARC:
			return geom_hyperbolic_arc_valid(&conic->data.hyperbolic_arc);

		default:
			return 0;
	}*/
	BUG();
}


/* geom_conic_dump */

void geom_conic_dump(struct geom_conic *conic)
{
	msg("Dumping geom_conic structure at %p", (void*)conic);
	if (!conic) {
		msg(".\n");
		return;
	}
	msg(": type=%d", conic->type);
	switch (conic->type) {
		case GEOM_CONIC_PARABOLIC_ARC:
			msg("\n");
			geom_parabolic_arc_dump(&conic->data.parabolic_arc);
			return;

		case GEOM_CONIC_ARC:
			msg("\n");
			geom_arc_dump(&conic->data.arc);
			return;

		case GEOM_CONIC_ELLIPTIC_ARC:
			msg("\n");
			geom_elliptic_arc_dump(&conic->data.elliptic_arc);
			return;

		case GEOM_CONIC_HYPERBOLIC_ARC:
			msg("\n");
			geom_hyperbolic_arc_dump(&conic->data.hyperbolic_arc);
			return;

		default:
			msg(" (invalid)\n");
			return;
	}
}


/* geom_conic_to_poly */

int geom_conic_to_poly(struct geom_conic *conic, real *poly)
{
	GEOM_CONIC_ASSERT(conic);
	BASSERT(poly != NULL);

	switch (conic->type) {
		case GEOM_CONIC_PARABOLIC_ARC:
			return geom_parabolic_arc_to_poly(&conic->data.parabolic_arc, poly);

		case GEOM_CONIC_ARC:
			return geom_arc_to_poly(&conic->data.arc, poly);

		case GEOM_CONIC_ELLIPTIC_ARC:
			return geom_elliptic_arc_to_poly(&conic->data.elliptic_arc, poly);

		case GEOM_CONIC_HYPERBOLIC_ARC:
			return geom_hyperbolic_arc_to_poly(&conic->data.hyperbolic_arc, poly);

		default:
			BUG();
	}
}


/* geom_conic_to_bezier */

int geom_conic_to_bezier(struct geom_conic *conic, struct geom_bezier *bezier)
{
	GEOM_CONIC_ASSERT(conic);
	BASSERT(bezier != NULL);

	switch (conic->type) {
		case GEOM_CONIC_PARABOLIC_ARC:
			return geom_parabolic_arc_to_bezier(&conic->data.parabolic_arc, bezier);

		case GEOM_CONIC_ARC:
			return geom_arc_to_bezier(&conic->data.arc, bezier);

		case GEOM_CONIC_ELLIPTIC_ARC:
			return geom_elliptic_arc_to_bezier(&conic->data.elliptic_arc, bezier);

		case GEOM_CONIC_HYPERBOLIC_ARC:
			return geom_hyperbolic_arc_to_bezier(&conic->data.hyperbolic_arc, bezier);

		default:
			BUG();
	}
}


/* geom_conic_to_path */

int geom_conic_to_path(struct geom_conic *conic, struct geom_path *path)
{
	GEOM_CONIC_ASSERT(conic);
	BASSERT(path != NULL);

	switch (conic->type) {
		case GEOM_CONIC_PARABOLIC_ARC:
			return geom_parabolic_arc_to_path(&conic->data.parabolic_arc, path);

		case GEOM_CONIC_ARC:
			return geom_arc_expand_to_tail(&conic->data.arc, path);

		case GEOM_CONIC_ELLIPTIC_ARC:
			return geom_elliptic_arc_expand(&conic->data.elliptic_arc, 0, path);

		case GEOM_CONIC_HYPERBOLIC_ARC:
			return geom_hyperbolic_arc_to_path(&conic->data.hyperbolic_arc, path);

		default:
			BUG();
	}
}


/* geom_conic_from_poly */

int geom_conic_from_poly(struct geom_conic *conic, const real *poly)
{
	real a, c, d, e, f, x, y, sin1, cos1, sin2, cos2, sincos;
	struct geom_point center;
	real p, rotation, r, ar, br;

#ifdef GEOM_DEBUG
	BASSERT(poly != NULL);
	for (uns i = 0; i < 6; i++)
		BASSERT(isfinite(poly[i]));
	BASSERT(conic != NULL);
#endif

	/* rotate to clear X*Y coefficient */
	if (IS_ZERO(poly[1])) {
		rotation = 0;
		a = poly[0];
		c = poly[2];
		d = poly[3];
		e = poly[4];
		f = poly[5];
		sin1 = 0;
		cos1 = 1;
	}
	else {
		rotation = 0.5 * atan(poly[1] / (poly[0] - poly[2]));
		if (!isfinite(rotation))
			rotation = M_PI / 4;
		sin1 = sin(rotation);
		cos1 = cos(rotation);
		sin2 = SQR(sin1);
		cos2 = SQR(cos1);
		sincos = sin1 * cos1;
		a = poly[0] * cos2 + poly[1] * sincos + poly[2] * sin2;
		c = poly[0] * sin2 - poly[1] * sincos + poly[2] * cos2;
		d = poly[3] * cos1 + poly[4] * sin1;
		e = -poly[3] * sin1 + poly[4] * cos1;
		f = poly[5];
		if (!isfinite(a) || !isfinite(c) || !isfinite(d) || !isfinite(e))
			return GEOM_ERR_NUMERIC;
	}

	if (IS_ZERO(a))
		a = 0;
	if (IS_ZERO(c))
		c = 0;
	if (IS_ZERO(d))
		d = 0;
	if (IS_ZERO(e))
		e = 0;

	//DBG("angle=%f   coefs=%f %f %f %f %f %f\n", rotation, a, 0.0, c, d, e, f);

	/* Dx + Ey + F = 0  => no conic */
	if (a == 0 && c == 0)
		return GEOM_ERR_CONIC_DEGENERATED;

	/* Cy^2 + Dx + F = 0  => parabola with axis parallel to x */
	if (a == 0) {
		y = -0.5 * e / c;
		f -= c * SQR(y);
		x = -f / d;
		center.x = x * cos1 - y * sin1;
		center.y = x * sin1 + y * cos1;
		p = -0.25 * d / c;
		if (!isfinite(center.x) || !isfinite(center.y) || !isfinite(p))
			return GEOM_ERR_NUMERIC;
		if (p < 0) {
			rotation += M_PI;
			p = -p;
		}
		rotation -= M_PI / 2;

		conic->type = GEOM_CONIC_PARABOLIC_ARC;
		geom_parabolic_arc_set_unit(&conic->data.parabolic_arc, &center, rotation, p);
		geom_parabolic_arc_normalize(&conic->data.parabolic_arc);
		return 0;
	}

	/* Ax^2 + Ey + F = 0  => parabola with axis parallel to y */
	if (c == 0) {
		x = -0.5 * d / a;
		f -= a * SQR(x);
		y = -f / e;
		center.x = x * cos1 - y * sin1;
		center.y = x * sin1 + y * cos1;
		p = -0.25 * e / a;
		if (!isfinite(center.x) || !isfinite(center.y) || !isfinite(p))
			return GEOM_ERR_NUMERIC;
		if (p < 0) {
			rotation += M_PI;
			p = -p;
		}
		if (rotation < 0)
			rotation += 2 * M_PI;

		conic->type = GEOM_CONIC_PARABOLIC_ARC;
		geom_parabolic_arc_set_unit(&conic->data.parabolic_arc, &center, rotation, p);
		geom_parabolic_arc_normalize(&conic->data.parabolic_arc);
		return 0;
	}

	/* find center point */
	y = 0;
	if (d == 0)
		x = 0;
	else {
		x = -0.5 * d / a;
		f -= a * SQR(x);
		d = 0;
	}
	if (e == 0)
		center.y = 0;
	else {
		y = -0.5 * e / c;
		f -= c * SQR(y);
		e = 0;
	}
	center.x = x * cos1 - y * sin1;
	center.y = x * sin1 + y * cos1;
	if (!isfinite(center.x) || !isfinite(center.y))
		return GEOM_ERR_NUMERIC;

	/* Ax^2 + By^2 + F = 0  => circle */
	if (IS_ZERO(a - c)){
		r = sqrt(-f / a);
		if (!isfinite(r) || r <= 0)
			return GEOM_ERR_NUMERIC;

		conic->type = GEOM_CONIC_ARC;
		geom_arc_set_closed(&conic->data.arc, &center, r);
		geom_arc_normalize(&conic->data.arc);
		return 0;
	}

	/* compute ellipse/hyperbola radius */
	ar = sqrt(fabs(f / a));
	br = sqrt(fabs(f / c));
	if (!isfinite(ar) || !isfinite(br))
		return GEOM_ERR_NUMERIC;

	/* normalize to: Ax^2 + By^2 + F = 0, where A > 0 */
	if (a < 0) {
		a = -a;
		c = -c;
		f = -f;
	}

	/* A > 0, B > 0  => ellipse */
	if (c > 0) {
		if (f > 0)
			return GEOM_ERR_CONIC_DEGENERATED;
		conic->type = GEOM_CONIC_ELLIPTIC_ARC;
		if (ar >= br) {
			if (rotation < 0)
				rotation += M_PI;
			geom_elliptic_arc_set_closed(&conic->data.elliptic_arc, &center, rotation, ar, br);
		}
		else
			geom_elliptic_arc_set_closed(&conic->data.elliptic_arc, &center, rotation + M_PI / 2, br, ar);
		geom_elliptic_arc_normalize(&conic->data.elliptic_arc);
		return 0;
	}

	conic->type = GEOM_CONIC_HYPERBOLIC_ARC;

	/* A > 0, B < 0, F > 0  => hyperbola */
	if (f > 0)
		geom_hyperbolic_arc_set_unit(&conic->data.hyperbolic_arc, &center, rotation + M_PI / 2, br, ar);

	/* A > 0, B < 0, F < 0  => hyperbola */
	else {
		if (rotation < 0)
			rotation += M_PI;
		geom_hyperbolic_arc_set_unit(&conic->data.hyperbolic_arc, &center, rotation, ar, br);
	}

	geom_hyperbolic_arc_normalize(&conic->data.hyperbolic_arc);
	return 0;
}


/* geom_conic_from_5_points */

int geom_conic_from_5_points(struct geom_conic *conic, const struct geom_point *points)
{
	uns i, j;
	int dim;
	struct geom_matrix matrix, ker;
	double buf[GEOM_MATRIX_BUF_SIZE(5, 6)], buf2[GEOM_MATRIX_BUF_SIZE(6, 6)];
	real poly[6];

	BASSERT(points != NULL);
	BASSERT(conic != NULL);

	geom_matrix_init_buf(buf, ARRAY_SIZE(buf), 5, 6, &matrix);
	geom_matrix_init_buf(buf2, ARRAY_SIZE(buf2), 6, 6, &ker);

	for (i = 0; i < 5; i++) {
		GEOM_POINT_ASSERT(&points[i]);
		GEOM_MATRIX_DATA(&matrix, i, 0) = SQR(points[i].x);
		GEOM_MATRIX_DATA(&matrix, i, 1) = points[i].x * points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 2) = SQR(points[i].y);
		GEOM_MATRIX_DATA(&matrix, i, 3) = points[i].x;
		GEOM_MATRIX_DATA(&matrix, i, 4) = points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 5) = 1;
		for (j = 0; j < 3; j++)
			if (!isfinite(GEOM_MATRIX_DATA(&matrix, i, j)))
				return GEOM_ERR_NUMERIC;
	}

	dim = geom_matrix_kernel(&matrix, &matrix, &ker);
	if (dim < 0)
		return GEOM_ERR_NUMERIC;
	else if (dim != 1)
		return GEOM_ERR_CONIC_AMBIGUOUS;

	for (i = 0; i < 6; i++) {
		poly[i] = GEOM_MATRIX_DATA(&ker, i, 0);
		if (!isfinite(poly[i]))
			return GEOM_ERR_NUMERIC;
	}

	return geom_conic_from_poly(conic, poly);
}


/* geom_conic_from_rotation_and_4_points */

int geom_conic_from_rotation_and_4_points(struct geom_conic *conic, const real rotation, const struct geom_point *points)
{
	uns i, j;
	int dim;
	struct geom_matrix matrix, ker;
	double buf[GEOM_MATRIX_BUF_SIZE(5, 6)], buf2[GEOM_MATRIX_BUF_SIZE(6, 6)];
	real poly[6], angle, tan1;

	BASSERT(points != NULL);
	BASSERT(conic != NULL);

	geom_matrix_init_buf(buf, ARRAY_SIZE(buf), 5, 6, &matrix);
	geom_matrix_init_buf(buf2, ARRAY_SIZE(buf2), 6, 6, &ker);

	for (i = 0; i < 4; i++) {
		GEOM_POINT_ASSERT(&points[i]);
		GEOM_MATRIX_DATA(&matrix, i, 0) = SQR(points[i].x);
		GEOM_MATRIX_DATA(&matrix, i, 1) = points[i].x * points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 2) = SQR(points[i].y);
		GEOM_MATRIX_DATA(&matrix, i, 3) = points[i].x;
		GEOM_MATRIX_DATA(&matrix, i, 4) = points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 5) = 1;
	}
	angle = rotation - (M_PI / 2) * round(rotation / (M_PI / 2));
	tan1 = tan(2 * angle);
	GEOM_MATRIX_DATA(&matrix, 4, 0) = tan1;
	GEOM_MATRIX_DATA(&matrix, 4, 1) = -1;
	GEOM_MATRIX_DATA(&matrix, 4, 2) = -tan1;
	GEOM_MATRIX_DATA(&matrix, 4, 3) = 0;
	GEOM_MATRIX_DATA(&matrix, 4, 4) = 0;
	GEOM_MATRIX_DATA(&matrix, 4, 5) = 0;
	for (i = 0; i < 5; i++)
		for (j = 0; j < 6; j++)
			if (!isfinite(GEOM_MATRIX_DATA(&matrix, i, j)))
				return GEOM_ERR_NUMERIC;

	dim = geom_matrix_kernel(&matrix, &matrix, &ker);
	if (dim < 0)
		return GEOM_ERR_NUMERIC;
	else if (dim != 1)
		return GEOM_ERR_CONIC_AMBIGUOUS;

	for (i = 0; i < 6; i++) {
		poly[i] = GEOM_MATRIX_DATA(&ker, i, 0);
		if (!isfinite(poly[i]))
			return GEOM_ERR_NUMERIC;
	}

	return geom_conic_from_poly(conic, poly);
}


/* geom_conic_from_2_tangents_and_point */

int geom_conic_from_2_tangents_and_point(struct geom_conic *conic, const real *angles, const struct geom_point *points)
{
	uns i, j;
	int dim;
	struct geom_matrix matrix, ker;
	double buf[GEOM_MATRIX_BUF_SIZE(5, 6)], buf2[GEOM_MATRIX_BUF_SIZE(6, 6)];
	real poly[6], angle, cos1, sin1, x, y;

	BASSERT(points != NULL);
	BASSERT(conic != NULL);

	geom_matrix_init_buf(buf, ARRAY_SIZE(buf), 5, 6, &matrix);
	geom_matrix_init_buf(buf2, ARRAY_SIZE(buf2), 6, 6, &ker);

	for (i = 0; i < 3; i++) {
		GEOM_POINT_ASSERT(&points[i]);
		GEOM_MATRIX_DATA(&matrix, i, 0) = SQR(points[i].x);
		GEOM_MATRIX_DATA(&matrix, i, 1) = points[i].x * points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 2) = SQR(points[i].y);
		GEOM_MATRIX_DATA(&matrix, i, 3) = points[i].x;
		GEOM_MATRIX_DATA(&matrix, i, 4) = points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 5) = 1;
	}
	for (i = 3; i < 5; i++) {
		x = points[i - 3].x;
		y = points[i - 3].y;
		angle = angles[i - 3];
		BASSERT(isfinite(angle));
		cos1 = cos(angle);
		sin1 = sin(angle);
		GEOM_MATRIX_DATA(&matrix, i, 0) = 2 * x * sin1;
		GEOM_MATRIX_DATA(&matrix, i, 1) = (sin1 * y - cos1 * x);
		GEOM_MATRIX_DATA(&matrix, i, 2) = -2 * y * cos1;
		GEOM_MATRIX_DATA(&matrix, i, 3) = sin1;
		GEOM_MATRIX_DATA(&matrix, i, 4) = -cos1;
		GEOM_MATRIX_DATA(&matrix, i, 5) = 0;
	}
	for (i = 0; i < 5; i++)
		for (j = 0; j < 6; j++)
			if (!isfinite(GEOM_MATRIX_DATA(&matrix, i, j)))
				return GEOM_ERR_NUMERIC;
	dim = geom_matrix_kernel(&matrix, &matrix, &ker);
	if (dim < 0)
		return GEOM_ERR_NUMERIC;
	else if (dim != 1)
		return GEOM_ERR_CONIC_AMBIGUOUS;

	for (i = 0; i < 6; i++) {
		poly[i] = GEOM_MATRIX_DATA(&ker, i, 0);
		if (!isfinite(poly[i]))
			return GEOM_ERR_NUMERIC;
	}

	return geom_conic_from_poly(conic, poly);
}


/* geom_arc_from_3_points */

/*int geom_arc_from_3_points(const struct geom_point *points, struct geom_arc *arc)
{
	uns i;
	struct geom_matrix matrix;
	double b[3], poly[6], buf[GEOM_MATRIX_BUF_SIZE(3, 4)], x, y;
	struct geom_conic conic;

#ifdef DEBUG
	ASSERT(points != NULL);
	for (i = 0; i < 3; i++)
		GEOM_POINT_ASSERT(&points[i]);
	ASSERT(arc != NULL);
#endif

	geom_matrix_init_buf(buf, ARRAY_SIZE(buf), 3, 3, &matrix);

	for (i = 0; i < 3; i++) {
		x = points[i].x;
		y = points[i].y;
		GEOM_MATRIX_DATA(&matrix, i, 0) = x;
		GEOM_MATRIX_DATA(&matrix, i, 1) = y;
		GEOM_MATRIX_DATA(&matrix, i, 2) = 1;
		b[i] = -(SQR(x) + SQR(y));
		if (!isfinite(b[i]))
			return -1;
	}
	if (geom_matrix_solve(&matrix, b, &matrix, b, NULL) != 0)
		return -1;

	poly[0] = 1;
	poly[1] = 0;
	poly[2] = 1;
	poly[3] = b[0];
	poly[4] = b[1];
	poly[5] = b[2];

	if (geom_conic_from_poly(poly, &conic) < 0)
		return -1;

	*arc = conic.data.arc;
	return 0;
}*/


#endif
