#ifndef _VRR_GEOMLIB_GROUP_H_
#define _VRR_GEOMLIB_GROUP_H_

/** @file
 *
 * GEOMLIB - Groups of items.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


#include "geomlib/item.h"


/* Class geom_group declaration.*/

#define geom_group_VMT    geom_item_VMT
#define geom_group_INIT   geom_item_INIT \
	.destroy = (void (*)(void *)) &geom_group_destroy, \
	.subdump = (void (*)(const void *, uns)) &geom_group_subdump, \
	.copy = (void (*)(void *, void *)) &geom_item_copy__fail,
GEOM_CLASS_HEAD(group, item);




/**
 * Default group constructor.
 * Resulting group belongs to no group and contains no item.
 *
 * @ao  group   Structure to initialize.
 */
static inline void geom_group_create(struct geom_group *group)
{
	GEOM_INSTANCE_IS_ASSERT(group, group);
	geom_item_create(&group->item);
	cclist_init(&group->list);
}


/**
 * Default group destructor.
 *
 * @ai  group  Structure to destroy.
 */
static inline void geom_group_destroy(struct geom_group *group)
{
	GEOM_INSTANCE_IS_ASSERT(group, group);
	GEOM_BASSERT(cclist_empty(&group->list));
	geom_item_destroy(&group->item);
}


void geom_group_subdump(const struct geom_group *group, uns flags);


/*----------------------------------------------*
 *                                              *
 *             ITEM LIST ROUTINES               *
 *                                              *
 *----------------------------------------------*/


/**
 * Returns whether a given item belongs to a group.
 *
 * @ai  item  Item.
 *
 * @rv  nonzero  Item belongs to a group.
 * @rv  zero     Item has no owner.
 */
static inline int geom_item_has_owner(void *item)
{ return GEOM_INSTANCE_IS_ASSERTX(item, item)->owner != NULL; }


/**
 * Returns owner group of given item.
 *
 * @ai  item  Item.
 *
 * @rv  NULL       Item does not belong to any group.
 * @rv  otherwise  Owner group.
 */
static inline void *geom_item_owner(void *item)
{ return GEOM_INSTANCE_IS_ASSERTX(item, item)->owner; }


/**
 * Returns whether a given group is empty.
 *
 * @ai  group  Group.
 *
 * @rv  nonzero  Group is empty.
 * @rv  zero     Group is not empty.
 */
static inline int geom_group_empty(void *group)
{ return cclist_empty(&GEOM_INSTANCE_IS_ASSERTX(group, group)->list); }


/**
 * Returns first item in a given group.
 *
 * @ai  group  Group.
 *
 * @rv  NULL       Group is empty.
 * @rv  otherwise  First group item.
 */
static inline void *geom_group_head(void *group)
{
	return (GEOM_INSTANCE_IS_ASSERTX(group, group)->list.head.next == &((struct geom_group *)group)->list.head) ?
		NULL : SKIP_BACK(struct geom_item, node, ((struct geom_group *)group)->list.head.next);
}


/**
 * Returns last item in a given group.
 *
 * @ai  group  Group.
 *
 * @rv  NULL       Group is empty.
 * @rv  otherwise  Last group index.
 */
static inline void *geom_group_tail(void *group)
{
	return (GEOM_INSTANCE_IS_ASSERTX(group, group)->list.head.next == &((struct geom_group *)group)->list.head) ?
		NULL : SKIP_BACK(struct geom_item, node, ((struct geom_group *)group)->list.head.prev); }


/**
 * Returns previous item to a given item in its owner group.
 *
 * @ai  item  Item.
 *
 * @rv  NULL       Item is first item in owner group.
 * @rv  otherwise  Previous item.
 */
static inline void *geom_group_prev(void *item)
{
	void *ptr = cclist_prev(&((struct geom_item *)item)->owner->list, &GEOM_INSTANCE_IS_ASSERTX(item, item)->node);
	return ptr ? SKIP_BACK(struct geom_item, node, ptr) : NULL;
}


/**
 * Multiple geom_group_prev. Skiped items must exist.
 * (interface for futer balanced tree optimalizations)
 *
 * @ai  item   Item.
 * @ai  dist   Distance of wanted item (greater than zero).
 *
 * @rv  NULL       There are only dist-1 items to the left of given item.
 * @rv  otherwise  Item to the left in given distace.
 */
void *geom_group_prev_skip(void *item, uns dist);


/**
 * Returns next item to a given item in its owner group.
 *
 * @ai  item  Item.
 *
 * @rv  NULL       Item is last item in owner group.
 * @rv  otherwise  Next item.
 */
static inline void *geom_group_next(void *item)
{
	void *ptr = cclist_next(&((struct geom_item *)item)->owner->list, &GEOM_INSTANCE_IS_ASSERTX(item, item)->node);
	return ptr ? SKIP_BACK(struct geom_item, node, ptr) : NULL;
}


/**
 * Multiple geom_group_next. Skiped items must exist.
 * (interface for futer balanced tree optimalizations)
 *
 * @ai  item   Item.
 * @ai  dist   Distance of wanted item (greater than zero).
 *
 * @rv  NULL       There are only dist-1 items to the right of given item.
 * @rv  otherwise  Item to the right in given distace.
 */
void *geom_group_next_skip(void *item, uns count);


/**
 * Inserts item to head of given group.
 *
 * @ai  group  Group, where to insert.
 * @ai   item   Item to insert (with NULL owner).
 */
static inline void geom_group_add_head(void *group, void *item)
{
	GEOM_BASSERT(GEOM_INSTANCE_IS_ASSERTX(item, item)->owner == NULL);
	cclist_add_head(&GEOM_INSTANCE_IS_ASSERTX(group, group)->list, &((struct geom_item *)item)->node);
	((struct geom_item *)item)->owner = group;
}


/**
 * Inserts item to tail of given group.
 *
 * @aio  group  Group, where to insert.
 * @ai   item   Item to insert (with NULL owner).
 */
static inline void geom_group_add_tail(void *group, void *item)
{
	GEOM_BASSERT(GEOM_INSTANCE_IS_ASSERTX(item, item)->owner == NULL);
	cclist_add_tail(&GEOM_INSTANCE_IS_ASSERTX(group, group)->list, &((struct geom_item *)item)->node);
	((struct geom_item *)item)->owner = group;
}


/**
 * Inserts item before given item.
 *
 * @ai  what    Item to insert (with NULL owner).
 * @ai  before  Item, before which we want to insert (must be in a group).
 */
static inline void geom_group_insert_before(void *what, void *before)
{
	GEOM_BASSERT(GEOM_INSTANCE_IS_ASSERTX(item, what)->owner == NULL);
	GEOM_PASSERT(GEOM_INSTANCE_IS_ASSERTX(item, before)->owner);
	cclist_insert_before(&(((struct geom_item *)before)->owner)->list, &((struct geom_item *)what)->node, &((struct geom_item *)before)->node);
	((struct geom_item *)what)->owner = ((struct geom_item *)before)->owner;
}


/**
 * Inserts item after given item.
 *
 * @ai  what   Item to insert (with NULL owner).
 * @ai  after  Item, after which we want to insert (must be in a group).
 */
static inline void geom_group_insert_after(void *what, void *after)
{
	GEOM_BASSERT(GEOM_INSTANCE_IS_ASSERTX(item, what)->owner == NULL);
	GEOM_PASSERT(GEOM_INSTANCE_IS_ASSERTX(item, after)->owner);
	cclist_insert_after(&(((struct geom_item *)after)->owner)->list, &((struct geom_item *)what)->node, &((struct geom_item *)after)->node);
	((struct geom_item *)what)->owner = ((struct geom_item *)after)->owner;
}


/**
 * Removes item from its owner group.
 *
 * @ai  item  Item to remove (must be in a group).
 */
static inline void geom_group_remove(void *item)
{
	GEOM_BASSERT(GEOM_INSTANCE_IS_ASSERTX(item, item)->owner);
	cclist_remove(&(((struct geom_item *)item)->owner)->list, &((struct geom_item *)item)->node);
	((struct geom_item *)item)->owner = NULL;
}


/**
 * Removes first item from a given group.
 *
 * @ai  group  Nonempty group.
 */
static inline void geom_group_remove_head(void *group)
{ geom_group_remove(geom_group_head(group)); }


/**
 * Removes last item from a given group.
 *
 * @ai  group  Nonempty group.
 */
static inline void geom_group_remove_tail(void *group)
{ geom_group_remove(geom_group_tail(group)); }


/**
 * Removes all items from a given group.
 *
 * @ai  group  Group (possibly empty).
 */
void geom_group_remove_all(void *group);


/**
 * Removes all items from a given group and destroys them.
 *
 * @ai  group  Group (possibly empty).
 */
void geom_group_destroy_all(void *group);


/**
 * Removes all items from a given group, destroys them and free them with xfree().
 *
 * @ai  group  Group (possibly empty).
 */
void geom_group_free_all(void *group);


/**
 * Returns number of items in a given group.
 *
 * @ai  group  Group.
 *
 * @r  Number of items.
 */
static inline uns geom_group_count(void *group)
{ return ((struct geom_group *)group)->list.count; }


/**
 * Returns item at a given group position.
 *
 * @ai  group  Nonempty group.
 * @ai  index  Item index (counted from zero, must be lower than number of items in the group).
 *
 * @r  Item at given position.
 */
void *geom_group_index(void *group, uns index);


/*----------------------------------------------*
 *                                              *
 *            GROUPING TREE ROUTINES            *
 *                                              *
 *----------------------------------------------*/


/**
 * Returns topmost ancestor of given item.
 *
 * @ai  item  Item.
 *
 * @r  Topmost ancestor group or pointer to item, if no owner exist.
 */
void *geom_tree_root(void *item);


void *geom_subtree_prefix_next(void *subtree_root, void *item);


void *geom_tree_prefix_next(void *item);


/**
 * Sort items in the same tree in prefix order.
 *
 * @ai   count  Number of items to sort.
 * @aio  items  Array of item/group pointers.
 */
void geom_tree_prefix_sort(uns count, void **items);


/*----------------------------------------------*
 *                                              *
 *                   WALK MACROS                *
 *                                              *
 *----------------------------------------------*/


#define GEOM_GROUP_WALK(group, item) \
	for (item = geom_group_head(group); item; item = geom_group_next(item))

#define GEOM_GROUP_WALK_DELSAFE(group, item, temp) \
	for (item = geom_group_head(group); temp = geom_group_next(item), item; item = (void *)temp)

#define GEOM_SUBTREE_PREFIX_WALK(subtree_root, item) \
	for (item = (subtree_root); item; item = geom_subtree_prefix_next(subtree_root, item))


#endif /* _VRR_GEOMLIB_ITEM_H_ */


