/*
 * GEOMLIB - Matrix manipulation functions.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/numeric.h"
#include "geomlib/matrix.h"


/* geom_matrix_dump */

void geom_matrix_dump(const struct geom_matrix *a)
{
	uns i, j;

	msg("Dumping geom_matrix structure at %p", (void*)a);

	if (!a) {
		msg(".\n");
		return;
	}
	msg(", m=%d, n=%d, w=%d", a->m, a->n, a->w);

	if (a->n > GEOM_MATRIX_MAX_SIZE || a->m > GEOM_MATRIX_MAX_SIZE || a->n > ((uns)1 << a->w)) {
		msg(" (invalid size).\n");
		return;
	}

	for (i = 0; i < a->m; i++)
		for (j = 0; j < a->n; j++)
			if (!isfinite(GEOM_MATRIX_DATA(a, i, j))) {
				msg(" (invalid data)");
				goto DUMP_DATA;
			}

DUMP_DATA:
	msg(":\n");
	for (i = 0; i < a->m; i++) {
		for (j = 0; j < a->n; j++)
			msg(" %12f", GEOM_MATRIX_DATA(a, i, j));
		msg("\n");
	}
}


/* geom_matrix_init */

void geom_matrix_init(const uns m, const uns n, struct geom_matrix *a)
{
	BASSERT(m > 0);
	BASSERT(m <= GEOM_MATRIX_MAX_SIZE);
	BASSERT(n > 0);
	BASSERT(n <= GEOM_MATRIX_MAX_SIZE);
	BASSERT(a != NULL);

	a->m = m;
	a->n = n;
	for (a->w = 0; a->n > ((uns)1 << a->w); a->w++);
	a->data = xmalloc(((1 << a->w) * (m - 1) + n) * sizeof(double));
}


/* geom_matrix_destroy */

void geom_matrix_destroy(struct geom_matrix *a)
{
	GEOM_MATRIX_ASSERT(a);

	xfree(a->data);
}


/* geom_matrix_init_buf */

void geom_matrix_init_buf(double *buf, const uns buf_size UNUSED, const uns m, const uns n, struct geom_matrix *a)
{
	BASSERT(buf != NULL);
	BASSERT(m > 0);
	BASSERT(m <= GEOM_MATRIX_MAX_SIZE);
	BASSERT(n > 0);
	BASSERT(n <= GEOM_MATRIX_MAX_SIZE);
	BASSERT(a != NULL);

	a->m = m;
	a->n = n;
	for (a->w = 0; a->n > ((uns)1 << a->w); a->w++);
	a->data = buf;

	BASSERT(buf_size >= ((1 << a->w) * (m - 1) + n));
}


/* geom_matrix_init_submatrix */

void geom_matrix_init_submatrix(const struct geom_matrix *a, const uns i, const uns j, const uns m, const uns n, struct geom_matrix *r)
{
	GEOM_MATRIX_ASSERT(a);
	BASSERT(r != NULL);
	BASSERT(i < a->m);
	BASSERT(j < a->n);
	BASSERT(m <= a->m - i);
	BASSERT(n <= a->n - j);

	r->m = m;
	r->n = n;
	r->w = a->w;
	r->data = &GEOM_MATRIX_DATA(a, i, j);
}


/* geom_matrix_set_zero */

void geom_matrix_set_zero(struct geom_matrix *a)
{
	uns i, j;
	double *d;

	GEOM_MATRIX_ASSERT(a);

	d = a->data;
	for (i = 0; i < a->m; i++) {
		for (j = 0; j < a->n; j++)
			d[j] = 0;
		d += 1 << a->w;
	}
}


/* geom_matrix_set_identity */

void geom_matrix_set_identity(struct geom_matrix *a)
{
	uns i, j;
	double *d;

	GEOM_MATRIX_ASSERT(a);
	BASSERT(a->m == a->n);

	d = a->data;
	for (i = 0; i < a->m; i++) {
		for (j = 0; j < a->n; j++)
			d[j] = 0;
		d[i] = 1;
		d += 1 << a->w;
	}
}


/* geom_matrix_copy */

void geom_matrix_copy(const struct geom_matrix *a, struct geom_matrix *r)
{
	uns i, j;
	double *ap, *rp;

	GEOM_MATRIX_ASSERT(a);
	GEOM_MATRIX_DATA_ASSERT(a);
	GEOM_MATRIX_ASSERT(r);
	BASSERT(a->m == r->m);
	BASSERT(a->n == r->n);

	if (a == r)
		return;

	ap = a->data;
	rp = r->data;
	for (i = 0; i < a->m; i++) {
		for (j = 0; j < a->n; j++)
			rp[j] = ap[j];
		ap += 1 << a->w;
		rp += 1 << r->w;
	}
}


/* geom_matrix_mul */

int geom_matrix_mul(const struct geom_matrix *a, const struct geom_matrix *b, struct geom_matrix *r)
{
	uns i, j, k;
	double *ap, *bp, *rp, x;

	GEOM_MATRIX_ASSERT(a);
	GEOM_MATRIX_DATA_ASSERT(a);
	GEOM_MATRIX_ASSERT(b);
	GEOM_MATRIX_DATA_ASSERT(b);
	GEOM_MATRIX_ASSERT(r);
	BASSERT(a->n == b->m);
	BASSERT(r->m == a->m);
	BASSERT(r->n == b->n);
	BASSERT(r != a);
	BASSERT(r != b);

	ap = a->data;
	rp = r->data;
	for (i = 0; i < r->m; i++) {
		for (j = 0; j < r->n; j++) {
			bp = b->data;
			x = 0;
			for (k = 0; k < a->m; k++) {
				x += ap[k] * bp[j];
				bp += 1 << b->w;
			}
			if (!isfinite(x))
				return GEOM_ERR_NUMERIC;
			rp[j] = x;
		}
		ap += 1 << a->w;
		rp += 1 << r->w;
	}

	return 0;
}


/* geom_matrix_plu */

int geom_matrix_plu(const struct geom_matrix *a, uns *p, uns *c, struct geom_matrix *l, struct geom_matrix *u)
{
	uns i, j, k, q;
	double x, y;

#ifdef GEOM_DEBUG
	GEOM_MATRIX_ASSERT(a);
	GEOM_MATRIX_DATA_ASSERT(a);
	if (l && l != u) {
		GEOM_MATRIX_ASSERT(l);
		BASSERT(l->m == a->m);
		BASSERT(l->n == a->m);
	}
	GEOM_MATRIX_ASSERT(u);
	BASSERT(u->m == a->m);
	BASSERT(u->n == a->n);
#endif

	if (u != a)
		geom_matrix_copy(a, u);

	/* current column */
	q = 0;

	/* for each row */
	for (j = 0; j < u->m; j++) {

		/* skip zero columns and select row with max absolute value (pivot) */
		while (1) {
			if (q == u->n)
				goto FINISH;

			k = j;
			x = fabs(GEOM_MATRIX_DATA(u, k, q));
			if (IS_ZERO(x))
				GEOM_MATRIX_DATA(u, k, q) = x = 0;
			for (i = j; i < u->m; i++) {
				y = fabs(GEOM_MATRIX_DATA(u, i, q));
				if (IS_ZERO(y))
					GEOM_MATRIX_DATA(u, i, q) = 0;
				else if (x < y) {
					k = i;
					x = y;
				}
			}
			if (x != 0)
				break;
			q++;
		}

		/* store permutation */
		if (p)
			p[j] = k;

		/* store diangonal */
		if (c)
			c[j] = q;

		/* swap current row and pivot row */
		if (j != k) {
			if (l)
				for (i = 0; i < j; i++) {
					y = GEOM_MATRIX_DATA(l, j, i);
					GEOM_MATRIX_DATA(l, j, i) = GEOM_MATRIX_DATA(l, k, i);
					GEOM_MATRIX_DATA(l, k, i) = y;
				}
			for (i = q; i < u->n; i++) {
				y = GEOM_MATRIX_DATA(u, j, i);
				GEOM_MATRIX_DATA(u, j, i) = GEOM_MATRIX_DATA(u, k, i);
				GEOM_MATRIX_DATA(u, k, i) = y;
			}
		}

		/* compute column in matrix L and update trailing matrix */
		if (l && l != u) {
			for (i = 0; i < j; i++)
				GEOM_MATRIX_DATA(l, i, j) = 0;
			GEOM_MATRIX_DATA(l, j, j) = 1;
		}
		x = 1 / GEOM_MATRIX_DATA(u, j, q);
		for (k = j + 1; k < u->m; k++) {
			y = GEOM_MATRIX_DATA(u, k, q) * x;
			if (!isfinite(y))
				return GEOM_ERR_NUMERIC;
			if (l)
				GEOM_MATRIX_DATA(l, k, j) = y;
			for (i = q + 1; i < u->n; i++) {
				GEOM_MATRIX_DATA(u, k, i) -= y * GEOM_MATRIX_DATA(u, j, i);
				if (!isfinite(GEOM_MATRIX_DATA(u, k, i)))
					return GEOM_ERR_NUMERIC;
			}
			if (l != u)
				GEOM_MATRIX_DATA(u, k, q) = 0;
		}

		q++;
	}

FINISH:
	if (l && l != u)
		for (k = j; k < u->m; k++) {
			for (i = 0; i < u->m; i++)
				GEOM_MATRIX_DATA(l, i, k) = 0;
			GEOM_MATRIX_DATA(l, k, k) = 1;
		}

	return j;
}


/* geom_matrix_rank */

int geom_matrix_rank(const struct geom_matrix *a, struct geom_matrix *temp)
{
	return geom_matrix_plu(a, NULL, NULL, NULL, temp);
}


/* geom_matrix_determinant */

double geom_matrix_determinant(const struct geom_matrix *a, struct geom_matrix *temp)
{
	int rank;
	uns i, p[GEOM_MATRIX_MAX_SIZE];
	double result, *tp;

	GEOM_MATRIX_ASSERT(a);
	BASSERT(a->m == a->n);

	rank = geom_matrix_plu(a, p, NULL, NULL, temp);
	if (rank < 0)
		return NAN;
	if ((uns)rank < a->n)
		return 0;
	result = 1;
	tp = temp->data;
	for (i = 0; i < a->n; i++) {
		if (p[i] != i)
			result = -result;
		result *= *tp;
		tp += 1 + (1 << temp->w);
	}
	return isfinite(result) ? result : NAN;
}


/* geom_matrix_plu_solve */

static int geom_matrix_plu_solve(const uns rank, const uns *p, const uns *c, const struct geom_matrix *l, const struct geom_matrix *u, const double *b, double *x)
{
	uns i, j;
	double y, v[GEOM_MATRIX_MAX_SIZE];

#ifdef GEOM_DEBUG
	BASSERT(p != NULL);
	BASSERT(c != NULL);
	GEOM_MATRIX_ASSERT(u);
	GEOM_MATRIX_DATA_ASSERT(u);
	if (l != u) {
		GEOM_MATRIX_ASSERT(l);
		GEOM_MATRIX_DATA_ASSERT(l);
		BASSERT(l->m == u->m);
		BASSERT(l->n == u->n);
	}
	BASSERT(b != NULL);
	BASSERT(x != NULL);
#endif

	/* permute vector b */
	for (i = 0; i < u->m; i++)
		v[i] = b[i];
	for (i = 0; i < rank; i++) {
		y = v[i];
		v[i] = v[p[i]];
		v[p[i]] = y;
	}

	/* forward substitution (solve L * z = b) */
	for (j = 0; j < rank; j++) {
		for (i = 0; i < j; i++)
			v[j] -= GEOM_MATRIX_DATA(l, j, i) * v[i];
		if (!isfinite(v[j]))
			return GEOM_ERR_NUMERIC;
	}
	for (; j < u->m; j++) {
		for (i = 0; i < rank; i++)
			v[j] -= GEOM_MATRIX_DATA(l, j, i) * v[i];
		if (!isfinite(v[j]))
			return GEOM_ERR_NUMERIC;
		if (IS_ZERO(v[j]))
			return GEOM_ERR_MATRIX_NO_SOLUTION;
	}

	/* backward substitution (solve U * x = z) */
	for (i = 0; i < u->n; i++)
		x[i] = 0;
	i = rank;
	while (i--) {
		x[c[i]] = v[i];
		for (j = i + 1; j < rank; j++)
			x[c[i]] -= GEOM_MATRIX_DATA(u, i, c[j]) * x[c[j]];
		x[c[i]] /= GEOM_MATRIX_DATA(u, i, c[i]);
		if (!isfinite(x[c[i]]))
			return GEOM_ERR_NUMERIC;
	}

	return u->n - rank;
}


/* geom_matrix_plu_solve_systems */

static int geom_matrix_plu_solve_systems(const uns rank, const uns *p, const uns *c, const struct geom_matrix *l, const struct geom_matrix *u, const struct geom_matrix *b, struct geom_matrix *x)
{
	uns i, j;
	int dim;
	double v[GEOM_MATRIX_MAX_SIZE];

#ifdef GEOM_DEBUG
	BASSERT(p != NULL);
	BASSERT(c != NULL);
	GEOM_MATRIX_ASSERT(u);
	GEOM_MATRIX_DATA_ASSERT(u);
	if (l != u) {
		GEOM_MATRIX_ASSERT(l);
		GEOM_MATRIX_DATA_ASSERT(l);
		BASSERT(l->n == u->n);
		BASSERT(l->m == u->m);
	}
	GEOM_MATRIX_ASSERT(b);
	GEOM_MATRIX_DATA_ASSERT(b);
	BASSERT(b->m == u->m);
	GEOM_MATRIX_ASSERT(x);
	BASSERT(x->m == u->n);
	BASSERT(x->n == b->n);
#endif

	for (i = 0; i < b->n; i++) {
		for (j = 0; j < u->m; j++)
			v[j] = GEOM_MATRIX_DATA(b, j, i);
		dim = geom_matrix_plu_solve(rank, p, c, l, u, v, v);
		if (dim < 0)
			return dim;
		for (j = 0; j < u->n; j++)
			GEOM_MATRIX_DATA(x, j, i) = v[j];
	}

	return u->n - rank;
}

/* geom_matrix_plu_ref */

static int geom_matrix_plu_ref(const uns rank, uns *c, const struct geom_matrix *u, struct geom_matrix *r)
{
	uns i, j, k, q;

	j = 0; /* not necessary - warning elimination */

	if (r != u)
		geom_matrix_copy(u, r);

	for (k = 0; k < rank; k++) {
		q = c[k];
		for (i = q + 1; i < u->n; i++) {
			GEOM_MATRIX_DATA(r, k, i) /= GEOM_MATRIX_DATA(r, k, q);
			if (!isfinite(GEOM_MATRIX_DATA(r, k, i)))
				return GEOM_ERR_NUMERIC;
		}
		GEOM_MATRIX_DATA(r, k, q) = 1;
		for (i = 0; i < q; i++)
			GEOM_MATRIX_DATA(r, k, i) = 0;
		for (j = 0; j < k; j++) {
			for (i = q + 1; i < r->n; i++) {
				GEOM_MATRIX_DATA(r, j, i) -= GEOM_MATRIX_DATA(r, j, q) * GEOM_MATRIX_DATA(r, k, i);
				if (!isfinite(GEOM_MATRIX_DATA(r, j, i)))
					return GEOM_ERR_NUMERIC;
			}
			GEOM_MATRIX_DATA(r, j, q) = 0;
		}
	}
	for (; k < r->m; k++)
		for (i = 0; i < r->n; i++)
			GEOM_MATRIX_DATA(r, j, i) = 0;

	return 0;
}


/* geom_matrix_ref_kernel */

static uns geom_matrix_ref_kernel(const uns rank, const uns *c, const struct geom_matrix *r, struct geom_matrix *ker)
{
	uns i, j, k, q, dim;

	BASSERT(c != NULL);
	GEOM_MATRIX_ASSERT(r);
	GEOM_MATRIX_DATA_ASSERT(r);
	BASSERT(rank <= MIN(r->n, r->m));
	GEOM_MATRIX_ASSERT(ker);
	BASSERT(ker->m == r->n);
	BASSERT(ker->n >= r->n - rank);

	/* nonpivot columns provide a basis for the nullspace */
	dim = r->n - rank;
	k = 0;
	q = 0;
	for (i = 0; i < dim; i++) {
		while (c[k] == q) {
			k++;
			q++;
		}
		for (j = 0; j < (uns)rank; j++)
			GEOM_MATRIX_DATA(ker, j, i) = -GEOM_MATRIX_DATA(r, j, q);
		for (; j < ker->m; j++)
			GEOM_MATRIX_DATA(ker, j, i) = 0;
		GEOM_MATRIX_DATA(ker, rank + i, i) = 1;
		q++;
	}
	for (; i < ker->n; i++)
		for (j = 0; j < ker->m; j++)
			GEOM_MATRIX_DATA(ker, j, i) = 0;

	return dim;
}


/* geom_matrix_solve */

int geom_matrix_solve(const struct geom_matrix *a, const double *b, struct geom_matrix *temp, double *x, struct geom_matrix *ker)
{
	int rank, dim;
	uns p[GEOM_MATRIX_MAX_SIZE], c[GEOM_MATRIX_MAX_SIZE];

	rank = geom_matrix_plu(a, p, c, temp, temp);
	if (rank < 0)
		return rank;
	dim = geom_matrix_plu_solve((uns)rank, p, c, temp, temp, b, x);
	if (ker) {
		if (dim < 0)
			return dim;
		if (geom_matrix_plu_ref((uns)rank, c, temp, temp) < 0)
			return GEOM_ERR_NUMERIC;
		return geom_matrix_ref_kernel((uns)rank, c, temp, ker);
	}
	return dim;
}


/* geom_matrix_solve_systems */

int geom_matrix_solve_systems(const struct geom_matrix *a, const struct geom_matrix *b, struct geom_matrix *temp, struct geom_matrix *x, struct geom_matrix *ker)
{
	int rank, dim;
	uns p[GEOM_MATRIX_MAX_SIZE], c[GEOM_MATRIX_MAX_SIZE];

	rank = geom_matrix_plu(a, p, c, temp, temp);
	if (rank < 0)
		return rank;
	dim = geom_matrix_plu_solve_systems((uns)rank, p, c, temp, temp, b, x);
	if (ker) {
		if (dim < 0)
			return dim;
		if (geom_matrix_plu_ref((uns)rank, c, temp, temp) < 0)
			return GEOM_ERR_NUMERIC;
		return geom_matrix_ref_kernel((uns)rank, c, temp, ker);
	}
	return dim;
}


/* geom_matrix_kernel */

int geom_matrix_kernel(const struct geom_matrix *a, struct geom_matrix *temp, struct geom_matrix *ker)
{
	int rank;
	uns c[GEOM_MATRIX_MAX_SIZE];

	GEOM_MATRIX_ASSERT(a);
	GEOM_MATRIX_DATA_ASSERT(a);
	GEOM_MATRIX_ASSERT(temp);
	BASSERT(temp->n == a->n);
	BASSERT(temp->m == a->m);
	GEOM_MATRIX_ASSERT(ker);
	BASSERT(ker->m == a->n);

	rank = geom_matrix_plu(a, NULL, c, NULL, temp);
	if (rank < 0)
		return rank;
	if (geom_matrix_plu_ref((uns)rank, c, temp, temp) < 0)
		return GEOM_ERR_NUMERIC;
	return geom_matrix_ref_kernel((uns)rank, c, temp, ker);
}


/* geom_matrix_inverse */

int geom_matrix_inverse(const struct geom_matrix *a, struct geom_matrix *temp, struct geom_matrix *inv)
{
	int rank;
	uns i, j, k, n, p[GEOM_MATRIX_MAX_SIZE];
	double x;

	GEOM_MATRIX_ASSERT(a);
	GEOM_MATRIX_DATA_ASSERT(a);
	BASSERT(a->m == a->n);
	GEOM_MATRIX_ASSERT(inv);
	BASSERT(inv->n == a->n);
	BASSERT(inv->m == a->n);

	/* TODO: zrychlit, osetrit numericke chyby */

	n = a->n;

	/* factorization: A = P * L * U */
	rank = geom_matrix_plu(a, p, NULL, temp, temp);
	if (rank < 0)
		return rank;

	/* test for singularity */
	if ((uns)rank < n)
		return GEOM_ERR_MATRIX_SINGULAR;

	/* solve L * B = I, overwrite L data with B (lower triangular matrix with 1's on diagonal) */
	for (i = 0; i < n; i++) {
		for (j = i + 1; j < n; j++) {
			x = -GEOM_MATRIX_DATA(temp, j, i);
			for (k = i + 1; k < j; k++)
				x -= GEOM_MATRIX_DATA(temp, j, k) * GEOM_MATRIX_DATA(temp, k, i);
			GEOM_MATRIX_DATA(temp, j, i) = x;
		}
	}

	/* solve U * C = B, where P * C = inv(A) */
	i = n;
	for (i = 0; i < n; i++) {
		j = n;
		while (j--) {
			if (j < i)
				x = 0;
			else if (j == i)
				x = 1;
			else
				x = GEOM_MATRIX_DATA(temp, j, i);
			for (k = j + 1; k < n; k++)
				x -= GEOM_MATRIX_DATA(inv, k, i) * GEOM_MATRIX_DATA(temp, j, k);
			GEOM_MATRIX_DATA(inv, j, i) = x / GEOM_MATRIX_DATA(temp, j, j);
		}
	}

	/* inv(A) = P * C */
	n = i;
	while (i--)
		if (i != p[i])
			for (j = 0; j < n; j++) {
				x = GEOM_MATRIX_DATA(inv, j, i);
				GEOM_MATRIX_DATA(inv, j, i) = GEOM_MATRIX_DATA(inv, j, p[i]);
				GEOM_MATRIX_DATA(inv, j, p[i]) = x;
			}

	return 0;
}


