#ifndef _VRR_GEOMLIB_NUMERIC_H_
#define _VRR_GEOMLIB_NUMERIC_H_

/** @file
 *
 * GEOMLIB - Auxilliary numeric routines.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "geomlib/base.h"

#ifndef M_PI
#define M_PI (3.14159265358979323846)
#endif

#define SQR(x) ((x)*(x))

#define IS_EQUAL(x, y) ((x) >= (y) - EQ_EPSILON && (x) <= (y) + EQ_EPSILON)
#define IS_ZERO(x) IS_EQUAL((x), 0)
#define IS_EQUAL_XY(p1, p2) (IS_EQUAL((p1).x, (p2).x) && IS_EQUAL((p1).y, (p2).y))

#define IS_EQUAL_L(x, y) ((x) >= (y) - LARGE_EPSILON && (x) <= (y) + LARGE_EPSILON)
#define IS_ZERO_L(x) IS_EQUAL_L((x), 0)
#define IS_EQUAL_XY_L(p1, p2) (IS_EQUAL_L((p1).x, (p2).x) && IS_EQUAL_L((p1).y, (p2).y))

#define GEOM_MAX_DEGREE 35


/* Constants and some macros for real type */

#define REAL_INF INFINITY
#define DBL_INF INFINITY

#ifdef FLOAT_REAL

#define REAL_SWITCH(double, float) float
#define REAL_SUFFIX(name) name##f

#define REAL_DIG FLG_DIG
#define REAL_MAX FLT_MAX
#define REAL_EPSILON FLT_EPSILON
#define REAL_MIN FLT_MIN
#define EQ_EPSILON (FLT_EPSILON * 10)
#define LARGE_EPSILON (EQ_EPSILON * 100)

#else /* FLOAT_REAL */

#define REAL_SWITCH(double, float) double
#define REAL_SUFFIX(name) name

#define REAL_DIG DBL_DIG
#define REAL_MAX DBL_MAX
#define REAL_EPSILON DBL_EPSILON
#define REAL_MIN DBL_MIN
#define EQ_EPSILON (DBL_EPSILON * 1000)
#define LARGE_EPSILON (EQ_EPSILON * 100)

#endif /* FLOAT_REAL */

/* Constants */
#define R_PI ((real)M_PI)
#define R_2PI ((real)(2 * M_PI))
#define R_PI2 ((real)(0.5 * M_PI))

/* Real type equivalents for C99 math functions */
static inline real fminr(real x, real y) { return REAL_SUFFIX(fmin)(x, y); }
static inline real fmaxr(real x, real y) { return REAL_SUFFIX(fmax)(x, y); }
static inline real fabsr(real x) { return REAL_SUFFIX(fabs)(x); }
static inline real cosr(real x) { return REAL_SUFFIX(cos)(x); }
static inline real sinr(real x) { return REAL_SUFFIX(sin)(x); }
static inline real tanr(real x) { return REAL_SUFFIX(tan)(x); }
static inline real acosr(real x) { return REAL_SUFFIX(acos)(x); }
static inline real asinr(real x) { return REAL_SUFFIX(asin)(x); }
static inline real atanr(real x) { return REAL_SUFFIX(atan)(x); }
static inline real atan2r(real x, real y) { return REAL_SUFFIX(atan2)(x, y); }
static inline real hypotr(real x, real y) { return REAL_SUFFIX(hypot)(x, y); }
static inline real sqrtr(real x) { return REAL_SUFFIX(sqrt)(x); }
static inline real fmodr(real x, real y) { return REAL_SUFFIX(fmod)(x, y); }
static inline real nextafterr(real x, real y) { return REAL_SUFFIX(nextafter)(x, y); }
static inline real fmar(real x, real y, real z) { return REAL_SUFFIX(fma)(x, y, z); }
static inline real floorr(real x) { return REAL_SUFFIX(floor)(x); }
static inline real roundr(real x) { return REAL_SUFFIX(round)(x); }

/* Aliases with automatic conversion to real type */
static inline int risfinite(real x) { return isfinite(x); }
static inline int risinf(real x) { return isinf(x); }
static inline int risnan(real x) { return isnan(x); }

/* New floating point functions */

static inline double nanset(double x, double y) { return likely(!isnan(x)) ? x : y; }
static inline real nansetr(real x, real y) { return likely(!isnan(x)) ? x : y; }

static inline double nfinset(double x, double y) { return likely(isfinite(x)) ? x : y; }
static inline real nfinsetr(real x, real y) { return likely(isfinite(x)) ? x : y; }


/**
 * Fits number to given range.
 *
 * @ai  x    Number to fit.
 * @ai  min  Left interval limit (NAN is the same as -INFINITY).
 * @ai  max  Right interval limit (NAN is the same as +INFINITY).
 *
 * @rv  min  x is lower than min.
 * @rv  max  x is greater than max.
 * @rv  NAN  x is NAN.
 * @rv  x    Otherwise.
 */
#define MINMAX(x,min,max) (unlikely((x)<(min))?(min):(unlikely((x)>(max))?(max):(x)))
static inline double fminmax(double x, double min, double max) { return MINMAX(x, min, max); }
static inline real fminmaxr(real x, real min, real max) { return MINMAX(x, min, max); }


/**
 * Fits number to given range.
 *
 * @ai  x    Number to fit.
 * @ai  a    First interval limit (NAN is ignored).
 * @ai  b    Second interval limit (NAN is ignored).
 *
 * @rv  a    x is outside the range.
 * @rv  b    x is outside the range.
 * @rv  NAN  x is NAN.
 * @rv  x    Otherwise.
 */
#define RNG(x,a,b) (!((a)>(b))?MINMAX(x,a,b):MINMAX(x,b,a))
static inline double frng(double x, double a, double b) { return RNG(x, a, b); }
static inline real frngr(real x, real a, real b) { return RNG(x, a, b); }


/**
 * Converts infinite number to finite with maximum possible absolute value.
 *
 * @ai  x  Number to convert.
 *
 * @rv  x     x is finite or NAN.
 * @rv  +MAX  x is possitive infinity.
 * @rv  -MAX  x is negative infinity.
 */
static inline double inf2fin(double x) { int i = isinf(x); return likely(!i) ? x : (i > 0 ? DBL_MAX : -DBL_MAX); }
static inline real inf2finr(real x) { int i = risinf(x); return likely(!i) ? x : (i > 0 ? REAL_MAX : -REAL_MAX); }


/**
 * Computes floating point remainder of x / y so that z = x - n * y.
 *
 * @ai  x  Any number.
 * @ai  y  Any number.
 *
 * @rv  NAN   x or y is NAN.
 * @rv  zero  y is zero.
 * @rv  z     z is nonnegative and with magnitute less than magnitude of y.
 */
static inline double fmodfloor(double x, double y)
{ if (unlikely(y == 0)) return 0; else if (x >= 0) return fmod(x, y);
	else { double a = fabs(y), r = fmod(x, y) + a;
		return likely(r < a) ? r : nextafter(r, 0); }}

static inline real fmodfloorr(real x, real y)
{ if (unlikely(y == 0)) return 0; else if (x >= 0) return fmodr(x, y);
	else { real a = fabsr(y), r = fmodr(x, y) + a;
		return likely(r < a) ? r : nextafterr(r, 0); }}


/**
 * Computes z = (x + y) / 2.
 *
 * @ai  x  Any number.
 * @ai  y  Any number.
 *
 * @rv  NAN        x and y are infinities with different signs or at least one is NAN.
 * @rv  +INFINITY  x and/or y is possitive infinity.
 * @rv  -INFINITY  x and/or y is negative infinity.
 * @rv  z          Result is in interval [x, y] or [y, x] and near (x + y) / 2.
 */
static inline double fmidrng(double x, double y) { return frng(0.5 * (x + y), x, y); }
static inline real fmidrngr(real x, real y) { return frngr(0.5 * (x + y), x, y); }


/**
 * Computes z = (x + y) / 2.
 *
 * @ai  x  Any number.
 * @ai  y  Any number.
 *
 * @rv  NAN        x and y are infinities with different signs or at least one is NAN.
 * @rv  z          Result is finite number near (x + y) / 2.
 */
static inline double fmidfin(double x, double y) { return inf2fin(0.5 * (x + y)); }
static inline real fmidfinr(real x, real y) { return inf2finr(0.5 * (x + y)); }


/**
 * Computes z = x + t * (y - x).
 *
 * @ai  x  Any number.
 * @ai  y  Any number.
 * @ai  t  Coefficient.
 *
 * @rv  NAN        x and y are infinities with different signs or at least one of 3 numbers is NAN.
 * @rv  +INFINITY  x and/or y is possitive infinity.
 * @rv  -INFINITY  x and/or y is negative infinity.
 * @rv  z          Result is in interval [x, y] or [y, x] and near x + t * (y - x).
 */
static inline double fcoefrng(double x, double y, double t) { real s = y - x, r = fma(t, s, x);
	return likely(!isnan(r)) ? frng(r, x, y) : (isnan(s) ? s : (!isfinite(y) ? y : x)); }
static inline real fcoefrngr(real x, real y, real t) { real s = y - x, r = fmar(t, s, x);
	return likely(!isnan(r)) ? frngr(r, x, y) : (isnan(s) ? s : (!isfinite(y) ? y : x)); }


/**
 * Computes z = x + t * (y - x).
 *
 * @ai  x  Any number.
 * @ai  y  Any number.
 * @ai  t  Coefficient.
 *
 * @rv  NAN        x and y are infinities with different signs or at least one of 3 numbers is NAN.
 * @rv  z          Result is finite number near x + t * (y - x).
 */
static inline double fcoeffin(double x, double y, double t) { real s = y - x, r = fma(t, s, x);
	return inf2fin(likely(!isnan(r)) ? r : (isnan(s) ? s : (!isfinite(y) ? y : x))); }
static inline real fcoeffinr(real x, real y, real t) { real s = y - x, r = fmar(t, s, x);
	return inf2finr(likely(!isnan(r)) ? r : (isnan(s) ? s : (!isfinite(y) ? y : x))); }


#define RSWAP(x, y) do{real _rswap = (x); x = (y); y = _rswap;}while(0)



#ifdef GEOM_DEBUG
#define GEOM_REALS_ASSERT(n, v) do{ GEOM_PASSERT(v); for(uns i=0;i<n;i++) GEOM_ASSERT(r,(v)[i]); }while(0)
#else
#define GEOM_REALS_ASSERT(n, v) do{}while(0)
#endif

int geom_reals_valid(uns n, real reals[]);
void geom_reals_msg(uns n, real reals[]);

#ifdef GEOM_DEBUG
#define GEOM_REALS_SORTED_ASSERT(n, v) do{ GEOM_REALS_ASSERT(n, v); \
	for(uns i=1;i<n;i++) GEOM_BASSERT((v)[i]>=(v)[i-1]); }while(0)
#else
#define GEOM_REALS_SORTED_ASSERT(n, v) do{}while(0)
#endif






int geom_factorial(int n);

int geom_binomial(int n, int k);

int geom_doubles_sort(int n, double seq[]);

int geom_reals_sort(int n, real seq[]);

int geom_doubles_merge(int n1, double seq1[], int n2, double seq2[],
  double result[]);

int geom_permutation_generate_init(int n, int p[], int x[]);

int geom_permutation_generate(int n, int p[], int x[]);

int geom_permutation_sign(int n, int p[]);

#endif
