/*
 * GEOMLIB - Line patterns
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */
#define GEOM_SOURCE
#define GEOM_DEBUG_FUNC 1000
#define GEOM_DEBUG_RET 1000
#define GEOM_DEBUG_TRACE 1000
#include "geomlib/bezier.h"
#include "geomlib/pattern.h"


#define GEOM_PATTERN_PERIOD_MAX 10000


/* geom_item_pattern_append 
 *
 * FIXME:
 * - dodefinovat geom_item_part (podobne jako geom_bezier_part)
 * - osetrit kratke/dlouhe krivky
 * - osetrit cyklicke krivky
 * - podpora nenatahujicich se vzoru (napr. vyznaceni stupnice)
 * - podpora specialnich objektu ve vzoru (kolma carka, puntik, ...)
 * - predelat na pathset s rychlym alokatorem (stejne jako fpath)
 * - cachovani
 * - rychlejsi vyhodnocovani bodu/derivaci u dlouhych cest (i s planovanym vyvazenym stromem je vyhledani TIME O(log N))
 * - moznost zarovnani period v ostrych zlomech
 */

static int geom_item_pattern_item_append(void *curve, real left_time, real right_time, struct geom_pattern_item *item, struct geom_path *result)
{
	FUNC;
	if (item->flags & GEOM_PATTERN_LINE) {
		struct geom_point left_point, right_point;
		/*struct geom_vector left_deriv, right_deriv; */
		struct geom_bezier *bezier;
		
		
		/* FIXME: hack dokud nenapisu geom_item_part */
		
		ETEST_RET(geom_item_point_at_time(curve, left_time, &left_point));
		ETEST_RET(geom_item_point_at_time(curve, right_time, &right_point));
		/*ETEST_RET(geom_item_derivation_at_time(curve, left_time, &right_deriv));
		ETEST_RET(geom_item_derivation_at_time(curve, right_time, &right_deriv));*/

		GEOM_INSTANCE_NEW_CREATE2(bezier, bezier, geom_bezier_set_linear_nonrational, &left_point, &right_point);
		geom_group_add_tail(result, bezier);
	}
	ERET(0);
}

int geom_item_pattern_append(void *curve, struct geom_pattern *pattern, struct geom_path *result)
{
	uns i, j, k, index;
	real alength, r1, r2;
	uns start_count, period_count, end_count;
	struct geom_pattern_item *start_items, *period_items, *end_items;
	real start_len, period_len, end_len;
	real start_stretch, period_stretch, end_stretch, stretch_coef;
	uns times_count;
	real *atimes, *times;

	FUNC;
	ITEM_ASSERT(item, curve);
	PASSERT(pattern);
	INSTANCE_IS_ASSERT(path, result);

	RTEST_NUM(alength = geom_item_alength(curve));

	/* Select used blocks */
	start_count = pattern->start_count;
	start_items = pattern->start_items;
	period_count = pattern->period_count;
	period_items = pattern->period_items;
	end_count = pattern->end_count;
	end_items = pattern->end_items;

	/* Prepare blocks */
	start_len = start_stretch = 0;
	for (i = 0; i < start_count; i++) {
		start_len += start_items[i].alength;
		start_stretch += start_items[i].alength * start_items[i].stretch;
	}
	period_len = period_stretch = 0;
	for (i = 0; i < period_count; i++) {
		period_len += period_items[i].alength;
		period_stretch += period_items[i].alength * period_items[i].stretch;
	}
	end_len = end_stretch = 0;
	for (i = 0; i < end_count; i++) {
		end_len += end_items[i].alength;
		end_stretch += end_items[i].alength * end_items[i].stretch;
	}

	/* Compute number of periods */
	r1 = alength - start_len - end_len; 
	RTEST_NUM(r2 = floorr(r1 / period_len));
	if (r2 <= 0)
		k = 0;
	else if (r2 >= GEOM_PATTERN_PERIOD_MAX)
		k = GEOM_PATTERN_PERIOD_MAX;
	else {
		k = r2;
		if (r1 - r2 * period_len > 0.5 * period_len)
			k++;
	}
	BASSERT(k <= GEOM_PATTERN_PERIOD_MAX);

	/* Compute stretch coefficient */
	r1 = alength - start_len - end_len - k * period_len;
	r2 = start_stretch + end_stretch + k * period_stretch;
	stretch_coef = r1 / r2;
	if (!RTEST(stretch_coef))
		stretch_coef = 0;

	/* Compute all wanted ATIMEs in non-decreasing order */
	times_count = start_count + end_count + k * period_count;
	atimes = alloca(times_count * sizeof(real));
	times = alloca(times_count * sizeof(real));
	index = 0;
	r1 = 0;
	for (i = 0; i < start_count; i++, index++) {
		r1 = fmaxr(r1, r1 + start_items[i].alength + start_items[i].stretch * stretch_coef);
		atimes[index] = r1;
	}
	for (j = 0; j < k; j++)
		for (i = 0; i < period_count; i++, index++) {
			r1 = fmaxr(r1, r1 + period_items[i].alength + period_items[i].stretch * stretch_coef);
			atimes[index] = r1;
		}
	for (i = 0; i < end_count; i++, index++) {
		r1 += fmaxr(r1, r1 + end_items[i].alength + end_items[i].stretch * stretch_coef);
		atimes[index] = r1;
	}
	RTEST_NUM(r1);

	/* ATIME -> TIME conversion */
	ETEST_RET(geom_item_atimes_to_times(curve, times_count, atimes, times));

	/* Generate output */
	index = 0;
	r1 = 0;
	for (i = 0; i < start_count; i++, index++) {
		r2 = times[index];
		ETEST_RET(geom_item_pattern_item_append(curve, r1, r2, &start_items[i], result));
		r1 = r2;
	}
	for (j = 0; j < k; j++)
		for (i = 0; i < period_count; i++, index++) {
			r2 = times[index];
			ETEST_RET(geom_item_pattern_item_append(curve, r1, r2, &period_items[i], result));
			r1 = r2;
		}
	for (i = 0; i < end_count; i++, index++) {
		r2 = times[index];
		ETEST_RET(geom_item_pattern_item_append(curve, r1, r2, &end_items[i], result));
		r1 = r2;
	}

	ERET(0);
}


