/*
 * 	GEOMLIB - Implementation of Rtree planar search data structure.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/rectangle.h"
#include "geomlib/rtree.h"


/* WALK_NODE_IN_RTREE_POSTFIX */

#define WALK_NODE_IN_RTREE_POSTFIX_BEGIN(_name, _rtree, _node) \
	{ if ((_rtree)->root) { \
		struct geom_rtree_node *_node, *_name##_stack[GEOM_RTREE_MAX_HEIGHT]; \
		int _name##_top, _name##_index[GEOM_RTREE_MAX_HEIGHT]; \
		_name##_index[_name##_top = 0] = 0; \
		_name##_stack[0] = (_rtree)->root; \
		do \
		{ \
			while ((_node = _name##_stack[_name##_top])->height && \
					_name##_index[_name##_top] < _node->count) { \
				_name##_stack[_name##_top + 1] = _node->child[_name##_index[_name##_top]++]; \
				_name##_index[++_name##_top] = 0; \
			}; {


#define WALK_NODE_IN_RTREE_POSTFIX_END(_name) } \
		} while (_name##_top--); }}


/* geom_rtree_assert */

void geom_rtree_assert(struct geom_rtree *rtree)
{
	int i;
	struct geom_rtree_obj *obj;
	BASSERT(rtree);
	WALK_NODE_IN_RTREE_BEGIN(walk, rtree, node)
	{
		BASSERT(node->count <= GEOM_RTREE_MAX);
		if (node == rtree->root)
		{
			BASSERT(!node->parent);
			BASSERT(node->count >= 1);
		}
		else
			BASSERT(node->count >= GEOM_RTREE_MIN);
		for (i = 0; i < node->count; i++)
			BASSERT(node->child[i]);
		if (node->height)
			for (i = 0; i < node->count; i++)
			{
				BASSERT(node->child[i]->parent == node);
				BASSERT(node->child[i]->height == node->height - 1);
				BASSERT(node->child[i]->bbox.ll.x >= node->bbox.ll.x);
				BASSERT(node->child[i]->bbox.ll.y >= node->bbox.ll.y);
				BASSERT(node->child[i]->bbox.ur.x <= node->bbox.ur.x);
				BASSERT(node->child[i]->bbox.ur.y <= node->bbox.ur.y);
			}
		else
			for (i = 0; i < node->count; i++)
			{
				obj = (struct geom_rtree_obj *)node->child[i];
				BASSERT(obj->parent == node);
				BASSERT(obj->bbox.ll.x <= obj->bbox.ur.x);
				BASSERT(obj->bbox.ll.y <= obj->bbox.ur.y);
				BASSERT(obj->bbox.ll.x >= node->bbox.ll.x);
				BASSERT(obj->bbox.ll.y >= node->bbox.ll.y);
				BASSERT(obj->bbox.ur.x <= node->bbox.ur.x);
				BASSERT(obj->bbox.ur.y <= node->bbox.ur.y);
			}
	}
	WALK_NODE_IN_RTREE_END(walk);
}


/* geom_rtree_obj_dump */

static void geom_rtree_obj_dump(struct geom_rtree_obj *obj UNUSED)
{
	BASSERT(obj);
	DBG("    obj: addr=%p parent=%p rect=(%.0f,%.0f,%.0f,%.0f)\n",
		obj, obj->parent,
		obj->bbox.ll.x, obj->bbox.ll.y, obj->bbox.ur.x, obj->bbox.ur.y);
}

/* geom_rtree_node_dump */

static void geom_rtree_node_dump(struct geom_rtree_node *node)
{
	BASSERT(node);
	DBG("    node: addr=%p parent=%p rect=(%.0f,%.0f,%.0f,%.0f) height=%d count=%d child=",
		node, node->parent,
		node->bbox.ll.x, node->bbox.ll.y, node->bbox.ur.x, node->bbox.ur.y,
		node->height, node->count);
	for (int i = 0; i < node->count; i++)
		DBG("%p ", node->child[i]);
	DBG("\n");
}

/* geom_rtree_dump */

void geom_rtree_dump(struct geom_rtree *rtree)
{
	BASSERT(rtree);
	DBG("  rtree dump:\n");
	if (rtree->root)
	{
		WALK_NODE_IN_RTREE_BEGIN(walk, rtree, node)
		{
			geom_rtree_node_dump(node);
			if (!node->height)
				for (int i = 0; i < node->count; i++)
					geom_rtree_obj_dump((struct geom_rtree_obj *)node->child[i]);
		}
		WALK_NODE_IN_RTREE_END(walk);
	}
	else
		DBG("    NULL\n");
}

/* geom_rtree_query_call_show */

static inline void geom_rtree_query_call_show(struct geom_rtree_query *query, struct geom_rtree_obj *obj)
{
	if (query->call_show)
		((geom_rtree_query_callback *)query->call_show)(query, obj, GEOM_RTREE_QUERY_SHOW);
}

/* geom_rtree_query_call_hide */

static inline void geom_rtree_query_call_hide(struct geom_rtree_query *query, struct geom_rtree_obj *obj)
{
	if (query->call_hide)
		((geom_rtree_query_callback *)query->call_hide)(query, obj, GEOM_RTREE_QUERY_HIDE);
}

/* geom_rtree_query_call_update */

static inline void geom_rtree_query_call_update(struct geom_rtree_query *query, struct geom_rtree_obj *obj, uns flags)
{
	if (query->call_update)
		((geom_rtree_query_callback *)query->call_update)(query, obj, GEOM_RTREE_QUERY_UPDATE | flags);
}

/*
 * ==================================================================================
 *
 *    WORKING WITH RECTANGLES
 *
 * ==================================================================================
 */


/* geom_rtree_box_equal
 *
 * - compares two rectangles
 * - parameters:
 *     r1 - pointer to first rectangle
 *     r2 - pointer to second rectangle
 * - returns:
 *     nonzero, when rectangles are equal
 **/

static inline int geom_rtree_box_equal(struct geom_rectangle *r1, struct geom_rectangle *r2)
{
	return (r1->ll.x == r2->ll.x && r1->ll.y == r2->ll.y &&
			r1->ur.x == r2->ur.x && r1->ur.y == r2->ur.y);
}


/* geom_rtree_box_overlap
 *
 * - returns overlapped area of two given rectangles
 * - result can be zero
 * - parameters:
 *     r1 - pointer to first rectangle
 *     r2 - pointer to second rectangle
 * - returns:
 *     area, where r1 and r2 overlap
 **/

static inline double geom_rtree_box_overlap(struct geom_rectangle *r1, struct geom_rectangle *r2)
{
	struct geom_point p;
	double result;
	if ((p.x = MIN(r1->ur.x, r2->ur.x) - MAX(r1->ll.x, r2->ll.x)) <= 0)
		return 0;
	if ((p.y = MIN(r1->ur.y, r2->ur.y) - MAX(r1->ll.y, r2->ll.y)) <= 0)
		return 0;
	return isfinite(result = p.x * p.y) ? result : 0;
}


/* geom_rtree_box_merge
 *
 * - extends first rectangle to contain the second
 * - parameters:
 *     result - pointer to rectangle to extend
 *     r - second rectangle
 **/

static inline void geom_rtree_box_merge(struct geom_rectangle *result, struct geom_rectangle *r)
{
	if (result->ll.x > r->ll.x)
		result->ll.x = r->ll.x;
	if (result->ll.y > r->ll.y)
		result->ll.y = r->ll.y;
	if (result->ur.x < r->ur.x)
		result->ur.x = r->ur.x;
	if (result->ur.y < r->ur.y)
		result->ur.y = r->ur.y;
}


/* geom_rtree_box_area
 *
 * - returns area of non-empty rectangle
 * - parameters:
 *     r - pointer to a rectangle
 * - returns:
 *     area of rectangle r
 **/

static inline double geom_rtree_box_area(struct geom_rectangle *r)
{
	double result;
	return isfinite(result = (r->ur.x - r->ll.x) * (r->ur.y - r->ll.y)) ? result : 0;
}


/* geom_rtree_box_enlargement
 *
 * - returns area enlargement when adding a rectangle to to bbox
 * - parameters:
 *     r1 - original bbox
 *     r2 - rectangle to add
 * - returns:
 *     bbox area enlargment after extending to contain r2
 **/

static inline double geom_rtree_box_enlargement(struct geom_rectangle *r1, struct geom_rectangle *r2)
{
	struct geom_rectangle r;
	double result;
	r = *r1;
	geom_rtree_box_merge(&r, r2);
	result = ((r.ur.x - r.ll.x) * (r.ur.y - r.ll.y)) - ((r1->ur.x - r1->ll.x) * (r1->ur.y - r1->ll.y));
	return isfinite(result) ? result : 0;
}


/* geom_rtree_box_update
 *
 * - updates node bbox as bbox of children
 * - parameters:
 *     node - node to update
 **/

static void geom_rtree_box_update(struct geom_rtree_node *node)
{
	int i;
	if (node->height)
	{
		node->bbox = node->child[0]->bbox;
		for (i = 1; i < node->count; i++)
			geom_rtree_box_merge(&node->bbox, &node->child[i]->bbox);
	}
	else
	{
		node->bbox = ((struct geom_rtree_obj *)node->child[0])->bbox;
		for (i = 1; i < node->count; i++)
			geom_rtree_box_merge(&node->bbox, &((struct geom_rtree_obj *)node->child[i])->bbox);
	}
}


/* geom_rtree_box_adjust
 *
 * - updates all bounding boxes on the path from given node to rtree root
 * - parameters:
 *     node - node, where to start
 **/

static void geom_rtree_box_adjust(struct geom_rtree_node *node)
{
	struct geom_rectangle box;
	while (node)
	{
		box = node->bbox;
		geom_rtree_box_update(node);
		if (geom_rtree_box_equal(&node->bbox, &box))
			return;
		node = node->parent;
	}
}


/* geom_rtree_box_adjust_enlarge
 *
 * - enlarges node's bbox to contain new given box and
 *   updates all necessary bboxes on the path to root
 * - parameters:
 *     node - node to update
 *     bbox - enlarging box
 **/

static void geom_rtree_box_adjust_enlarge(struct geom_rtree_node *node, struct geom_rectangle *box)
{
	struct geom_rectangle old;
	while (node)
	{
		old = node->bbox;
		geom_rtree_box_merge(&node->bbox, box);
		if (geom_rtree_box_equal(&node->bbox, &old))
			return;
		node = node->parent;
	}
}


/* geom_rtree_box_adjust_shrink
 *
 * - called after something was removed from node's children
 * - updates all necessary bboxes on the path to root
 * - parameters:
 *     node - node to update
 */

static inline void geom_rtree_box_adjust_shrink(struct geom_rtree_node *node)
{
	geom_rtree_box_adjust(node);
}


/*
 * ==================================================================================
 *
 *    MEMORY MANAGEMENT
 *
 * ==================================================================================
 */

/* geom_rtree_node_alloc */

static struct geom_rtree_node *geom_rtree_node_alloc(struct geom_rtree *rtree UNUSED)
{
	DBG("  geom_rtree_node_alloc\n");
	BASSERT(rtree);
	return xmalloc_zero(sizeof(struct geom_rtree_node));
}

/* geom_rtree_node_free */

static void geom_rtree_node_free(struct geom_rtree *rtree UNUSED, struct geom_rtree_node *node)
{
	DBG("  geom_rtree_node_free\n");
	BASSERT(rtree);
	BASSERT(node);
	xfree(node);
}

/*
 * ==================================================================================
 *
 *    INSERT / DELETE
 *
 * ==================================================================================
 */


/* headers */

static void geom_rtree_split(struct geom_rtree *rtree, struct geom_rtree_node *node);
static void geom_rtree_reinsert(struct geom_rtree *rtree, struct geom_rtree_node *node, int count, int *overflow_height);
static void geom_rtree_insert_data(struct geom_rtree *rtree, struct geom_rtree_node *n, int height, int *overflow_height);
static void geom_rtree_delete_data(struct geom_rtree *rtree, struct geom_rtree_obj *obj);


/* geom_rtree_split
 *
 * - splits node with GEOM_RTREE_MAX+1 children
 * - can cause overflow in parent node
 * - if node was root, new root is created
 * - node's bbox is not important
 * - parameters:
 *     rtree - pointer to rtree
 *     node - pointer to node with GEOM_RTREE_MAX+1 children
 **/

#define GEOM_RTREE_DISTRIB (GEOM_RTREE_MAX - 2 * GEOM_RTREE_MIN + 2) // possible distributions

struct geom_rtree_split_dim // for each dimension
{
	double sort1[GEOM_RTREE_MAX + 1]; // primary sorting value
	double sort2[GEOM_RTREE_MAX + 1]; // secondary sorting value
	int order[GEOM_RTREE_MAX + 1]; // sort permutation
	double value; // choose axis algorithm value
	struct geom_rtree_split_dist // for each distribution
	{
		struct geom_rectangle bbox[2]; // bounding boxes of two sets
	} dist[GEOM_RTREE_DISTRIB];
};

#define ASORT_PREFIX(x) geom_rtree_split_##x
#define ASORT_KEY_TYPE int
#define ASORT_ELT(i) dim->order[i]
#define ASORT_LT(x,y) (dim->sort1[x] < dim->sort1[y]) || \
	(dim->sort1[x] == dim->sort1[y] && dim->sort2[x] < dim->sort2[y])
#define ASORT_EXTRA_ARGS , struct geom_rtree_split_dim *dim
#include "lib/arraysort.h"

static void geom_rtree_split(struct geom_rtree *rtree, struct geom_rtree_node *node)
{
	struct geom_rtree_node *nodes[GEOM_RTREE_MAX + 1]; // copied childpointers of node
	struct geom_rtree_node *new, *root; // new nodes
	struct geom_rectangle *bbox[GEOM_RTREE_MAX + 1]; // node->child[] bounding boxes
	struct geom_rtree_split_dim dims[2], *dim;
	double value, best_value;
	int i, k, best_count, best[GEOM_RTREE_MAX + 1];

	/* not necessary - warning elimination */
	best_value = 0; best_count = 0;

	DBG("  geom_rtree_split\n");
	BASSERT(rtree && node && node->count == GEOM_RTREE_MAX + 1);

	/* copy childpointers and get bounding boxes */
	if (node->height)
		for (i = 0; i < GEOM_RTREE_MAX + 1; i++)
			bbox[i] = &(nodes[i] = node->child[i])->bbox;
	else
		for (i = 0; i < GEOM_RTREE_MAX + 1; i++)
			bbox[i] = &((struct geom_rtree_obj *)(nodes[i] = node->child[i]))->bbox;

	/* in each axis sort nodes by rectangle minimum, then by maximum */
	for (i = 0; i < GEOM_RTREE_MAX + 1; i++)
	{
		dims[0].order[i] = dims[1].order[i] = i;
		dims[0].sort1[i] = bbox[i]->ll.x;
		dims[0].sort2[i] = bbox[i]->ur.x;
		dims[1].sort1[i] = bbox[i]->ll.y;
		dims[1].sort2[i] = bbox[i]->ur.y;
	}
	geom_rtree_split_sort(GEOM_RTREE_MAX + 1, &dims[0]);
	geom_rtree_split_sort(GEOM_RTREE_MAX + 1, &dims[1]);

	/* compute boundix boxes and margin values of distributions */
	for (dim = dims; dim < &dims[2]; dim++)
	{
		dim->value = 0;
		for (k = 0; k < GEOM_RTREE_DISTRIB; k++)
		{
			/* compute splitted sets bounding boxes */
			dim->dist[k].bbox[0] = *bbox[dim->order[0]];
			for (i = 1; i < GEOM_RTREE_MIN + k; i++)
				geom_rtree_box_merge(&dim->dist[k].bbox[0], bbox[dim->order[i]]);
			dim->dist[k].bbox[1] = *bbox[dim->order[GEOM_RTREE_MIN + k]];
			for (i = GEOM_RTREE_MIN + k + 1; i < GEOM_RTREE_MAX + 1; i++)
				geom_rtree_box_merge(&dim->dist[k].bbox[1], bbox[dim->order[i]]);

			/* margin value */
			for (i = 0; i < 2; i++)
				dim->value += dim->dist[k].bbox[i].ur.x - dim->dist[k].bbox[i].ll.x +
				              dim->dist[k].bbox[i].ur.y - dim->dist[k].bbox[i].ll.y;
		}
	}
	/* choose axis with the minimum sum of all margin values of the different distributions */
	dim = dims[0].value <= dims[1].value ? &dims[0] : &dims[1];

	/* choose distribution with minimal overlap value */
	for (k = 0; k < GEOM_RTREE_DISTRIB; k++)
	{
		value = geom_rtree_box_overlap(&dim->dist[k].bbox[0], &dim->dist[k].bbox[1]);
		if (!k || value < best_value)
		{
			best[0] = k;
			best_count = 1;
			best_value = value;
		}
		else if (value == best_value)
			best[best_count++] = k;
	}

	/* resolve ties by choosing the distribution with minimal area value */
	if (best_count > 1)
		for (k = 0; k < best_count; k++)
		{
			value = geom_rtree_box_area(&dim->dist[best[k]].bbox[0]) +
				    geom_rtree_box_area(&dim->dist[best[k]].bbox[1]);
			if (!k || value < best_value)
			{
				best[0] = best[k];
				best_value = value;
			}
		}

	/* compute new counts and bboxes  */
	node->count = GEOM_RTREE_MIN + best[0];
	node->bbox = dim->dist[best[0]].bbox[0];

	new = geom_rtree_node_alloc(rtree);
	new->count = GEOM_RTREE_MAX + 1 - node->count;
	new->bbox = dim->dist[best[0]].bbox[1];
	new->height = node->height;

	/* distribute childpointers */
	if (node->height)
	{
		for (i = 0; i < node->count; i++)
			(node->child[i] = nodes[dim->order[i]])->parent = node;
		for (i = 0; i < new->count; i++)
			(new->child[i] = nodes[dim->order[i + node->count]])->parent = new;
	}
	else
	{
		for (i = 0; i < node->count; i++)
			((struct geom_rtree_obj *)(node->child[i] = nodes[dim->order[i]]))->parent = node;
		for (i = 0; i < new->count; i++)
			((struct geom_rtree_obj *)(new->child[i] = nodes[dim->order[i + node->count]]))->parent = new;
	}

	/* update parent pointer */
	if (node->parent)
	{
		DBG("    updating parent\n");
		new->parent = node->parent;
		node->parent->child[node->parent->count++] = new;

/**/		geom_rtree_box_adjust(new->parent);
	}

	/* create new root */
	else
	{
		DBG("    creating new root\n");
		rtree->root = root = geom_rtree_node_alloc(rtree);
		root->count = 2;
		root->height = node->height + 1;
		(root->child[0] = node)->parent = root;
		(root->child[1] = new)->parent = root;
		geom_rtree_box_update(root);
	}
}


/* geom_rtree_reinsert
 *
 * - deletes count far children from given node and reinserts them to same rtree level
 * - necessary: GEOM_RTREE_MIN + count <= node->count <= GEOM_RTREE_MAX + 1,
 * - node->bbox is not important
 * - if reinsert cause overflows in same level
 *   (for example if node had been overflowed and children were reinserted
 *   to same position), splits are used to solve this
 * - parameters:
 *     tree - pointer to rtree
 *     node - pointer to internal node
 *     count - (> 0) number of node's children to reinsert
 *     overflow_height - maximal height of node, where overflow threatment was inwaked
 */

#define ASORT_PREFIX(x) geom_rtree_reinsert_##x
#define ASORT_KEY_TYPE int
#define ASORT_ELT(x) perm[x]
#define ASORT_LT(x,y) (dist[x]<dist[y])
#define ASORT_EXTRA_ARGS , int *perm, double *dist
#include "lib/arraysort.h"

static void geom_rtree_reinsert(struct geom_rtree *rtree, struct geom_rtree_node *node, int count, int *overflow_height)
{
	int i, j; // iterators
	int order[GEOM_RTREE_MAX + 1]; // sort permutation
	double dist[GEOM_RTREE_MAX + 1]; // bbox centers distances
	struct geom_point center, point; // bboxes centers
	struct geom_rtree_node *nodes[GEOM_RTREE_MAX + 1]; // copy of node childpointers

	DBG("  geom_rtree_reinsert\n");

	/* sort children by distances of bbox centeres to node bbox center */
	center.x = 0.5 * (node->bbox.ll.x + node->bbox.ur.x);
	center.y = 0.5 * (node->bbox.ll.y + node->bbox.ur.y);
	if (node->height)
		for (i = 0; i < node->count; i++)
		{
			nodes[i] = node->child[i];
			point.x = 0.5 * (nodes[i]->bbox.ur.x + nodes[i]->bbox.ll.x) - center.x;
			point.y = 0.5 * (nodes[i]->bbox.ur.y + nodes[i]->bbox.ll.y) - center.y;
			dist[i] = point.x * point.x + point.y * point.y;
		}
	else
		for (i = 0; i < node->count; i++)
		{
			nodes[i] = node->child[i];
			point.x = 0.5 * (((struct geom_rtree_obj *)nodes[i])->bbox.ur.x +
					         ((struct geom_rtree_obj *)nodes[i])->bbox.ll.x) - center.x;
			point.y = 0.5 * (((struct geom_rtree_obj *)nodes[i])->bbox.ur.y +
					         ((struct geom_rtree_obj *)nodes[i])->bbox.ll.y) - center.y;
			dist[i] = point.x * point.x + point.y * point.y;
		}
	for (i = 0; i < node->count; i++)
		if (!isfinite(dist[order[i] = i]))
			dist[i] = 0;
	geom_rtree_reinsert_sort(node->count, order, dist);

	/* delete count far nodes from node */
	node->count -= count;
	for (i = 0; i < node->count; i++)
		node->child[i] = nodes[order[i]];

	/* update bounding boxes */
	geom_rtree_box_adjust_shrink(node);

	/* reinsert removed nodes */
	j = node->count;
	for (i = 0; i < count; i++)
		geom_rtree_insert_data(rtree, nodes[order[i + j]], node->height, overflow_height);
}


/* geom_rtree_insert_data
 *
 * - inserts node n to parent with given rtree level
 * - if level is zero, n must be geom_rtree_obj, else geom_rtree_node
 * - boundig box of n must be initialized and non-empty
 * - parameters:
 *     rtree - pointer to rtree
 *     n - pointer to inserting object
 *     height - height of resulting n's parent
 *     overflow_height - maximal level, where reinsert could be used
 **/

static void geom_rtree_insert_data(struct geom_rtree *rtree, struct geom_rtree_node *n, int height, int *overflow_height)
{
	int i, j; // iterators
	struct geom_rectangle *box; // inserting object bounding box
	int best_count; // choose subtree candidates count
	double value, best_value; // choose subtree values
	struct geom_rtree_node *best[GEOM_RTREE_MAX + 1]; // choose subtree candidates
	struct geom_rtree_node *node; // node iterator

	/* not necessary - warning elimination */
	best_count = 0; best_value = 0;

	DBG("  geom_rtree_insert_data (height = %d)\n", height);

	/* insert to empty RTree*/
	if (!rtree->root)
	{
		DBG("  inserting to empty RTree, creating root node\n");
		rtree->root = geom_rtree_node_alloc(rtree);
		rtree->root->bbox = ((struct geom_rtree_obj *)n)->bbox;
		rtree->root->count = 1;
		rtree->root->height = 0;
		rtree->root->child[0] = n;
		((struct geom_rtree_obj *)n)->parent = rtree->root;
		return;
	}

	/* geom inserting object bounding box */
	if (height)
		box = &n->bbox;
	else
		box = &((struct geom_rtree_obj *)n)->bbox;

	/* choose subtree, where to insert the object */
	{
		/* start in root node, choose the best subtree in each level */
		for (node = rtree->root; node->height > height; )
		{
			/* children contain leaves */
			if (node->height == 1)
			{
				/* choose entry whose rectangle needs least overlap enlargment */
				for (i = 0; i < node->count; i++)
				{
					value = 0;
					for (j = 0; j < node->child[i]->count; j++)
						value += geom_rtree_box_overlap(box,
							&((struct geom_rtree_obj *)node->child[i]->child[j])->bbox);
					if (!i || value < best_value)
					{
						best[0] = node->child[i];
						best_count = 1;
						best_value = value;
					}
					else if (value == best_value)
						best[best_count++] = node->child[i];
				}
				/* resolve ties by choosing the entry whose area needs least area enlargement */
				if (best_count > 1)
				{
					j = best_count;
					for (i = 0; i < j; i++)
					{
						value = geom_rtree_box_enlargement(&best[i]->bbox, box);
						if (!i || value < best_value)
						{
							best[0] = best[i];
							best_count = 1;
							best_value = value;
						}
						else if (value == best_value)
							best[best_count++] = best[i];
					}
				}
			}
			/* childpointers do not point to leaves */
			else
			{
				/* choose the entry whose area needs least area enlargement */
				for (i = 0; i < node->count; i++)
				{
					value = geom_rtree_box_enlargement(&node->child[i]->bbox, box);
					if (!i || value < best_value)
					{
						best[0] = node->child[i];
						best_count = 1;
						best_value = value;
					}
					else if (value == best_value)
						best[best_count++] = node->child[i];
				}
			}
			/* resolve ties by choosing the entry with the rectangle of smallest area */
			if (best_count > 1)
				for (i = 0; i < best_count; i++)
				{
					value = geom_rtree_box_area(&best[i]->bbox);
					if (!i || value < best_value)
					{
						best[0] = best[i];
						best_value = value;
					}
				}
			/* go to next level */
			node = best[0];
		}
	}

	/* insert object to node childpointers */
	node->child[node->count++] = n;
	if (height)
	{
		n->parent = node;
		n->height = height - 1;
	}
	else
		((struct geom_rtree_obj *)n)->parent = node;

	geom_rtree_box_adjust_enlarge(node, box);

	/* if there is no overflow, update bbox and exit */
	if (node->count <= GEOM_RTREE_MAX)
	{
		/*geom_rtree_box_adjust_enlarge(node, box);*/
		return;
	}

	/* overflow treatment */
	do
	{
		DBG("  overflow treatment\n");
		/* if level is not root level and this is the first
		 * overflow treatment in this level, invoke reinsert */
		if (GEOM_RTREE_REINSERT && node->height != rtree->root->height && node->height > *overflow_height)
		{
			*overflow_height = MAX(*overflow_height, node->height);
			geom_rtree_reinsert(rtree, node, GEOM_RTREE_REINSERT, overflow_height);

			/* if there is no overflow, exit */
			if (node->count <= GEOM_RTREE_MAX)
				return;
		}
		else
			*overflow_height = MAX(*overflow_height, node->height);

		/* split the node */
		geom_rtree_split(rtree, node);
		node = node->parent;
	}
	while (node->count > GEOM_RTREE_MAX);
}


/* geom_rtree_delete_data */

static void geom_rtree_delete_data(struct geom_rtree *rtree, struct geom_rtree_obj *obj)
{
	struct geom_rtree_node *parent, *node;
	int i, overflow_height = -1;

	DBG("geom_rtree_delete_data\n");

	/* remove obj from parent childpointers and update bboxes */
	node = obj->parent;
	for (i = 0; i < node->count; i++)
		if (node->child[i] == (struct geom_rtree_node *)obj)
		{
			node->child[i] = node->child[--node->count];
			break;
		}
	geom_rtree_box_adjust_shrink(node);

	/* mark obj to be unassociated object */
	obj->parent = NULL;

	/* underfull treatment cycle */
	if (parent = node->parent)
		while (parent && node->count < GEOM_RTREE_MIN)
		{
			DBG("  underfull treatment, reinserting childpointers\n");

			/* remove tree level */
			if (parent->count == 1)
			{
				DBG("  root underfull, removing tree level\n");
				geom_rtree_node_free(rtree, parent);
				node->parent = NULL;
				rtree->root = node;
				break;
			}

			/* remove node from parent childpointers and update bboxes */
			for (i = 0; i < parent->count; i++)
				if (parent->child[i] == node)
				{
					parent->child[i] = parent->child[--parent->count];
					break;
				}
			geom_rtree_box_adjust_shrink(parent);

			/* reinsert node childpointers to the same level */
			for (i = 0; i < node->count; i++)
				geom_rtree_insert_data(rtree, node->child[i], node->height, &overflow_height);

			/* free node */
			geom_rtree_node_free(rtree, node);
			parent = (node = parent)->parent;
		}

	/* free root after last object deletion */
	else if (!node->count)
	{
		DBG("  root node empty, deleting\n");
		geom_rtree_node_free(rtree, node);
		rtree->root = NULL;
	}
}

/* geom_rtree_insert */

void geom_rtree_insert(struct geom_rtree *rtree, struct geom_rtree_obj *obj, struct geom_rectangle *bbox)
{
	int overflow_height = -1;

	DBG("geom_rtree_insert\n");
	BASSERT(rtree);
	BASSERT(obj);
	BASSERT(bbox);

	/* insert object */
	obj->bbox = *bbox;
	geom_rtree_insert_data(rtree, (struct geom_rtree_node *)obj, 0, &overflow_height);

	/* queries messages */
	WALK_QUERY_IN_RTREE(rtree, query)
		if (geom_rtree_box_intersect(bbox, &query->box))
			geom_rtree_query_call_show(query, obj);
}


/* geom_rtree_delete */

void geom_rtree_delete(struct geom_rtree *rtree, struct geom_rtree_obj *obj)
{

	DBG("geom_rtree_delete\n");
	BASSERT(rtree);
	BASSERT(obj);
	BASSERT(obj->parent);

	/* delete object */
	geom_rtree_delete_data(rtree, obj);

	/* queries messages */
	WALK_QUERY_IN_RTREE(rtree, query)
		if (geom_rtree_box_intersect(&obj->bbox, &query->box))
			geom_rtree_query_call_hide(query, obj);
}


/* geom_rtree_init
 *
 * - initialize RTree main structure
 * - parameters:
 *     rtree - pointer to the structure
 **/

void geom_rtree_init(struct geom_rtree *rtree)
{
	DBG("geom_rtree_init\n");
	BASSERT(rtree);
	rtree->root = NULL;
	clist_init(&rtree->lquery);
}


/* geom_rtree_destroy
 *
 * - cleans up internal RTree data
 * - parameters:
 *     rtree - pointer to the structure
 **/

void geom_rtree_destroy(struct geom_rtree *rtree)
{
	DBG("geom_rtree_done\n");
	BASSERT(rtree);
	BASSERT(clist_empty(&rtree->lquery));
	geom_rtree_clear(rtree);
}


/* geom_rtree_clear
 *
 * - removes all objects from given RTree
 * - parameters:
 *     rtree - pointer to RTree structure
 **/

void geom_rtree_clear(struct geom_rtree *rtree)
{
	DBG("geom_rtree_clear\n");
	BASSERT(rtree);

	/* send delete messages to dynamic queries */
	WALK_OBJ_IN_RTREE_BEGIN(walk, rtree, obj)
	{
		WALK_QUERY_IN_RTREE(rtree, query)
			if (geom_rtree_box_intersect(&obj->bbox, &query->box))
				geom_rtree_query_call_hide(query, obj);
		obj->parent = NULL;
	}
	WALK_OBJ_IN_RTREE_END(walk);

	/* free all allocated nodes */
	WALK_NODE_IN_RTREE_POSTFIX_BEGIN(walk, rtree, node)
	{
		geom_rtree_node_free(rtree, node);
	}
	WALK_NODE_IN_RTREE_POSTFIX_END(walk);
	rtree->root = NULL;
}


/* geom_rtree_update */

void geom_rtree_update(struct geom_rtree *rtree, struct geom_rtree_obj *obj, uns flags, struct geom_rectangle *bbox)
{
	struct geom_rectangle old_bbox;
	int overflow_height = -1;

	DBG("geom_rtree_update\n");
	BASSERT(rtree);
	BASSERT(obj);
	BASSERT(obj->parent);

	if (bbox && !geom_rtree_box_equal(bbox, &obj->bbox))
	{
		old_bbox = obj->bbox;
		geom_rtree_delete_data(rtree, obj);
		obj->bbox = *bbox;
		geom_rtree_insert_data(rtree, (struct geom_rtree_node *)obj, 0, &overflow_height);

		WALK_QUERY_IN_RTREE(rtree, query)
			if (geom_rtree_box_intersect(&obj->bbox, &query->box))
				if (geom_rtree_box_intersect(&old_bbox, &query->box))
					geom_rtree_query_call_update(query, obj, flags | GEOM_RTREE_QUERY_UPDATE_BBOX);
				else
					geom_rtree_query_call_show(query, obj);
			else if (geom_rtree_box_intersect(&old_bbox, &query->box))
				geom_rtree_query_call_hide(query, obj);
	}
	else
		WALK_QUERY_IN_RTREE(rtree, query)
			if (geom_rtree_box_intersect(&obj->bbox, &query->box))
				geom_rtree_query_call_update(query, obj, flags);
}


/* geom_rtree_query_init */

void geom_rtree_query_init(struct geom_rtree_query *query, struct geom_rtree *rtree, struct geom_rectangle *box, void *show, void *hide, void *update)
{
	BASSERT(query);
	BASSERT(rtree);
	BASSERT(box);

	query->rtree = rtree;
	query->box = *box;
	query->call_show = show;
	query->call_hide = hide;
	query->call_update = update;
	clist_add_head(&rtree->lquery, &query->nquery);

	if (show)
		GEOM_RTREE_RECT_INTERSECT_QUERY_BEGIN(q1, rtree, box, obj)
		{
			geom_rtree_query_call_show(query, obj);
		}
		GEOM_RTREE_RECT_INTERSECT_QUERY_END(q1);
}


/* geom_rtree_query_destroy */

void geom_rtree_query_destroy(struct geom_rtree_query *query)
{
	BASSERT(query);

	if (query->call_hide)
		GEOM_RTREE_RECT_INTERSECT_QUERY_BEGIN(q1, query->rtree, &query->box, obj)
		{
			geom_rtree_query_call_hide(query, obj);
		}
		GEOM_RTREE_RECT_INTERSECT_QUERY_END(q1);

	clist_remove(&query->nquery);
}


