/*
 * 	GEOMLIB - Transformations alternatives.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/numeric.h"
#include "geomlib/transform.h"
#include "geomlib/transform2.h"


/* transform2 transforms from inner coords to outer coords,
   inverted matrix is for reverse transformation */

int
geom_transform2_is_valid(const struct geom_transform2 *t)
{
  return (geom_transform_valid  (&(((struct geom_transform2 *)t) -> primary)) &&
	  geom_transform_valid  (&(((struct geom_transform2 *)t) -> inverted)));
}

int
geom_transform2_set (struct geom_transform2 *t, real a, real b, real c, real d, real e, real f)
{
  int i = 0;
  i = i || geom_transform_set (&(t -> primary), a, b, c, d, e, f);
  i = i || geom_transform_invert (&(t -> primary), &(t -> inverted));
  return (i ? -1 : 0);
}

int
geom_transform2_set_identity (struct geom_transform2 *t)
{
  t -> primary = geom_transform_identity;
  t -> inverted = geom_transform_identity;

  return (0);
}

int
geom_transform2_set_move (struct geom_transform2 *t, const double x, const double y)
{
  int i = 0;
  i = i || geom_transform_set_move (&(t -> primary), x, y);
  i = i || geom_transform_set_move (&(t -> inverted), -x, -y);
  return (i ? -1 : 0);
}

int
geom_transform2_set_rotation (struct geom_transform2 *t, const double angle)
{
  int i = 0;
  i = i || geom_transform_set_rotation (&(t -> primary), angle);
  i = i || geom_transform_set_rotation (&(t -> inverted), -angle);
  return (i ? -1 : 0);
}

int
geom_transform2_set_scale (struct geom_transform2 *t, const double x, const double y)
{
  int i = 0;
  if (IS_ZERO (x) || IS_ZERO (y)) return (-1);
  i = i || geom_transform_set_scale (&(t -> primary), x, y);
  i = i || geom_transform_set_scale (&(t -> inverted), 1/x , 1/y);
  return (i ? -1 : 0);
}

int
geom_transform2_set_basis (struct geom_transform2 *t, real p1x, real p1y,
			   real p2x, real p2y, real p3x, real p3y)
{
  t -> primary . flags = 0;
  t -> primary . coef [0] [0] = p2x - p1x;
  t -> primary . coef [0] [1] = p3x - p1x;
  t -> primary . coef [0] [2] = p1x;
  t -> primary . coef [1] [0] = p2y - p1y;
  t -> primary . coef [1] [1] = p3y - p1y;
  t -> primary . coef [1] [2] = p1y;
  for (uns i = 0; i < 2; i++)
	  for (uns j = 0; j < 3; j++)
		  if (!GEOM_RTEST(t -> primary . coef [i] [j]))
			  return -1;
  return (geom_transform_invert (&(t -> primary), &(t -> inverted)));
}

int
geom_transform2_invert (const struct geom_transform2 *t,
			struct geom_transform2 *result)
{
  result -> primary = t -> inverted;
  result -> inverted = t -> primary;

  return (0);
}

int
geom_transform2_merge (const struct geom_transform2 *t1,
		       const struct geom_transform2 *t2,
		       struct geom_transform2 *result)
{
  int i = 0;
  i = i || geom_transform_merge (&(t1 -> primary), &(t2 -> primary), &(result -> primary));
  i = i || geom_transform_merge (&(t2 -> inverted), &(t1 -> inverted), &(result -> inverted));
  return (i ? -1 : 0);
}

int
geom_transform2_translate (struct geom_transform2 *t, const double x, const double y)
{
  struct geom_transform2 arg;
  int i = 0;
  i = i || geom_transform2_set_move (&arg, x, y);
  i = i || geom_transform2_merge (t, &arg, t);
  return (i ? -1 : 0);
}

int
geom_transform2_rotate (struct geom_transform2 *t, const double angle)
{
  struct geom_transform2 arg;
  int i = 0;
  i = i || geom_transform2_set_rotation (&arg, angle);
  i = i || geom_transform2_merge (t, &arg, t);
  return (i ? -1 : 0);
}


int
geom_transform2_scale (struct geom_transform2 *t, const double x, const double y)
{
  struct geom_transform2 arg;
  int i = 0;
  i = i || geom_transform2_set_scale (&arg, x, y);
  i = i || geom_transform2_merge (t, &arg, t);
  return (i ? -1 : 0);
}

int geom_transform2_set_rotate_center (struct geom_transform2 *t, const real x, const real y, const real angle)
{
	real cos1, sin1;

	cos1 = cos(angle);
	sin1 = sin(angle);

	t->primary.flags = GEOM_TRANSFORM_SIMILAR;
	t->primary.coef[0][0] = cos1;
	t->primary.coef[0][1] = -sin1;
	t->primary.coef[0][2] = -cos1 * x + sin1 * y + x;
	t->primary.coef[1][0] = sin1;
	t->primary.coef[1][1] = cos1;
	t->primary.coef[1][2] = -sin1 * x - cos1 * y + y;

	t->inverted.flags = GEOM_TRANSFORM_SIMILAR;
	t->inverted.coef[0][0] = cos1;
	t->inverted.coef[0][1] = sin1;
	t->inverted.coef[0][2] = -cos1 * x - sin1 * y + x;
	t->inverted.coef[1][0] = -sin1;
	t->inverted.coef[1][1] = cos1;
	t->inverted.coef[1][2] = sin1 * x - cos1 * y + y;

	if (!isfinite(t->primary.coef[0][2]) || !isfinite(t->primary.coef[1][2]) ||!isfinite(t->inverted.coef[0][2]) ||!isfinite(t->inverted.coef[1][2]))
		return GEOM_ERR_NUMERIC;

	return 0;
}

int geom_transform2_set_scale_center (struct geom_transform2 *t, const real x, const real y, const real sx, const real sy)
{
	real sxi, syi;

	t->primary.flags = t->inverted.flags = (sx == sy || sx == -sy) ? GEOM_TRANSFORM_SIMILAR : 0;
	t->primary.coef[0][0] = sx;
	t->primary.coef[0][1] = 0;
	t->primary.coef[0][2] = -sx * x + x;
	t->primary.coef[1][0] = 0;
	t->primary.coef[1][1] = sy;
	t->primary.coef[1][2] = -sy * y + y;

	sxi = 1 / sx;
	syi = 1 / sy;

	t->inverted.coef[0][0] = sxi;
	t->inverted.coef[0][1] = 0;
	t->inverted.coef[0][2] = -sxi * x + x;
	t->inverted.coef[1][0] = 0;
	t->inverted.coef[1][1] = syi;
	t->inverted.coef[1][2] = -syi * y + y;

	if (!isfinite(t->primary.coef[0][2]) || !isfinite(t->primary.coef[1][2]) ||!isfinite(t->inverted.coef[0][2]) ||!isfinite(t->inverted.coef[1][2]))
		return GEOM_ERR_NUMERIC;

	return 0;
}

static int matrix33_invert (double * mi, double * mo)
{
  double det = mi [0] * mi [4] * mi [8] + mi [1] * mi [5] * mi [6]
    + mi [2] * mi [3] * mi [7] - mi [2] * mi [4] * mi [6]
    - mi [1] * mi [3] * mi [8] - mi [0] * mi [5] * mi [7];

  if (IS_ZERO_L (det)) return GEOM_ERR_NUMERIC;
  if (!isfinite(det)) return GEOM_ERR_NUMERIC;

  det = 1 / det;
  if (IS_ZERO_L (det)) return GEOM_ERR_NUMERIC;
  if (!isfinite(det)) return GEOM_ERR_NUMERIC;
  mo [0] = (mi [4] * mi [8] - mi [5] * mi [7]) * det;
  mo [1] = (mi [2] * mi [7] - mi [1] * mi [8]) * det;
  mo [2] = (mi [1] * mi [5] - mi [2] * mi [4]) * det;
  mo [3] = (mi [5] * mi [6] - mi [3] * mi [8]) * det;
  mo [4] = (mi [0] * mi [8] - mi [2] * mi [6]) * det;
  mo [5] = (mi [2] * mi [3] - mi [0] * mi [5]) * det;
  mo [6] = (mi [3] * mi [7] - mi [4] * mi [6]) * det;
  mo [7] = (mi [1] * mi [6] - mi [0] * mi [7]) * det;
  mo [8] = (mi [0] * mi [4] - mi [1] * mi [3]) * det;

  int i;
  for ( i=0; i<9; i++ ) if (!isfinite(mo[i])) return GEOM_ERR_NUMERIC;
  return 0;
}

static void matrix_mult_vector (double * mi, double * src, struct geom_transform *t, int row)
{
  double tmp;
  for (int i = 0; i < 3; i ++)
    {
      tmp = 0;
      for (int j = 0; j < 3; j ++)
	tmp += src [j] * mi [j +  3 * i];
      t -> coef [row] [i] = tmp;
    }
}

int geom_transform2_set_skew (struct geom_transform2 *t,
			      real x1, real y1, real x2, real y2,
			      real xf, real yf, real xt, real yt)
{
  double mi [9] = {x1, y1, 1, x2, y2, 1, xf, yf, 1};
  double mo [9];
  double src1 [3] = {x1, x2, xt};
  double src2 [3] = {y1, y2, yt};

  if (matrix33_invert (mi, mo) != 0) return (-1);

  t -> primary . flags = 0;
  matrix_mult_vector (mo, src1, &(t -> primary), 0);
  matrix_mult_vector (mo, src2, &(t -> primary), 1);

  if (! geom_transform_valid (&(t -> primary))) return (-1);
  if (geom_transform_invert (&(t -> primary), &(t -> inverted)) != 0) return (-1);
  return (0);
}

int geom_transform2_set_skew_normal (struct geom_transform2 *t, const real zx, const real zy,
	const real ax, const real ay, const real fx, const real fy, const real tx, const real ty)
{
	BASSERT(t);
	int errval;
	/* b is the transition matrix from the axis basis to the standard basis,
	   bi is the inverted transition matrix */
	struct geom_transform bi, b = { 0, {{ ax - zx, ay - zy, zx }, { ay - zy, zx - ax, zy } } };
	if (errval = geom_transform_invert( &b, &bi )) return errval;

	struct geom_point v1 = { fx, fy };
	struct geom_point v2 = { tx, ty };

	if (errval = geom_point_transform( &v1, &bi, &v1 )) return errval;
	if (errval = geom_point_transform( &v2, &bi, &v2 )) return errval;

	/* tt is the transformation matrix (regarding to the axis basis) in which
	   (x,0) becomes (x,0)
	   (v1.x,1) becomes (v2.x,1) ... computed by hand :) */
	struct geom_transform tt = { 0, {{1.0, v1.x - v2.x, 0.0}, {0.0, 1.0, 0.0} } };

	/* now find the matrix of t regarding to the standard basis */
	geom_transform_merge( &bi, &tt, &t->primary );
	geom_transform_merge( &t->primary, &b, &t->primary );

	if (!geom_transform_valid(&t->primary)) return GEOM_ERR_NUMERIC;
	if (errval = geom_transform_invert( &t->primary, &t->inverted )) return errval;
	return 0;
}
