/*
 *  GUI - The command manager.
 *
 *  Structures, types, enums, functions etc. for commads and menu items
 *  management.
 *
 * Copyright (c) 2004-2006    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "kernel/kernel.h"

#include "gui/cmdmgr.h"
#include "gui/properties.h"
#include "build/path.h"

/** \file gui/cmdmgr.c
 * The implementation of the command manager.
 */

static void on_command_activated( GtkWidget * widget, struct cmd * command );
static addr_int_t stamp;
static void change_stamp(void){ stamp = (stamp + 1) % 256 + 1; }
static int sel_req_ok( uns mask, uns cmask, int cnt, int ccnt );
static int context_matches( struct cmd * cmd );
static void load_icons_in_category( struct cmd * cmd );
static void fill_parent_pointers( struct cmd * cmd, struct cmd * parent );
static void expand_category( struct cmd * cmd, uns expand );

static int sync_in_progress;
static void fill_menu_with_category( GtkWidget * menu, struct cmd * cmd, unsigned int location, struct cmddata cd );
static void sync_menu_by_category( GtkWidget * menuitem, struct cmd * cmd, unsigned int location );
static void fill_toolbar_with_category( GtkBox * tb, struct cmd * cmd, unsigned int location, struct cmddata cd );
static void sync_toolbar_by_category( struct cmd * cmd, unsigned int location, GList ** list );
static void sync_cmd_item( struct cmd * cmd, GtkWidget * command, unsigned int location );
static void calculate_selection(void);

GtkAccelGroup * accel_group;
static GtkTooltips * tooltips;
static struct cmd * expanded_category;

static int plugin_menu_count;
#define PLUGIN_ICON_PATH LIB_PATH"obj_properties.png"
static GdkPixbuf * icon_plugin;
garr plugin_menu_recs;

void command_manager_init(void)
{
	accel_group = gtk_accel_group_new();
	tooltips = GTK_TOOLTIPS(gtk_tooltips_new());
	load_icons_in_category( &main_ctg );
	icon_plugin = icon_load(PLUGIN_ICON_PATH);
	struct cmd * c;
	for(c = create_ctg.spec.ctg.first; c && c->type != CT_CATEGORY; c = c->next);
	if (c) expanded_category = c;
	fill_parent_pointers( &main_ctg, NULL );

	plugin_menu_recs = garr_new(1, sizeof(struct plugin_menu_rec*));

	change_stamp();
}

static void fill_parent_pointers( struct cmd * cmd, struct cmd * parent )
{
	struct cmd * co;
	for ( co = cmd; co; co = co->next )
	{
		if (co->type == CT_CATEGORY && co->spec.ctg.first)
			fill_parent_pointers( co->spec.ctg.first, co );
		co->parent = parent;
	}
}
static void load_icons_in_category( struct cmd * cmd )
{
	ASSERT( cmd );

	struct cmd * co;
	for ( co = cmd; co; co = co->next )
	{
		if ( co->type == CT_CATEGORY )
		{
			if (co->spec.ctg.first) load_icons_in_category(co->spec.ctg.first);
		}
		if (co->icon_path && !(co->flags & CC_STOCK) && !(co->icon_pixbuf = icon_load(co->icon_path)))
		{
			co->flags |= CC_STOCK; //set the icon of missing image instead
			co->icon_path = "gtk-missing-image";
		}
	}
}

void command_manager_finish(void)
{
	// destroy commands allocated for plugins
	for (int i = 0; i < plugin_menu_count; i++)
		plugin_menu_unregister(i);

	garr_destroy(plugin_menu_recs);
	plugin_menu_count = 0;
}

/* CONTEXT AND SELECTION */

void change_context( struct window * window, struct go_group * group, struct o * o )
{
	struct context old_context = context;
	ASSERT( !window || window->statusbar );
	if (o)
		ASSERT_O(o);
	if (group)
		ASSERT_GROUP(group);
	context.w = window;
	context.o = o;
	if ( context.group != group )
	{
		context.group = group;
		context.tlo = group ? go_get_tlo(&group->go) : NULL;
		context.doc = (context.tlo && context.tlo->obj.father && context.tlo->obj.father->type == OT_DOCUMENT ) ?
				obj2doc(context.tlo->obj.father)  : NULL;
		context.sel_valid = 0;
	}

	if (old_context.w != context.w || old_context.group != context.group ||
		old_context.o != context.o || old_context.tlo != context.tlo || old_context.doc != context.doc)
	{

		ub_context_changed(&old_context);
		prop_window_context_changed(&old_context);
		refresh_gui();
	}
}

static void calculate_selection(void)
{
	if ( context.sel_valid && context.msel_valid ) return;

	if ( !context.msel_valid )
	{
		context.msel_mask = ( obj_sel_count[OT_DOCUMENT] > 0 ) * VTM_DOCUMENT;
		context.msel_cnt_all = obj_sel_count[OT_DOCUMENT];
		context.msel_mask |= ( obj_sel_count[OT_TLO] > 0 ) * VTM_TLO;
		context.msel_cnt_all += obj_sel_count[OT_TLO];
		context.msel_valid = 1;
	}

	if ( !context.sel_valid )
	{
		struct go * go;
		context.sel_mask = 0;
		context.sel_cnt_all = 0;
		for (uns u = 0; u < GOT_MAX; u++) context.sel_cnt[u] = 0;

		if ( context.group ) WALK_GROUP_GO(context.group, go)
			if ( go_is_selected(go) )
			{
				context.sel_cnt[go->type]++;
				context.sel_mask |= GOT_TO_VTM(go->type);
				context.sel_cnt_all++;
			}
		context.sel_valid = 1;
	}
}

/** cmask (=command mask) must be a bitwise supset of the context mask */
static int sel_req_ok( uns mask, uns cmask, int cnt, int ccnt )
{
	int ok = ( mask & ~cmask ) == 0 && ( ccnt == 0 || (ccnt == -1 && cnt > 0) || ccnt == cnt );
	//DBG("Hit=%d, Context: M=%x, C=%d, Cmd: M=%x, C=%d\n", ok, mask, cnt, cmask, ccnt );
	return ok;
}

static int context_matches( struct cmd * cmd )
{
	ASSERT( cmd->type == CT_FUNC );
	
	/* An opened view determines a view, TLO and document */
	if ( cmd->spec.func.request_cnt <= 1 &&
		(context.w && WT_TO_VTM(context.w->type) == cmd->spec.func.request_mask) ||
		((cmd->spec.func.request_mask & VTM_CT_GRP) && context.group ) ||
		((cmd->spec.func.request_mask & VTM_CT_TLO) && context.tlo) ||
		((cmd->spec.func.request_mask & VTM_CT_DOC) && context.doc ))
		return 1;

	/* If there is a selection, use selection; if not, use current o (under cursor) */
	int sel_cnt;
	uns mask;
	switch ( cmd->spec.func.request_type )
	{
	case RT_META:
		sel_cnt = context.msel_cnt_all;
		mask = context.msel_mask;
		break;
	case RT_GROUP:
		sel_cnt = context.sel_cnt_all;
		mask = context.sel_mask;
		break;
	case RT_ANY:
		sel_cnt = context.msel_cnt_all + context.sel_cnt_all;
		mask = context.msel_mask | context.sel_mask;
		break;
	default: BUG();
	}
	if ( sel_cnt > 0 || cmd->spec.func.request_cnt == 0 )
		return sel_req_ok( mask, cmd->spec.func.request_mask, sel_cnt, cmd->spec.func.request_cnt );
	
	/* This is a bit weird ... */
	if ( context.o )
		return sel_req_ok( ( context.o->kind == T_GO ) ? GOT_TO_VTM(context.o->type) : OT_TO_VTM(context.o->type),
			cmd->spec.func.request_mask, 1, cmd->spec.func.request_cnt );
	return 0;
}

/* CREATING MENUS AND TOOLBARS */

static void sync_cmd_item( struct cmd * cmd, GtkWidget * widget, uns location )
{
	switch ( cmd->type )
	{
	case CT_FUNC:
		COMBINE_FLAGS( context_matches(cmd), CS_ENABLED );
		if ( cmd->spec.func.modify_state_f ) cmd->spec.func.modify_state_f(cmd, widget);
		DBG("Sync func item %s, vis = %d, en = %d, check = %d\n",
			cmd->title, cmd->flags & CS_VISIBLE, cmd->flags & CS_ENABLED, cmd->flags & CS_CHECKED );
		break;
	case CT_FACTORY_OP:
		COMBINE_FLAGS(cmd->spec.op.of_state == factory.start_state, CS_CHECKED);
		COMBINE_FLAGS(!(factory.flags & OFF_PATH_ON ) || cmd->spec.op.of_state->flags & OFF_PATHABLE, CS_ENABLED);
		break;
	case CT_SEPARATOR:
		return;
	case CT_CATEGORY:
		COMBINE_FLAGS( cmd == expanded_category, CS_CHECKED );
		break;
	default:;
	}

	if ( (cmd->flags & CS_VISIBLE) && (((location & CL_ALL_TBARS) == 0) || !(cmd->flags & CS_UNEXPANDED)) )
		gtk_widget_show_all(widget);
	else
		gtk_widget_hide_all(widget);

	gtk_widget_set_sensitive( widget, cmd->flags & CS_ENABLED );

	if(!GTK_IS_TEAROFF_MENU_ITEM(widget))
	{
		if ( cmd->flags & CC_CHECKABLE )
		{
			sync_in_progress = 1;
			if ( GTK_IS_TOGGLE_BUTTON(widget) )
				gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(widget), cmd->flags & CS_CHECKED );
			else	
				gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(widget), cmd->flags & CS_CHECKED );
			sync_in_progress = 0;
		}
	}
	if (GTK_IS_BUTTON(widget))
		gtk_tooltips_set_tip( tooltips, widget, cmd->description, NULL );
}

GtkWidget * create_menu( uns location, uns popup, struct cmddata cd ){

	GtkWidget * msh;
	ASSERT( main_ctg.type == CT_CATEGORY && main_ctg.spec.ctg.first );
	calculate_selection();
	msh = popup ? gtk_menu_new() : gtk_menu_bar_new();
	fill_menu_with_category( msh, main_ctg.spec.ctg.first, location, cd );
	g_object_set_data( G_OBJECT(msh), "vstamp", (gpointer)stamp );
	return msh;
}

GtkWidget * create_toolbar( uns location, uns vertical, struct cmddata cd )
{
	GtkBox * tb = GTK_BOX( vertical ? gtk_vbox_new( FALSE, 1 ) : gtk_hbox_new( FALSE, 1 ));
	ASSERT( main_ctg.type == CT_CATEGORY && main_ctg.spec.ctg.first );
	fill_toolbar_with_category( tb, main_ctg.spec.ctg.first, location, cd );
	g_object_set_data( G_OBJECT(tb), "vstamp", (gpointer)stamp );
	gtk_widget_show( GTK_WIDGET(tb) );
	return GTK_WIDGET(tb);
}
static void fill_menu_with_category( GtkWidget * menu, struct cmd * cmd, uns location, struct cmddata cd )
{
	ASSERT( cmd && menu );
	DBG("Fill menu with cmd %s\n", cmd->title );

	if ( cmd != main_ctg.spec.ctg.first ){
		GtkWidget * tear = gtk_tearoff_menu_item_new();
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), tear);
		gtk_widget_show(tear);
	}

	GtkWidget * mi;
	for ( struct cmd * co = cmd; co; co = co->next ){
		if (!(co->flags & location))
			continue;

		switch (co->type)
		{
		case  CT_CATEGORY:;
			mi = gtk_menu_item_new_with_label( co->title );
			GtkWidget * submenu = gtk_menu_new();
			gtk_menu_item_set_submenu( GTK_MENU_ITEM(mi), submenu );
			if (co->spec.ctg.first) fill_menu_with_category( submenu, co->spec.ctg.first, location, cd );
			gtk_widget_show(mi);
			gtk_widget_show(submenu);
			break;
		case CT_SEPARATOR:
			mi = gtk_separator_menu_item_new();
			gtk_widget_show(mi);
			break;
		default:
			mi = (co->flags & CC_CHECKABLE) ? gtk_check_menu_item_new_with_label(co->title) : gtk_menu_item_new_with_label(co->title);
			if ( co->accel)
				gtk_widget_add_accelerator( mi, "activate", accel_group, co->accel, co->modifier, GTK_ACCEL_VISIBLE );
			sync_cmd_item( co, mi, location );
			g_signal_connect( G_OBJECT(mi), ( co->flags & CC_CHECKABLE ) ? "toggled" : "activate",
				G_CALLBACK(on_command_activated), (gpointer)(co) );
		}
		gtk_menu_shell_append( GTK_MENU_SHELL( menu ), mi );
	}
}
#define CREATE_BUTTON \
	DBG("Add tool item %s\n", co->title ); \
	GtkWidget * widget = ( co->flags & CC_CHECKABLE ) ? gtk_toggle_button_new() : gtk_button_new(); \
	if (co->icon_path){ \
		if ( co->flags & CC_STOCK ) \
			gtk_container_add( GTK_CONTAINER(widget), \
				gtk_image_new_from_stock( co->icon_path, GTK_ICON_SIZE_SMALL_TOOLBAR) ); \
		else gtk_container_add( GTK_CONTAINER(widget), gtk_image_new_from_pixbuf( co->icon_pixbuf ) ); \
	} \
	gtk_box_pack_start( tb, widget, FALSE, FALSE, 0 ); \
	gtk_tooltips_set_tip( tooltips, widget, co->description, NULL ); \
	if (cd.window) g_object_set_data( G_OBJECT(widget), "vrrwindow", (gpointer)cd.window ); \
	if (cd.group) g_object_set_data( G_OBJECT(widget), "vrrgroup", (gpointer)cd.group ); \
	g_signal_connect( G_OBJECT(widget), "clicked", G_CALLBACK(on_command_activated), (gpointer)co ); \
	sync_cmd_item( co, widget, location )

static void fill_toolbar_with_category( GtkBox * tb, struct cmd * cmd, uns location, struct cmddata cd )
{
	ASSERT( cmd && tb );
	DBG("Fill toolbar with command %s\n", cmd->title );

	for ( struct cmd * co = cmd; co; co = co->next ){
		if (!(co->flags & location))
			continue;

		switch (co->type)
		{
		case CT_CATEGORY:
			if (co->spec.ctg.flags & location)
			{
				CREATE_BUTTON;
				expand_category(co, co == expanded_category);
			}
			if (co->spec.ctg.first) fill_toolbar_with_category( tb, co->spec.ctg.first, location, cd );
			break;
		case CT_SEPARATOR:
			break;
		default:
			CREATE_BUTTON;
			//g_signal_connect( G_OBJECT(widget), "button-press-event", G_CALLBACK(command_icon_press), (gpointer)co );
		}
	}
}

int sync_menu( GtkWidget * menu, unsigned int location ){
	calculate_selection();
	if ( (addr_int_t)g_object_get_data( G_OBJECT(menu), "vstamp") != stamp ) return 0;
	ASSERT( main_ctg.type == CT_CATEGORY && main_ctg.spec.ctg.first );
	sync_menu_by_category( menu, main_ctg.spec.ctg.first, location );
	return 1;
}

int sync_toolbar( GtkWidget * toolbar, unsigned int location )
{
	GList * list, *origlist;
	calculate_selection();
	if ( (addr_int_t)g_object_get_data( G_OBJECT(toolbar), "vstamp") != stamp ) return 0;
	origlist = list = gtk_container_get_children( GTK_CONTAINER( toolbar ) );
	ASSERT( main_ctg.type == CT_CATEGORY && main_ctg.spec.ctg.first );
	sync_toolbar_by_category( main_ctg.spec.ctg.first, location, &list );
	g_list_free(origlist);
	return 1;
}
static void sync_menu_by_category( GtkWidget * menu, struct cmd * cmd, unsigned int location )
{
	ASSERT( cmd && menu );

	GList * list = gtk_container_get_children( GTK_CONTAINER(menu) );

	GList * cur;
	struct cmd * co;
	for( cur = list, co = cmd; cur && co; ){
		if (!(co->flags & location))
		{ co = co->next; continue; }

		GtkMenuItem * mi = GTK_MENU_ITEM(cur->data);
		if(!GTK_IS_TEAROFF_MENU_ITEM(mi))
		{	
			if ( co->type == CT_CATEGORY )
			{
				if (co->spec.ctg.first) sync_menu_by_category( gtk_menu_item_get_submenu(mi), co->spec.ctg.first, location );
			}
			else
				sync_cmd_item( co, GTK_WIDGET(cur->data), location );
			co = co->next;
		}
		cur = cur->next; //proceed in the existing items list
	}
	g_list_free(list);
}

static void sync_toolbar_by_category( struct cmd * cmd, unsigned int location, GList ** list )
{
	ASSERT( cmd && list );

	for ( struct cmd * co = cmd; co && *list; co = co->next ){
		if (!(co->flags & location))
			continue;

		switch (co->type)
		{
		case CT_CATEGORY:
			if (co->spec.ctg.flags & location)
			{
				sync_cmd_item( co, GTK_WIDGET((*list)->data), location );
				*list = (*list)->next; //proceed in the existing items list
			}
			if (co->spec.ctg.first) sync_toolbar_by_category( co->spec.ctg.first, location, list );
			break;
		case CT_SEPARATOR:
			break;
		default:
			sync_cmd_item( co, GTK_WIDGET( (*list)->data), location );
			*list = (*list)->next; //proceed in the existing items list
		}
	}
}

/** Set the CS_UNEXPANDED flag recursively to category cmd. */
static void expand_category( struct cmd * cmd, uns expand )
{
	ASSERT(cmd && cmd->type == CT_CATEGORY );
	for ( cmd = cmd->spec.ctg.first; cmd; cmd = cmd->next )
	{
		COMBINE_FLAGS( !expand, CS_UNEXPANDED );
		if ( cmd->type == CT_CATEGORY )
			expand_category( cmd, expand );
	}
}

/* ACTIVATING COMMANDS */

static void on_command_activated( GtkWidget * widget, struct cmd * command )
{
	if ( sync_in_progress ) return; //ignore callbacks caused by checking items manually

	struct window * w;
	ASSERT( widget );
	ASSERT( command );
	void * v = g_object_get_data( G_OBJECT(widget), "vrrgroup" );
	struct go_group * group = NULL;
	if ( v )
	{
		group = (struct go_group*)void2go(v);
		ASSERT_GROUP(group);
	}
	w = (struct window *)g_object_get_data( G_OBJECT(widget), "vrrwindow" );
	if (!group)
		group = context.group;
	if (!w)
		w = context.w;

	if ( w && w != context.w || group && group != context.group )
		change_context( w, group, (context.o && context.o->kind == T_GO &&
			go_get_group(o2go(context.o)) == group ) ? context.o : NULL );
	execute_command_item( command );
}

void execute_command_item( struct cmd * cmd )
{
	if ( sync_in_progress ) return; //ignore callbacks caused by checking items manually
	ASSERT( cmd );

	if ( cmd->type != CT_CATEGORY && cmd->parent && cmd->parent->parent == &create_ctg )
	{
		if (expanded_category) expand_category( expanded_category, 0 );
		expand_category( expanded_category = cmd->parent, 1 );
		cmd->parent->spec.ctg.selected = cmd;
	}
	
	switch (cmd->type)
	{
	case CT_FACTORY_OP:
		ASSERT(cmd->spec.op.of_state);
		if (!context.group)
			break;
		if (!(cmd->flags & CC_SAFE))
			factory_op_break();
		factory_op_start( context.group, cmd->spec.op.of_state, cmd->description );
		break;

	case CT_FUNC:
		ASSERT(cmd->spec.func.action_f);
		
		if (!(cmd->flags & CC_SAFE))
			factory_op_break();
		
		if (!context_matches(cmd))
		{
			DBG("Context does not match!\n");
			break;
		}
		void * param = NULL;
		if ( context.w && WT_TO_VTM(context.w->type) == cmd->spec.func.request_mask )
			param = context.w; //window context
		else
		{
			if ( context.group && cmd->spec.func.request_cnt <= 1 ) //local context
			{
				if ( !param && (cmd->spec.func.request_mask & VTM_CT_GRP) ) param = context.group;
				if ( !param && (cmd->spec.func.request_mask & VTM_CT_TLO) ) param = context.tlo;
				if ( !param && (cmd->spec.func.request_mask & VTM_CT_DOC) ) param = context.doc;
			}
		}
		int count = 0;
		switch (cmd->spec.func.request_type)
		{
		case RT_META: count = context.msel_cnt_all; break;
		case RT_GROUP: count = context.sel_cnt_all; break;
		case RT_ANY: count = context.sel_cnt_all + context.msel_cnt_all;
		default:;
		}
		if ( !param && context.o && count == 0 && cmd->spec.func.request_cnt != 0 )
			param = context.o; //The context matched, so the context.o must be all it needs

		cmd->spec.func.action_f(param);
		break;
	case CT_CATEGORY:
		if (expanded_category) expand_category( expanded_category, 0 );
		expand_category( expanded_category = cmd, 1 );

		if ( cmd != &plugin_ctg &&
			(cmd->spec.ctg.selected || (cmd->spec.ctg.selected = cmd->spec.ctg.first)) &&
				(cmd->spec.ctg.selected->flags & CS_ENABLED) )
			execute_command_item( cmd->spec.ctg.selected );
		break;
	default:;
	}

	refresh_gui();
}

void on_context_dump( void * subj UNUSED )
{
	msg("\nCurrent context is:\nVIEW = %p, GROUP = %p, O = %p\n",
		context.w, context.group, context.o );
	msg("msel_mask = %x, msel_cnt_all = %u; sel_mask = %x, sel_cnt_all = %u\n",
		context.msel_mask, context.msel_cnt_all, context.sel_mask, context.sel_cnt_all );
	for ( uns u = 0; u < GOT_MAX; u++ )
		if ( context.sel_cnt[u] ) msg("%u %s\n", context.sel_cnt[u], string_go_type[u]->text);
	msg("\n");
}

/* ADDING AND REMOVING COMMANDS DYNAMICALLY */

void add_new_command_into( struct cmd * cmd, struct cmd * category )
{
	ASSERT( cmd && category && category->type == CT_CATEGORY );
	struct cmd * first = category->spec.ctg.first;
	category->spec.ctg.first = cmd;
	cmd->next = first;
	change_stamp();
}

void add_new_command_after( struct cmd * cmd, struct cmd * after )
{
	ASSERT( cmd && after );
	struct cmd * next = after->next;
	after->next = cmd;
	cmd->next = next;
	change_stamp();
}

void remove_command( struct cmd * cmd, struct cmd * from )
{
	ASSERT( cmd && from && from->type == CT_CATEGORY );
	if ( cmd == &main_ctg || cmd == &create_ctg || cmd == &plugin_ctg || !(from->spec.ctg.first) )
		return;

	if ( cmd == from->spec.ctg.first )
		from->spec.ctg.first = cmd->next;
	else
	{
		for (from = from->spec.ctg.first; from->next && from->next != cmd; from = from->next);
		if (from->next)
			from->next = cmd->next;

	}
	cmd->next = NULL;
	change_stamp();
	refresh_gui(); // this is maybe too slow, but we need to remove the icons pointing to invalid memory
}

int plugin_menu_register( char * desc )
{
	struct plugin_menu_rec *r, ** rr = (struct plugin_menu_rec**)garr_get(plugin_menu_recs, plugin_menu_count);
	r = *rr = xmalloc_zero(sizeof(struct plugin_menu_rec));
	
	r->category = (struct cmd){ .type = CT_CATEGORY, .flags = CL_EVERYWHERE | CS_VIS_EN };
	r->category.title = xstrdup(desc);
	r->category.description = r->category.title;
	slist_init(&r->cmd_list);

	add_new_command_into( &r->category, &plugin_ctg );

	return plugin_menu_count++;
}

int plugin_menu_unregister( int plugin_menu_id )
{
	if (plugin_menu_id < 0 || plugin_menu_id >= plugin_menu_count)
		return 1;
	
	struct plugin_menu_rec * r, **rr = (struct plugin_menu_rec**)garr_get(plugin_menu_recs, plugin_menu_id);
	r = *rr;
	if (!r)
		return 1;
	struct plugin_menu_cmd_rec * p;
	while (p = (struct plugin_menu_cmd_rec*)slist_remove_head(&r->cmd_list)) {
		remove_command(&p->cmd, &r->category);
		xfree(p->cmd.title);
		xfree(p->cmd.description);
		xfree(p);
	}
	remove_command( &r->category, &plugin_ctg );
	xfree(r->category.title);
	xfree(r);
	*rr = NULL;
	
	return 0;
}

int plugin_menu_command_register( int plugin_menu_id, struct cmd * cmd )
{
	if (plugin_menu_id < 0 || plugin_menu_id >= plugin_menu_count)
		return 1;
	struct plugin_menu_rec * r = *(struct plugin_menu_rec**)garr_get(plugin_menu_recs, plugin_menu_id);
	if (!r || !cmd)
		return 1;
	if (!(cmd->icon_pixbuf))
		cmd->icon_pixbuf = icon_plugin;

	add_new_command_into( cmd, &r->category );
	return 0;
}

int plugin_menu_command_register_conv( int plugin_menu_id, char * title, char * desc, uns request_mask,
					int request_cnt, uns request_type, void (*action_f)(void *) )
{
	if (plugin_menu_id < 0 || plugin_menu_id >= plugin_menu_count)
		return 1;
	struct plugin_menu_rec * r = *(struct plugin_menu_rec**)garr_get(plugin_menu_recs, plugin_menu_id);
	if (!r)
		return 1;

	struct plugin_menu_cmd_rec * p = xmalloc_zero(sizeof(struct plugin_menu_cmd_rec));
	p->cmd = (struct cmd){ .title = xstrdup(title), .description = xstrdup(desc), .type = CT_FUNC,
			.flags = CL_VIEW_MENU | CL_VIEW_TBAR2 | CS_VIS_EN,
			.icon_path = PLUGIN_ICON_PATH, .icon_pixbuf = icon_plugin, .spec = { .func = { .request_mask = request_mask,
			.request_cnt = request_cnt, .request_type = request_type, .action_f = action_f }} };

	slist_add( &r->cmd_list, &p->n );
	add_new_command_into( &p->cmd, &r->category );

	return 0;
}

void plugin_menu_dump(void * v UNUSED)
{
	struct plugin_menu_rec * r;
	msg("\nThere are %d registered menu plugins\n", plugin_menu_count );
	for (int i=0; i < plugin_menu_count; i++)
	{
		r = *(struct plugin_menu_rec**)garr_get(plugin_menu_recs, i);
		if (!r)
			msg("%d deleted\n", i );
		else
		{
			struct plugin_menu_cmd_rec * p;
			msg("%d. \"%s\", \"%s\", address = %p\n", i,
				r->category.title, r->category.description, &r->category );
			SLIST_WALK(p, r->cmd_list)
			{
				msg("\t - \"%s\", \"%s\", f = %p\n", p->cmd.title, p->cmd.description,
					p->cmd.spec.func.action_f );
			}
		}
	}
}
