/** @file
 * GUI - The implementation of the GO factory, snap and internal
 * GO factory callbacks.
 * 
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"
#include "kernel/kernel.h"
#include "font/font.h"
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/visualisation.h"
#include "gui/properties.h"

struct factory factory;
static void factory_set_state( struct of_state * state );
static void factory_op_finish( void );
static void factory_inner_moving_start(struct geom_point po);
static void factory_inner_moving_end(void);

static void factory_prop_edit_create(struct of_state * state);
static void factory_prop_edit_destroy(void);

static void factory_path_end(int close);
static int factory_path_new(void);

static void factory_clean_state( struct of_state * state );
static void factory_undo_delete(struct obj_tlo *tlo, struct undo_gui* ug, struct undo_gui *prev_item, struct trans_hook* h);
static struct trans_hook trans_hook;

static struct go_hook go_hook;
static void factory_prop_change( struct o *o, struct go_hook* h );

// non-public factory states (the public ones are in main.h).
static struct of_state ofs_point_inters_2;
static struct of_state ofs_ngon_2, ofs_ngon_3;
static struct of_state ofs_segment_2;
static struct of_state ofs_bez_cub_2, ofs_bez_cub_3, ofs_bez_cub_4;
static struct of_state ofs_bez_quad_2, ofs_bez_quad_3;
static struct of_state ofs_arc_3p_2, ofs_arc_3p_3;
static struct of_state ofs_circ_3p_2, ofs_circ_3p_3;
static struct of_state ofs_circ_center_pt_2;
static struct of_state ofs_ellipse_3p_smallest_2, ofs_ellipse_3p_smallest_3;
static struct of_state ofs_ellipse_3p_ecc_2, ofs_ellipse_3p_ecc_3;
static struct of_state ofs_ellipse_foci_pt_2, ofs_ellipse_foci_pt_3;
static struct of_state ofs_ellipse_xy1ecc_2;
static struct of_state ofs_transform_2;

//creating functions
static void create_point( struct of_state * state );
static void create_decorator_pt( struct of_state * state );
static void create_arrow( struct of_state * state );
static void create_ngon( struct of_state * state );
static void ngon_check_apex_number( struct of_state * state );
static void create_intersection( struct of_state * state );
static void create_segment_1( struct of_state * state );
static void create_bez_cub_1( struct of_state * state );
static void create_bez_quad_1( struct of_state * state );
static void create_arc_3p_1( struct of_state * state );
static void create_circ_3p_1( struct of_state * state );
static void create_circ_center_pt_1( struct of_state * state );
static void create_arc_center_radius_1( struct of_state * state );
static void create_ellipse_3p_smallest_2( struct of_state * state );
static void create_ellipse_3p_ecc_1( struct of_state * state );
static void create_ellipse_segment( struct of_state * state );
static void create_ellipse_foci_pt_2( struct of_state * state );
static void fix_ellipse_anchor_2( struct of_state * state );
static void create_ellarc_center_radii_1( struct of_state * state );
static void create_ellipse_xy1ecc_1( struct of_state * state );
static void create_text_1( struct of_state * state );
static void create_tex_text_1( struct of_state * state );
static void do_transform_1( struct of_state * s );
static void do_transform_2( struct of_state * s );
static void do_rehang_1( struct of_state * s );
static void do_rehang_2( struct of_state * s );

// cleanup functions
static void clean_transform( struct of_state * state );
// rehanging functions
static void fix_anchor_2( struct of_state * state );
static void fix_anchor_3( struct of_state * state );
static void fix_anchor_4( struct of_state * state );

#define VAL_INIT {.dpr = { DIST_PASS_POINT, {.point = {0.0,{0.0, 0.0}}} }}
#define TEMP_INITS NULL, { NULL }

//! a special state which means end of operation
struct of_state ofs_end = { 0, OFIK_NONE, 0, 0, VAL_INIT, "", NULL, NULL, &ofs_end, &ofs_end, TEMP_INITS };
//! a special state which means before start of an operation
struct of_state ofs_start = { 0, OFIK_NONE, 0, 0, VAL_INIT, "", NULL, NULL, &ofs_start, &ofs_start, TEMP_INITS };

struct of_state ofs_transform_x = { OFF_HIDE_HANGERS, OFIK_TRANSFORM, 0, 0, VAL_INIT, "",
				do_transform_2, NULL, &ofs_select, &ofs_end, TEMP_INITS };

struct of_state ofs_select = { OFF_HIDE_HANGERS, OFIK_SEL, 0, 0, VAL_INIT, "Select",
				do_transform_1, clean_transform, &ofs_start, &ofs_transform_x, TEMP_INITS };

static struct of_state ofs_transform_2 = { OFF_HIDE_HANGERS, OFIK_TF, 0, 0, VAL_INIT, "Transform",
				do_transform_2, NULL, &ofs_transform_1, &ofs_end, TEMP_INITS };
struct of_state ofs_transform_1 = { OFF_HIDE_HANGERS, OFIK_TF, 0, 0, VAL_INIT,
				"Edit the transformation points by shift-click and drag them",
				do_transform_1, clean_transform, &ofs_start, &ofs_transform_2, TEMP_INITS };

struct of_state ofs_rehang_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the new position of the anchor",
				do_rehang_2, NULL, &ofs_rehang_1, &ofs_end, TEMP_INITS };
struct of_state ofs_rehang_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_ANCHOR, 0, VAL_INIT, "Choose an anchor",
				do_rehang_1, NULL, &ofs_start, &ofs_rehang_2, TEMP_INITS };

static struct of_state ofs_point_inters_2 = { 0, OFIK_DPR, DIST_PASS_GO, 0, VAL_INIT, "Choose the second object for the new intersection",
				create_intersection, NULL, &ofs_point_inters, &ofs_end, TEMP_INITS };
struct of_state ofs_point_inters = { 0, OFIK_DPR, DIST_PASS_GO, 0, VAL_INIT, "Choose the first object for the new intersection",
				NULL, NULL, &ofs_start, &ofs_point_inters_2, TEMP_INITS };

struct of_state ofs_point = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose a position for the new point",
				create_point, NULL, &ofs_start, &ofs_end, TEMP_INITS };

struct of_state ofs_arrow = { 0, OFIK_DPR, DIST_PASS_GO, 0, VAL_INIT, "Choose a position for the arrow",
				create_arrow, NULL, &ofs_start, &ofs_end, TEMP_INITS };

struct of_state ofs_decorator_pt = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose a position for the new decoration point",
				create_decorator_pt, NULL, &ofs_start, &ofs_end, TEMP_INITS };

static struct of_state ofs_ngon_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose an apex for the new n-gon",
				create_ngon, NULL, &ofs_ngon_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_ngon_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center point for the new n-gon",
				NULL, NULL, &ofs_ngon_1, &ofs_ngon_3, TEMP_INITS };
struct of_state ofs_ngon_1 = { 0, OFIK_PROP, PT_UNS, PTU_UNSPECIFIED, {.prop = {.value = {.u = 3}}}, "Enter the number of apices:",
				ngon_check_apex_number, NULL, &ofs_start, &ofs_ngon_2, TEMP_INITS };

static struct of_state ofs_segment_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the endpoint of the new segment",
				fix_anchor_2, NULL, &ofs_segment_1, &ofs_end, TEMP_INITS };
struct of_state ofs_segment_1 = { OFF_START_MOVE | OFF_PATHABLE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT,
				"Choose the startpoint of the new segment",
				create_segment_1, NULL, &ofs_start, &ofs_segment_2, TEMP_INITS };

static struct of_state ofs_bez_cub_4 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the endpoint of the new cubic Bezier curve",
				fix_anchor_4, NULL, &ofs_bez_cub_3, &ofs_end, TEMP_INITS };
static struct of_state ofs_bez_cub_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the second control point of the new cubic Bezier curve",
				fix_anchor_3, NULL, &ofs_bez_cub_2, &ofs_bez_cub_4, TEMP_INITS };
static struct of_state ofs_bez_cub_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the first control point of the new cubic Bezier curve",
				fix_anchor_2, NULL, &ofs_bez_cub_1, &ofs_bez_cub_3, TEMP_INITS };
struct of_state ofs_bez_cub_1 = { OFF_START_MOVE | OFF_PATHABLE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT,
				"Choose the startpoint of the new cubic Bezier curve",
				create_bez_cub_1, NULL, &ofs_start, &ofs_bez_cub_2, TEMP_INITS };

static struct of_state ofs_bez_quad_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the endpoint of the new quadratic Bezier curve",
				fix_anchor_3, NULL, &ofs_bez_quad_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_bez_quad_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the control point of the new quadratic Bezier curve",
				fix_anchor_2, NULL, &ofs_bez_quad_1, &ofs_bez_quad_3, TEMP_INITS };
struct of_state ofs_bez_quad_1 = { OFF_START_MOVE | OFF_PATHABLE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT,
				"Choose the startpoint of the new quadratic Bezier curve",
				create_bez_quad_1, NULL, &ofs_start, &ofs_bez_quad_2, TEMP_INITS };

static struct of_state ofs_arc_3p_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the endpoint of the new circular arc",
				fix_anchor_3, NULL, &ofs_arc_3p_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_arc_3p_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the midpoint of the new circular arc",
				fix_anchor_2, NULL,  &ofs_arc_3p_1, &ofs_arc_3p_3, TEMP_INITS };
struct of_state ofs_arc_3p_1 = { OFF_START_MOVE | OFF_PATHABLE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT,
				"Choose the startpoint of the new circular arc",
				create_arc_3p_1, NULL,  &ofs_start, &ofs_arc_3p_2, TEMP_INITS };

static struct of_state ofs_circ_3p_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the third point on the new circle",
				fix_anchor_3, NULL, &ofs_circ_3p_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_circ_3p_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the second point on the new circle",
				fix_anchor_2, NULL, &ofs_circ_3p_1, &ofs_circ_3p_3, TEMP_INITS };
struct of_state ofs_circ_3p_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the first point on the new circle",
				create_circ_3p_1, NULL, &ofs_start, &ofs_circ_3p_2, TEMP_INITS };

static struct of_state ofs_circ_center_pt_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the point on the new circle",
				fix_anchor_2, NULL, &ofs_circ_center_pt_1, &ofs_end, TEMP_INITS };
struct of_state ofs_circ_center_pt_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center point of the new circle",
				create_circ_center_pt_1, NULL, &ofs_start, &ofs_circ_center_pt_2, TEMP_INITS };

struct of_state ofs_arc_center_radius_1 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center of the new arc",
				create_arc_center_radius_1, NULL, &ofs_start, &ofs_end, TEMP_INITS };

struct of_state ofs_circ_center_radius_1 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center of the new circle",
				create_arc_center_radius_1, NULL, &ofs_start, &ofs_end, TEMP_INITS };


static struct of_state ofs_ellipse_3p_smallest_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the third point on the new ellipse",
				fix_ellipse_anchor_2, NULL, &ofs_ellipse_3p_smallest_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_ellipse_3p_smallest_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the second point on the new ellipse",
				create_ellipse_3p_smallest_2, NULL, &ofs_ellipse_3p_smallest_1, &ofs_ellipse_3p_smallest_3,
				TEMP_INITS };
struct of_state ofs_ellipse_3p_smallest_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the first point on the new ellipse",
				create_ellipse_segment, NULL, &ofs_start, &ofs_ellipse_3p_smallest_2, TEMP_INITS };

static struct of_state ofs_ellipse_3p_ecc_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the third point on the new ellipse",
				fix_anchor_3, NULL, &ofs_ellipse_3p_ecc_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_ellipse_3p_ecc_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the second point on the new ellipse",
				fix_anchor_2, NULL, &ofs_ellipse_3p_ecc_1, &ofs_ellipse_3p_ecc_3, TEMP_INITS };
struct of_state ofs_ellipse_3p_ecc_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the first point on the new ellipse",
				create_ellipse_3p_ecc_1, NULL, &ofs_start, &ofs_ellipse_3p_ecc_2, TEMP_INITS };

struct of_state ofs_ellarc_center_radii_1 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center of the new elliptic arc",
				create_ellarc_center_radii_1, NULL, &ofs_start, &ofs_end, TEMP_INITS };

struct of_state ofs_ellipse_center_radii_1 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center of the new ellipse",
				create_ellarc_center_radii_1, NULL, &ofs_start, &ofs_end, TEMP_INITS };

static struct of_state ofs_ellipse_foci_pt_3 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose a point on the new ellipse",
				fix_ellipse_anchor_2, NULL, &ofs_ellipse_foci_pt_2, &ofs_end, TEMP_INITS };
static struct of_state ofs_ellipse_foci_pt_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the second focus of the new ellipse",
				create_ellipse_foci_pt_2, NULL, &ofs_ellipse_foci_pt_1, &ofs_ellipse_foci_pt_3, TEMP_INITS };
struct of_state ofs_ellipse_foci_pt_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the first focus of the new ellipse",
				create_ellipse_segment, NULL, &ofs_start, &ofs_ellipse_foci_pt_2, TEMP_INITS };

static struct of_state ofs_ellipse_xy1ecc_2 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the point on the new ellipse",
				fix_anchor_2, NULL, &ofs_ellipse_xy1ecc_1, &ofs_end, TEMP_INITS };
struct of_state ofs_ellipse_xy1ecc_1 = { OFF_START_MOVE, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the center point of the new ellipse",
				create_ellipse_xy1ecc_1, NULL, &ofs_start, &ofs_ellipse_xy1ecc_2, TEMP_INITS };

struct of_state ofs_text_1 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the refpoint of the new text",
				create_text_1, NULL, &ofs_start, &ofs_end, TEMP_INITS };

struct of_state ofs_tex_text_1 = { 0, OFIK_DPR, DIST_PASS_HANGER, 0, VAL_INIT, "Choose the refpoint of the new TeX text",
				create_tex_text_1, NULL, &ofs_start, &ofs_end, TEMP_INITS };

static string noname_op, string_apices;

static int tsort_on;
#define TSORT_ON { tsort_hanger_presort( factory.hanger_cursor, factory.tlo ); tsort_on = 1; }
#define TSORT_OFF { tsort_presort_end( factory.tlo ); tsort_on = 0; }
#define LINK_HEAD(_go) go_link_head( _go, factory.group )

void factory_init(void)
{
	DBG("Factory init\n");
	factory.start_state = &ofs_select;
	trans_hook.data = &factory;
	trans_hook.undo_delete = factory_undo_delete;

	go_hook.data = &factory;
	go_hook.go_prop_change = factory_prop_change;

	factory.snap_tol_pix = prop_lookup_uns( obj2o(obj_universum), string_tolerance, TOLERANCE_PIX);
	factory.snap_set = SNAP_HANGERS | SNAP_DEPENDENCY;

	factory_set_state( &ofs_select );
	factory.desc = noname_op = string_ref(string_lookup_ref("Anonymous operation"));
	ofs_ngon_1.value.prop.key = string_apices = string_lookup_ref("number-of-apices");
}
void factory_finish(void)
{
	DBG("Factory finish\n");
	if ( factory.flags & OFF_PATH_ACTIVE )
		factory_path_end(0);
	if ( factory.desc) string_unref(factory.desc);
	string_unref( noname_op );
	string_unref( string_apices );
	if ( factory.group ) go_unref( &factory.group->go );
	if ( factory.tlo )
	{
		obj_unref( &factory.tlo->obj );
		trans_hook_remove( factory.tlo, &trans_hook );
	}
}

void factory_op_start( struct go_group * group, struct of_state * state, char * desc )
{
	string s;
	ASSERT_GROUP(group);
	ASSERT(factory.desc);
	ASSERT(state && state->prev == &ofs_start );
	DBG("factory_op_start called for group %p, state %p\n", group, state );

	if ( factory.flags & OFF_OPERATION_ACTIVE ) factory_op_finish();
	if ( factory.group != group && (factory.flags & OFF_PATH_ACTIVE) )
		factory_path_end(0);
	if ( factory.group ) go_unref( &factory.group->go );
	if ( factory.tlo )
	{
		obj_unref( &factory.tlo->obj );
		trans_hook_remove( factory.tlo, &trans_hook );
	}
	factory.group = group;
	factory.tlo = go_get_tlo( &group->go );
	go_ref( &group->go );
	obj_ref( &factory.tlo->obj );
	trans_hook_add( factory.tlo, &trans_hook );

	if ((factory.flags & OFF_PATH_ON) && !(factory.flags & OFF_PATH_ACTIVE) )
		if (factory_path_new())
			return;

	factory.start_state = state;

	s = factory.desc;
	factory.desc = ( !desc ) ? string_ref(factory.desc) : string_lookup_ref(desc);
	string_unref(s);

	factory_set_state( state );
}

static void factory_set_state( struct of_state * state )
{
	if ( factory.state && factory.state->input_kind == OFIK_PROP )
	{
		DBG("Last state used property widgets, destroying\n");
		factory_prop_edit_destroy();
	}
	
	DBG("factory_set_state, last state = %p, new state = %p\n", factory.state, state );
	factory.state = state;
	factory.state->prev_undo_item = factory.tlo ? factory.tlo->undo_present : NULL;

	if ( state != &ofs_end )
		for ( GSList * l = view_list; l; l = l->next )
			factory_msg( factory.state->description );

	if ( state == &ofs_end ) factory_op_finish();
	else refresh_gui();

	statusbar_msg(NULL, " ");
	
	if ( state->input_kind == OFIK_PROP )
	{
		DBG("This state uses property widgets, creating\n");
		factory_prop_edit_create(state);
	}

	//wait for input from gui
}

void factory_moving_step(struct geom_point po)
{
	if (moving_mode != MOVE_FACTORY) return; //ignore unwanted events
	char err[TRANS_ERR_SIZE];
	volatile int errval = 0;
	TRANS_BEGIN_MAIN(factory.tlo, factory.desc->text, err)
	{
		if(hanger_mc_move( factory.hanger_cursor, &po ))
			tsort_presorted_recompute( factory.tlo );
	}
	TRANS_FAILED  //ignore
		errval = 1;
		statusbar_err(NULL, "%s", err);
	TRANS_END;
	if (!errval) factory_merge_trans();
}

static void factory_inner_moving_start(struct geom_point po)
{
	ASSERT( moving_mode == MOVE_NONE && !factory.hanger_cursor );
	DBG("Factory inner moving start\n");
	struct geom_point pi;
	ASSERT( context.w->type == WT_VIEW );
	struct view * vw = (struct view*)context.w;
	factory.hanger_cursor = hanger_new_mc(&po);
	hanger_ref( factory.hanger_cursor );
	TSORT_ON; // all newly created things must hang on the cursor hanger
	visualisation_coords_to_pixels( &(vw->vis), &po, &pi );
	view_moving_start( pi, vw, MOVE_FACTORY, GDK_CROSSHAIR );
}

static void factory_inner_moving_end(void)
{
	if ( moving_mode != MOVE_FACTORY ) { DBG("Warning: move mode not factory!\n"); return; }
	DBG("Factory inner moving end\n");
	view_moving_end();
	TSORT_OFF;
	hanger_unref( factory.hanger_cursor );
	factory.hanger_cursor = NULL;
}

static int factory_merge_trans_from(struct of_state * state)
{
	// merge all new partial transactions into one
	ASSERT( factory.flags & OFF_OPERATION_ACTIVE );
	ASSERT( state->prev_undo_item );
	struct undo_gui * ug, * start = state->prev_undo_item;
	if ( start == factory.tlo->undo_present )
	{
		DBG("merge trans: nothing to merge\n");
		return 0;
	}
	DBG("Merge trans from state \"%s\", start = %p\n", state->description, start );
	ug = trans_get_next_gui( factory.tlo, start );
	while( ug && ug != factory.tlo->undo_present )
	{
		ASSERT( trans_get_next_gui(factory.tlo, ug) );
		DBG("Merging item %p with item %p\n", ug, trans_get_next_gui(factory.tlo, ug) );
		ug = trans_merge_undo( factory.tlo, ug, factory.desc );
	}
	return ug != NULL;
}

/** Merge all transactions of the current factory operation state -- i.e., after
 factory.state->prev_undo_item of the current state -- into one.
 Returns nonzero if there was at least one new transaction since that. */
int factory_merge_trans(void)
{ return factory_merge_trans_from( factory.state ); }

static void factory_op_finish( void )
{
	DBG("factory_op_finish called\n");
	ASSERT( factory.flags & OFF_OPERATION_ACTIVE );
	ASSERT( factory.tlo );

	//cleanup after all actions
	for ( struct of_state * ofs = factory.start_state; ofs != &ofs_end; ofs = ofs->next )
		factory_clean_state(ofs);

	//if ( moving_mode != MOVE_NONE ) view_moving_end();
	ASSERT( moving_mode == MOVE_NONE );

	// merge all new partial transactions into one
	if ( factory_merge_trans_from( factory.start_state ) )
		trans_rename_undo( factory.tlo, factory.tlo->undo_present, factory.desc );
	factory.start_state->prev_undo_item = factory.tlo->undo_present;

	factory.flags &= ~OFF_OPERATION_ACTIVE;
	factory_set_state( factory.start_state );
}

void factory_op_break( void )
{
	char s[TRANS_ERR_SIZE];
	DBG("factory_op_break called\n");
	if ( ! (factory.flags & OFF_OPERATION_ACTIVE) ) return;

	//do the cleanup first - must be before undo because of tsort!
	for ( struct of_state * ofs = factory.state->prev; ofs != &ofs_start; ofs = ofs->prev )
		factory_clean_state(ofs);

	// merge all new partial transactions into one
	if ( factory_merge_trans_from( factory.start_state ) )
	{
		if (!trans_undo( factory.tlo, s )) statusbar_err(NULL, "%s", s);
		trans_clear_one_redo( factory.tlo );
	}
	factory.flags &= ~OFF_OPERATION_ACTIVE;
	factory_set_state( factory.start_state );
}

void on_step_back( void * subj UNUSED ) { factory_op_step_back(); }
void on_break_op( void * subj UNUSED ) { factory_op_break(); }

void factory_op_step_back(void)
{
	char s[TRANS_ERR_SIZE];
	DBG("factory_op_step_back called\n");

	if ( factory.state == factory.start_state ) { statusbar_err(NULL, "Cannot step back."); return; }
	if ( factory.state->prev == factory.start_state ) { factory_op_break(); return; }

	struct geom_point po;  //if moving, save the last position
	if ( moving_mode == MOVE_FACTORY && factory.hanger_cursor ) po = hanger_get_position(factory.hanger_cursor);

	int tsort_active = tsort_on;
	if ( tsort_active ) TSORT_OFF;

	struct of_state * prev = factory.state->prev;
	factory_clean_state(prev);

	// undo all new partial transactions
	if ( factory_merge_trans_from( factory.state->prev ) )
	{
		if (!trans_undo( factory.tlo, s )) statusbar_err(NULL, "%s", s);
		trans_clear_one_redo( factory.tlo );
	}
	factory_set_state( prev );

	if ( tsort_active ) TSORT_ON;

	//restore the moving position
	if ( moving_mode == MOVE_FACTORY ) factory_moving_step(po);
}

static void factory_undo_delete(struct obj_tlo *tlo UNUSED, struct undo_gui* ug, struct undo_gui *prev_item UNUSED, struct trans_hook* h UNUSED)
{
	for ( struct of_state * ofs = factory.start_state; ofs != &ofs_end; ofs = ofs->next )
		if (ofs->prev_undo_item == ug)
		{
			DBG("Undo present item %p of state \"%s\" was deleted\n", ug, ofs->description);
			ofs->prev_undo_item = trans_get_first_gui(factory.tlo);
			break;
		}
}

static void reform_input_to_hanger( void )
{
	ASSERT(factory.state->input_kind == OFIK_DPR);
	struct dist_pass_result * dpr = &factory.state->value.dpr;
	switch ( dpr->type )
	{
	case DIST_PASS_POINT:
		dpr->data.hanger.hanger = hanger_new_mc( &dpr->data.point.point );
		break;
	case DIST_PASS_GO:
	{
		if ( factory.snap_set & SNAP_DEPENDENCY )
		{
			//create a parametric point
			DBG("Factory: creating a parametric point at %p, time %.3f\n", dpr->data.go.go, dpr->data.go.geom.time );
			struct go_parametric_point * gp;
			if (factory.hanger_cursor)
				TSORT_OFF;
			gp = go_parametric_point_new( dpr->data.go.go, dpr->data.go.geom.time );
			go_link_head( &gp->go, factory.group );
			dpr->data.hanger.hanger = go_get_hanger( &gp->go, 0 );
			if (factory.hanger_cursor)
				TSORT_ON;
		} else
			dpr->data.hanger.hanger = hanger_new_mc( &dpr->data.go.geom.point );
		break;
	}
	case DIST_PASS_INTERSECTION:
	{
		if ( factory.snap_set & SNAP_DEPENDENCY )
		{
			//create an intersection
			DBG("Factory: creating an intersection of %p and %p, param = %.3f\n", dpr->data.intersection.go1,
				dpr->data.intersection.go2, dpr->data.intersection.geom.time1 );
			if (factory.hanger_cursor)
				TSORT_OFF;
			struct go_intersection * gp = go_intersection_new( dpr->data.intersection.go1,
				dpr->data.intersection.go2, dpr->data.intersection.geom.time1 );
			go_link_head( &gp->go, factory.group );
			dpr->data.hanger.hanger = go_get_hanger( &gp->go, 0 );
			if (factory.hanger_cursor)
				TSORT_ON;
		} else
			dpr->data.hanger.hanger = hanger_new_mc( &dpr->data.intersection.geom.point );
		break;
	}
	case DIST_PASS_HANGER:
		if (!( factory.snap_set & SNAP_DEPENDENCY ))
		{
			struct geom_point pt = hanger_get_position( dpr->data.hanger.hanger );
			dpr->data.hanger.hanger = hanger_new_mc( &pt );
		}
		break;
	default: BUG();
	}
	ASSERT( dpr->data.hanger.hanger );
	hanger_ref( dpr->data.hanger.hanger );
	dpr->type = DIST_PASS_HANGER;
}

void factory_op_step( void )
{
	char err[TRANS_ERR_SIZE];
	volatile int success = 1;
	DBG("factory_op_step\n");
	ASSERT( factory.state->input_kind == OFIK_DPR &&
		(factory.state->input_type == DIST_PASS_HANGER || factory.state->input_type == factory.state->value.dpr.type ) ||
		factory.state->input_kind == OFIK_PROP && factory.state->input_type == factory.state->value.prop.type &&
		factory.state->input_subtype == factory.state->value.prop.subtype ||
		factory.state->input_kind == OFIK_SEL || factory.state->input_kind == OFIK_TRANSFORM ||
		factory.state->input_kind == OFIK_TF );
	//TODO: lepsi assert, jestli jsou nenulove prisl. pointry a tak

	factory.state->flags |= OFF_FULL;
	
	if ( !(factory.flags & OFF_OPERATION_ACTIVE) )
	{
		factory.flags |= OFF_OPERATION_ACTIVE; //operation becomes active after the first step
		// update the pointer in case that someone has changed undo history since the last mode switch
		factory.start_state->prev_undo_item = factory.tlo->undo_present;
	}

	TRANS_BEGIN_MAIN( factory.tlo, factory.desc->text, err )
		// Some not yet existing hangers may be passed as something else
		if ( (factory.state->input_kind == OFIK_DPR && factory.state->input_type == DIST_PASS_HANGER) )
			reform_input_to_hanger();

		if ( factory.state->action_func ) factory.state->action_func( factory.state );
	TRANS_FAILED
		success = 0;
		statusbar_err(NULL, "Invalid thing chosen: %s Try another.", err );
		if ( factory.hanger_cursor && !tsort_on )
			TSORT_ON; //the trans_fail came while tsort was temporarily switched off
		factory_clean_state(factory.state);
	TRANS_END

	if ( success )
	{
		factory_merge_trans_from( factory.state ); //leave just one item after the state
		factory_set_state( factory.state->next ); //all succeeded, go to next state
	}
}

void factory_clean_state( struct of_state * state )
{
	if ( state->cleanup_func )
		state->cleanup_func( state );

	if ( state->flags & OFF_FULL )
	{
		//clean the temporary gos
		for (uns u = 0; u < OFR_CNT; u++ )
			if ( state->tmp_go[u] )
			{
				go_unref( state->tmp_go[u] );
				state->tmp_go[u] = NULL;
			}
		if ( state->input_kind == OFIK_DPR )
			switch ( state->value.dpr.type )
			{
			case DIST_PASS_HANGER:
				ASSERT( state->value.dpr.data.hanger.hanger );
				hanger_unref( state->value.dpr.data.hanger.hanger );
				state->value.dpr.data.hanger.hanger = NULL;
				break;
			case DIST_PASS_GO:
				ASSERT( state->value.dpr.data.go.go );
				go_unref( state->value.dpr.data.go.go );
				state->value.dpr.data.go.go = NULL;
				break;
			case DIST_PASS_ANCHOR:
				ASSERT( state->value.dpr.data.anchor.anchor );
				go_unref( anchor_get_go(state->value.dpr.data.anchor.anchor) );
				state->value.dpr.data.anchor.anchor = NULL;
			default:;
			}
		state->flags &= ~OFF_FULL;
	}

	if ( state->flags & OFF_START_MOVE )
		if ( moving_mode == MOVE_FACTORY ) factory_inner_moving_end();	
}

//FIXME: fuuuuj, co je tohle za hack?!?
struct go * factory_get_created_go(void)
{
	if ( !(factory.flags & OFF_OPERATION_ACTIVE) ) return NULL;
	if ( factory.start_state->next->tmp_go[OFR_DISPLAY] ) return factory.start_state->next->tmp_go[OFR_DISPLAY];
	if ( factory.start_state->tmp_go[OFR_DISPLAY] ) return factory.start_state->tmp_go[OFR_DISPLAY];
	return NULL;
}

#define WALK_FACTORY_VIEW_START \
	for ( GSList * l = view_list; l; l = l->next ) \
	{ \
		struct view * vw = (struct view*)l->data; \
		if ( vw->tlo != factory.tlo ) continue

#define WALK_FACTORY_VIEW_END }


/** Print the message into the factory statusbar in all views. */
void factory_msg(const char * text)
{
	WALK_FACTORY_VIEW_START;
		gtk_statusbar_pop(GTK_STATUSBAR(vw->factory_statusbar), vw->context_id);
		gtk_statusbar_push(GTK_STATUSBAR(vw->factory_statusbar), vw->context_id, text);
	WALK_FACTORY_VIEW_END;
}

static void on_factory_prop_ok( GtkWidget * w UNUSED, gpointer p UNUSED )
{
	ASSERT(factory.state->input_kind == OFIK_PROP && factory.state->value.prop.key);
	struct o * o = PROP_STORE_O(ps_global);
	struct prop * prop = prop_find(o, factory.state->value.prop.key);
	ASSERT(prop);
	factory.state->value.prop.value = prop_get_value( o, prop );
	factory.state->value.prop.type = prop->type;
	factory.state->value.prop.subtype = prop->subtype;
	factory_op_step();
}

static void factory_prop_change( struct o *o, struct go_hook* h UNUSED )
{
	if (o == PROP_STORE_O(ps_global))
	{ WALK_FACTORY_VIEW_START;
		ASSERT(vw->pi);
		prop_sync(vw->pi);
	WALK_FACTORY_VIEW_END; }
}
static void factory_prop_edit_create(struct of_state * state)
{
	ASSERT(state->input_kind == OFIK_PROP && state->value.prop.key);
	struct o * o = PROP_STORE_O(ps_global);
	go_hook_add( PROP_STORE_TLO(ps_global), &go_hook );
	struct prop * prop;
	if ( !(prop = prop_find(o, state->value.prop.key)) ||
		prop->type != state->input_type || prop->subtype != state->input_subtype )
	{
		char buf[TRANS_ERR_SIZE];
		TRANS_BEGIN_MAIN( PROP_STORE_TLO(ps_global), "Create recycler property", buf )
			prop_create( o, state->value.prop.key, state->input_type, state->input_subtype, PROP_UNIT_MAX, state->value.prop.value );
		TRANS_FAILED
			statusbar_err(NULL, "%s\n", buf);
		TRANS_END
		prop = prop_find(o, state->value.prop.key);
	}
	ASSERT(prop);

	WALK_FACTORY_VIEW_START;

		ASSERT(!(vw->factory_box_full));
		ASSERT(vw->pi);
		prop_item_init(vw->pi, o, state->value.prop.key );
		vw->pi->flags |= PIF_FACTORY;

		gtk_box_pack_start( GTK_BOX(vw->factory_box), vw->pi->value_edit, FALSE, FALSE, 0 );
		if (GTK_IS_ENTRY(vw->pi->value_edit))
			g_object_set( G_OBJECT(vw->pi->value_edit), "activates_default", TRUE, NULL );
		gtk_widget_grab_focus(vw->pi->value_edit);
		if ( vw->pi->unit_edit) gtk_box_pack_start( GTK_BOX(vw->factory_box), vw->pi->unit_edit, FALSE, FALSE, 0 );

		GtkWidget * button = gtk_button_new_with_label( "OK" );
		g_signal_connect( G_OBJECT(button), "clicked", G_CALLBACK(on_factory_prop_ok), NULL );
		gtk_box_pack_start( GTK_BOX(vw->factory_box), button, FALSE, FALSE, 0 );
		GTK_WIDGET_SET_FLAGS( button, GTK_CAN_DEFAULT );
		gtk_widget_grab_default( button );
		/*button = gtk_button_new_with_label( "Back" );
		gtk_widget_set_sensitive( button, state->prev != &ofs_start );
		g_signal_connect( G_OBJECT(button), "clicked", G_CALLBACK(on_step_back), NULL );
		gtk_box_pack_start( GTK_BOX(vw->factory_box), button, FALSE, FALSE, 0 );*/

		gtk_widget_show_all( vw->factory_box );
		vw->factory_box_full = 1;
		// remove temporarily the accelerators because of conflicts with Delete and
		// other keys.
		DBG("Remove accel group for window %p\n", vw->window );
		gtk_window_remove_accel_group( GTK_WINDOW(vw->window), accel_group );

	WALK_FACTORY_VIEW_END;
}

static void factory_prop_widget_destroy( GtkWidget * w, gpointer data UNUSED ){ gtk_widget_destroy(w); }
static void factory_prop_edit_destroy(void)
{
	go_hook_remove( PROP_STORE_TLO(ps_global), &go_hook );
	WALK_FACTORY_VIEW_START;
		if (vw->factory_box_full)
		{
			gtk_container_foreach( GTK_CONTAINER(vw->factory_box), factory_prop_widget_destroy, NULL );
			// restore the original accel group removed by factory_prop_edit_create.
			DBG("Add accel group for window %p\n", vw->window );
			gtk_window_add_accel_group( GTK_WINDOW(vw->window), accel_group );
			vw->factory_box_full = 0;
		}
	WALK_FACTORY_VIEW_END;
}

/*** PATHS *****/
static void factory_path_end(int close)
{
	if (!factory.path)
		return;

	ASSERT( factory.path && (factory.flags & OFF_PATH_ACTIVE) );
	factory_op_break();

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( factory.tlo, "End path", buf )
		
		if ( !group_get_first_go( &factory.path->group ) )
			go_unlink( &factory.path->group.go );
		else
		{
			if (close)
			{
				DBG("Close path\n");
				struct hanger * start = go_get_hanger_start(group_get_first_go(&factory.path->group));
				struct hanger * end = go_get_hanger_end(group_get_last_go(&factory.path->group));
				ASSERT( start && end );
				go_link_tail( &go_segment_new(end, start)->go, &factory.path->group );
			}
			change_context( context.w, factory.group, go2o(&factory.path->group.go) );
		}

	TRANS_FAILED
		statusbar_err( NULL, "Cannot end path: %s", buf );
	TRANS_END
	go_unref( &factory.path->group.go );
	factory.path = NULL;
	factory.flags &= ~OFF_PATH_ACTIVE;
}
static int factory_path_new(void)
{
	ASSERT( !factory.path && !(factory.flags & OFF_PATH_ACTIVE) );
	char buf[TRANS_ERR_SIZE];
	volatile int success = 1;
	TRANS_BEGIN_MAIN( factory.tlo, "New path", buf )
		factory.path = go_path_new();
		gui_prop_recycle( go2o(&factory.path->group.go) );
		LINK_HEAD( &factory.path->group.go );
		go_ref(&factory.path->group.go);
	
		if (factory.path)
			factory.flags |= OFF_PATH_ACTIVE;
	TRANS_FAILED
		statusbar_err( NULL, "Cannot create a new path: %s", buf );
		factory.path = 0;
		success = 0;
	TRANS_END

	return !success;
}
void on_path_switch( void * subj UNUSED )
{
	ASSERT( factory.tlo && factory.group );
	ASSERT( !(factory.flags & OFF_OPERATION_ACTIVE) );
	
	if ( factory.flags & OFF_PATH_ON )
	{
		factory_path_end(0);
		factory.flags &= ~OFF_PATH_ON;
	}
	else
	{
		if (factory_path_new())
			return;
		factory.flags |= OFF_PATH_ON;
		if ( factory.path && !(factory.start_state->flags & OFF_PATHABLE ))
			factory_op_start( factory.group, &ofs_segment_1, "Create a new segment" );
		
	}
}
void on_path_end_close( void * subj UNUSED )
{ factory_path_end(1); factory_path_new(); }

void on_path_end( void * subj UNUSED )
{ factory_path_end(0); factory_path_new(); }

/**** CREATION OF GO ****/

void snap_point( struct geom_point * pos, struct dist_pass_result * dpr, struct geom_transform2 * grid, struct obj_tlo * tlo, real maxdist )
{
	volatile struct geom_point minp;
	volatile real mindist = 0.0, dist;
	volatile int isnotset = 1;

	struct dist_pass_result info;
	struct center_pass cp;
	char err[TRANS_ERR_SIZE];
	int valid;
	volatile uns pass_flags = 0;
	if ( factory.snap_set & SNAP_HANGERS ) pass_flags |= DIST_PASS_HANGERS;
	if ( factory.snap_set & SNAP_LINES ) pass_flags |= DIST_PASS_GOS;
	if ( factory.snap_set & SNAP_INTERS ) pass_flags |= DIST_PASS_INTERSECTIONS;

	if ( pass_flags )
	{
		TRANS_BEGIN_MAIN( tlo, "Snap the point", err )
			center_pass_init(&cp, CENTER_PASS_FLAGS | pass_flags, NULL, pos, maxdist, tlo);
			while ( valid = center_pass_next( &cp, &info ) )
			{
				switch (info.type)
				{
				case DIST_PASS_GO:
					if (tsort_is_go_presorted(info.data.go.go)) continue;
					break;
				case DIST_PASS_HANGER:
					if (!hanger_is_position(info.data.hanger.hanger)) continue;
					if (tsort_is_go_presorted(hanger_get_go(info.data.hanger.hanger))) continue;
					break;
				case DIST_PASS_INTERSECTION:
					if (tsort_is_go_presorted(info.data.intersection.go1) ||
						tsort_is_go_presorted(info.data.intersection.go2)) continue;
					break;
				default:;
				}
				break; // if none of the bad cases matches, we have found a good one
			}
			if (valid)
			{
				isnotset = 0;
				switch (info.type)
				{
				case DIST_PASS_GO:
					minp = info.data.go.geom.point;
					mindist = info.data.go.geom.dist;
					DBG("Snapping to go %p, time = %.3f\n", info.data.go.go, info.data.go.geom.time );
					break;
				case DIST_PASS_HANGER:
					minp = hanger_get_position(info.data.hanger.hanger);
					mindist = info.data.hanger.dist;
					DBG("Snapping to hanger %p\n", info.data.hanger.hanger );
					break;
				case DIST_PASS_INTERSECTION:
					minp = info.data.intersection.geom.point;
					mindist = info.data.intersection.dist;
					DBG("Snapping to intersection of %p and %p\n", info.data.intersection.go1, info.data.intersection.go2);
					break;
				default:;
				}
			}
			center_pass_destroy(&cp);
		TRANS_FAILED
			statusbar_err(NULL, "Cannot snap: %s", err);
		TRANS_END
	}
	
	if ( factory.snap_set & SNAP_GRID ){
		struct geom_point gridpt;
		int res = geom_point_transform( pos, &grid->inverted, &gridpt );
		if ( res == 0 )
		{
			gridpt.x = floor( gridpt.x + 0.5 );
			gridpt.y = floor( gridpt.y + 0.5 );
			res = geom_point_transform( &gridpt, &grid->primary, &gridpt );
		} else
			statusbar_err(NULL, "snapping to grid: geom_transform failed");

		if ( res == 0 )
		{
			dist = hypot( pos->x - gridpt.x, pos->y - gridpt.y );
			if ( isnotset && dist < maxdist || !isnotset && dist < mindist )
			{
				mindist = dist; minp = gridpt;
				isnotset = 0;
				info.type = DIST_PASS_POINT;
				info.data.point.point = gridpt; info.data.point.dist = dist;
				DBG("Snapping to grid point %.3f, %.3f\n", gridpt.x, gridpt.y );
			}
		} else
			statusbar_err(NULL, "snapping to grid: geom_transform back failed");
	}
	
	// If no suitable result has been found, fall back to the original position.
	if ( isnotset )
	{
		mindist = 0.0; minp = *pos;
		info.type = DIST_PASS_POINT; info.data.point.point = minp;
		info.data.point.dist = 0.0;
	}
	DBG("Resulting point [%.3f, %.3f] in dist = %.3f\n", minp.x, minp.y, mindist);
	if (dpr) *dpr = info;
	*pos = minp;
}

void snap_to_go( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist )
{
	struct dist_pass_result info;
	struct center_pass cp;
	char err[TRANS_ERR_SIZE];
	int valid;
	TRANS_BEGIN_MAIN( tlo, "Find the nearest go", err )
		center_pass_init(&cp, CENTER_PASS_FLAGS | DIST_PASS_GOS, NULL, pos, maxdist, tlo);
		while ( (valid = center_pass_next( &cp, &info )) && tsort_is_go_presorted(info.data.go.go) );
		if (valid)
		{
			struct geom_point ap = info.data.go.geom.point;
			DBG("Snapped point %.3f, %.3f to %.3f, %.3f (go %p), time %.3f\n", pos->x, pos->y, ap.x, ap.y,
				info.data.go.go, info.data.go.geom.time );
			*pos = ap;
		}
		else
		{
			DBG("No near go found.\n");
			info.type = DIST_PASS_POINT;
			info.data.point.point = *pos; info.data.point.dist = 0.0;
		}
		if (dpr) *dpr = info;
		center_pass_destroy(&cp);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot snap: %s", err);
	TRANS_END
}

void snap_to_anchor( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist )
{
	struct dist_pass_result info;
	struct center_pass cp;
	char err[TRANS_ERR_SIZE];
	int valid;
	TRANS_BEGIN_MAIN( tlo, "Find the nearest anchor", err )
		center_pass_init(&cp, CENTER_PASS_FLAGS | DIST_PASS_ANCHORS, NULL, pos, maxdist, tlo);
		while ( (valid = center_pass_next( &cp, &info )) &&
			(tsort_is_go_presorted(anchor_get_go(info.data.anchor.anchor)) ||
			 !go_is_selected(anchor_get_go(info.data.anchor.anchor)) ||
			 !hanger_is_position(info.data.anchor.anchor->h))); // FIXME: is this necessary?
		if (valid)
		{
			struct geom_point ap = hanger_get_position(info.data.anchor.anchor->h);
			DBG("Snapped point %.3f, %.3f to %.3f, %.3f (anchor %p) over distance %.3f\n",
				pos->x, pos->y, ap.x, ap.y, info.data.anchor.anchor, info.data.anchor.dist );
			*pos = ap;
		}
		else
		{
			DBG("No anchor found.\n");
			info.type = DIST_PASS_POINT;
			info.data.point.point = *pos; info.data.point.dist = 0.0;
		}
		if (dpr) *dpr = info;
		center_pass_destroy(&cp);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot snap: %s", err);
	TRANS_END
}

void fix_anchor_2( struct of_state * state )
{
	ASSERT( state->prev->tmp_go[OFR_DISPLAY] );
	anchor_rehang( go_get_anchor(state->prev->tmp_go[OFR_DISPLAY], 1), state->value.dpr.data.hanger.hanger );
}
void fix_anchor_3( struct of_state * state )
{
	ASSERT( state->prev->prev->tmp_go[OFR_DISPLAY] );
	anchor_rehang( go_get_anchor(state->prev->prev->tmp_go[OFR_DISPLAY], 2), state->value.dpr.data.hanger.hanger );
}
void fix_anchor_4( struct of_state * state )
{
	ASSERT( state->prev->prev->prev->tmp_go[OFR_DISPLAY] );
	anchor_rehang( go_get_anchor(state->prev->prev->prev->tmp_go[OFR_DISPLAY], 3), state->value.dpr.data.hanger.hanger );
}
static void perturb_cursor(void)
{
	struct geom_point po = hanger_get_position(factory.hanger_cursor);
	po.x += factory.snap_tolerance;
	po.y += factory.snap_tolerance;
	hanger_mc_move( factory.hanger_cursor, &po );
}

/* transformation */

void do_transform_1( struct of_state * s UNUSED )
{
	ASSERT(factory.group);
	tsort_selected_flag_presort( factory.group, factory.tlo );
}

void do_transform_2( struct of_state * s UNUSED ) { }

static void clean_transform( struct of_state * s UNUSED )
{
	view_moving_end();
	tsort_presort_end( factory.tlo );
}

/* Rehanging of anchors */

void do_rehang_1( struct of_state * s UNUSED )
{
	struct anchor * a = ofs_rehang_1.value.dpr.data.anchor.anchor;
	struct go * go = anchor_get_go(a);
	go_ref(go);
	factory_inner_moving_start( hanger_get_position(a->h) );
	tsort_presort_end( factory.tlo ); // FIXME: Volat tsort_presort na prazdnou mnozinu (na nove vytvorene mysi nic nevisi) a vzdy hned tsort zrusit mi pripada divne.
	tsort_go_presort( go, factory.tlo );
	anchor_rehang( a, factory.hanger_cursor );
	tsort_presorted_recompute( factory.tlo ); // Pokud je mys na stejnem miste jako puvodni hanger, tak je to mozna zbytecne.
}

void do_rehang_2( struct of_state * s UNUSED )
{
	anchor_rehang( ofs_rehang_1.value.dpr.data.anchor.anchor, ofs_rehang_2.value.dpr.data.hanger.hanger );
}

#define GET_HANGER(_h) struct hanger * _h = state->value.dpr.data.hanger.hanger

#define STICK_HANGER_TO_PATH_TAIL(_h) \
	struct hanger * path_tail = NULL; \
	struct go * path_last_go; \
	if ((factory.flags & OFF_PATH_ACTIVE) && (path_last_go = group_get_last_go(&factory.path->group)) && \
		(path_tail = go_get_hanger_end(path_last_go))) \
	{ \
		ASSERT(state->input_kind == OFIK_DPR && state->input_type == DIST_PASS_HANGER); \
		ASSERT(state->next->input_kind == OFIK_DPR && state->next->input_type == DIST_PASS_HANGER); \
		state->next->value = state->value; \
		_h = path_tail; \
	}

#define MS(_type, _h) \
	struct _type * g; \
	ASSERT( state->flags & OFF_START_MOVE ); \
	factory_inner_moving_start( hanger_get_position(_h) ); \
	struct hanger * hc = factory.hanger_cursor

static void factory_link( struct go * go )
{
	if ( factory.flags & OFF_PATH_ON )
	{
		ASSERT(factory.start_state->flags & OFF_PATHABLE );
		ASSERT(factory.path);
		go_link_tail( go, &factory.path->group );
	} else
		LINK_HEAD(go);
}

#define ML \
	state->tmp_go[OFR_DISPLAY] = &g->go; \
	factory_link( go_ref( &g->go ) ); \
	gui_prop_recycle( go2o(&g->go) ); \
	change_context( context.w, factory.group, go2o(&g->go) )

#define MLT \
	state->tmp_go[OFR_DISPLAY] = &g->go; \
	TSORT_OFF; \
	factory_link( go_ref( &g->go ) ); \
	TSORT_ON; \
	gui_prop_recycle( go2o(&g->go) ); \
	change_context( context.w, factory.group, go2o(&g->go) )

/* Point */
void create_point( struct of_state * state )
{
	struct go_point * g = go_point_new( ofs_point.value.dpr.data.hanger.hanger );
	ML;
}

void create_decorator_pt( struct of_state * state )
{
	struct go_decorator_point * g = go_decorator_point_new( state->value.dpr.data.hanger.hanger, 0, 2.0, 0.0 );
	ML;
}
static void create_arrow( struct of_state * state )
{
	struct go_arrow * g = go_arrow_new(state->value.dpr.data.go.go, state->value.dpr.data.go.geom.time, 10.0, 0.0 );
	ML;
}
static void create_intersection( struct of_state * state UNUSED )
{
	struct go_intersection * g = go_intersection_new( ofs_point_inters.value.dpr.data.go.go,
		ofs_point_inters_2.value.dpr.data.go.go, ofs_point_inters.value.dpr.data.go.geom.time );
	ML;
}

/* Experimental ngon */
void ngon_check_apex_number( struct of_state * state )
{
	uns n = state->value.prop.value.u;
	if ( n < 3 )
		trans_fail("Invalid number of apices.");
}

void create_ngon( struct of_state * state )
{
	uns n = ofs_ngon_1.value.prop.value.u;
	ASSERT( n >= 3 );

	//a temporary circle with n parametric points spread equally along the circle
	struct go_elarc * circ = go_elarc_new_xy1ecc( state->prev->value.dpr.data.hanger.hanger,
		state->value.dpr.data.hanger.hanger, 0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
	LINK_HEAD(&circ->go);
	
	uns h = 0;
	if ( n == 56109 ) { h = 1; n = 5; }
	if ( n >= 100 )
	{
		gui_prop_recycle( go2o(&circ->go) );
		change_context( context.w, factory.group, go2o(&circ->go) );
		return;
	}

	struct go * apices[n];
	for (uns u = 0; u < n; u++ )
	{
		apices[u] = &go_parametric_point_new( &circ->go, ((real)u) / n )->go;
		LINK_HEAD(apices[u]);
	}
	struct go_path * path = go_path_new();
	LINK_HEAD(&path->group.go);

	if (h)
	{
		struct go * go;
		go_link_tail( go = &go_bezier_new_cubic( go_get_hanger(apices[0], 0), 1.0, go_get_hanger(apices[1], 0), 1.5,
				go_get_hanger(apices[2], 0), 1.0, go_get_hanger_by_name(&circ->go, string_lookup("center")), 1.0 )->go,
				&path->group );
		gui_prop_recycle( go2o(go) );
		go_link_tail( go = &go_bezier_new_cubic( go_get_hanger_end(go), 1.0, go_get_hanger(apices[3], 0), 1.0,
				go_get_hanger(apices[4], 0), 1.5, go_get_hanger(apices[0], 0), 1.0 )->go, &path->group );
		gui_prop_recycle( go2o(go) );
	} else {
		struct go * segment = NULL;
		struct hanger * h1, * h2;
		for (int i = n - 1; i >= 0; i-- )
		{
			h1 = ((uns)i == n - 1) ? go_get_hanger(apices[0], 0) : go_get_hanger_end(segment);
			h2 = go_get_hanger(apices[i], 0);
			ASSERT( h1 && h2 );
			go_link_tail( segment = &go_segment_new( h1, h2 )->go, &path->group );
			gui_prop_recycle( go2o(segment) );
		}
	}

	go_remove_rehang(&circ->go);
	for (uns u = 0; u < n; u++ ) go_remove_rehang(apices[u]);
	gui_prop_recycle( go2o(&path->group.go) );
	change_context( context.w, factory.group, go2o(&path->group.go) );
}

/* Segment */
void create_segment_1( struct of_state * state )
{
	GET_HANGER(h);
	STICK_HANGER_TO_PATH_TAIL(h);
	MS(go_segment, h);
	g = go_segment_new( h, hc );
	MLT;
}

/* Cubic Bezier */
void create_bez_cub_1( struct of_state * state )
{
	GET_HANGER(h);
	STICK_HANGER_TO_PATH_TAIL(h);
	MS(go_bezier, h);
	g = go_bezier_new_cubic( h, 1.0, hc, 1.0, hc, 1.0, hc, 1.0 );
	MLT;
}

/* Quadratic Bezier */
void create_bez_quad_1( struct of_state * state )
{
	GET_HANGER(h);
	STICK_HANGER_TO_PATH_TAIL(h);
	MS(go_bezier, h);
	g = go_bezier_new_quadratic( h, 1.0, hc, 1.0, hc, 1.0 );
	MLT;
}

/* Arc by three points */
void create_arc_3p_1( struct of_state * state UNUSED )
{
	GET_HANGER(h);
	STICK_HANGER_TO_PATH_TAIL(h);
	MS(go_elarc, h);
	perturb_cursor();
	g = go_elarc_new_3ecc( h, hc, hc, 0.0, 0.0, GEOM_CONIC_ARC_MIDDLE );
	MLT;
}
/* Circle/circular arc by center and radius */
void create_arc_center_radius_1( struct of_state * state )
{
	struct go_elarc * g = go_elarc_new_xyr( state->value.dpr.data.hanger.hanger, 0.0, 10.0, 10.0, 0.0,
		(state == &ofs_arc_center_radius_1) ? 2.0 : R_2PI, GEOM_CONIC_ARC_START_DIF );
	ML;
}
/* Circle by three points */
void create_circ_3p_1( struct of_state * state UNUSED )
{
	GET_HANGER(h);
	MS(go_elarc, h);
	//perturb_cursor();
	g = go_elarc_new_3ecc( h, hc, hc, 0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}

/* Circle by center and point */
void create_circ_center_pt_1( struct of_state * state UNUSED )
{
	GET_HANGER(h);
	MS(go_elarc, h);
	g = go_elarc_new_xy1ecc( h, hc, 0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}

/* Smallest ellipse by 3 points */
void create_ellipse_3p_smallest_2( struct of_state * state UNUSED )
{
	TSORT_OFF;
	go_unlink( state->prev->tmp_go[OFR_DISPLAY] );//unlink the old go
	perturb_cursor();
	struct go_elarc * g = go_elarc_new_3small( state->prev->value.dpr.data.hanger.hanger,
		state->value.dpr.data.hanger.hanger, factory.hanger_cursor, GEOM_CONIC_ARC_POINT_ENTIRE );
//	go_link( go_ref( &g->go ), factory.group, NULL );
//	change_context( context.w, factory.group, go2o(&g->go) );
	ML;
	TSORT_ON;
}
/* Ellipse by 3 points, rotation and eccentricity */
void create_ellipse_3p_ecc_1( struct of_state * state UNUSED )
{
	GET_HANGER(h);
	MS(go_elarc, h);
	perturb_cursor();
	g = go_elarc_new_3ecc( h, hc, hc, 0, 0.5, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}
/* Ellipse/elliptic arc by center and radiii */
void create_ellarc_center_radii_1( struct of_state * state )
{
	struct go_elarc * g = go_elarc_new_xyr( state->value.dpr.data.hanger.hanger, 0.0, 20.0, 10.0, 0.0,
		(state == &ofs_ellarc_center_radii_1) ? 2.0 : R_2PI, GEOM_CONIC_ARC_START_DIF );
	ML;
}
/* Ellipse by center, point, rotation and eccentricity */
void create_ellipse_xy1ecc_1( struct of_state * state UNUSED )
{
	GET_HANGER(h);
	MS(go_elarc, h);
	perturb_cursor();
	g = go_elarc_new_xy1ecc( h, hc, 0.0, 0.5, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}
/* Ellipse by foci and point */
void create_ellipse_segment( struct of_state * state )
{
	GET_HANGER(h);
	MS(go_segment, h);
	g = go_segment_new( h, hc );
	state->tmp_go[OFR_DISPLAY] = &g->go;
	TSORT_OFF;
	go_link_head( go_ref( &g->go ), factory.group );
	TSORT_ON;
}

void create_ellipse_foci_pt_2( struct of_state * state )
{
	TSORT_OFF;
	go_unlink( state->prev->tmp_go[OFR_DISPLAY] );//unlink the old go
	struct go_elarc * g = go_elarc_new_foci( state->prev->value.dpr.data.hanger.hanger,
		state->value.dpr.data.hanger.hanger, factory.hanger_cursor, GEOM_CONIC_ARC_START_ENTIRE );
	ML;
	TSORT_ON;
}
void fix_ellipse_anchor_2( struct of_state * state )
{
	struct go_elarc * c = (struct go_elarc*)state->prev->tmp_go[OFR_DISPLAY];
	anchor_rehang( go_get_anchor(&c->go, 2), state->value.dpr.data.hanger.hanger );
}

/* Text and TeX text */
void create_text_1( struct of_state * state )
{
	struct go_text * g = go_text_new( state->value.dpr.data.hanger.hanger, 0.5, 0.5, 0.0, 0.0,
		string_lookup(""), fonts_get_default(), 10);
	ML;
	text_edit_create( go2o(&g->go), string_text );
}
void create_tex_text_1( struct of_state * state )
{
	struct go_tex_text * g = go_tex_text_new( state->value.dpr.data.hanger.hanger,
		0.5, 0.5, 0.0, 0.0, string_lookup("") );
	ML;
	text_edit_create( go2o(&g->go), string_text );
}

/****************  Random things **************/

static inline real random_coord(void)
{ return ((random () * 8.0) / RAND_MAX) - 4; }

static inline struct hanger* random_hanger(void)
{ return hanger_new_mc_xy( random_coord(), random_coord() ); }

static void create_random_point( struct go_group * group, char * err )
{
	struct go_point *pt;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Random point", err )
		pt = go_point_new( random_hanger() );
		go_link_head( &pt->go, group );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot add a random point: %s", err);
	TRANS_END
}

static void create_random_bezier_quad( struct go_group * group, char * err )
{
	struct go_bezier * gs;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Random quadratic bezier", err )
		gs = go_bezier_new_quadratic( random_hanger(), 1, random_hanger(), 1, random_hanger(), 1 );
		go_link_head( &gs->go, group );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot add a random bezier: %s", err);
	TRANS_END
}

static void create_random_bezier_cubic( struct go_group * group, char * err )
{
	struct go_bezier * gs;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Random cubic bezier", err )
		gs = go_bezier_new_cubic( random_hanger(), 1, random_hanger(), 1, random_hanger(), 1, random_hanger(), 1 );
		go_link_head( &gs->go, group );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot add a random bezier: %s", err);
	TRANS_END
}

static void create_random_segment( struct go_group * group, char * err )
{
	struct go_segment * gs;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Random segment", err )
		gs = go_segment_new( random_hanger(), random_hanger() );
		go_link_head( &gs->go, group );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot add a random segment: %s", err);
	TRANS_END
}

#define RANDOM_THING(f,n) \
void f( void * subj ) \
{ \
	struct go * grp; \
	char err[TRANS_ERR_SIZE]; \
	if ( subj ) \
	{ \
		grp = void2go(subj); \
		ASSERT_GROUP(grp); \
		n((struct go_group*)grp, err); \
	} else WALK_GROUP_SELECTED_GO(tlo_get_tlg(context.tlo), grp) n((struct go_group*)grp, err); \
}

RANDOM_THING(on_random_point, create_random_point)
RANDOM_THING(on_random_bezier_quad, create_random_bezier_quad)
RANDOM_THING(on_random_bezier_cubic, create_random_bezier_cubic)
RANDOM_THING(on_random_segment, create_random_segment)
