/** @file
 *
 * GUI - Maintainance of the font list. Implementation of the Text and TeX text editors.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <values.h>
#include <fontconfig/fontconfig.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <iconv.h>
#include <errno.h>
#include <langinfo.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "font/font.h"
#include "gui/main.h"
#include "gui/dialogs.h"
#include "gui/properties.h"

GtkListStore * font_list;

GSList * text_edit_list;
static int last_used_font;

static void text_edit_destroy( GtkWidget * w, struct text_edit * te );
static void text_edit_go_del( struct o * o, struct o *owner UNUSED, struct o *prev UNUSED, struct go_hook* h );
static void text_edit_obj_del( struct o *o, struct o *owner UNUSED, struct o *prev UNUSED, struct obj_hook* h );
static void text_edit_go_prop_changed( struct o *o, struct go_hook* h );
static void text_edit_obj_prop_changed( struct o *o, struct obj_hook* h );
static void text_edit_obj_prop_deleted(struct o *o, struct prop* p, struct obj_hook* h);
static void text_edit_go_prop_deleted(struct o *o, struct prop* p, struct go_hook* h);
static void text_edit_refresh_text( GtkButton * b, struct text_edit * te );
static void text_edit_refresh_widgets( struct text_edit * te );

enum { TEM_TEXT, TEM_TEX, TEM_OTHER };
static uns prop_counts[] = { 8, 6, 0 };

/* FONT NAME LIST etc. */

void gui_fonts_finish(void)
{
	if (font_list)
	{
		g_object_unref( G_OBJECT(font_list) );
		font_list = NULL;
	}
}

/** Use fontconfig to get the list of installed fonts; fill
 * the font_list with them.
 * @rv zero if succesful
 */
int fill_font_list(void)
{
	if (font_list)
		return 1;

	font_list = gtk_list_store_new( 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING );
	gtk_tree_sortable_set_sort_column_id( GTK_TREE_SORTABLE(font_list), 1, GTK_SORT_ASCENDING );

	FcPattern * pattern = FcPatternCreate();
	if (!pattern)
		return 0;
	FcPatternAddBool (pattern, FC_SCALABLE, FcTrue);

	FcObjectSet * os = FcObjectSetCreate ();
	if (!os)
		return 0;
	FcObjectSetAdd (os, FC_FILE);
	FcObjectSetAdd (os, FC_STYLE);
	FcObjectSetAdd (os, FC_FAMILY);

	FcFontSet * set = FcFontList (0, pattern, os);
	FcPatternDestroy (pattern);
	FcObjectSetDestroy (os);
	if (!set)
		return 0;

	FcChar8 *family, *style, *fontpath;
	for (int i = 0; i < set->nfont; i++)
	{
		FcPattern * font = set->fonts[i];
		if (FcPatternGetString (font, FC_FILE, 0, &fontpath) == FcResultNoMatch)
			continue;
		if (FcPatternGetString (font, FC_FAMILY, 0, &family) == FcResultNoMatch )
			continue;
		if (FcPatternGetString (font, FC_STYLE, 0, &style) == FcResultNoMatch)
			continue;
		DBG("FILL_FONTS: inserting %s, %s, %s\n", fontpath, family, style);

		GtkTreeIter ti;
		char buf[strlen(family) + strlen(style) + 10];
		sprintf( buf, "%s (%s)", family, style );
		DBG("Buf = %s\n", buf);

		gtk_list_store_append( font_list, &ti );
		gtk_list_store_set( font_list, &ti, 0, fontpath, 1, buf, -1 );
	}
	return 1;
}

int fonts_get_default(void)
{
	if ( last_used_font )
		return last_used_font;

	if (!font_list && !fill_font_list())
		trans_fail("Could not load fonts.");

	//get the first font we have found
	GtkTreeIter ti;
	if (!gtk_tree_model_get_iter_first( GTK_TREE_MODEL(font_list), &ti ))
		trans_fail("No fonts available.");

	int font_id = fonts_iter_to_font_id(&ti);
	if (!font_id_valid(font_id))
		trans_fail("No fonts available.");
	else
		return font_id;
}

uns fonts_iter_to_font_id( GtkTreeIter * ti )
{
	SAFE_G_VALUE(fp);
	gtk_tree_model_get_value( GTK_TREE_MODEL(font_list), ti, 0, &fp);
	uns font_id = font_load_file((char *)g_value_get_string(&fp));
	g_value_unset(&fp);

	return font_id;
}

int fonts_compare_iter_with( GtkTreeIter * ti, char * filename )
{
	SAFE_G_VALUE(fp);
	gtk_tree_model_get_value( GTK_TREE_MODEL(font_list), ti, 0, &fp);
	int result =  strcmp( (char *)g_value_get_string(&fp), filename );
	g_value_unset(&fp);

	return result;
}

/* TEXT BUFFER SAVE, LOAD */

static char *get_text_from_text_buffer (GtkTextBuffer *text_buf)
{
	GtkTextIter startiter, enditer;

	gtk_text_buffer_get_start_iter (text_buf, &startiter);
	gtk_text_buffer_get_end_iter (text_buf, &enditer);
	return gtk_text_buffer_get_text (text_buf, &startiter, &enditer, 1);
}

static char *convert_text (char *text, int exporting)
{
	iconv_t conv;
	char *local_charset = nl_langinfo (CODESET);
	if (exporting)
		conv = iconv_open (local_charset, "UTF8");
	else
		conv = iconv_open ("UTF8", local_charset);
	if (conv == (iconv_t) -1) {
		statusbar_err (NULL, "Cannot initialize charset conversion.");
		return NULL;
	}
	size_t in_len = strlen (text);
   	size_t out_len = 4*in_len + 1;
	char *output = xmalloc (out_len);
	char *outpos = output;
	int res = iconv (conv, &text, &in_len, &outpos, &out_len);
	iconv_close(conv);
	if (res < 0) {
		statusbar_err (NULL, "Charset conversion failed: %s", strerror(errno));
		xfree (output);
		return NULL;
	}
	*outpos = 0;
	return output;
}

/** Save the text to a file. */
static int save_text (char *text, const char *filename)
{
	text = convert_text (text, 1);
	if (!text)
		return 1;
	int fd = open(filename, O_WRONLY | O_TRUNC | O_CREAT, 0666);
	int result = 1;
	if (fd < 0)
		statusbar_err (NULL, "Cannot open %s for writing: %s", filename, strerror (errno));
	else {
		int len = strlen (text);
		if (write (fd, text, len) == len)
			result = 0;
		else
			statusbar_err (NULL, "Error writing %s", filename);
		close (fd);
	}
	xfree (text);
	return result;
}

/** Load the text from a file. */
static char *load_text (const char *filename)
{
	int fd = open(filename, O_RDONLY);
	if (fd < 0) {
		statusbar_err (NULL, "Cannot open %s for reading: %s", filename, strerror (errno));
		return NULL;
	}
	struct stat st;
	char *result = NULL;
	if (stat (filename, &st) < 0)
		statusbar_err (NULL, "Cannot stat %s: %s", filename, strerror (errno));
	else {
		char *buf = xmalloc (st.st_size + 1);
		if (read (fd, buf, st.st_size) != st.st_size)
			statusbar_err (NULL, "Error reading %s", filename);
		else {
			buf[st.st_size] = 0;
			result = convert_text (buf, 0);
		}
		xfree (buf);
	}
	close (fd);
	return result;
}

static int save_text_edit (struct text_edit *te, const char *filename)
{
	char *text = get_text_from_text_buffer (te->text_buf);
	int res = save_text (text, filename);
	g_free (text);
	return res;
}

static int load_text_edit (struct text_edit *te, const char *filename)
{
	char *text = load_text (filename);
	if (!text)
		return 1;
	gtk_text_buffer_set_text (te->text_buf, text, -1);
	xfree (text);
	return 0;
}

static void on_save_text (GtkWidget *button UNUSED, struct text_edit *te)
{
	SAVE_DLG_START("Save text", NULL, NULL)
	{
		if (save_text_edit (te, filename))
			statusbar_err ( (struct window*)te, "Couldn't save text into the file.");
		else
			statusbar_msg ( (struct window*)te, "File saved successfully.");
	}
	SAVE_DLG_END;
}

static void on_load_text (GtkWidget *button UNUSED, struct text_edit *te)
{
	OPEN_DLG_START( "Open text", NULL, NULL )
	{
		if (load_text_edit (te, filenames[0]))
			statusbar_err ( (struct window*)te, "Couldn't load text from the file.");
		else
			statusbar_msg ( (struct window*)te, "File loaded successfully.");
		break; //load only the first selected file
	}
	OPEN_DLG_END;
}

/** Open external editor on the text contained in the buffer, wait for the
 * editor to finish and load the text again.
 */
static void open_external_editor (GtkWidget *button UNUSED, struct text_edit *te)
{
	const char * editor = prop_lookup_text( obj2o(obj_universum), string_editor );
	if (!editor) { statusbar_err( (struct window*)te, "Please set the text editor in the Settings first!"); return; }
	uns run_in_term = prop_lookup_uns( obj2o(obj_universum), string_editor_in_term, 2 );
	if (run_in_term == 2) { statusbar_err( (struct window*)te, "Please set if to run the text editor in a terminal first!"); return; }
	const char * term = NULL;
	if (run_in_term)
	{
		term = prop_lookup_text( obj2o(obj_universum), string_terminal );
		if (!term) { statusbar_err((struct window*)te, "Please set the terminal in the Settings first!"); return; }
	}

	char *tmpdir = getenv ("TMPDIR");
	if (!tmpdir)
		tmpdir = "/tmp";
	char *filename = alloca (strlen (tmpdir) + 11);
	sprintf (filename, "%s/vrrXXXXXX", tmpdir);
	int fd = mkstemp(filename);
	if (fd < 0) { statusbar_err((struct window*)te, "Cannot create temporary file."); return; }
	close (fd);

	if (save_text_edit (te, filename))
		goto remove_file;

	pid_t pid;
	switch (pid = fork())
	{
	case -1: statusbar_err((struct window*)te, "Cannot fork!"); break;
	case 0:
		msg("About to run editor on file %s", filename );
		if (run_in_term)
			execlp( term, term, "-e", editor, filename, NULL );
		else
			execlp( editor, editor, filename, NULL );
		statusbar_err((struct window*)te, "Could not run the editor %s on %s: %s", editor, filename, strerror(errno));
		exit(1);

	default:;
		int status; waitpid(pid, &status, 0); // TODO: something better! This makes vrr freeze completely!
		if (!WIFEXITED(status) || WEXITSTATUS(status))
			statusbar_err((struct window*)te, "Editor failed (exited with code %d)", status);
		else
			load_text_edit (te, filename);
	}

remove_file:
	unlink (filename);
}

/* TEXT EDITOR WINDOW */

static void text_edit_refresh_text( GtkButton * b UNUSED, struct text_edit * te )
{
	ASSERT( te );
	char err[TRANS_ERR_SIZE];

	statusbar_clear((struct window*)te);
	char *s = get_text_from_text_buffer (te->text_buf);

	TRANS_BEGIN_MAIN( te->parent_tlo, "Refresh text", err)
		prop_set_text( te->o, te->text_key, s);
	TRANS_FAILED
		statusbar_err((struct window*)te, "Cannot refresh text: %s", err);
	TRANS_END

	g_free(s);
}

static void text_edit_set_window_title( struct text_edit * te )
{
	string s = o_get_name(te->o);
	char * name = (s == string_noname) ? "" : s->text;
	const char * tn = ((te->o->kind == T_OBJ) ? string_obj_type : string_go_type)[te->o->type]->text;
	char buf[strlen(te->text_key->text) + strlen(tn) + strlen(name) + 20];
	sprintf( buf, "%s of %s %s", te->text_key->text, tn, name );
	gtk_window_set_title( GTK_WINDOW(te->window), buf );
}

static void text_edit_refresh_widgets( struct text_edit * te )
{
	const char * text  = prop_lookup_text( te->o, te->text_key );

	if ( text )
		gtk_text_buffer_set_text( te->text_buf, text, -1 );

	for (uns u = 0; u < te->prop_count; u++ )
		prop_sync( te->pi + u );
	text_edit_set_window_title(te);
}

static void text_edit_go_prop_changed( struct o *o, struct go_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( te->o == o ) text_edit_refresh_widgets(te);
}
static void text_edit_obj_prop_changed( struct o *o, struct obj_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( te->o == o ) text_edit_refresh_widgets(te);
}

static void text_edit_obj_prop_deleted(struct o *o, struct prop* p, struct obj_hook* h)
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( o == te->o && p->key == te->text_key )
		gtk_widget_destroy( te->window );
}
static void text_edit_go_prop_deleted(struct o *o, struct prop* p, struct go_hook* h)
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( o == te->o && p->key == te->text_key )
		gtk_widget_destroy( te->window );
}
static void text_edit_obj_del( struct o *o, struct o *owner UNUSED, struct o *prev UNUSED, struct obj_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( o_is_ancestor(o, obj2o(&te->parent_tlo->obj))) gtk_widget_destroy( te->window );
}

static void text_edit_go_del( struct o * o, struct o *owner UNUSED, struct o *prev UNUSED, struct go_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( te->o == o ) gtk_widget_destroy( te->window );
}

static void text_edit_destroy( GtkWidget * w UNUSED, struct text_edit * te )
{
	o_unref(te->o);

	go_hook_remove( te->parent_tlo, &te->go_hook );
	obj_hook_remove( &te->obj_hook );
	obj_unref( &te->parent_tlo->obj );

	string_unref(te->text_key);

	text_edit_list = g_slist_remove( text_edit_list, (gpointer)te );
	win_o_finish( (struct window*)te );
	xfree(te);
}

#define TE_ADD_PROP( _left, _top, _key, _desc ) \
	ASSERT(counter < te->prop_count); \
	pi = te->pi + counter++; \
	prop_item_init( pi, te->o, _key ); \
	label = gtk_label_new(_desc); \
	g_object_set( G_OBJECT(label), "xalign", 0.0, NULL ); \
	gtk_table_attach( table, label, _left, _left+1, _top, _top+1, GTK_FILL, GTK_FILL, 5, 0 ); \
	if (pi->unit_edit) \
	{ \
		gtk_table_attach( table, pi->value_edit, _left+1, _left+2, _top, _top+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 ); \
		gtk_table_attach( table, pi->unit_edit, _left+2, _left+3, _top, _top+1, GTK_FILL, GTK_FILL, 0, 0 ); \
	} else \
		gtk_table_attach( table, pi->value_edit, _left+1, _left+3, _top, _top+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 )

/** Create the text editor window. The window has three forms according to what
 * object o is edited and what is the key identifier of the property.
 * For the "text" property of a go_text and go_tex_text, it displays additional
 * property editors.
 */
void text_edit_create( struct o * o, string text_key )
{
	ASSERT( o && text_key );

	struct text_edit * te = xmalloc_zero(sizeof(struct text_edit));
	te->mode = TEM_OTHER;
	if (o->kind == T_GO && text_key == string_text)
	{
		if (o->type == GOT_TEXT)
			te->mode = TEM_TEXT;
		if (o->type == GOT_TEX_TEXT)
			te->mode = TEM_TEX;
	}
	te->prop_count = prop_counts[te->mode];
	te->pi = xmalloc_zero(sizeof(struct prop_item) * te->prop_count);
	GtkTable *table;
	GtkWidget * widget, * scw, * vbox, * butbox;
	GtkWidget *button_save, *button_load, *button_external_editor;
	GtkWidget *label_text, *text_hbox;
	GtkTextView * tw;
	GtkAccelGroup * ag = gtk_accel_group_new();
	GtkTooltips *tooltips = gtk_tooltips_new();

	win_o_init( (struct window*)te, WT_TEXT_EDITOR );
	
	te->o = o;
	te->text_key = string_ref(text_key);

	te->window = gtk_window_new( GTK_WINDOW_TOPLEVEL );
	text_edit_set_window_title(te);
	gtk_window_add_accel_group( GTK_WINDOW(te->window), ag );
	g_signal_connect( G_OBJECT(te->window), "destroy", G_CALLBACK(text_edit_destroy), te );

	gtk_container_add( GTK_CONTAINER(te->window), vbox = gtk_vbox_new( FALSE, 0 ) );

	if (te->mode != TEM_OTHER)
	{
		table = GTK_TABLE(gtk_table_new( 6, 4, FALSE ));
		gtk_box_pack_start( GTK_BOX(vbox), GTK_WIDGET(table), FALSE, TRUE, 0 );

		uns counter = 0;
		GtkWidget * label;
		struct prop_item * pi;
		if (te->mode == TEM_TEXT)
		{
			TE_ADD_PROP( 0, 0, string_font_size, "Font size:" );
			TE_ADD_PROP( 3, 0, string_font, "Font:" );
		}

		TE_ADD_PROP( 0, 1, string_alignment_x, "Align-x:" );
		TE_ADD_PROP( 3, 1, string_alignment_y, "Align-y:" );
		TE_ADD_PROP( 0, 2, string_rpos_x, "Rel-x:" );
		TE_ADD_PROP( 3, 2, string_rpos_y, "Rel-y:" );
		TE_ADD_PROP( 0, 3, string_apos_x, "Abs-x:" );
		TE_ADD_PROP( 3, 3, string_apos_y, "Abs-y:" );

		gtk_box_pack_start( GTK_BOX(vbox), gtk_hseparator_new(), FALSE, TRUE, 0 );
	}

	// label text and buttons Save & Load
	text_hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), text_hbox, FALSE, TRUE, 0);
	label_text = gtk_label_new ("Text:");
	gtk_box_pack_start (GTK_BOX (text_hbox), label_text, FALSE, TRUE, 0);
	button_save = gtk_button_new_with_label ("Save ...");
	gtk_tooltips_set_tip (tooltips, button_save, "Save text to a file.", NULL);
	g_signal_connect( G_OBJECT(button_save), "clicked", G_CALLBACK(on_save_text), te );
	gtk_box_pack_end (GTK_BOX (text_hbox), button_save, FALSE, TRUE, 0);
	button_load = gtk_button_new_with_label ("Load ...");
	gtk_tooltips_set_tip (tooltips, button_load, "Load text from a file.", NULL);
	g_signal_connect( G_OBJECT(button_load), "clicked", G_CALLBACK(on_load_text), te );
	gtk_box_pack_end (GTK_BOX (text_hbox), button_load, FALSE, TRUE, 3);
	button_external_editor = gtk_button_new_with_label ("Edit with external editor ...");
	gtk_tooltips_set_tip (tooltips, button_external_editor, "Use external editor to edit text.", NULL);
	g_signal_connect (G_OBJECT (button_external_editor), "clicked", G_CALLBACK(open_external_editor), te);
	gtk_box_pack_end (GTK_BOX (text_hbox), button_external_editor, FALSE, TRUE, 0);

	// Text view
	scw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(scw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
	gtk_box_pack_start( GTK_BOX(vbox), scw, TRUE, TRUE, 0 );

	tw = GTK_TEXT_VIEW( gtk_text_view_new() );
	gtk_text_view_set_editable( tw, 1 );
	gtk_text_view_set_wrap_mode( tw, GTK_WRAP_WORD );
	gtk_container_add( GTK_CONTAINER(scw), GTK_WIDGET(tw) );
	te->text_buf = gtk_text_view_get_buffer(tw);

	// Buttons
	gtk_box_pack_start( GTK_BOX(vbox), butbox = gtk_hbox_new( FALSE, 5 ), FALSE, TRUE, 0 );
	widget = gtk_button_new_with_label( "Refresh" );
	//gtk_widget_set_size_request (widget, 84, -1);
	gtk_tooltips_set_tip (tooltips, widget, "Apply changes made to the text.", NULL);
	g_signal_connect( G_OBJECT(widget), "clicked", G_CALLBACK(text_edit_refresh_text), te );
	gtk_widget_add_accelerator( widget, "clicked", ag, GDK_KP_Enter, GDK_CONTROL_MASK, 0 );
	gtk_widget_add_accelerator( widget, "clicked", ag, GDK_Return, GDK_CONTROL_MASK, 0 );
	gtk_box_pack_end( GTK_BOX(butbox), widget, FALSE, TRUE, 0 );

	gtk_box_pack_start( GTK_BOX(vbox), te->statusbar, FALSE, TRUE, 0 );

	gtk_widget_show_all( te->window );
	gtk_widget_grab_focus( GTK_WIDGET(tw) );

	text_edit_list = g_slist_prepend( text_edit_list, (gpointer)te );

	o_ref(te->o);
	if (te->o->kind == T_GO)
		te->parent_tlo = go_get_tlo(o2go(te->o));
	else
		te->parent_tlo = tlo_universum;
	text_edit_refresh_widgets(te);
	obj_ref( &te->parent_tlo->obj );

	te->go_hook.data = te;
	te->go_hook.go_unlink = text_edit_go_del;
	te->go_hook.go_prop_change = text_edit_go_prop_changed;
	te->go_hook.go_prop_delete = text_edit_go_prop_deleted;
	go_hook_add( te->parent_tlo, &te->go_hook );
	te->obj_hook.data = te;
	te->obj_hook.obj_unlink = text_edit_obj_del;
	te->obj_hook.obj_prop_change = text_edit_obj_prop_changed;
	te->obj_hook.obj_prop_delete = text_edit_obj_prop_deleted;
	obj_hook_add( &te->obj_hook );
}
