/*
 * GUI - Horizontal ruler
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include <string.h>
#include "gui/ruler.h"

#define RULER_HEIGHT          14

#define ROUND(x) ((int) ((x) + 0.5))

static void horiz_ruler_class_init    (HorizRulerClass *klass);
static void horiz_ruler_init          (HorizRuler      *horiz_ruler);
int horiz_ruler_motion_notify (GtkWidget      *widget,
				      GdkEventMotion *event);
static void horiz_ruler_draw_ticks    (Ruler       *ruler);
static void horiz_ruler_draw_pos      (Ruler       *ruler);
static gint horiz_ruler_expose (GtkWidget *widget, GdkEventExpose *event UNUSED);

GtkType horiz_ruler_get_type (void)
{
  static GtkType horiz_ruler_type = 0;

  if (!horiz_ruler_type)
    {
      static const GtkTypeInfo horiz_ruler_info = {
	"HorizRuler",
	sizeof(HorizRuler),
	sizeof (HorizRulerClass),
	(GtkClassInitFunc) horiz_ruler_class_init,
	(GtkObjectInitFunc) horiz_ruler_init,
	0, 0, 0 };

      horiz_ruler_type = gtk_type_unique (ruler_get_type(),
		      &horiz_ruler_info);
    }
    DBG("Horiz_ruler gettype finished\n");

  return horiz_ruler_type;
}

static void horiz_ruler_class_init (HorizRulerClass *klass)
{
  GtkWidgetClass *widget_class;
  RulerClass *ruler_class;

  widget_class = (GtkWidgetClass*) klass;
  ruler_class = (RulerClass*) klass;

  widget_class->motion_notify_event = horiz_ruler_motion_notify;
  widget_class->expose_event = horiz_ruler_expose;

  ruler_class->draw_ticks = horiz_ruler_draw_ticks;
  ruler_class->draw_pos = horiz_ruler_draw_pos;
}

static void horiz_ruler_init (HorizRuler *horiz_ruler)
{
  GtkWidget *widget;

  widget = GTK_WIDGET (horiz_ruler);
  widget->requisition.width = widget->style->xthickness * 2 + 1;
  widget->requisition.height = widget->style->ythickness * 2 + RULER_HEIGHT;
}


GtkWidget* horiz_ruler_new (void)
{
  return g_object_new (TYPE_HORIZ_RULER, NULL);
}

int horiz_ruler_motion_notify (GtkWidget *widget,
			  GdkEventMotion *event)
{
  Ruler *ruler;
  gint x;

  ruler = RULER (widget);

  if (event->is_hint)
    gdk_window_get_pointer (widget->window, &x, NULL, NULL);
  else
    x = event->x;

  ruler->position = x;
  g_object_notify (G_OBJECT (ruler), "position");

  /*  Make sure the ruler has been allocated already  */
  if (ruler->backing_store)
    ruler_draw_pos (ruler);

  return FALSE;
}

static void print_label (Ruler *ruler, PangoLayout *layout, double pos, int digit_width UNUSED, char *text, int digit_offset UNUSED)
{
	PangoRectangle logical_rect;
	GtkWidget *widget = GTK_WIDGET (ruler);

	pango_layout_set_text (layout, text, -1);
	pango_layout_get_extents (layout, &logical_rect, NULL);
	gtk_paint_layout (widget->style,
                   	ruler->backing_store,
 	    			GTK_WIDGET_STATE (widget),
					FALSE,
                    NULL,
                    widget,
                    "horiz_ruler",
                    pos + 2, -1,
					layout);
}

static double draw_line (Ruler *ruler, GdkGC *gc, double cur, int height)
{
	double pos;
	int ruler_height;
	GtkWidget *widget = GTK_WIDGET (ruler);
	int ythickness;

	ythickness = widget->style->ythickness;
	ruler_height = widget->allocation.height - 2 * ythickness;
	pos = ROUND (ruler_coords_to_pixels (ruler, cur));
	gdk_draw_line (ruler->backing_store, gc,
			  pos, ruler_height, pos, ruler_height - height);

	return pos;
}

static void horiz_ruler_draw_ticks (Ruler *ruler)
{
  GtkWidget *widget;
  GdkGC *gc, *bg_gc;
  gint width, height;
  gint ruler_height;
  gint xthickness;
  gint ythickness;
  gint digit_width;
  PangoLayout *layout;
  PangoRectangle logical_rect, ink_rect;

  if (!GTK_WIDGET_DRAWABLE (ruler))
    return;

  widget = GTK_WIDGET (ruler);

  gc = widget->style->fg_gc[GTK_STATE_NORMAL];
  bg_gc = widget->style->bg_gc[GTK_STATE_NORMAL];

  xthickness = widget->style->xthickness;
  ythickness = widget->style->ythickness;

  layout = gtk_widget_create_pango_layout (widget, "012456789");
  pango_layout_get_extents (layout, &ink_rect, &logical_rect);

  digit_width = PANGO_PIXELS (ink_rect.width) / 10 + 1;

  width = widget->allocation.width;
  height = widget->allocation.height - ythickness * 2;
  ruler_height = widget->allocation.height;

  gtk_paint_box (widget->style, ruler->backing_store,
		 GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		 NULL, widget, "horiz_ruler",
		 0, 0,
		 widget->allocation.width, widget->allocation.height);


  gdk_draw_line (ruler->backing_store, gc,
		 xthickness,
		 height,
		 widget->allocation.width - xthickness,
		 height);

  compute_and_draw_ticks (ruler, layout, digit_width, width, height, draw_line, print_label, 0);

  g_object_unref (layout);
}

static void horiz_ruler_draw_pos (Ruler *ruler)
{
  GtkWidget *widget;
  GdkGC *gc;
  int i;
  gint x, y;
  gint width, height;
  gint bs_width, bs_height;
  gint xthickness;
  gint ythickness;
  gdouble increment;

  if (GTK_WIDGET_DRAWABLE (ruler))
    {
      widget = GTK_WIDGET (ruler);

      gc = widget->style->fg_gc[GTK_STATE_NORMAL];
      xthickness = widget->style->xthickness;
      ythickness = widget->style->ythickness;
      width = widget->allocation.width;
      height = widget->allocation.height - ythickness * 2;

      bs_width = height / 2;
      bs_width |= 1;  /* make sure it's odd */
      bs_height = bs_width / 2 + 1;

      if ((bs_width > 0) && (bs_height > 0))
	{
	  /*  If a backing store exists, restore the ruler  */
	  if (ruler->backing_store && ruler->non_gr_exp_gc)
	    gdk_draw_drawable (ruler->widget.window,
			       ruler->non_gr_exp_gc,
			       ruler->backing_store,
			       ruler->xsrc, ruler->ysrc,
			       ruler->xsrc, ruler->ysrc,
			       bs_width, bs_height);

	  increment = ruler->resolution;

	  x = ruler -> position + (xthickness - bs_width) / 2 - 1;
	  y = bs_height + 2 * ythickness + 1;

	  for (i = 0; i < bs_height; i++)
	    gdk_draw_line (widget->window, gc,
			   x + i, y + i,
			   x + bs_width - i - 1, y + i);


	  ruler->xsrc = x;
	  ruler->ysrc = y;
	}
    }
}

static int horiz_ruler_expose (GtkWidget *widget, GdkEventExpose *event)
{
	ruler_expose(widget, event);
	return 0;
}
