/** @file
 * GUI - Implementataion of the moving routines in the view windows and
 * of the Select/Transform view tool.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <gtk/gtk.h>

#include "kernel/kernel.h"

#include "gui/main.h"
#include "gui/visualisation.h"
#include "gui/properties.h"

#include "vcl/vcl.h"

enum transform_display_state
{
	TDM_HIDDEN = 0, TDM_VISIBLE = 1, TDM_NONEMPTY = 1 << 1, TDM_WIDE = 1 << 2, TDM_HIGH = 1 << 3, TDM_SHIFT = 1 << 4,
	TDM_SHAPE_MASK = TDM_SHIFT - 1, TDM_RECT = (TDM_WIDE | TDM_HIGH)
};

static struct
{
	int events_pending;
	int last_pix_x, last_pix_y;
	struct geom_point start_coords, start_origin, last_coords, last_vcoords;
	struct view * last_view, * first_view;
	int first_step;
	gulong hook_id;
	GdkEventButton press_event;
} moving;
enum moving_mode moving_mode;

static void transform_redisplay( struct view * vw );
static void transform_hide( struct view * vw );

static gboolean process_last_me( void * u UNUSED );
static int process_transform_step_x( struct view * vw, struct geom_point po );
static int process_transform_step( struct view * vw, struct geom_point po );

static struct vcl_path * tmp_rect;

static void calculate_selection( struct view * vw );
enum tf_device
{
	TFD_NONE,
	TFD_MOVE,
	TFD_ROTATE,
	TFD_SCALE_EQ,
	TFD_SCALE_X,
	TFD_SCALE_Y,
	TFD_MOVE_ORIGIN,
	TFD_SKEW,
	TFD_AXIS,
	TFD_CNT
};
static enum tf_device snap_to_tf_device( struct view * vw, struct geom_point * pt, int mod1, int mod2 UNUSED );

static GdkCursorType tf_device_cursor[TFD_CNT] = { GDK_LEFT_PTR, GDK_FLEUR, GDK_EXCHANGE, GDK_SIZING,
	GDK_SB_H_DOUBLE_ARROW, GDK_SB_V_DOUBLE_ARROW, GDK_FLEUR, GDK_DIAMOND_CROSS, GDK_FLEUR };

static char * tf_device_desc[TFD_CNT] = { "", "Move", "Rotate", "Scale equally", "Scale horizontally",
	"Scale vertically", "Move the transformation centerpoint", "Skew", "Move the skew axis point" };

static enum tf_device transform_device;
static uns idle_func_on;

static uns fifi_on;
static uns fifi_enabled;

static struct geom_point orig_sant_pts[3];

#define SET_IDLE_FUNC_ON ASSERT(!idle_func_on); g_idle_add( (GSourceFunc)process_last_me, &idle_func_on ); idle_func_on = 1
#define UNSET_IDLE_FUNC_ON idle_func_on = 0
#define SET_IDLE_FUNC_OFF ASSERT(idle_func_on); g_idle_remove_by_data( &idle_func_on ); idle_func_on = 0

void view_moving_start( struct geom_point pix, struct view * vw, enum moving_mode mode, GdkCursorType cursor_type )
{
	ASSERT( vw->navigator || mode != MOVE_PANNING_NAV );
	ASSERT( moving_mode == MOVE_NONE  && mode != MOVE_NONE );
	DBG("Moving start, mode = %d\n", mode);
	
	moving.last_pix_x = pix.x; moving.last_pix_y = pix.y;
	moving.events_pending = 0;
	moving.start_coords = moving.last_coords;
	moving.start_origin = vw->transform_origin;
	moving.first_view = moving.last_view = vw;
	moving.first_step = 1;
	moving_mode = mode;
	for (int i = 0; i<3; i++) orig_sant_pts[i] = vw->vis.transform_points[i];

	GdkCursor * cursor = gdk_cursor_new( cursor_type );
	switch ( mode )
	{
	case MOVE_PANNING:
	case MOVE_PANNING_IMG:
		gdk_window_set_cursor( vw->vis.canvas->gtk->widget->window, cursor );
		break;
	case MOVE_PANNING_NAV: break;
	default:
		for ( GSList * l = view_list; l; l = l->next )
		{
			struct view * vvv = (struct view*)l->data;
			if ( vvv->tlo == vw->tlo ) gdk_window_set_cursor( vvv->vis.canvas->gtk->widget->window, cursor );
		}
	}
	gdk_cursor_unref( cursor );
}

void view_moving_end(void)
{
	DBG("Moving end, mode = %d\n", moving_mode );
	if ( moving.first_view && moving.first_view->tf_drag_on )
		moving.first_view->tf_drag_on = 0;
	//unset the cursor for all relevant views
	switch ( moving_mode )
	{
	case MOVE_PANNING:
	case MOVE_PANNING_IMG:
		if ( moving.first_view && moving.first_view->vis.canvas->gtk->widget->window )
			gdk_window_set_cursor( moving.first_view->vis.canvas->gtk->widget->window, NULL );
		break;
	case MOVE_PANNING_NAV: break;
	default:
		for ( GSList * l = view_list; l; l = l->next )
		{
			struct view * vvv = (struct view*)l->data;
			if ( moving.first_view && vvv->tlo == moving.first_view->tlo && vvv->vis.canvas->gtk->widget->window )
				gdk_window_set_cursor( vvv->vis.canvas->gtk->widget->window, NULL );
		}
	}
	if (idle_func_on)
	{ SET_IDLE_FUNC_OFF; }

	moving_mode = MOVE_NONE;
	moving.first_view = moving.last_view = NULL;
}

static gboolean process_last_me( void * u UNUSED )
{
	DBG("Process last moving event\n");
	if (unlikely(!moving.events_pending))
	{
		DBG("No pending events, as one would expect\n");
		UNSET_IDLE_FUNC_ON;
		return FALSE;
	}
	moving.events_pending = 0;

	if (!moving.last_view)
	{
		DBG("moving.last_view == NULL!\n");
		return FALSE;
	}

	struct geom_point pi = { moving.last_pix_x, moving.last_pix_y };
	struct geom_point po, vp;
	visualisation_pixels_to_coords( &moving.last_view->vis, &pi, &po );

	if (fifi_on && fifi_enabled)
	{
		struct geom_point snapped = po;
		if ( factory.state == &ofs_rehang_1 )
			snap_to_anchor( &snapped, NULL, moving.last_view->tlo, moving.last_view->tolerance );
		else
			snap_point( &snapped, NULL, &moving.last_view->grid, moving.last_view->tlo, moving.last_view->tolerance );
		visualisation_gui_fifi_update( &moving.last_view->vis, &moving.last_view->fifi, snapped );
	}

	switch ( moving_mode )
	{
		case MOVE_FACTORY:
			if (moving.last_view->tlo == moving.first_view->tlo)
				factory_moving_step(po);
			break;
		case MOVE_TRANSFORM:
			process_transform_step_x(moving.last_view, po);
			break;
		case MOVE_TRANSFORM_2:
			process_transform_step(moving.last_view, po);
			break;
		case MOVE_PANNING:
		case MOVE_PANNING_IMG:
		case MOVE_PANNING_NAV:
			if ( moving.last_view != moving.first_view )
				break;
			struct geom_point delta;
			if ( moving_mode == MOVE_PANNING_NAV )
			{
				visualisation_pixels_to_coords( &moving.last_view->navigator->vis, &pi, &vp );
				visualisation_coords_to_vcoords( &moving.last_view->vis, &vp, &vp );
			}
			else
				visualisation_pixels_to_vcoords( &moving.last_view->vis, &pi, &vp );
			delta.x = moving.last_vcoords.x - vp.x;
			delta.y = moving.last_vcoords.y - vp.y;

			if ( moving_mode != MOVE_PANNING_NAV )
				moving.last_vcoords = vp;
			if ( moving_mode == MOVE_PANNING_IMG )
			{ delta.x = -delta.x; delta.y = -delta.y; }
			visualisation_move( &moving.last_view->vis, delta.x, delta.y );
			break;
		case MOVE_SELECT:
		{
			moving.last_coords = po;
			struct geom_point corner = { moving.last_coords.x, moving.start_coords.y };
			visualisation_gui_rectangle_update( &moving.first_view->vis, &tmp_rect,
				moving.last_coords, corner, moving.start_coords );
			break;
		}
		default:;
	}
	UNSET_IDLE_FUNC_ON;
	return FALSE;
}

void view_moving_motion( GdkEventMotion *event, struct view *vw )
{
	DBG("mouse event: motion\n");
	if ( moving_mode != MOVE_NONE && moving_mode != MOVE_FACTORY &&
		!(event->state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) )
		// the button release event was lost, we need to end the move
		view_moving_release( NULL, vw );

	DBG("Pixel coords: %.3f, %.3f\n", event->x, event->y );
	moving.last_pix_x = event->x; moving.last_pix_y = event->y;
	moving.last_view = vw;
	moving.events_pending = 1;

	if (!idle_func_on)
	{ SET_IDLE_FUNC_ON; }
}

void view_clicked( GdkEventButton * event, struct view * vw )
{
	DBG("mouse event: button press\n");
	moving.last_view = vw;
	moving.last_pix_x = event->x; moving.last_pix_y = event->y;
	moving.press_event = *event;
}

void view_moving_release( GdkEventButton *event, struct view * vw )
{
	DBG("mouse event: button release, event = %p, mode = %d\n", event, moving_mode);
	// if the mode is MOVE_FACTORY or MOVE_NONE, the release event is ignored,
	// which is good, because in does not have to match any press event
	// (that could happen somewhere else)

	if (event)
	{
		view_clicked(event, vw);
		moving.events_pending = 1;
		process_last_me(NULL); //make sure that the move continues up to here
		// (the last event might have been skipped)
	}

	switch ( moving_mode )
	{
		case MOVE_TRANSFORM:
		case MOVE_TRANSFORM_2:
			transform_recalc_dimen(moving.last_view);
			factory_op_step();
			break;
		case MOVE_PANNING:
		case MOVE_PANNING_IMG:
		case MOVE_PANNING_NAV:
			view_moving_end(); break;
		case MOVE_SELECT:
			visualisation_gui_rectangle_destroy( &moving.first_view->vis, &tmp_rect );
			calculate_selection(moving.last_view);
			view_moving_end();
			refresh_gui();
			break;
		default:;
	}

	if (!event)
	{
		// unset the view poiter, because it becomes invalid now
		if ( moving.last_view == vw )
			moving.last_view = NULL;
		if ( moving.first_view == vw )
			moving.first_view = NULL;
		moving.events_pending = 0;

	}
}

/* THE SELECT/TRANSFORM TOOL */

void view_selection_start( struct geom_point coords, struct geom_point coords_pix, struct view * vw )
{
	DBG("Selection start\n");
	//(rectangular) selection
	visualisation_gui_rectangle_update( &vw->vis, &tmp_rect, coords, coords, coords );
	moving.last_coords = coords;
	view_moving_start( coords_pix, vw, MOVE_SELECT, GDK_TOP_LEFT_ARROW );
}

int view_transform_start( GdkEventButton *event, struct geom_point coords, struct geom_point coords_pix, struct view * vw )
{
	DBG("Transform start\n");
	if ( transform_device = snap_to_tf_device( vw, &coords, event->state & GDK_SHIFT_MASK, event->state & GDK_CONTROL_MASK ) )
	{
		factory_op_step();
		factory_msg( tf_device_desc[transform_device] );
		moving.last_coords = coords;
		view_moving_start( coords_pix, vw, MOVE_TRANSFORM, tf_device_cursor[transform_device] );
		return 1;
	} else
		return 0;
}

void view_panning_start( struct geom_point coords_pix, struct view * vw )
{
	DBG("Panning start\n");
	visualisation_pixels_to_vcoords( &vw->vis, &coords_pix, &moving.last_vcoords );
	uns pan_img = prop_lookup_uns( obj2o(obj_universum), string_drag_image, PANNING_DRAG_IMAGE );
	view_moving_start( coords_pix, vw, pan_img ? MOVE_PANNING_IMG : MOVE_PANNING, GDK_HAND2 );
}

void view_panning_nav_start( struct geom_point coords, struct geom_point coords_pix, struct view * vw )
{
	DBG("Panning navigator start\n");
	visualisation_coords_to_vcoords( &vw->vis, &coords, &moving.last_vcoords );
	view_moving_start( coords_pix, vw, MOVE_PANNING_NAV, GDK_FLEUR );
}

#define HIDE_FIFI \
	for ( GSList * l = view_list; l; l = l->next ) \
	{ \
		struct view * vw = (struct view*)l->data; \
		if (vw->fifi) \
			visualisation_gui_fifi_destroy(&vw->vis, &vw->fifi); \
	}

void set_fifi_on(uns on)
{
	if (on == fifi_on)
		return;
	fifi_on = on;
	if (!fifi_on || !fifi_enabled)
		HIDE_FIFI;
}
void set_fifi_enabled(uns enabled)
{
	if ( fifi_enabled == enabled )
		return;
	fifi_enabled = enabled;
	if (!fifi_on || !fifi_enabled)
		HIDE_FIFI;
}

#define SNAP_DEVICE(d,a,b) tmpt.x = (a); tmpt.y = (b);			\
	dist = hypot( pt->x - tmpt.x, pt->y - tmpt.y );			\
	if ( dist < maxdist && ( mindist < 0 || mindist > dist ))	\
	{ dist = mindist; dev = (d); dev_pt = tmpt;}

static enum tf_device snap_to_tf_device( struct view * vw, struct geom_point * pt, int mod1, int mod2 UNUSED )
{
        struct geom_rectangle bb;
	enum tf_device dev = TFD_NONE;
	struct geom_point dev_pt;
	struct geom_point tmpt;
	real dist, mindist = -1.0, maxdist = vw->tolerance;
        visualisation_get_selector_bbox( &vw->vis, &bb );
	visualisation_vcoords_to_coords( &vw->vis, &bb.ll, &bb.ll );
	visualisation_vcoords_to_coords( &vw->vis, &bb.ur, &bb.ur );
	if ( !(vw->transform_state & TDM_NONEMPTY) ) return TFD_NONE;

	if ( vw->transform_state & TDM_NONEMPTY )
	{
		// origin
		SNAP_DEVICE( mod1 ? TFD_MOVE_ORIGIN : TFD_MOVE, vw->transform_origin.x, vw->transform_origin.y );
	}
	if ( (vw->transform_state & TDM_RECT) == TDM_RECT )
	{
		// scale/skew at midpoints of the edges
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_X, bb.ur.x, (bb.ll.y + bb.ur.y)/2 );
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_X, bb.ll.x, (bb.ll.y + bb.ur.y)/2 );
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_Y, (bb.ll.x + bb.ur.x)/2, bb.ll.y );
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_Y, (bb.ll.x + bb.ur.x)/2, bb.ur.y );
	}
	if ( vw->transform_state & TDM_WIDE )
	{
		// scale equally or rotate at corners
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ll.x, bb.ll.y );
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ur.x, bb.ur.y );
	}
	if ( vw->transform_state & TDM_HIGH )
	{
		// scale equally or rotate at the other two corners
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ll.x, bb.ur.y );
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ur.x, bb.ll.y );
	}
	if ( mod1 && (vw->transform_state & TDM_RECT) == TDM_RECT )
	{
		SNAP_DEVICE( TFD_AXIS, vw->axis.x, vw->axis.y ); //edit axis
	}
	if ( dev == TFD_NONE && // vertical bbox lines
		(fabs(bb.ll.x - pt->x) < maxdist || fabs(bb.ur.x - pt->x) < maxdist) &&
		(pt->y < bb.ur.y + maxdist) && (pt->y > bb.ll.y - maxdist) )
	{
		dev = TFD_MOVE; dev_pt = *pt;
	}
	if ( dev == TFD_NONE && // horizontal bbox lines
		(fabs(bb.ll.y - pt->y) < maxdist || fabs(bb.ur.y - pt->y) < maxdist) &&
		(pt->x < bb.ur.x + maxdist) && (pt->x > bb.ll.x - maxdist) )
	{
		dev = TFD_MOVE; dev_pt = *pt;
	}
	if ( dev != TFD_NONE ) *pt = dev_pt;
	return dev;
}

static int process_transform_step_x( struct view * vw, struct geom_point po )
{
	struct geom_transform2 t;
	volatile int errval;
	struct geom_point zero = vw->transform_origin;
	struct geom_point start = moving.start_coords;
	switch ( transform_device ){
		case TFD_NONE:
			BUG(); break;
		case TFD_MOVE:
			errval = geom_transform2_set_move( &t, po.x - start.x, po.y - start.y );
			break;
		case TFD_SCALE_X:
			DBG("zero = [ %f, %f ], po = [ %f, %f ], start = [ %f, %f ]\n", zero.x, zero.y, po.x, po.y, start.x, start.y);
			errval = geom_transform2_set_scale_center( &t, zero.x, zero.y, (po.x - zero.x) / (start.x - zero.x), 1.0 );
			break;
		case TFD_SCALE_Y:
			DBG("zero = [ %f, %f ], po = [ %f, %f ], start = [ %f, %f ]\n", zero.x, zero.y, po.x, po.y, start.x, start.y);
			errval = geom_transform2_set_scale_center( &t, zero.x, zero.y, 1.0, (po.y - zero.y) / (start.y - zero.y) );
			break;
		case TFD_SCALE_EQ:
		{
			real r = hypot( po.x - zero.x, po.y - zero.y ) / hypot( start.x - zero.x, start.y - zero.y );
			errval = geom_transform2_set_scale_center( &t, zero.x, zero.y, 
				((po.x - zero.x) * (start.x - zero.x) < 0) ? -r : r, ((po.y - zero.y) * (start.y - zero.y) < 0) ? -r : r );
			break;
		}
		case TFD_ROTATE:
		{
			real angle = atan2( start.x - zero.x, start.y - zero.y ) - atan2( po.x - zero.x, po.y - zero.y );
			errval = geom_transform2_set_rotate_center( &t, zero.x, zero.y, angle );
			break;
		}
		case TFD_SKEW:
		{
			errval = geom_transform2_set_skew_normal( &t, zero.x, zero.y, vw->axis.x, vw->axis.y,
				start.x, start.y, po.x, po.y );
			break;
		}
		case TFD_MOVE_ORIGIN:
			vw->axis.x += po.x - vw->transform_origin.x;
			vw->axis.y += po.y - vw->transform_origin.y;
			vw->transform_origin = po;
			visualisation_set_center_gadget( &vw->vis, vw->transform_origin.x, vw->transform_origin.y );
			visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
			return 0;
		case TFD_AXIS:
			vw->axis = po;
			visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
			return 0;
		default: BUG();
	}
	if (errval == 0)
  	{
    		char errs[TRANS_ERR_SIZE];
		errval = !tsort_presorted_transform_safe(vw->tlo, &t, moving.first_step, factory.desc->text, errs );
		moving.first_step = 0;
		if ( !errval )
		{
			if ( transform_device == TFD_MOVE )
			{
				// move the center and axes accordingly
				struct geom_point dxa = { vw->axis.x - vw->transform_origin.x, vw->axis.y - vw->transform_origin.y };
				vw->transform_origin.x = moving.start_origin.x + po.x - start.x;
				vw->transform_origin.y = moving.start_origin.y + po.y - start.y;
				vw->axis.x = vw->transform_origin.x + dxa.x;
				vw->axis.y = vw->transform_origin.y + dxa.y;
				visualisation_set_center_gadget( &vw->vis, vw->transform_origin.x, vw->transform_origin.y );
				visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
			}
		}
		else statusbar_err( NULL, "%s", errs );
	}
	else statusbar_err( NULL, "Numeric error in transformation" );
	return !errval;
}

static void calculate_selection( struct view * vw )
{
	if ( fabs(moving.last_coords.x - moving.start_coords.x) <= vw->tolerance ||
		fabs(moving.last_coords.y - moving.start_coords.y) <= vw->tolerance )
	{
		// individual selection
		struct dist_pass_result dpr;
		struct geom_point po = moving.last_coords;
		snap_to_go( &po, &dpr, context.tlo, vw->tolerance );
		if ( dpr.type == DIST_PASS_GO )
		{
			modify_selection( dpr.data.go.go, go_get_tlo(dpr.data.go.go), &moving.press_event, NULL, vw->group );
			change_context( (struct window*)vw, vw->group, (struct o*)go_get_owner_and_member(dpr.data.go.go, vw->group) );
		} else {
			char err[TRANS_ERR_SIZE];
			TRANS_BEGIN_MAIN( context.tlo, "Unselect all", err );
				group_unselect_all_go(context.group);
			TRANS_FAILED
	       			die("Cannot unselect: %s", err);
			TRANS_END
		}
	} else {
		//rectangular selection
		struct geom_rectangle rect;
		if ( moving.last_coords.x < moving.start_coords.x )
		{ rect.ll.x = moving.last_coords.x; rect.ur.x = moving.start_coords.x; }
		else
		{ rect.ll.x = moving.start_coords.x; rect.ur.x = moving.last_coords.x; }
		if ( moving.last_coords.y < moving.start_coords.y )
		{ rect.ll.y = moving.last_coords.y; rect.ur.y = moving.start_coords.y; }
		else
		{ rect.ll.y = moving.start_coords.y; rect.ur.y = moving.last_coords.y; }

		modify_selection( NULL, context.tlo, &moving.press_event, &rect, vw->group );
		//change_context( (struct window*)vw, vw->group, (struct o*)group_get_first_selected_go(vw->group) );
	}
}

void transform_set_shift( int on, struct view * vw )
{
	if (on)
		vw->transform_state |= TDM_SHIFT;
	else
		vw->transform_state &= ~TDM_SHIFT;
	transform_redisplay(vw);
}

/** Show or hide the transform tool according to the current factory state
  in the given view */
void transform_update( struct view *vw )
{
	if ( factory.start_state == &ofs_select || factory.start_state == &ofs_transform_x )
	{
		if (!(vw->transform_state & TDM_SHAPE_MASK))
		{ vw->transform_state = TDM_VISIBLE; transform_recalc_all(vw); }
	}
	else
	{
		if ( vw->transform_state & TDM_SHAPE_MASK )
		{ transform_hide(vw); vw->transform_state = TDM_HIDDEN; }
	}
	// show/hide hangers
	if ( !(factory.start_state->flags & OFF_HIDE_HANGERS) && (factory.snap_set & SNAP_HANGERS) )
		visualisation_set_show_hangers_mode( &vw->vis, VIS_SHOW_ALL, NULL );
	else
		visualisation_set_show_hangers_mode( &vw->vis, VIS_SHOW_NONE, NULL );

	set_fifi_on( !(factory.start_state->flags & OFF_HIDE_HANGERS) &&
		(factory.state == &ofs_rehang_1 || factory.snap_set & (SNAP_HANGERS | SNAP_GRID | SNAP_LINES | SNAP_INTERS)));

	// show/hide anchors
	if ( factory.start_state == &ofs_rehang_1 )
		visualisation_set_show_anchors_mode( &vw->vis, VIS_SHOW_SELECTED, NULL );
	else
	{
		struct go * g; //show anchors for the currently craeted go
		if ( factory.tlo == vw->tlo && (g = factory_get_created_go()))
			visualisation_set_show_anchors_mode( &vw->vis, VIS_SHOW_SPECIFIC, g );
		else
			visualisation_set_show_anchors_mode( &vw->vis, VIS_SHOW_NONE, NULL );
	}
}
static void transform_hide( struct view * vw )
{
	int i;
	for ( i=0; i<=22; i++ ) visualisation_set_gadget_visible( &vw->vis, i, 0 );
}
/** Recalculate the dimension mode of the transformation bbox (empty,
 * wide, high) and redisplay it. */
void transform_recalc_dimen( struct view * vw )
{
	if ( vw->transform_state == TDM_HIDDEN ) return;

	struct geom_rectangle bbox;
	uns new_mode;
	visualisation_get_selector_bbox( &vw->vis, &bbox );
	DBG("Selector bbox is %.3f, %.3f, %.3f, %.3f\n", bbox.ll.x, bbox.ll.y, bbox.ur.x, bbox.ur.y );
	visualisation_vcoords_to_coords( &vw->vis, &bbox.ll, &bbox.ll );
	visualisation_vcoords_to_coords( &vw->vis, &bbox.ur, &bbox.ur );
	if ( bbox.ll.x > bbox.ur.x ) //empty bbox
	{
		vw->transform_state = TDM_VISIBLE;
		transform_hide(vw);
		return;
	}
	else new_mode = TDM_NONEMPTY;
	if ( bbox.ur.x - bbox.ll.x > 2 * vw->tolerance ) new_mode |= TDM_WIDE;
	if ( bbox.ur.y - bbox.ll.y > 2 * vw->tolerance ) new_mode |= TDM_HIGH;

	if ( (new_mode & TDM_SHAPE_MASK) != (vw->transform_state & TDM_SHAPE_MASK) )
	{
		vw->transform_state = vw->transform_state & TDM_SHIFT | new_mode;
		transform_redisplay(vw);
	}
}

/** Recalculate the transform tool coordinates of the center and axis point and
 * update the gadgets in the view; recalculate the dimenstion mode of the transformation
 * tool as well. */
void transform_recalc_all( struct view * vw )
{
	if ( vw->transform_state == TDM_HIDDEN ) return;

	struct geom_rectangle bbox;
	visualisation_get_selector_bbox( &vw->vis, &bbox );
	visualisation_vcoords_to_coords( &vw->vis, &bbox.ll, &bbox.ll );
	visualisation_vcoords_to_coords( &vw->vis, &bbox.ur, &bbox.ur );

	vw->transform_origin.x = (bbox.ll.x + bbox.ur.x)/2;
	vw->transform_origin.y = (bbox.ll.y + bbox.ur.y)/2;
	vw->axis.x = 2 * bbox.ur.x - vw->transform_origin.x;
	vw->axis.y = vw->transform_origin.y;
	visualisation_set_center_gadget( &vw->vis, vw->transform_origin.x, vw->transform_origin.y );
	visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
	
	transform_recalc_dimen(vw);
}
/** Show the gadgets of the transform tool in the particular mode according to vw->transform_state */
static void transform_redisplay( struct view * vw )
{
	int i = (vw->transform_state & TDM_SHIFT) ? 0 : 1;
	for ( ; i<=17; i+=2 ) visualisation_set_gadget_visible( &vw->vis, i, 0 );

	i = (vw->transform_state & TDM_SHIFT) ? 1 : 0;
	visualisation_set_gadget_visible( &vw->vis, 8, vw->transform_state & TDM_NONEMPTY );
	visualisation_set_gadget_visible( &vw->vis, 18, vw->transform_state & TDM_NONEMPTY ); // the center points

	visualisation_set_gadget_visible( &vw->vis, i, vw->transform_state & TDM_WIDE ); // two opposite corners
	visualisation_set_gadget_visible( &vw->vis, i + 16, vw->transform_state & TDM_WIDE );

	visualisation_set_gadget_visible( &vw->vis, i + 4, vw->transform_state & TDM_HIGH ); // the other two opposite corners
	visualisation_set_gadget_visible( &vw->vis, i + 12, vw->transform_state & TDM_HIGH );

	visualisation_set_gadget_visible( &vw->vis, i + 2, (vw->transform_state & TDM_RECT) == TDM_RECT ); // the edge midpoints
	visualisation_set_gadget_visible( &vw->vis, i + 6, (vw->transform_state & TDM_RECT) == TDM_RECT );
	visualisation_set_gadget_visible( &vw->vis, i + 10, (vw->transform_state & TDM_RECT) == TDM_RECT );
	visualisation_set_gadget_visible( &vw->vis, i + 14, (vw->transform_state & TDM_RECT) == TDM_RECT );

	int axis_visible = ((vw->transform_state & TDM_RECT) == TDM_RECT) && (vw->transform_state & TDM_SHIFT);
	visualisation_set_gadget_visible( &vw->vis, 19, axis_visible );
	visualisation_set_gadget_visible( &vw->vis, 21, axis_visible );
}


/* SANTIAGO'S TRANSFORM TOOL */

static real tmp_hypot (real x, real y, struct geom_point p)
{
  x -= p . x;
  y -= p . y;
  return (x * x + y * y);
}

static int get_near_tf_point (struct view *vw, real x, real y)
{
  real old_delta = vw -> tolerance * vw -> tolerance;
  real new_delta;
  int retval = 0;

  if ((vw -> vis . flags & VIS_FLAGS_TF_MOVE) &&
      ((new_delta = tmp_hypot (x, y, vw -> vis . transform_points [0])) < old_delta))
    {
      retval = 1;
      old_delta = new_delta;
    }
  else if ((vw -> vis . flags & VIS_FLAGS_TF_SECOND) &&
	   ((new_delta = tmp_hypot (x, y, vw -> vis . transform_points [1])) < old_delta))
    {
      retval = (vw -> vis . flags & VIS_FLAGS_TF_RESIZE) ? 2 : 3;
      old_delta = new_delta;
    }
  else if ((vw -> vis . flags & VIS_FLAGS_TF_SKEW) &&
	   ((new_delta = tmp_hypot (x, y, vw -> vis . transform_points [2])) < old_delta))
    {
      retval = 4;
      old_delta = new_delta;
    }

  return (retval);
}

static int add_new_tf_point (struct view *vw, real x, real y)
{
  if (! (vw -> vis . flags & VIS_FLAGS_TF_MOVE))
    {
      visualisation_set_tf_move (&(vw -> vis), x, y);
      return 1;
    }
  else if (! (vw -> vis . flags & VIS_FLAGS_TF_SECOND))
    {
      visualisation_set_tf_resize (&(vw -> vis), x, y);
      return 2;
    }
  else // if (! (vw -> vis . flags & VIS_FLAGS_TF_SKEW))
    {
      visualisation_set_tf_skew (&(vw -> vis), x, y);
      return 4;
    }
}

static int compute_transformed_points (struct view *vw, struct geom_transform *t, struct geom_point *pts)
{
  int i = 0;
  if (vw -> vis . flags & VIS_FLAGS_TF_MOVE)
    i = i || geom_point_transform (orig_sant_pts + 0, t, pts + 0);
  if (vw -> vis . flags & VIS_FLAGS_TF_SECOND)
    i = i || geom_point_transform (orig_sant_pts + 1, t, pts + 1);
  if (vw -> vis . flags & VIS_FLAGS_TF_SKEW)
    i = i || geom_point_transform (orig_sant_pts + 2, t, pts + 2);

  return (i);
}

static void update_transformed_points (struct view *vw, struct geom_point *pts)
{
  if (vw -> vis . flags & VIS_FLAGS_TF_MOVE)
    visualisation_set_tf_move (&(vw -> vis), pts [0] . x, pts [0] . y);
  if (vw -> vis . flags & VIS_FLAGS_TF_RESIZE)
    visualisation_set_tf_resize (&(vw -> vis), pts [1] . x, pts [1] . y);
  if (vw -> vis . flags & VIS_FLAGS_TF_ROTATE)
    visualisation_set_tf_rotate (&(vw -> vis), pts [1] . x, pts [1] . y);
  if (vw -> vis . flags & VIS_FLAGS_TF_SKEW)
    visualisation_set_tf_skew (&(vw -> vis), pts [2] . x, pts [2] . y);
}



static int transform_mode_1 (struct geom_transform2 *t, const struct geom_point *pf,
			     const struct geom_point *pt)
{ return geom_transform2_set_move (t, pt -> x - pf -> x, pt -> y - pf -> y); }

static int transform_mode_2 (struct geom_transform2 *t, const struct geom_point *p1,
			     const struct geom_point *pf, const struct geom_point *pt)
{
  int i = 0;
  i = i || geom_transform2_set_identity (t);
  i = i || geom_transform2_translate (t, - p1 -> x, - p1 -> y);
  i = i || geom_transform2_scale (t, (pt -> x - p1 -> x) / (pf -> x - p1 -> x),
				 (pt -> y - p1 -> y) / (pf -> y - p1 -> y));
  i = i || geom_transform2_translate (t, p1 -> x,  p1 -> y);
  return (i ? -1 : 0);
}

static int transform_mode_3 (struct geom_transform2 *t, const struct geom_point *p1,
			     const struct geom_point *pf, const struct geom_point *pt)
{
  int i = 0;
  double pfx = pf -> x - p1 -> x;
  double pfy = pf -> y - p1 -> y;
  double ptx = pt -> x - p1 -> x;
  double pty = pt -> y - p1 -> y;
  double scale = hypot (ptx, pty) / hypot (pfx, pfy);
  double angle = atan2 (pty, ptx) - atan2 (pfy, pfx);
  i = i || geom_transform2_set_identity (t);
  i = i || geom_transform2_translate (t, - p1 -> x, - p1 -> y);
  i = i || geom_transform2_scale (t, scale, scale);
  i = i || geom_transform2_rotate (t, angle);
  i = i || geom_transform2_translate (t, p1 -> x,  p1 -> y);
  return (i ? -1 : 0);
}

static int transform_mode_4 (struct geom_transform2 *t,
			     const struct geom_point *p1, const struct geom_point *p2,
			     const struct geom_point *pf, const struct geom_point *pt)
{ return (geom_transform2_set_skew (t, p1 -> x, p1 -> y, p2 -> x, p2 -> y,
				    pf -> x, pf -> y, pt -> x, pt -> y)); }
void view_s_transform_start( GdkEventButton *event, struct geom_point coords, struct geom_point coords_pix, struct view * vw )
{
	if ( event->state & GDK_SHIFT_MASK ) //edit transformation points
	{
		DBG("Edit transformation point\n");
		switch (get_near_tf_point (vw, coords.x, coords.y))
		{
		case 0: add_new_tf_point (vw, coords.x, coords.y);
			break;
		case 1: visualisation_unset_tf_move (&(vw -> vis));
			break;
		case 2: // because user clicked on resize img, then there are
			// valid data in visualisation . transform_points [1]
			visualisation_set_tf_rotate (&(vw -> vis), vw -> vis . transform_points [1] . x,
							vw -> vis . transform_points [1] . y);
	   		break;
		case 3: visualisation_unset_tf_second (&(vw -> vis));
			break;
			case 4: visualisation_unset_tf_skew (&(vw -> vis));
			break;
		}
	} else {
		if ( vw -> tf_drag_on = get_near_tf_point (vw, coords.x, coords.y) )
		{
			factory_op_step();
			view_moving_start( coords_pix, vw, MOVE_TRANSFORM_2, GDK_FLEUR );
		}
	}
}

static int process_transform_step( struct view * vw, struct geom_point po )
{
	struct geom_transform2 t;
	volatile int errval;

	switch (vw -> tf_drag_on)
	{
	case 1: errval = transform_mode_1 (&t, orig_sant_pts + 0, &po);
		break;
	case 2: errval = transform_mode_2 (&t, orig_sant_pts + 0,
					       orig_sant_pts + 1, &po);
   		break;
  	case 3: errval = transform_mode_3 (&t, orig_sant_pts + 0,
					       orig_sant_pts + 1, &po);
    		break;
  	case 4: errval = transform_mode_4 (&t, orig_sant_pts + 0,
					       orig_sant_pts + 1,
					       orig_sant_pts + 2, &po);
		break;
	default: BUG ();
	}

	if (errval == 0)
  	{
    		char errs[TRANS_ERR_SIZE];
		errval = !tsort_presorted_transform_safe( vw->tlo, &t, moving.first_step, factory.desc->text, errs );
		moving.first_step = 0;
		if (!errval)
		{
			TRANS_BEGIN_ANONYMOUS(errs)
			{
				struct geom_point p [3];
				int i = compute_transformed_points (vw, &(t . primary), p);
				if (i != 0) trans_fail ("Numeric error in transformation");
				update_transformed_points (vw, p);
			}
			TRANS_FAILED
				statusbar_err( NULL, "Cannot transform: %s", errs );
				errval = 1;
			TRANS_END;
		}
		else statusbar_err( NULL, "%s", errs );
	}
	else statusbar_err( NULL, "Cannot transform: Numeric error in transformation" );
	return !errval;
}
