#ifndef __VRR_PROPERTIES_H__
#define __VRR_PROPERTIES_H__

/** @file
 * GUI - The property window showing properties of a certain object, property
 * editors etc, the unit window.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "gui/main.h"

/** The Property window structure */
struct prop_window
{
	WIN_O_MAGIC;
	GtkWidget * window;	//!< the GTK window
	struct o * o;		//!< the object whose properties are displayed in the window
	struct obj_tlo * tlo;	//!< the parent tlo, if the object displayed is a go
	struct go_hook go_hook;
	struct obj_hook obj_hook;
	struct unit_hook unit_hook;
	int redraw;		//!< should gtk widgets be redrawn?
	GtkTable *table;	//!< table for properties
	GtkWidget *button_add;	//!< button for adding new properties
	GtkWidget *button_delete; //!< button for deleting an existing property
	uns prop_count;		//!< the count of displayed properties
	struct prop_item * items; //!< The array of prop_item structures for all displayed properties
	uns flags;
};

/** The global list of all existing property windows */
extern GSList * prop_window_list;
/** The one and the only context window, if exists. */
extern struct prop_window * context_property_window;

#define PIF_SELECTED	1
#define PIF_FACTORY	2

/** The property item structure used in property window for individual properties */
struct prop_item
{
	struct o * o;		//!< The owner object of the property
	string key;		//!< The key identifier of the property
	uns type, subtype;	//!< The type and subtype
	GtkWidget * value_edit;	//!< The widget editor of the property value
	GtkWidget * unit_edit;	//!< The widget editor of the property unit, if the property has any
	uns flags;		//!< If the property item is selected
	struct window * pw;	//!< A pointer to the parent window
};

/** Do some initialization: initialize the unit lists etc. */
void gui_prop_init(void);

/** Free the global unit lists. */
void gui_prop_finish(void);

/** Create and open the property window.
 \param o A pointer to the object whose properties are to be displayed; or NULL,
 if the craeted window should show properties of context objects. */
void prop_window_create(struct o * o);

/** A callback to inform the property window that the context has changed. */
void prop_window_context_changed(struct context * old_context);

/** Create a widget to edit the property.
 \param pi THe structure filled with data needed to create the widget.
 Sets the pointer to thw widget editor. */
void prop_value_edit_create( struct prop_item * pi );

/** Create a widget to edit the property unit.
 \param pi THe structure filled with data needed to create the widget.
 Sets the pointer to thw widget editor. */
void prop_unit_edit_create( struct prop_item * pi );

/** Initialize the property item -- set the given values, create editor
 widgets and synchronize them with the current values. */
void prop_item_init( struct prop_item * pi, struct o * o, string key );

/** Sychronize the property widgets in the prop_item struct with
 the current value of property prop. If the struct pi contains nonzero
 pointer to a unit editor, it will be synchronized, too. */
void prop_sync( struct prop_item * pi );

/** The definition of the property store with the given identifier */
#define PROP_STORE_DEFINE(_id) struct obj_tlo * _id
/** Create a property store with the given identifier */
#define PROP_STORE_NEW(_id) _id = tlo_new(); obj_ref(&_id->obj)
/** Destroy a property store with the given identifier */
#define PROP_STORE_DESTROY(_id) obj_unref(&_id->obj); _id = NULL
/** Get the property store object */
#define PROP_STORE_O(_id) go2o(&tlo_get_tlg(_id)->go)
/** Get the transaction TLO of the property store object */
#define PROP_STORE_TLO(_id) _id
/** Get a property value from the property store */
#define PROP_STORE_GET(_id, _name, _type, _t, _default) \
prop_named_lookup( PROP_STORE_O(_id), _name, _type, (prop_value){ ._t = _default })._t

/** Set a property value to the property store */
#define PROP_STORE_SET(_id, _name, _type, _t, _val) \
{ prop_value v = { ._t = _val }; prop_store_set( _id, _name, _type, v); }

/** Get an unsigned property value from the property store */
#define PROP_STORE_GET_UNS(_id, _name, _default) PROP_STORE_GET(_id, _name, PT_UNS, u, _default)
/** Set an unsigned property value from to property store */
#define PROP_STORE_SET_UNS(_id, _name, _val) PROP_STORE_SET(_id, _name, PT_UNS, u, _val)
/** Set a general property value to the property store */
void prop_store_set( PROP_STORE_DEFINE(ps), const char * name, uns type, prop_value pv );

/** The global property store for various purposes (saving last values of some
 * option widgets etc. */
extern PROP_STORE_DEFINE(ps_global);
/** The recycler property store for recycling GO properties */
extern PROP_STORE_DEFINE(ps_recycler);

/** Go through the saved properties and if there are some that
 match, change their values in o accordingly.
 Must be called in a transaction.  */
void gui_prop_recycle( struct o * o );

/** Create a new property in the recycler according to the given arguments.
 * If a non-compatible property with the same key exists, replace the old one. */
void gui_prop_recycler_set( string key, uns type, uns subtype, uns unit, prop_value val );

#define GUI_STRINGS \
	string_tolerance, string_editor, string_editor_in_term, string_browser, string_browser_in_term, \
	string_terminal, string_grid_x1, string_grid_x2, string_grid_y1, string_grid_y2, \
	string_rulerx_res, string_rulery_res, string_drag_image, string_enable_fifi

extern string GUI_STRINGS;

/* ***** UNIT WINDOW ******/
/** The Unit manager window structure */
struct wnd_units
{
	WIN_O_MAGIC;
	GtkWidget * window;
	int visible;
	GtkComboBox * quantity_combo;
	GtkWidget * unit_view;
	GtkTreeViewColumn * unit_view_mult_col;	//!< The view column which shows the multipliers
	GtkWidget * delete_button;
	GtkWidget * default_button;
	GtkWidget * edit_button;
};
/** The one and the only Unit manager window instance */
extern struct wnd_units wnd_units;
/** The lists of units per quantities */
extern GtkListStore ** unit_quant_lists[];

void create_unit_quant_lists(void);
void destroy_unit_quant_lists(void);

enum { UNIT_NAME_COL, UNIT_MULT_COL, UNIT_INDEX_COL, UNIT_WEIGHT_COL, UNIT_COL_CNT };

int tree_iter_to_unit( GtkTreeModel * tm, GtkTreeIter * ti, uns * unit );
int unit_to_tree_iter( GtkTreeModel * tm, uns unit, GtkTreeIter * ti );

#endif
