/*
 * IMPORT - SVG
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <ctype.h>
#include <stdarg.h>
#include "kernel/kernel.h"
#include "import/import.h"
#include "font/font.h"

#define UNDEFINED_VALUE         NAN

#define SVG_CUBIC_BEZIER        0
#define SVG_QUADRATIC_BEZIER    1

#define ABSOLUTE_POSITION       0
#define RELATIVE_POSITION       1

static real height = 0;

#define DEFAULT_SVG_FONT_SIZE	15
static const char default_font_family[] = "Verdana";

#define SUCCESS 		0
#define ERROR 			1
#define READ_AGAIN 		2

struct svg_color
{
	char *name;
	uns rgb[3];
};

static const struct svg_color svg_def_color[] =
	{ {"aqua", {0, 255, 255}},
	  {"black", {0, 0, 0}},
	  {"blue", {0, 0, 255}},
	  {"fuchsia", {255, 0, 255}},
	  {"gray", {128, 128, 128}},
	  {"green", {0, 128, 0}},
	  {"lime", {0, 255, 0}},
	  {"maroon", {128, 0, 0}},
	  {"navy", {0, 0, 128}},
	  {"olive", {128, 128, 0}},
	  {"purple", {128, 0, 128}},
	  {"red", {255, 0, 0}},
	  {"silver", {192, 192, 192}},
	  {"teal", {0, 128, 128}},
	  {"white", {255, 255, 255}},
	  {"yellow", {255, 255, 0}} };

static real lastx, lasty;
	
struct go_style
{
	uns invis;
	uns stroke;
	real stroke_width;
	uns fill;
	uns stroke_linecap;
	uns stroke_join;
};

struct svg_smooth_data
{
	real controlx, controly;
	int type;
};

static real transform_x (real x)
{
	return x * POINT2MM;
}

static real transform_y (real y)
{
	return (height - y) * POINT2MM;
}

static real transform_len (real len)
{
	return len * POINT2MM;
}
		
static uns default_color (real opacity)
{
		return rgba_uns2uns(0, 0, 0, (uns) (255 * opacity));
}

static int sscanf_real (char *str, real *x)
{
	double g;
	int ret;

	ret = sscanf (str, "%lf", &g);
	if (ret > 0)
		*x = g;
	
	return ret;
}

static int get_str_attribute(xmlTextReaderPtr ptr, const char *attr, char **val)
{
	*val = (char *) xmlTextReaderGetAttribute (ptr, (xmlChar *) attr);
	if (!*val)
		return ERROR;
	return SUCCESS;
}

static int get_invisibility (xmlTextReaderPtr ptr)
{
	char *visibility;
	int ret = 0;
		
	if (get_str_attribute (ptr, "visibility", &visibility) == SUCCESS)
	{
		if (!strcmp (visibility, "hidden"))
			ret = 1;
		free (visibility);
	}

	return ret;
}

static int get_real_attribute(xmlTextReaderPtr ptr, const char *attr, real *val)
{
	char *tmp;
	
	if (get_str_attribute(ptr, attr, &tmp) == SUCCESS)
	{
		if (sscanf_real (tmp, val) > 0)
		{
			free (tmp);
			return SUCCESS;
		}
		free (tmp);
	}
	return ERROR; 
}

static int get_realset_attribute (xmlTextReaderPtr ptr, int n, ...)
{
	va_list ap;
	
	va_start (ap, n);
	for (int i = 0; i < n; i++)
	{
		char *str;
		real *r;

		str = va_arg(ap, char *);
		r = va_arg(ap, real *);	
		if (get_real_attribute (ptr, str, r) != SUCCESS)
		{
			va_end(ap);
			return ERROR;
		}
	}
	va_end(ap);
	return SUCCESS;
}

static int get_int_attribute(xmlTextReaderPtr ptr, const char *attr, int *val)
{
	char *tmp;
 
	if (SUCCESS == get_str_attribute(ptr, attr, &tmp))
	{
		if (sscanf (tmp, "%d", val) > 0)
		{
			free (tmp);
			return SUCCESS;
		}
		free (tmp);
	}
	return ERROR;
}

static char *skip_number (char *str)
{
	if(*str == '-' || *str == '+')
		str++;
	while (*str && (isdigit(*str) || (*str == '.')))
		str++;
	return str;
}

static char *skip_delimiters (char *str)
{
	while (*str && (isspace(*str) || (*str == ',') || (*str == ';')))
		str++;
	return str;
}

static int get_real_from_str (char **str, real *x)
{
	if (str)
	{
		*str = skip_delimiters (*str);
		if (sscanf_real (*str, x) == 1)
		{
			*str = skip_number (*str);
			return SUCCESS;
		}
	}
	return ERROR;
}

static int get_realset_from_str (char **str, int n, ...)
{
	va_list ap;
	
	va_start (ap, n);
	for (int i = 0; i < n; i++)
	{
		real *r;

		r = va_arg (ap, real *);	
		if (get_real_from_str (str, r) != SUCCESS)
		{
			va_end(ap);
			if (i)
				return ERROR;
			else
				return READ_AGAIN;
		}
	}
	va_end(ap);
	return SUCCESS;
}

static int is_delimiter (char c)
{
	switch (c)
	{
		case ' ':
		case '\t':
		case '\n':
		case ';':
		case '\0':
			return 1;
		default:
			return 0;
	}
}

static uns get_color_from_code (char *color, int len, real opacity)
{
	int n = len / 3;
	uns tmp, r, g, b;
	
	if ((len % 3) || (sscanf(color, "%x", &tmp) < 1))
		return default_color (opacity);
	if (n == 2)
	{
		b = tmp & 0xFF;
		g = (tmp >> 8) & 0xFF;
		r = tmp >> 16;
	}
	else if (n == 1)
	{
		b = tmp & 0xF;
		g = (tmp >> 4) & 0xF;
		r = tmp >> 8;
	}
	else
		return default_color (opacity);
	return rgba_uns2uns (r, g, b, opacity);
}

static uns get_predefined_svg_color (char *color, real opacity)
{
	for (uns i = 0; i < ARRAY_SIZE(svg_def_color); i++)
		if(!strcmp(color, svg_def_color[i].name))
			return rgba_uns2uns (svg_def_color[i].rgb[0],
					svg_def_color[i].rgb[1],
					svg_def_color[i].rgb[2], 255 * opacity);
	if (!strcmp(color, "none"))
		return rgba_uns2uns (0, 0, 0, opacity);
	return default_color (opacity);
}

static uns get_color(char *color, real opacity)
{
	uns r, g, b;

	if (!color || !*color)
		return default_color (opacity);
	if (sscanf(color, "rgb(%u,%u,%u)", &r, &g, &b) == 3)
		return rgba_uns2uns (r, g, b, (uns) (255 * opacity));
	if (*color == '#')
	{
		color++;
		return get_color_from_code (color, strlen(color), (uns) (255 * opacity));
	}
	return get_predefined_svg_color (color, (uns) (255 * opacity));
}

static int get_defined_str_value_from_str (char *str, char *type, char *value)
{
	char *beg = str, *tmp = str;
	int len = strlen (type);

	while (beg && *beg)
	{
		if (!strncmp (beg, type, len))
		{
			tmp += len + 1;
			while (*tmp && !is_delimiter (*tmp))
				*(value++) = *(tmp++);
			return SUCCESS;
		}
		// set new "block" beginning
		while (*tmp && !is_delimiter (*tmp)) // skip value
			tmp++;
		while (*tmp && is_delimiter (*tmp))	// and delimiters
			tmp++;
		beg = tmp;
	}
	return ERROR;
}

static int get_defined_int_value_from_str (char *str, char *type, int *value)
{
	char strvalue[1024];

	if (get_defined_str_value_from_str (str, type, strvalue) == SUCCESS)
	{
		if (sscanf (strvalue, "%d", value))
			return SUCCESS;
	}
	return ERROR;
}

static void fill_go_style (xmlTextReaderPtr ptr, struct go_style *gs)
{	
	char *color;
	char *linecap, *join;
	char tmp[1024];
	real stroke_opacity, fill_opacity;

	// invisibility
	gs->invis = get_invisibility (ptr);

	// stroke-opacity
	if (get_real_attribute (ptr, "stroke-opacity", &stroke_opacity) == ERROR)
		stroke_opacity = 1;

	// stroke-color
	if (get_str_attribute (ptr, "stroke", &color) == SUCCESS)
	{
		gs->stroke = get_color (color, stroke_opacity);
		free (color);
	}
	else if (get_str_attribute (ptr, "style", &color) == SUCCESS)
	{
		if (get_defined_str_value_from_str (color, "color", tmp) == SUCCESS)
			gs->stroke = get_color (tmp, stroke_opacity);
		else if (get_defined_str_value_from_str (color, "stroke", tmp) == SUCCESS)
			gs->stroke = get_color (tmp, stroke_opacity);
		else
			gs->stroke = default_color (stroke_opacity);
		free (color);
	}
	else
		gs->stroke = default_color (stroke_opacity);

	// stroke-width
	if (get_real_attribute (ptr, "stroke-width", &gs->stroke_width) == ERROR)
		gs->stroke_width = 0.0;
	
	// fill-opacity
	if (get_real_attribute (ptr, "fill-opacity", &fill_opacity) == ERROR)
		fill_opacity = 1.0;

	// fill-color
	if (get_str_attribute (ptr, "fill", &color) == SUCCESS)
	{
		gs->fill = get_color (color, fill_opacity);
		free (color);
	}
	else if (get_str_attribute (ptr, "fill", &color) == SUCCESS)
	{
		if (get_defined_str_value_from_str (color, "fill", tmp) == SUCCESS)
			gs->fill = get_color (tmp, fill_opacity);
		else
			gs->fill = default_color (fill_opacity);
		free (color);
	}
	else
		gs->fill = default_color (fill_opacity);

	// stroke-linecap
	if (get_str_attribute (ptr, "stroke-linecap", &linecap) == SUCCESS)
	{
		switch (linecap[0])
		{
			case 'b': 	// butt
				gs->stroke_linecap = 0;
				break;
			case 'r':	// round
				gs->stroke_linecap = 1;
				break;
			case 's':	// square
				gs->stroke_linecap = 2;
				break;
			default:
				gs->stroke_linecap = 0;
		}
		free (linecap);
	}
	else
		gs->stroke_linecap = 0;

	// stroke-join
	if (get_str_attribute (ptr, "stroke-join", &join) == SUCCESS)
	{
		switch (join[0])
		{
			case 'm':	// miter
				gs->stroke_join = 0;
				break;
			case 'r':	// round
				gs->stroke_join = 1;
				break;
			case 'b':	// bevel
				gs->stroke_join = 2;
				break;
			default:	
				gs->stroke_join = 0;
		}
		free (join);
	}
	else
		gs->stroke_join = 0;
}

static void set_go_style_to_o (struct o *o, struct go_style *gs)
{	
	prop_set_uns (o, string_invisible, (uintptr_t) gs-> invis);
	prop_set_uns (o, string_stroke_color, (uintptr_t) gs->stroke);
	prop_set_real (o, string_stroke_width, (uintptr_t) transform_len (gs->stroke_width));
	prop_set_uns (o, string_fill_color, (uintptr_t) gs->fill);
	prop_set_uns (o, string_stroke_cap, (uintptr_t) gs->stroke_linecap);
	prop_set_uns (o, string_stroke_join, (uintptr_t) gs->stroke_join);
}

static void null_bezier_smooth_data (struct svg_smooth_data *smooth)
{
	smooth->controlx = smooth->controly = UNDEFINED_VALUE;
}

static int get_bezier_smooth_data (struct svg_smooth_data *smooth,
		real *x, real *y, int type)
{
	if ((type != smooth->type) || 
		(smooth->controlx == UNDEFINED_VALUE) ||
		(smooth->controly == UNDEFINED_VALUE))
		return ERROR;
	*x = lastx + (lastx - smooth->controlx);
	*y = lasty + (lasty - smooth->controly);
	return SUCCESS;
}

static void set_bezier_smooth_data (struct svg_smooth_data *smooth,
		real x, real y, int type)
{
	smooth->controlx = x;
	smooth->controly = y;
	smooth->type = type;
}

static void create_line (real x1, real y1, real x2, real y2, struct go_style *gst)
{
	struct go_segment *gs;

	DBG("Creating line: %f %f %f %f\n", transform_x(x1), transform_y(y1), transform_x(x2), transform_y(y2));
	gs = go_segment_new (hanger_new_mc_xy(transform_x(x1), transform_y(y1)), 
			hanger_new_mc_xy(transform_x(x2), transform_y(y2)));
	go_link_head (&gs->go, &trans_tlo->tlg->group);
	
	set_go_style_to_o (go2o(&gs->go), gst);
}

static void create_circle (real cx, real cy, real r, struct go_style *gs)
{
	struct go_elarc *ge;

	ge = go_elarc_new_xyr (hanger_new_mc_xy(transform_x(cx), transform_y(cy)), 0, 
			transform_len(r), transform_len(r), 0, 2 * R_PI, 
			GEOM_CONIC_ARC_START_ENTIRE);
	go_link_head (&ge->go, &trans_tlo->tlg->group);

	set_go_style_to_o (go2o(&ge->go), gs);
}

static void create_cubic_bezier (real x1, real y1, real x2, real y2, real x3, real y3, real x4, real y4, struct go_style *gs)
{
	struct go_bezier *gb;

	gb = go_bezier_new_cubic (hanger_new_mc_xy(transform_x(x1), transform_y(y1)), 1.0, 
		hanger_new_mc_xy(transform_x(x2), transform_y(y2)), 1.0, 
		hanger_new_mc_xy(transform_x(x3), transform_y(y3)), 1.0, 
		hanger_new_mc_xy(transform_x(x4), transform_y(y4)), 1.0);
	go_link_head (&gb->go, &trans_tlo->tlg->group);

	set_go_style_to_o (go2o(&gb->go), gs);
}

static void create_quadratic_bezier (real x1, real y1, real x2, real y2, real x3, real y3, struct go_style *gs)
{
	struct go_bezier *gb;

	gb = go_bezier_new_quadratic (hanger_new_mc_xy(transform_x(x1), transform_y(y1)), 1.0, 
		hanger_new_mc_xy(transform_x(x2), transform_y(y2)), 1.0, 
		hanger_new_mc_xy(transform_x(x3), transform_y(y3)), 1.0);
	go_link_head (&gb->go, &trans_tlo->tlg->group);

	set_go_style_to_o (go2o (&gb->go), gs);
}

static void create_elarc (real x1 UNUSED, real y1 UNUSED, real x2 UNUSED, real y2 UNUSED, real rx UNUSED, real ry UNUSED, real rotation UNUSED, struct go_style *gs UNUSED)
{
	DBG("Creating elarc..\n");
}

static void create_text (real x, real y, int font_id, int font_size, char *text, uns clr, uns invis, struct geom_transform *tr)
{
	struct go_text *gt;
	
	gt = go_text_new (hanger_new_mc_xy(transform_x(tr->coef[0][2] + x), transform_y(tr->coef[1][2] + y)),
			0, 0,	// refp_x, refp_y
			0, 0,	// marg_x, marg_y
	string_lookup(text), font_id, transform_len(font_size)); 
	go_link_head (&gt->go, &trans_tlo->tlg->group);
	prop_set_uns (go2o (&gt->go), string_alignment_x, 1);	// center
	prop_set_uns (go2o (&gt->go), string_fill_color, (uintptr_t) clr);
	prop_set_uns (go2o (&gt->go), string_invisible, (uintptr_t) invis);
	prop_set_uns (go2o (&gt->go), string_alignment_x, PTAX_REF_LEFT);
	prop_set_uns (go2o (&gt->go), string_alignment_y, PTAY_BASELINE);
	prop_set (go2o (&gt->go), string_transform, (prop_value) { .t = tr });
}

static void set_last_point_coords (real x, real y)
{
	lastx = x;
	lasty = y;
}

static void make_absolute_coords (real *x, real *y)
{
	if (x)
		*x += lastx;
	if (y)
		*y += lasty;
}

static int path_parse (char **str, char c, int non_paired_args, int coord_pairs, ...)
{
	va_list ap;
	real *x, *y;

	if ((lastx == UNDEFINED_VALUE) || (lasty == UNDEFINED_VALUE))
		return ERROR;

	va_start (ap, coord_pairs);
	// non paired args
	for (int i = 0; i < non_paired_args; i++)
	{
		x = va_arg (ap, real *);
		if (get_real_from_str (str, x) != SUCCESS)
			return ERROR;
	}

	// paired coords
	for (int i = 0; i < coord_pairs; i++)
	{
		x = va_arg (ap, real *);
		y = va_arg (ap, real *);
		if (x && (get_real_from_str (str, x) != SUCCESS))
		{
			va_end(ap);
			if (!i)
				return READ_AGAIN;
			return ERROR;
		}
		if (y && (get_real_from_str (str, y) != SUCCESS))
		{
			va_end (ap);
			return ERROR;
		}
		if (c >= 'a')	// don't have to check, wheter it's a letter...
			make_absolute_coords(x, y);
	}

	va_end(ap);
	return SUCCESS;
}

static char *process_path_from_str (char *str, real *firstx, real *firsty, struct svg_smooth_data *smooth, struct go_style *gs)
{
	char c;
	int ret;
	real x1, y1, x2, y2, x3, y3;
	real rx, ry, rotation, flag1, flag2;
	
	if (!str)
		return NULL;
	while (*str && isspace(*str))
		str++;
	if (!*str)
		return NULL;
	DBG("path: %s\n",str);
	c = *(str++);
	switch(c)
	{
		case 'M':	// moveto
		case 'm':	// moveto (relatively positioned)
			null_bezier_smooth_data (smooth);
			if (get_realset_from_str (&str, 2, &lastx, &lasty) == SUCCESS)
			{
				*firstx = lastx;
				*firsty = lasty;
				ret = SUCCESS;
			}
			else
				ret =  ERROR;
			break;
		case 'L':	// lineto (absolutely positioned)
		case 'l':	// lineto (relatively positioned)
			null_bezier_smooth_data (smooth);
			while ((ret = path_parse (&str, c, 0, 1, &x1, &y1)) == SUCCESS)
			{
				create_line (lastx, lasty, x1, y1, gs);
				set_last_point_coords (x1, y1);
			}	
			break;
		case 'V':	// vertical lineto (absolutely positioned)
		case 'v':	// vertical lineto (relatively positioned)
			null_bezier_smooth_data (smooth);
			while ((ret = path_parse (&str, c, 0, 1, &y1)) == SUCCESS)
			{
				create_line (lastx, lasty, lastx, y1, gs);
				set_last_point_coords (lastx, y1);
			}
			break;
		case 'H': // horizontal lineto (absolutely positioned)
		case 'h': // horizontal lineto (relatively positioned)
			null_bezier_smooth_data (smooth);
			while ((ret = path_parse (&str, c, 0, 1, &x1, 0)) == SUCCESS)
			{
				create_line (lastx, lasty, x1, lasty, gs);
				set_last_point_coords (x1, lasty);
			}
			break;
		case 'C':	// curveto (cubic bezier) (absolutely positioned)
		case 'c':	// curveto (cubic bezier) (relatively positioned)
			while ((ret = path_parse (&str, c, 0, 3, &x1, &y1, &x2, &y2, &x3, &y3)) == SUCCESS)
			{
				create_cubic_bezier (lastx, lasty, x1, y1, x2, y2, x3, y3, gs);
				set_bezier_smooth_data (smooth, x2, y2, SVG_CUBIC_BEZIER);
				set_last_point_coords (x3, y3);
			}
			break;
		case 'S':	// smooth curveto (absolutely positioned)
		case 's':	// smooth curveto (relatively positioned)
			while ((ret = path_parse (&str, c, 0, 2, &x2, &y2, &x3, &y3)) == SUCCESS)
			{
				if (get_bezier_smooth_data (smooth, &x1, &y1, SVG_CUBIC_BEZIER)== SUCCESS)
				{
					create_cubic_bezier (lastx, lasty, x1, y1, x2, y2, x3, y3, gs); 
					set_bezier_smooth_data (smooth, x2, y2, SVG_CUBIC_BEZIER);
					set_last_point_coords (x3, y3);
				}
				else
					return NULL;
			}
			break;
		case 'Q':	// quadratic bezier curve (absolutely positioned)
		case 'q':	// quadratic bezier curve (relatively positioned)
			while ((ret = path_parse (&str, c, 0, 2, &x1, &y1, &x2, &y2)) == SUCCESS)
			{
				create_quadratic_bezier (lastx, lasty, x1, y1, x2, y2, gs);
				set_bezier_smooth_data (smooth, x1, y1, SVG_QUADRATIC_BEZIER);
				set_last_point_coords (x2, y2);
			}
			break;
		case 'T':	// smooth quadratic bezier curveto (absolutely positioned)
		case 't':	// smooth quadratic bezier curveto (relatively positioned)
			while ((ret = path_parse (&str, c, 0, 1, &x2, &y2)) == SUCCESS)
			{
				if (get_bezier_smooth_data (smooth, &x1, &y1, SVG_QUADRATIC_BEZIER)== SUCCESS)
				{
					create_quadratic_bezier (lastx, lasty, x1, y1, x2, y2, gs); 
					set_bezier_smooth_data (smooth, x1, y1, SVG_QUADRATIC_BEZIER);
					set_last_point_coords (x2, y2);
				}
				else
					return NULL;
			}
			break;
		case 'A':	// elliptical arc (absolutely positioned)
		case 'a':	// elliptical arc (relatively positioned)
			null_bezier_smooth_data (smooth);
			while ((ret = path_parse (&str, c, 5, 1, &rx, &ry, &rotation, &flag1, &flag2, &x1, &y1)) == SUCCESS)
			{
				create_elarc (lastx, lasty, x1, y1, rx, ry, rotation, gs);
				set_last_point_coords (x1, y1);
			}
			break;
		case 'Z':	// closepath
		case 'z':
			null_bezier_smooth_data (smooth);
			if ((lastx == UNDEFINED_VALUE) || (lasty == UNDEFINED_VALUE))
				return NULL;

			create_line (lastx, lasty, *firstx, *firsty, gs);
			set_last_point_coords (*firstx, *firsty);
			ret = SUCCESS;
			break;
		default:	// default : lineto
			null_bezier_smooth_data (smooth);
			str--;
			while ((ret = path_parse (&str, 'L', 2, 0, &x1, &y1)) == SUCCESS)
			{
				create_line (lastx, lasty, x1, y1, gs);
				set_last_point_coords (x1, y1);
			}	
	}
	if (ret == ERROR)
		return NULL;
	return str;
}

static void import_path (xmlTextReaderPtr ptr)
{
	char *d;
	real firstx, firsty;
	struct svg_smooth_data smooth;

	firstx = firsty = lastx = lasty = UNDEFINED_VALUE;
	null_bezier_smooth_data (&smooth);

	if (get_str_attribute(ptr, "d", &d) == SUCCESS)
	{
		DBG("Importing path - d: %s\n", d);
		char *tmp;

		struct go_style gs;
		fill_go_style (ptr, &gs);
		tmp = d;
		while (tmp)
			tmp = process_path_from_str (tmp, &firstx, &firsty,
					&smooth, &gs);
		free (d);
	}
}

static void import_rect (xmlTextReaderPtr ptr)
{
	real width, height;
	real x, y;

	if (get_realset_attribute (ptr, 4, "x", &x, "y", &y, "width", &width, "height", &height) == SUCCESS)
	{
		DBG("Importing rectangle - width: %f, height: %f)\n", width, height);
	
		struct go_style gs;
		fill_go_style (ptr, &gs);
		create_line (x, y, x + width, y, &gs);
		create_line (x + width, y, x + width, y + width, &gs);
		create_line (x + width, y + width, x, y + width, &gs);
		create_line (x, y + width, x, y, &gs);
	}	
}

static void import_ellipse (xmlTextReaderPtr ptr)
{
	real cx, cy, rx, ry;
	struct go_elarc *ge;

	DBG("ELLIPSE\n");
	if (get_realset_attribute (ptr, 4, "cx", &cx, "cy", &cy,
				"rx", &rx, "ry", &ry) == SUCCESS)
	{
		struct go_style gs;

		DBG("Importing ellipse - cx: %f, cy: %f, rx: %f, ry: %f\n", cx, cy, rx, ry);

		fill_go_style (ptr, &gs);
		ge = go_elarc_new_xyr (hanger_new_mc_xy(transform_x(cx), transform_y(cy)), 0, transform_len(rx), transform_len(ry), 0, 2 * R_PI, GEOM_CONIC_ARC_START_ENTIRE);
		go_link_head (&ge->go, &trans_tlo->tlg->group);
		set_go_style_to_o (go2o (&ge->go), &gs);
	}
}

static void import_circle (xmlTextReaderPtr ptr)
{
	real cx, cy, r;

	if (get_realset_attribute (ptr, 3, "cx", &cx, "cy", &cy, "r", &r) == SUCCESS)
	{
		struct go_style gs;

		fill_go_style (ptr, &gs);
		create_circle (cx, cy, r, &gs);
	}
}

static void import_line (xmlTextReaderPtr ptr)
{
	real x1, y1, x2, y2;

	if (get_realset_attribute (ptr, 4, "x1", &x1, "y1", &y1,
				"x2", &x2, "y2", &y2) == SUCCESS)
	{
		DBG("Importing line - x1: %f, y1: %f, x2: %f, y2: %f\n", x1, y1, x2, y2);

		struct go_style gs;
		fill_go_style (ptr, &gs);
		create_line (x1, y1, x2, y2, &gs);
	}
}

static void import_polyline (xmlTextReaderPtr ptr, int close_path)
{
	char *points, *str;
	real x1, y1, x2, y2;
	real x0, y0;
	struct go_style gs;

	if (get_str_attribute(ptr, "points", &points) == ERROR)
		return;
	DBG("Importing polygon - points: %s\n", points);
	str = points;
	fill_go_style (ptr, &gs);
	if (get_realset_from_str (&str, 2, &x1, &y1) == ERROR)
	{
		free (points);
		return;
	}
	if (close_path)
	{
		x0 = x1;
		y0 = y1;
	}

	while (get_realset_from_str (&str, 2, &x2, &y2) == SUCCESS)
	{
		create_line (x1, y1, x2, y2, &gs);
		x1 = x2;
		y1 = y2;
	}
	if (close_path)
		create_line (x2, y2, x0, y0, &gs);
	free (points);
}

static int get_font (xmlTextReaderPtr ptr, int *fontsize)
{
	char *style, *fontfamily = NULL, temp[1024] = {0, }, outbuf[1024] = {0, };
	
	if (get_str_attribute (ptr, "style", &style) == SUCCESS)
	{
		get_defined_str_value_from_str(style, "font-family", temp);
		get_defined_int_value_from_str(style, "font-size", fontsize);
		free (style);
	}
	if (!*temp)
		get_str_attribute(ptr, "font-family", (char **) &fontfamily);
	else
		fontfamily = strdup (temp);
	if (!fontfamily)
		fontfamily = strdup (default_font_family);
	if (!fontfamily)
	{
		DBG("Couldn't allocate memory for strdup!\n");
		return 0;
	}	
	if (!*fontsize)
		get_int_attribute (ptr, "font-size", fontsize);
	if (!*fontsize)
		*fontsize = DEFAULT_SVG_FONT_SIZE;	// default value

	if (!font_search (fontfamily, NULL, outbuf, 1024))
	{
		DBG("Couldn't load fontid from fontfamily\n");
		free (fontfamily);
		return 0;
	}
	if (fontfamily)
		free (fontfamily);
	return font_load_file (outbuf);
}

static void import_text (xmlTextReaderPtr ptr)
{
	struct geom_point pos;
	int font_size = 0, font_id;
	char *text = NULL;
	struct geom_transform *tr;
	real stroke_opacity;
	char *color, *transform;
	char tmp[1024];
	uns clr;

	DBG("Importing text\n");
	if (get_realset_attribute (ptr, 2, "x", &pos.x, "y", &pos.y) == ERROR)
		return;
	uns invis = get_invisibility (ptr);
	
	tr = xmalloc (sizeof (struct geom_transform));

	// stroke-opacity
	if (get_real_attribute (ptr, "stroke-opacity", &stroke_opacity) == ERROR)
		stroke_opacity = 1.0;
	
	// stroke-color
	if (get_str_attribute (ptr, "stroke", &color) == SUCCESS)
	{
		clr = get_color (color, (uns) stroke_opacity);
		free (color);
	}
	else if (get_str_attribute (ptr, "style", &color) == SUCCESS)
	{
		if (get_defined_str_value_from_str (color, "color", tmp) == SUCCESS)
			clr = get_color (tmp, stroke_opacity);
		else if (get_defined_str_value_from_str (color, "stroke", tmp) == SUCCESS)
			clr = get_color (tmp, stroke_opacity);
		else
			clr = default_color (stroke_opacity);
		free (color);
	}
	else
		clr = default_color (stroke_opacity);

	font_id = get_font (ptr, &font_size);
	if (get_str_attribute (ptr, "transform", &transform) == SUCCESS)
	{
		real a, b, c, d, e, f;
		if (sscanf (transform, "matrix(%f %f %f %f %f %f)", &a, &b, &c, &d, &e, &f) == 6)
			geom_transform_set (tr, a, -c, e, -b, d, f);
		else
			geom_transform_set (tr, 1, 0, 0, 0, 1, 0);
		free (transform);
	}
	else
		geom_transform_set (tr, 1, 0, 0, 0, 1, 0);
	
	geom_point_transform (&pos, tr, &pos);
	tr->coef[0][2] = tr->coef[1][2] = 0;

	if (!((xmlTextReaderRead (ptr) == 1) &&
		(xmlTextReaderNodeType (ptr) == XML_TEXT_NODE) &&
		(text = xstrdup ((char *) xmlTextReaderConstValue (ptr)))))
			DBG("Couldn't read text string from <text></text>\n");
	else if (!font_id)
		DBG("Couldn't get fontid from font file\n");
	else
		create_text (pos.x ,pos.y, font_id, font_size, text, clr, invis, tr);

	if (text)
		xfree (text);
}

void import_svg (char *filename)
{
	xmlTextReaderPtr ptr;
	char *tmp;

	ptr = xmlReaderForFile (filename, 0, 1 + 128);
	if (!ptr)
		return;
	while (xmlTextReaderRead (ptr) == 1)
	{
		if (xmlTextReaderNodeType (ptr) == XML_ELEMENT_NODE)
		{
#define TEST_IMPORT(str) else if (!strcmp (tmp, #str)) import_##str(ptr)
			tmp = (char *) xmlTextReaderName (ptr);
			if (!strcmp (tmp, "svg"))
				get_real_attribute(ptr, "height", &height);
			else if (!strcmp (tmp, "rect"))
				import_rect (ptr);
			else if (!strcmp (tmp, "ellipse"))
				import_ellipse (ptr);
			else if (!strcmp (tmp, "circle"))
				import_circle (ptr);
			else if (!strcmp (tmp, "line"))
				import_line (ptr);
			else if (!strcmp (tmp, "polygon"))
				import_polyline (ptr, 1);
			else if (!strcmp (tmp, "polyline"))
				import_polyline (ptr, 0);
			else if (!strcmp (tmp, "path"))
				import_path (ptr);
			else if (!strcmp (tmp, "text"))
				import_text (ptr);
		}
	}
	xmlFreeTextReader(ptr);
}
