/*
 * Kernel - GEOMLIB expansions
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"
#include "lib/hashfunc.h"

#ifdef DEBUG_EXPAND
#define DBG_EXPAND(...) msg(__VA_ARGS__)
#else
#define DBG_EXPAND(...) do { } while(0)
#endif

#define HASH_NODE struct expand_rtree
#define HASH_PREFIX(x) expand_sht_ ## x
#define HASH_KEY_ATOMIC go
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_INSERT
#define HASH_WANT_REMOVE
#define HASH_GIVE_ALLOC
static inline void expand_sht_free(struct expand_sht_table *table UNUSED, void *p UNUSED) { }
#define HASH_GIVE_HASHFN
static inline uns expand_sht_hash(struct expand_sht_table *table UNUSED, struct go *go) { return hash_int((intptr_t)go); }
#define HASH_CONSERVE_SPACE
#define HASH_ATOMIC_TYPE struct go*
#define HASH_TABLE_DYNAMIC
#include "lib/sht.h"

static inline struct expand_sht_bucket* rtree2bucket(struct expand_rtree *b)
	{ return SKIP_BACK(struct expand_sht_bucket, n, b); }

/*
 * Compute the box enclosing all visible parts of GO, hangers and anchors.
 */

#define EXPAND_BBOX(_id) case GO_TYPE(_id): ret = GLUEE(GO_NAME(_id), _bbox) ((GO_STRUCT(_id)*)go, bbox); break;

static int expand_get_bbox(struct go *go, struct geom_rectangle *bbox)
{
	int volatile ret = 0;
	DBG_EXPAND("expand_get_bbox: go %s at %p\n", go_lookup_name(go), go);
	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_ANONYMOUS(buf)
		GO_TYPE_SWITCH(GO_FOR_ALL_ID(EXPAND_BBOX));
	TRANS_FAILED
		DBG_EXPAND("Cannot compute bbox of go %s at %p: %s\n", go_lookup_name(go), go, buf);
	TRANS_END
	return ret;
}

const struct geom_rectangle* expand_insert(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo);
	struct expand_sht_bucket *b;
	struct geom_rectangle bbox;
	if(!tlo->expand || go->type == GOT_GROUP)
		return NULL;
	DBG_EXPAND("expand_insert: tlo %s at %p, go %s at %p\n", obj_lookup_name(&tlo->obj), tlo, go_lookup_name(go), go);
	ASSERT(!expand_sht_find(tlo->expand, go));
	b = xmalloc_zero(sizeof(struct expand_sht_bucket));
	if(!expand_get_bbox(go, &bbox))
	{
		xfree(b);
		return NULL;
	}
	geom_rtree_insert(&tlo->rtree, &b->n.rtree, &bbox);
	expand_sht_insert(tlo->expand, go, b);
	return &b->n.rtree.bbox;
}

const struct geom_rectangle* expand_update(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo);
	struct expand_rtree *b;
	struct geom_rectangle bbox;
	if(!tlo->expand)
		return NULL;
	if(!(b = expand_sht_find(tlo->expand, go)))
		return expand_insert(tlo, go);
	DBG_EXPAND("expand_update: tlo %s at %p, go %s at %p\n", obj_lookup_name(&tlo->obj), tlo, go_lookup_name(go), go);
	if(!expand_get_bbox(go, &bbox))
	{
		geom_rtree_delete(&tlo->rtree, &b->rtree);
		expand_sht_remove(tlo->expand, b);
		xfree(rtree2bucket(b));
		return NULL;
	}
	geom_rtree_update(&tlo->rtree, &b->rtree, 0, &bbox);
	return &b->rtree.bbox;
}

int expand_remove(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo || go_get_tlo(go) == tlo_zombie);
	struct expand_rtree *b;
	if(!tlo->expand || !(b = expand_sht_find(tlo->expand, go)))
		return 0;
	DBG_EXPAND("expand_remove: tlo %s at %p, go %s at %p\n", obj_lookup_name(&tlo->obj), tlo, go_lookup_name(go), go);
	geom_rtree_delete(&tlo->rtree, &b->rtree);
	expand_sht_remove(tlo->expand, b);
	if(b->usage <= 0)
		xfree(rtree2bucket(b));
	return 1;
}

void expand_ref(struct obj_tlo *tlo)
{
	ASSERT_TLO(tlo);
	tlo->expand_ref_count++;
	DBG_EXPAND("  Increase expand reference: tlo %s at %p, new ref_count %d\n", obj_lookup_name(&tlo->obj), tlo, tlo->expand_ref_count);
	if(tlo->expand_ref_count > 1)
		return;

	ASSERT(!tlo->expand);
	struct go *go;
	struct go_group *tlg = &tlo->tlg->group, *owner;
	geom_rtree_init(&tlo->rtree);
	tlo->expand = xmalloc(sizeof(struct expand_sht_table));
	expand_sht_init(tlo->expand);

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_ANONYMOUS(buf)
		WALK_GROUP_SMART(tlg, 1, go, owner, , expand_insert(tlo, go), ); // expand_insert(tlo, &owner->go));
	TRANS_FAILED
		die(buf);
	TRANS_END
}

void expand_unref(struct obj_tlo *tlo)
{
	ASSERT_TLO(tlo);
	ASSERT(tlo->expand_ref_count);
	tlo->expand_ref_count--;
	DBG_EXPAND("  Decrease expand reference: tlo %s at %p, new ref_count %d\n", obj_lookup_name(&tlo->obj), tlo, tlo->expand_ref_count);
	if(tlo->expand_ref_count > 0)
		return;

	HASH_FOR_ALL_DYNAMIC(expand_sht, tlo->expand, b)
		DANGEROUS(b->usage <= 0, "Somebody has %u reference to expansion %p of go %s at %p.\n", b->usage, b, go_lookup_name(b->go), b->go);
//		msg("free: %p, go %s at %p\n", b, go_lookup_name(b->go), b->go);
//		expand_sht_remove(tlo->expand, b);
//		xfree(rtree2bucket(b));
	HASH_END_FOR;

	geom_rtree_clear(&tlo->rtree);
	expand_sht_cleanup(tlo->expand);
	xfree(tlo->expand);
	tlo->expand = NULL;
}

/**
 * Returns expansion box of given GO.
 *
 * @ai tlo  You must have expand reference of this TLO to use this function.
 * @ai go   GO which belong into TLO.
 *
 * @rv nonzero Box enclosing all visible parts of GO, hangers and anchors. Read only.
 * @rv NULL    Numeric error.
 */

const struct geom_rectangle* go_get_expand_bbox(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo && tlo->expand);
	struct expand_rtree *b = expand_sht_find(tlo->expand, go);
//	o_dump(go2o(go)); msg("\t"); geom_rectangle_msg(b ? &b->rtree.bbox : NULL); msg("\n");
	return b ? &b->rtree.bbox : NULL;
}

/**
 * Returns expansion box of given TLO.
 *
 * @ai tlo  You must have expand reference of this TLO to use this function.
 *
 * @rv Box enclosing all visible parts in TLO, hangers and anchors. Read only.
 */

const struct geom_rectangle* tlo_get_expand_bbox(struct obj_tlo *tlo)
{
	ASSERT(tlo->expand);
	return &tlo->rtree.root->bbox;
}
