/*
 * VRR plugin mechanism
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "kernel/kernel.h"
#include "lib/clists.h"
#include "lib/garr.h"
#include <string.h>
#include <dlfcn.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdio.h>

void (*plugin_change_callback)(void);

clist plugin_list;
uns plugin_count;

void plugin_init(void)
{
	clist_init(&plugin_list);
	plugin_change_callback = NULL;
	plugin_count = 0;
}

void plugin_finish(void)
{
	struct plugin_rec *pr;
	struct plugin_function *fr;

	while (pr = clist_head(&plugin_list)) {
		DBG("Destroying plugin '%s' records\n", pr->name);
		void (*f)(void);

		f = dlsym(pr->handle, "plugin_stop");
		if (f)
			f();

		while (fr = clist_head(&pr->func_list)) {
			clist_remove(&fr->n);
			xfree(fr);
		}
		clist_remove(&pr->n);
		dlclose(pr->handle);
		xfree(pr);
	}
}

static void *current_plugin;

int plugin_load(const char *filename)
{
	void *r;
	u32 (*func)(void);
	char buf[4096];

	if (filename[0] != '/') {
		if (!getcwd(buf, sizeof(buf)-1)) {
			trans_fail("getcwd() failed");
			return -1;
		}
		strcat(buf, "/");
		strncat(buf, filename, sizeof(buf) - strlen(buf) - 1);
		filename = buf;
	}
	DBG("Loading plugin '%s'...\n", filename);
	r = dlopen(filename, RTLD_GLOBAL | RTLD_NOW);
	if (!r) {
		trans_fail("Plugin load error: %s", dlerror());
		return -1;
	}

	struct plugin_rec *pr = xmalloc_zero(sizeof(*pr) + strlen(filename));
	clist_init(&pr->func_list);
	pr->handle = r;
	strcpy(pr->name, filename);

	func = dlsym(r, "plugin_start");
	u32 flags = 0;
	if (func) {
		DBG("Calling start function...\n");
		current_plugin = pr;
		flags = func();
		current_plugin = NULL;
		DBG("Returned flags 0x%x\n", flags);
	}
	pr->flags = flags;
	clist_add_tail(&plugin_list, &pr->n);
	plugin_count++;
	DBG("Done.\n");
	if (plugin_change_callback)
		plugin_change_callback();

	return 0;
}

extern void register_plugin_fn (const char *name, unsigned long specific);

int plugin_function_register(const char *funcname, const char *comment, uns retval, uns argc, ...)
{
	struct plugin_rec *pr = current_plugin;
	struct plugin_function *fr;
	void *f;
	va_list va;

	DBG("Registering scheme function '%s' of plugin '%s', argc=%d\n", funcname, pr->name, argc);
	f = dlsym(pr->handle, funcname);
	if (!f) {
		trans_fail("Symbol '%s' load failure", funcname);
		return -1;
	}
	if (argc > PLUGIN_MAX_ARGS) {
		trans_fail("Too many arguments");
		return -1;
	}

	uns com_size = comment ? strlen(comment) : 0;

	fr = xmalloc_zero(sizeof(*fr) + com_size);
	fr->func = f;
	strncpy(fr->name, funcname, sizeof(fr->name)-1);
	if (comment)
		strcpy(fr->comment, comment);

	fr->retval = retval;
	fr->argc = argc;
	va_start(va, argc);
	uns t;
	for (uns i=0; i<argc; i++) {
		t = va_arg(va, uns);
		DBG("arg %d is 0x%x\n", i, t);
		fr->prototype[i] = t;
	}
	va_end(va);

	ASSERT (sizeof (unsigned long) >= sizeof (void *));
	register_plugin_fn (funcname, (unsigned long) fr);

	clist_add_tail(&pr->func_list, &fr->n);
	pr->func_count++;
	DBG("Done.\n");

	return 0;
}

int plugin_unload(void *plugin)
{
	struct plugin_rec *pr = plugin;
	struct plugin_function *fr;
	void (*f)(void);

	DBG("Unloading plugin '%s'...\n", pr->name);
	if (!(pr->flags & PLUGIN_UNLOADABLE)) {
		DBG("Not unloadable\n");
		return -1;
	}

	f = dlsym(pr->handle, "plugin_stop");
	if (f)
		f();
	
	while (fr = clist_head(&pr->func_list)) {
		clist_remove(&fr->n);
		xfree(fr);
	}
	clist_remove(&pr->n);
	dlclose(pr->handle);
	xfree(pr);
	plugin_count--;
	DBG("Done.\n");

	if (plugin_change_callback)
		plugin_change_callback();

	return 0;
}

static const char *typename(uns type)
{
	char *s;

	switch (type) {
	case PLUGIN_T_VOID:
		s = "void";
		break;
	case PLUGIN_T_INT:
		s = "int";
		break;
	case PLUGIN_T_UNS:
		s = "uns";
		break;
	case PLUGIN_T_REAL:
		s = "real";
		break;
	case PLUGIN_T_BOOL:
		s = "bool";
		break;
	case PLUGIN_T_O:
		s = "struct o *";
		break;
	case PLUGIN_T_OBJ:
		s = "struct obj *";
		break;
	case PLUGIN_T_OBJ_TLO:
		s = "struct obj_tlo *";
		break;
	case PLUGIN_T_OBJ_DOC:
		s = "struct obj_doc *";
		break;
	case PLUGIN_T_GO:
		s = "struct go *";
		break;
	case PLUGIN_T_GO_BEZIER:
		s = "struct go_bezier *";
		break;
	case PLUGIN_T_GO_ELARC:
		s = "struct go_elarc *";
		break;
	case PLUGIN_T_GO_SEGMENT:
		s = "struct go_segment *";
		break;
	case PLUGIN_T_GO_POINT:
		s = "struct go_point *";
		break;
	case PLUGIN_T_GO_TEXT:
		s = "struct go_text *";
		break;
	case PLUGIN_T_GO_TEX_TEXT:
		s = "struct go_tex_text *";
		break;
	case PLUGIN_T_GO_PARAMETRIC_POINT:
		s = "struct go_parametric_point *";
		break;
	case PLUGIN_T_GO_INTERSECTION:
		s = "struct go_intersection *";
		break;
	case PLUGIN_T_GO_GROUP:
		s = "struct go_group *";
		break;
	case PLUGIN_T_STRING_C:
	case PLUGIN_T_SYMBOL_C:
		s = "const char *";
		break;
	case PLUGIN_T_STRING_S:
	case PLUGIN_T_SYMBOL_S:
		s = "string";
		break;
	case PLUGIN_T_HANGER:
		s = "struct hanger *";
		break;
	case PLUGIN_T_ANCHOR:
		s = "struct anchor *";
		break;
	default:
		s = "???";
	}

	return s;
}

char *plugin_function_prototype(struct plugin_function *f, char *buf, uns bufsize)
{
	uns n = bufsize;
	const char *s;
	char *t;
	uns l, i;

	l = snprintf(buf, n, "%s %s(", typename(f->retval), f->name);
	if (l >= n) {
		*buf = 0;
		return buf;
	}
	t = buf + l;

	for (i = 0; i < f->argc; i++) {
		l = snprintf(t, n - (t-s), "%s%s", typename(f->prototype[i]), (i == f->argc-1 ? "" : ", "));
		if (l >= n - (t-s)) {
			*t = 0;
			return buf;
		}
		t += l;
	}
	if ((int)n > (t-s)) {
		*t++ = ')';
		*t = 0;
	}
	return buf;
}

/* Debugging & testing only */

void plugin_dbg(const char *plugname)
{
	struct plugin_rec *pl;
	struct plugin_function *fc;

	if (plugin_load(plugname)<0)
		err("Plugin '%s' load failed\n", plugname);
	else {
		msg("Loaded %d plugins\n", plugin_count);
		CLIST_WALK(pl, plugin_list) {
			msg("Plugin %s:\n", pl->name);
			DBG("Registered %d functions\n", pl->func_count);
			CLIST_WALK(fc, pl->func_list) {
				char buf[1024];

				msg("function %s\n", plugin_function_prototype(fc, buf, sizeof(buf)));
			}
		}
	}
}

