/*
 * Kernel - Property and unit operation
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

//#define LOCAL_DEBUG
#include "kernel/internal.h"
#include "lib/hashfunc.h"


/* The last item is used for searching with holdback. */
struct prop_unit_struct unit_array[PROP_UNIT_MAX + 1] =
{
	{ NULL, 1, PQ_LENGTH },
	{ NULL, 1, PQ_ANGLE },
	{ NULL, 1, PQ_REFERENCE },
};

uns unit_default[PQ_MAX] =
{
	[PQ_NONE] = PROP_UNIT_MAX,
	[PQ_LENGTH] = 0,
	[PQ_ANGLE] = 1,
	[PQ_REFERENCE] = 2,
};

uns prop_subtype2quantity[PT_MAX][PROP_SUBTYPE_MAX] =
{
	[PT_UNS] = { PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE },
	[PT_POINTER] = { PQ_NONE, PQ_NONE, },
	[PT_STRING] = { PQ_NONE, PQ_NONE, PQ_NONE, },
	[PT_REAL] =
	{
		[PTR_UNSPECIFIED] = PQ_NONE,
		[PTR_POSITIVE] = PQ_NONE,
		[PTR_COORDINATE] = PQ_LENGTH,
		[PTR_LENGTH] = PQ_LENGTH,
		[PTR_ANGLE_PI] = PQ_ANGLE,
		[PTR_ANGLE_2PI] = PQ_ANGLE,
		[PTR_ANGLE_4PI] = PQ_ANGLE,
		[PTR_REFERENCE] = PQ_REFERENCE,
	},
};


struct prop_name
{
	struct o* o;
	string name;
};


/*
 * Hash table name -> o
 */

#define HASH_NODE struct prop_name
#define HASH_PREFIX(x) pn_name_##x
#define HASH_KEY_ATOMIC name
#define HASH_ATOMIC_TYPE string
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_INSERT
#define HASH_WANT_REMOVE
#define HASH_GIVE_ALLOC
static inline void pn_name_free(void *p UNUSED) { }
#define HASH_GIVE_HASHFN
static uns pn_name_hash(string name) { return hash_ptr(name); }
#define HASH_CONSERVE_SPACE
#include "lib/sht.h"


/*
 * Hash table o -> name
 */

struct prop_o  // FIXME: ugly hack
{
	struct pn_name_bucket *next;
	struct o* o;
	string name;
};

#define HASH_NODE struct prop_o
#define HASH_PREFIX(x) pn_o_##x
#define HASH_KEY_ATOMIC o
#define HASH_ATOMIC_TYPE struct o*
#define HASH_WANT_CLEANUP
#define HASH_WANT_INSERT
#define HASH_WANT_FIND
#define HASH_WANT_REMOVE
#define HASH_GIVE_ALLOC
static inline void pn_o_free(void *p UNUSED) { }
#define HASH_GIVE_HASHFN
static uns pn_o_hash(struct o* o) { return hash_ptr(o); }
#define HASH_CONSERVE_SPACE
#include "lib/sht.h"


/* request for initialization fixed-size allocator */
const uns prop_hash_name_size = sizeof(struct pn_o_bucket);


/*
 * type-cast functions
 */
static inline struct prop_o* name2o(struct prop_name* pn)
	{ return SKIP_BACK(struct prop_o, o, &pn->o); }

static inline struct prop_name* o2name(struct prop_o* po)
	{ return SKIP_BACK(struct prop_name, o, &po->o); }

static inline struct pn_name_bucket* name2bucket(struct prop_name *pn)
	{ return SKIP_BACK(struct pn_name_bucket, n, pn); }

static inline struct pn_o_bucket* o2bucket(struct prop_o *po)
	{ return SKIP_BACK(struct pn_o_bucket, n, po); }


/*
 * Inicialization and clean-up functions
 */
void property_init(void)
{
	unit_array[0].name = string_lookup_ref("mm");
	unit_array[1].name = string_lookup_ref("rad");
	unit_array[2].name = string_lookup_ref("param");
	pn_name_init();
	pn_o_init();
}

void property_finish(void)
{
	DANGEROUS(pn_o_table.hash_count == 0 && pn_name_table.hash_count == 0, "Error in property hash table\n");

	pn_name_cleanup();
	pn_o_cleanup();

	for(uns i = 0; i < PROP_UNIT_MAX; i++)
		if(unit_array[i].name)
			string_unref(unit_array[i].name);
}

void prop_finish(struct o *o)
{
	ASSERT_KERNEL(o->prop.head.next);
	struct prop *p;
	while((p = prop_get_first(o)) && !(p->flags & PTF_VIRTUAL))
	{
		string_unref(p->key);
		prop_value_unref(p->type, p->value);
		slist_remove(&o->prop, &p->n);
		ffree(p);
	}
	struct prop_o* po = pn_o_find(o);
	if(po)
	{
		pn_o_remove(po);
		pn_name_remove(o2name(po));
		string_unref(po->name);
		ffree(SKIP_BACK(struct pn_o_bucket, n, po));
	}
#ifdef DEBUG_KERNEL
	o->prop.head.next = NULL;
#endif
}


/*
 * Create property
 */

void prop_create(struct o *o, string key, uns type, uns subtype, uns unit, prop_value value)
{
	ASSERT(type != PT_REAL || isfinite (value.r));
	ASSERT_O_TRANS(o);
	struct prop* p = prop_find(o, key);

	if(p)
	{
		if(type != p->type || subtype != p->subtype)
			trans_fail("Property exists with different type or subtype.");
		if(p->flags & PTF_VIRTUAL) 										// change a virtual property
		{
			if(prop_subtype2quantity[type][subtype] != PQ_NONE)
				prop2virt(p)->set_unit(o, unit);
			if(!prop_value_compare(type, value, prop2virt(p)->get(o)))
				prop2virt(p)->set(o, value);
		}
		else																			// change a non-virtual property
		{
			p->unit = unit;
			if(!prop_value_compare(type, p->value, value))
			{
		 		TRANS_NEW_UNDO(TRANS_PROP_CHANGE, prop, o_ref(o), p, p->value); // If type is PT_STRING then reference of value is moved from property to undo.
				p->value = prop_value_ref(type, value);
			}
			hook_call_prop_change(o);
		}
	}
	else																			// create a new propety
	{
		p = fmalloc(trans_tlo->fprop);
		*p = (struct prop) { .key = string_ref(key), .value = prop_value_ref(type, value), .type = type, .subtype = subtype, .flags = PTF_SAVE | PTF_UNDO_CREATE, .unit = unit };
		slist_add(&o->prop, &p->n);
 		TRANS_NEW_UNDO(TRANS_PROP_CREATE, prop, .o = o_ref(o), .prop = p);
		hook_call_prop_create(o, p);
	}

	if(o->kind == T_OBJ)
		obj_changed(o2obj(o));
}

/*
 * Properties searching.
 */

struct prop* prop_find(struct o *o, string key)
{
	struct prop* p;
	WALK_PROP(o, p)
		if(p->key == key)
			return p;
	return NULL;
}

prop_value prop_get_value(struct o *o, struct prop *p)
{
	prop_value v = (p->flags & PTF_VIRTUAL) ? prop2virt(p)->get(o) : p->value;
	ASSERT(p->type != PT_POINTER || p->subtype != PTP_TRANSFORM || geom_transform_valid(v.t));
	return v;
}

real prop_lookup_real_unit(struct o *o, string key, real err)
{
	struct prop *p = prop_find(o, key);

	if(!p || p->type != PT_REAL || prop_subtype2quantity[p->type][p->subtype] == PQ_NONE)
		return err;

	real value = prop_get_value(o, p).r / prop_get_unit_multiplier(o, p);
	if(!isfinite(value))
		return err;

	return value;
}


/*
 * Change value of a property.
 */

void prop_set(struct o *o, string key, prop_value value)
{
	ASSERT_O_TRANS(o);
	struct prop* p = prop_find(o, key);

	if(!p)
		trans_fail("Property %s does not exist.", key->text);
	ASSERT(p->type != PT_REAL || isfinite (value.r));

	if(p->flags & PTF_VIRTUAL) 		// change a virtual property
	{
		if(!prop_value_compare(p->type, value, prop2virt(p)->get(o)))
			prop2virt(p)->set(o, value);
	}
	else					// change a non-virtual property
	{
		if(prop_value_compare(p->type, p->value, value))
			return;
  		TRANS_NEW_UNDO(TRANS_PROP_CHANGE, prop, o_ref(o), p, p->value);  // If type is PT_STRING then reference of value is moved from property to undo.
		p->value = prop_value_ref(p->type, value); // Reference for property
		hook_call_prop_change(o);
	}
	if(o->kind == T_OBJ)
		obj_changed(o2obj(o));
}

void prop_set_real_unit(struct o *o, string key, real value)
{
	ASSERT(trans_present);
	struct prop *p = prop_find(o, key);

	if(!p)
		trans_fail("Property %s does not exist.", key->text);
	if(p->type != PT_REAL || prop_subtype2quantity[p->type][p->subtype] == PQ_NONE)
		trans_fail("Invalid property type.");

	value *= prop_get_unit_multiplier(o, p);
	if(!isfinite(value))
		trans_fail("Numeric error occured.");

	prop_set(o, key, (prop_value){.r = value});
}

void prop_named_set_real_unit(struct o *o, const char* key, real value)
{
	string s = string_find(key);
	if(!s)
		trans_fail("Property %s does not exist.", key);
	prop_set_real_unit(o, s, value);
}

static inline struct geom_transform* transform_copy(const struct geom_transform *t)
{
	ASSERT(geom_transform_valid(t));
	struct geom_transform *c = memcpy(xmalloc(sizeof(struct geom_transform)), t, sizeof(struct geom_transform));
	ASSERT(geom_transform_valid(c));
	return c;
}

void prop_saving_copy(struct o* source, struct o* target)
{
	struct prop *p;
	
	WALK_PROP(source, p)
		if(p->flags & PTF_SAVE)
		{
			if(p->type == PT_POINTER && p->subtype == PTP_TRANSFORM)
				prop_create(target, p->key, p->type, p->subtype, 0, (prop_value){.t = transform_copy(prop_get_value(source, p).t)});
			else 
				prop_create(target, p->key, p->type, p->subtype,
					prop_subtype2quantity[p->type][p->subtype] == PQ_NONE ? 0 : prop_get_unit(source, p),
					prop_get_value(source, p));
		}
}

/*
 * Properties deletion.
 */

void prop_delete(struct o *o, struct prop* p)
{
	ASSERT_O_TRANS(o);
	if(p->flags & PTF_VIRTUAL)
		TRANS_FAIL1("Cannot delete virtual property %s.", p->key->text);

	slist_remove(&o->prop, &p->n);
	p->n.next = NULL;
	hook_call_prop_delete(o, p);

	TRANS_NEW_UNDO(TRANS_PROP_DELETE, prop, .o = o_ref(o), .prop = p);
	p->flags |= PTF_UNDO_DELETE;	
	if(o->kind == T_OBJ)
		obj_changed(o2obj(o));
}

void prop_find_delete(struct o *o, string key)
{
	ASSERT_O_TRANS(o);
	struct prop* p = prop_find(o, key);
	if(!p)
		TRANS_FAIL1("Property %s does not exist.", key->text);
	prop_delete(o, p);
}


/*
 * Property unit functions
 */

uns unit_find(string name)
{
	uns i;
	unit_array[PROP_UNIT_MAX].name = name;
	for(i = 0; unit_array[i].name != name; i++);
	return i;
}

uns prop_get_unit(struct o *o, struct prop *prop)
{
	if((prop->type != PT_REAL) || (prop_subtype2quantity[prop->type][prop->subtype] == PQ_NONE))
		return PROP_UNIT_MAX;
	uns u = (prop->flags & PTF_VIRTUAL) ? prop2virt(prop)->get_unit(o) : prop->unit;
	if(u == PROP_UNIT_MAX || !unit_array[u].name || prop_get_quantity(prop) != unit_array[u].quantity)
		u = unit_default[prop_get_quantity(prop)];
	return u;
}

void prop_set_unit(struct o *o, struct prop *prop, uns unit)
{
	ASSERT((unit < PROP_UNIT_MAX && unit_array[unit].name && prop_get_quantity(prop) == unit_array[unit].quantity) || unit == PROP_UNIT_MAX);
	if(prop->flags & PTF_VIRTUAL)
		prop2virt(prop)->set_unit(o, unit);
	else
		prop->unit = unit;
}

uns unit_set(string name, real mult, uns quantity)
{
	ASSERT(name && trans_present && quantity != PQ_NONE && quantity < PQ_MAX);
	
	if(!isfinite(mult))
		trans_fail("Invalid multiplier.");
	if(mult < EQ_EPSILON)
		trans_fail("Too small multiplier.");
	if(mult > 1/EQ_EPSILON)
		trans_fail("Too large multiplier.");
	
	static uns last_set = 0;
	uns i = unit_find(name);

	if(i == PROP_UNIT_MAX)
	{
		unit_array[PROP_UNIT_MAX].name = NULL;	
		for(i = last_set + 1; unit_array[i].name; i++);
		if(i == PROP_UNIT_MAX)
			for(i = 0; unit_array[i].name; i++);
		if(i == PROP_UNIT_MAX)
			trans_fail("Too many units.");
		last_set = i;
		unit_array[i] = (struct prop_unit_struct){ string_ref(name), mult, quantity };
		hook_call_unit_add(i);
	}
	else
	{
		unit_array[i].multiplier = mult;
		unit_array[i].quantity = quantity;
		hook_call_unit_changed(i);
	}

	return i;
}

void unit_delete(uns index)
{
	ASSERT(index < PROP_UNIT_MAX && trans_present);
	if(unit_default[unit_array[index].quantity] == index)
		trans_fail("This unit is used as default unit.");
	if(!unit_array[index].name)
		trans_fail("This unit does not exist.");
 	string_unref(unit_array[index].name);
	unit_array[index].name = NULL;
	hook_call_unit_deleted(index);
}

void unit_change_name(uns index, string name)
{
	ASSERT(index < PROP_UNIT_MAX && trans_present);
	
	if(!unit_array[index].name)
		trans_fail("This unit does not exist.");
	if(unit_find(name) != PROP_UNIT_MAX)
		trans_fail("This unit already exists.");
	
	string_unref(unit_array[index].name);
	unit_array[index].name = string_ref(name);
	hook_call_unit_changed(index);
}

void unit_default_set(uns quantity, uns index)
{
	ASSERT(quantity != PQ_NONE && quantity < PQ_MAX && index < PROP_UNIT_MAX && trans_present);
	if(!unit_array[index].name)
		trans_fail("This unit does not exist.");
	if(unit_array[index].quantity != quantity)
		trans_fail("Different quantity.");
	unit_default[quantity] = index;
	hook_call_unit_default(quantity);
}


/*
 * Property name functions
 */

void prop_name_set_low(struct o* o, string name, struct obj_tlo *alloc)
{
	ASSERT(name != string_noname && !pn_o_find(o));
	struct pn_o_bucket *pn_new = fmalloc(alloc->fname);
	pn_o_insert(o, pn_new);
	pn_name_insert(name, name2bucket(o2name(&pn_new->n)));
}

prop_value o_get_prop_name(struct o *o)
{
	ASSERT_O(o);
	struct prop_o * po = pn_o_find(o);
	return (prop_value){.s = po ? po->name : string_noname};
}

static void o_set_prop_name_low(struct o *o, prop_value name)
{
	DBG("o_set_prop_name_low, o %p, name %s\n", o, name.s->text);
	struct prop_o *po = pn_o_find(o);

	if(name.s == string_noname)
	{
		ASSERT(po);		// This function is not called if old and new name is same.
		pn_o_remove(po);
		pn_name_remove(o2name(po));
		string_unref(po->name);
		ffree(o2bucket(po));
		return;
	}
	if(po)
	{
		ASSERT(po->name != name.s);
		string_unref(po->name);
		pn_name_remove(o2name(po));
		pn_name_insert(string_ref(name.s), name2bucket(o2name(po)));
	}
	else
		prop_name_set_low(o, string_ref(name.s), trans_tlo);
	hook_call_prop_change(o);
}

static void o_set_prop_name(struct o *o, prop_value name)
{
	trans_new_undo_prop_virtual(o, &o_prop_name, o_get_prop_name(o));
	o_set_prop_name_low(o, name);
}

struct prop_virtual o_prop_name = {
	{ { NULL }, NULL, {0}, PT_STRING, PTS_UNSPECIFIED, PTF_VIRTUAL | PTF_SAVE, 0 },
	UINT_MAX, o_get_prop_name, o_set_prop_name, o_set_prop_name_low,
	prop_unit_get_bug, prop_unit_set_bug
};

struct o* o_lookup(string name)
{
	struct prop_name *pn = pn_name_find(name);
	if(pn)
		return pn->o;

	DBG("o_lookup: %s not found\n", name->text);
	return NULL;
}


/*
 * Compare function for property value
 */

uns prop_value_compare(uns prop_type, prop_value v1, prop_value v2)
{
	switch(prop_type)
	{
		case PT_UNS:
			return v1.u == v2.u;
		case PT_POINTER:
			return v1.v == v2.v;
		case PT_STRING:
			return v1.s == v2.s;
		case PT_REAL:
			return v1.r == v2.r;
		default:
			die("Invalid property type.");
	}
}


/*
 * Virtual property functions.
 */

void prop_virtual_init(uns start, ...)
{
	struct prop_virtual *v;
	va_list ap;
	va_start(ap, start);
	while((v = va_arg(ap, struct prop_virtual*)))
	{
		ASSERT(v->p.key == NULL);
		v->p.key = string_lookup_ref(va_arg(ap, char*));
	}
	va_end(ap);
}

void prop_virtual_finish(uns start, ...)
{
	struct prop_virtual *v;
	va_list ap;
	va_start(ap, start);
	for(; (v = va_arg(ap, struct prop_virtual*)); va_arg(ap, char*))
		string_unref(v->p.key);
	va_end(ap);
}
