/*
 * VRR kernel - 2D search algorithms.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

//#define LOCAL_DEBUG
//#define SHOW_STATS
#include "kernel/internal.h"

/* dist_pass_dump */

void dist_pass_dump(struct dist_pass *pass)
{
	uns i, node_count, go_count, go_bbox_count/*, bezier_bbox_count, bezier_count*/;
	struct dist_pass_item *item;

	ASSERT(pass);

	node_count = go_count = go_bbox_count = /*bezier_bbox_count = bezier_count = */0;
	for (i = 0; i < pass->count; i++) {
		item = &pass->items[pass->heap[i]];
		switch (item->type) {
			case DIST_PASS_NODE:
				node_count++;
				break;

			case DIST_PASS_GO:
				go_count++;
				break;

			case DIST_PASS_GO_BBOX:
				go_bbox_count++;
				break;

			/*case DIST_PASS_BEZIER_BBOX:
				bezier_bbox_count++;
				break;

			case DIST_PASS_BEZIER:
				bezier_count++;
				break;*/

			default:
				BUG();
		}
	}

	msg("Dumping dist_pass structure at %p:\n", pass);
	msg("  flags=%x size=%d used=%d count=%d", pass->flags, pass->size, pass->used, pass->count);
#ifdef DIST_PASS_STAT
	msg(" tlo=%p returned=%d", pass->stat_tlo, pass->stat_returned);
#endif
	msg("\n");
	if (pass->count) {
		msg("  items count in heap:");
		if (node_count)
			msg(" node=%d", node_count);
		if (go_count)
			msg(" go=%d", go_count);
		if (go_bbox_count)
			msg(" go_bbox=%d", go_bbox_count);
		/*if (bezier_bbox_count)
			msg(" bezier_bbox=%d", bezier_bbox_count);
		if (bezier_count)
			msg(" bezier=%d", bezier_count);*/
		msg("\n");
	}
#ifdef DIST_PASS_STAT
	uns stat_total_dist = pass->stat_box_dist + pass->stat_point_dist + pass->stat_segment_dist + pass->stat_bezier_dist;
	if (stat_total_dist) {
		msg("  distance call count:");
		if (pass->stat_box_dist)
			msg(" box=%d", pass->stat_box_dist);
		if (pass->stat_point_dist)
			msg(" point=%d", pass->stat_point_dist);
		if (pass->stat_segment_dist)
			msg(" segment=%d", pass->stat_segment_dist);
		if (pass->stat_bezier_dist)
			msg(" bezier=%d", pass->stat_bezier_dist);
		msg("\n");
	}
#endif
}


/* dist_pass_assert */

void dist_pass_assert(struct dist_pass *pass)
{
	char *buf;
	uns i, k;
	int j;
	struct dist_pass_item *item;

	ASSERT(pass);
	ASSERT(pass->type == DIST_PASS_CENTER_PASS || pass->type == DIST_PASS_DIRECTION_PASS);
	ASSERT(!(pass->flags & ~DIST_PASS_FLAGS));
	if (pass->flags & DIST_PASS_MAX_DIST)
		ASSERT(isfinite(pass->max_dist));
	if (pass->flags & DIST_PASS_BOX)
		GEOM_ASSERT(rectangle, &pass->box);
	ASSERT(pass->size);
	ASSERT(pass->used <= pass->size);
	ASSERT(pass->count <= pass->used);
	ASSERT(pass->items);
	ASSERT(pass->heap);
	buf = xmalloc_zero(pass->used);
	j = pass->free;
	for (i = 0; i < pass->used - pass->count; i++) {
		ASSERT(j >= 0 && (uns)j < pass->used);
		ASSERT(!buf[j]);
		buf[j] = 1;
		j = pass->items[j].data.free.next;
	}
	ASSERT(pass->free == -1);
	for (i = 0; i < pass->count; i++) {
		k = pass->heap[i];
		ASSERT(k < pass->used);
		ASSERT(!buf[k]);
		buf[k] = 1;
		item = &pass->items[k];
		ASSERT(item->index == i);
		GEOM_ASSERT(r, item->dist);
		if (i)
			ASSERT(item->dist >= pass->items[pass->heap[(i + 1) / 2 - 1]].dist);
		switch (item->type) {
			case DIST_PASS_NODE:
				ASSERT(item->data.node.node);
				break;

			case DIST_PASS_GO:
			case DIST_PASS_GO_BBOX:
				ASSERT(item->data.go.go);
				break;

			/*case DIST_PASS_BEZIER_BBOX:
			case DIST_PASS_BEZIER:
				ASSERT(item->data.bezier.bezier);
				break;*/

			default:
				BUG();
		}
	}
	xfree(buf);
}


/* dist_pass_alloc_item
 * - allocate new item
 * - parameters:
 *     pass - pass pointer
 * - returns:
 *     index of newly allocated item */

static inline uns dist_pass_alloc_item(struct dist_pass *pass)
{
	uns result;

	ASSERT(pass);

	if (pass->free >= 0) {
		result = pass->free;
		pass->free = pass->items[result].data.free.next;
		return result;
	}

	if (pass->used == pass->size) {
		DBG("reallocating\n");
		pass->size *= 2;
		pass->items = (struct dist_pass_item *)xrealloc(pass->items, pass->size * sizeof(struct dist_pass_item));
		pass->heap = (uns *)xrealloc(pass->heap, pass->size * sizeof(uns));
	}

	return pass->used++;
}


/* dist_pass_free_item
 * - free item previously allocated by dist_pass_alloc_item
 * - parameters:
 *     pass - pass pointer
 *     index - index of item to free */

static inline void dist_pass_free_item(struct dist_pass *pass, uns index)
{
	ASSERT(pass);
	ASSERT(index < pass->used);

	pass->items[index].data.free.next = pass->free;
	pass->free = index;
}


/* dist_pass_insert
 * - insert item to heap
 * - parameters:
 *     pass - pass pointer
 *     index - index of item to insert (with valid distance) */

static void dist_pass_insert(struct dist_pass *pass, uns index)
{
	uns i, j;
	real dist;

	ASSERT(pass);
	ASSERT(index < pass->used);
	GEOM_ASSERT(r, pass->items[index].dist);

	i = pass->count++;
	dist = pass->items[index].dist;
	while (i) {
		j = (i + 1) / 2 - 1;
		if (pass->items[pass->heap[j]].dist <= dist)
			break;
		pass->items[pass->heap[i] = pass->heap[j]].index = i;
		i = j;
	}
	pass->items[pass->heap[i] = index].index = i;
}


/* dist_pass_delete
 * - remove item from heap
 * - parameters:
 *     pass - pass pointer
 *     index - index of item previeusly inserted to heap */

static void dist_pass_delete(struct dist_pass *pass, uns index)
{
	uns i, j, l;
	real dist;

	ASSERT(pass);
	ASSERT(index < pass->used);

	l = pass->heap[--pass->count];
	dist = pass->items[l].dist;
	i = pass->items[index].index;
	while (1) {
		j = i * 2 + 1;
		if (j >= pass->count)
			break;
		if (j + 1 < pass->count && pass->items[pass->heap[j]].dist > pass->items[pass->heap[j + 1]].dist)
			j++;
		if (dist <= pass->items[pass->heap[j]].dist)
			break;
		pass->items[pass->heap[i] = pass->heap[j]].index = i;
		i = j;
	}
	pass->items[pass->heap[i] = l].index = i;
}


/* dist_pass_update_min
 * - update min item position in heap after distance change
 * - parameters:
 *     pass - pass pointer */

static inline void dist_pass_update_min(struct dist_pass *pass)
{
	uns index, i, j;
	double dist;

	ASSERT(pass);
	ASSERT(pass->count);

	index = pass->heap[0];
	dist = pass->items[index].dist;
	i = 0;
	while (1) {
		j = i * 2 + 1;
		if (j >= pass->count)
			break;
		if (j + 1 < pass->count && pass->items[pass->heap[j]].dist > pass->items[pass->heap[j + 1]].dist)
			j++;
		if (dist <= pass->items[pass->heap[j]].dist)
			break;
		pass->items[pass->heap[i] = pass->heap[j]].index = i;
		i = j;
	}
	pass->items[pass->heap[i] = index].index = i;
}


/* dist_pass_insert_temp */

static inline void dist_pass_insert_temp(struct dist_pass *pass)
{
	dist_pass_insert(pass, pass->temp_index);
	pass->temp_index = dist_pass_alloc_item(pass);
	pass->temp = &pass->items[pass->temp_index];
}


/* center_pass_box_dist */

static int center_pass_box_dist(struct dist_pass *pass, struct dist_pass_item *item, const struct geom_rectangle *b)
{
	struct geom_point *c = &SKIP_BACK(struct center_pass, pass, pass)->center;

	if (c->x < b->ll.x) {
		item->point.x = b->ll.x;
		if (c->y < b->ll.y) {
			item->point.y = b->ll.y;
			item->dist = hypot(b->ll.x - c->x, b->ll.y - c->y);
		}
		else if (c->y > b->ur.y) {
			item->point.y = b->ur.y;
			item->dist = hypot(b->ll.x - c->x, c->y - b->ur.y);
		}
		else {
			item->point.y = c->y;
			item->dist = b->ll.x - c->x;
		}
	}
	else if (c->x > b->ur.x) {
		item->point.x = b->ur.x;
		if (c->y < b->ll.y) {
			item->point.y = b->ll.y;
			item->dist = hypot(c->x - b->ur.x, b->ll.y - c->y);
		}
		else if (c->y > b->ur.y) {
			item->point.y = b->ur.y;
			item->dist = hypot(c->x - b->ur.x, c->y - b->ur.y);
		}
		else {
			item->point.y = c->y;
			item->dist = c->x - b->ur.x;
		}
	}
	else {
		item->point.x = c->x;
		if (c->y < b->ll.y) {
			item->point.y = b->ll.y;
			item->dist = b->ll.y - c->y;
		}
		else if (c->y > b->ur.y) {
			item->point.y = b->ur.y;
			item->dist = c->y - b->ur.y;
		}
		else {
			item->point.y = c->y;
			item->dist = 0;
			return 1;
		}
	}
	return GEOM_RTEST(item->dist);
}


/* center_pass_point_dist */

static int center_pass_point_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_point *point)
{
	struct geom_point *c = &SKIP_BACK(struct center_pass, pass, pass)->center;

	item->point = *point;
	item->dist = hypot(point->x - c->x, point->y - c->y);
	return GEOM_RTEST(item->dist);
}


/* center_pass_curve_dist */

static int center_pass_curve_dist(struct dist_pass *pass, struct dist_pass_item *item, void *ptr)
{
	struct geom_point *c = &SKIP_BACK(struct center_pass, pass, pass)->center;
	/*struct geom_point n;
	struct geom_nearest nearest;*/

	// FIXME
	int result = GEOM_ETEST(geom_item_nearest_to_point(ptr, c, GEOM_NEAREST_ALL, &item->data.go.geom));
	item->dist = item->data.go.geom.dist;
	return result;

	/*item->nearest.time = nearest.time;
	n = nearest.point;*/
	/* TODO: jaka parametrizace? */
	/*if (!isfinite(item->time = geom_path_arc_time(path, item->time)))
		return 0;*/
	/*item->nearest.point = n;
	item->nearest.dist = hypot(c->x - n.x, c->y - n.y);
	return isfinite(item->nearest.dist) && isfinite(item->nearest.time);*/
}


/* center_pass_segment_dist */

/*static int center_pass_segment_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_point *a, struct geom_point *b)
{
	struct geom_point *c = &SKIP_BACK(struct center_pass, pass, pass)->center;
	double t, abx, aby, acx, acy;

	abx = b->x - a->x;
	aby = b->y - a->y;
	acx = c->x - a->x;
	acy = c->y - a->y;
	t = (acx * abx + acy * aby) / (SQR(abx) + SQR(aby));
	if (!isfinite(t) || t < 0) {
		item->nearest.point = *a;
		item->nearest.time = 0;
		item->nearest.dist = hypot(acx, acy);
	}
	else if (t > 1) {
		item->nearest.point = *b;
		item->nearest.time = 1;
		item->nearest.dist = hypot(c->x - b->x, c->y - b->y);
	}
	else {
		item->nearest.point.x = a->x + t * abx;
		item->nearest.point.y = a->y + t * aby;
		item->nearest.time = t;
		item->nearest.dist = hypot(c->x - pass->temp->nearest.point.x, c->y - pass->temp->nearest.point.y);
	}
	return isfinite(item->nearest.dist);
}*/


/* center_pass_bezier_dist */

/*static int center_pass_bezier_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_bezier *bezier)
{
	struct geom_point *c = &SKIP_BACK(struct center_pass, pass, pass)->center;
	struct geom_point n;
	struct geom_nearest nearest;
	real t;

	if (geom_bezier_nearest_to_point(bezier, c, GEOM_NEAREST_POINT | GEOM_NEAREST_TIME, &nearest) < 0)
		return 0;
	t = nearest.time;
	n = nearest.point;
	if (!isfinite(item->nearest.time = geom_bezier_time_to_atime(bezier, item->nearest.time)))
		return 0;
	item->nearest.point = n;
	item->nearest.dist = hypot(c->x - n.x, c->y - n.y);
	return isfinite(item->nearest.dist) && isfinite(item->nearest.time);
}*/


/* center_pass_path_dist */

/*static int center_pass_path_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_path *path)
{
	struct geom_point *c = &SKIP_BACK(struct center_pass, pass, pass)->center;
	struct geom_point n;
	struct geom_nearest nearest;

	if (geom_path_nearest_to_point(path, c, GEOM_NEAREST_TIME | GEOM_NEAREST_POINT, &nearest) < 0)
		return 0;
	item->nearest.time = nearest.time;
	n = nearest.point;
	// TODO: jaka parametrizace?
	//if (!isfinite(item->time = geom_path_arc_time(path, item->time)))
	//	return 0;
	item->nearest.point = n;
	item->nearest.dist = hypot(c->x - n.x, c->y - n.y);
	return isfinite(item->nearest.dist) && isfinite(item->nearest.time);
}*/


/* direction_pass_box_dist */

/*static int direction_pass_box_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_rectangle *box)
{
	struct geom_vector *v = &SKIP_BACK(struct direction_pass, pass, pass)->vector;

	if (v->dx > 0) {
		item->nearest.point.x = box->ll.x;
		if (v->dy > 0) {
			item->nearest.point.y = box->ll.y;
			item->nearest.dist = v->dx * box->ll.x + v->dy * box->ll.y;
		}
		else {
			item->nearest.point.y = box->ur.y;
			item->nearest.dist = v->dx * box->ll.x + v->dy * box->ur.y;
		}
	}
	else {
		item->nearest.point.x = box->ur.x;
		if (v->dy > 0) {
			item->nearest.point.y = box->ll.y;
			item->nearest.dist = v->dx * box->ur.x + v->dy * box->ll.y;
		}
		else {
			item->nearest.point.y = box->ur.y;
			item->nearest.dist = v->dx * box->ur.x + v->dy * box->ur.y;
		}
	}
	return isfinite(item->nearest.dist);
}*/


/* direction_pass_point_dist */

/*static inline int direction_pass_point_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_point *point)
{
	struct geom_vector *v = &SKIP_BACK(struct direction_pass, pass, pass)->vector;

	item->nearest.point = *point;
	item->nearest.dist = v->dx * point->x + v->dy * point->y;
	return isfinite(item->nearest.dist);
}*/


/* direction_pass_segment_dist */

/*static int direction_pass_segment_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_point *a, struct geom_point *b)
{
	struct geom_vector *v = &SKIP_BACK(struct direction_pass, pass, pass)->vector;
	double da, db;

	da = v->dx * a->x + v->dy * a->y;
	if (!isfinite(da))
		return 0;
	db = v->dx * b->x + v->dy * b->y;
	if (!isfinite(db))
		return 0;
	if (da <= db) {
		item->nearest.point = *a;
		item->nearest.time = 0;
		item->nearest.dist = da;
	}
	else {
		item->nearest.point = *b;
		item->nearest.time = 1;
		item->nearest.dist = db;
	}
	return 1;
}*/


/* direction_pass_bezier_dist */

/*static int direction_pass_bezier_dist(struct dist_pass *pass UNUSED, struct dist_pass_item *item UNUSED, struct geom_bezier *bezier UNUSED)
{
	BUG();
	// TODO
}*/


/* direction_pass_path_dist */

/*static int direction_pass_path_dist(struct dist_pass *pass UNUSED, struct dist_pass_item *item UNUSED, struct geom_path *path UNUSED)
{
	BUG();
	// TODO
}*/


/* dist_pass_box_dist */

static inline int dist_pass_box_dist(struct dist_pass *pass, struct dist_pass_item *item, const struct geom_rectangle *box)
{
	static int (*tab[DIST_PASS_TYPE_COUNT])(struct dist_pass *, struct dist_pass_item *, const struct geom_rectangle *) =
		{[DIST_PASS_CENTER_PASS] = &center_pass_box_dist};
		 //[DIST_PASS_DIRECTION_PASS] = &direction_pass_box_dist};
#ifdef DIST_PASS_STAT
	pass->stat_box_dist++;
#endif
	return (tab[pass->type])(pass, item, box) && (!(pass->flags & DIST_PASS_MAX_DIST) || item->dist <= pass->max_dist);
}


/* dist_pass_point_dist */

static inline int dist_pass_point_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_point *point)
{
	static int (*tab[DIST_PASS_TYPE_COUNT])(struct dist_pass *, struct dist_pass_item *, struct geom_point *) =
		{[DIST_PASS_CENTER_PASS] = &center_pass_point_dist};
		 //[DIST_PASS_DIRECTION_PASS] = &direction_pass_point_dist};
#ifdef DIST_PASS_STAT
	pass->stat_point_dist++;
#endif
	return (tab[pass->type])(pass, item, point) && (!(pass->flags & DIST_PASS_MAX_DIST) || item->dist <= pass->max_dist);
}


/* dist_pass_curve_dist */

static inline int dist_pass_curve_dist(struct dist_pass *pass, struct dist_pass_item *item, void *ptr)
{
	static int (*tab[DIST_PASS_TYPE_COUNT])(struct dist_pass *, struct dist_pass_item *, void *) =
		{[DIST_PASS_CENTER_PASS] = &center_pass_curve_dist};
		 //[DIST_PASS_DIRECTION_PASS] = &direction_pass_point_dist};
/*#ifdef DIST_PASS_STAT
	pass->stat_point_dist++;
#endif*/
	return (tab[pass->type])(pass, item, ptr) && (!(pass->flags & DIST_PASS_MAX_DIST) || item->dist <= pass->max_dist);
}


/* dist_pass_segment_dist */

/*static inline int dist_pass_segment_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_point *a, struct geom_point *b)
{
	static int (*tab[DIST_PASS_TYPE_COUNT])(struct dist_pass *, struct dist_pass_item *, struct geom_point *, struct geom_point *) =
		{[DIST_PASS_CENTER_PASS] = &center_pass_segment_dist};
		 //[DIST_PASS_DIRECTION_PASS] = &direction_pass_segment_dist};
#ifdef DIST_PASS_STAT
	pass->stat_segment_dist++;
#endif
	return (tab[pass->type])(pass, item, a, b) && (!(pass->flags & DIST_PASS_MAX_DIST) || item->nearest.dist <= pass->max_dist);
}*/


/* dist_pass_bezier_dist */

/*static inline int dist_pass_bezier_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_bezier *bezier)
{
	static int (*tab[DIST_PASS_TYPE_COUNT])(struct dist_pass *, struct dist_pass_item *, struct geom_bezier *) =
		{[DIST_PASS_CENTER_PASS] = &center_pass_bezier_dist};
		 //[DIST_PASS_DIRECTION_PASS] = &direction_pass_bezier_dist};
#ifdef DIST_PASS_STAT
	pass->stat_bezier_dist++;
#endif
	return (tab[pass->type])(pass, item, bezier) && (!(pass->flags & DIST_PASS_MAX_DIST) || item->nearest.dist <= pass->max_dist);
}*/


/* dist_pass_path_dist */

/*static inline int dist_pass_path_dist(struct dist_pass *pass, struct dist_pass_item *item, struct geom_path *path)
{
	static int (*tab[DIST_PASS_TYPE_COUNT])(struct dist_pass *, struct dist_pass_item *, struct geom_path *) =
		{[DIST_PASS_CENTER_PASS] = &center_pass_path_dist};
		 //[DIST_PASS_DIRECTION_PASS] = &direction_pass_path_dist};
#ifdef DIST_PASS_STAT
	pass->stat_bezier_dist += path->count;
#endif
	return (tab[pass->type])(pass, item, path) && (!(pass->flags & DIST_PASS_MAX_DIST) || item->nearest.dist <= pass->max_dist);
}*/


/* dist_pass_insert_go
 * - insert go to pass heap
 * - parameters:
 *     pass - pass pointer
 *     go - go pointer */

static void dist_pass_insert_go(struct dist_pass *pass, struct go *go)
{
	struct hanger *hanger;
	struct anchor *anchor;
	int i;
	int count;
	struct go *go2;
	struct geom_intersection *intersection;
	struct geom_point gp;
	struct obj_tlo *tlo = go_get_tlo(go);
	const struct geom_rectangle *bbox = go_get_rtree_bbox(go);

	if(!bbox)
		return;

	if (pass->flags & DIST_PASS_HANGERS)
		WALK_GO_HANGER_BEGIN(go, hanger)
			if (hanger_is_position(hanger) && dist_pass_point_dist(pass, pass->temp, hanger_get_position_ptr(hanger, &gp))) {
				pass->temp->type = DIST_PASS_HANGER;
				pass->temp->data.hanger.hanger = hanger;
				dist_pass_insert_temp(pass);
			}
		WALK_GO_HANGER_END

	if (pass->flags & DIST_PASS_ANCHORS)
		WALK_GO_ANCHOR(go, anchor, i) {
			if (hanger_is_position(anchor->h) && dist_pass_point_dist(pass, pass->temp, hanger_get_position_ptr(anchor->h, &gp))) {
				pass->temp->type = DIST_PASS_ANCHOR;
				pass->temp->data.anchor.anchor = anchor;
				dist_pass_insert_temp(pass);
			}
		}

	if (pass->flags & DIST_PASS_INTERSECTIONS)
	{
		GEOM_RTREE_RECT_INTERSECT_QUERY_BEGIN(walk1, &tlo->rtree, bbox, leaf);
		go2 = SKIP_BACK(struct rtree_bbox, rtree, leaf)->go;

		WALK_GO_RECURSIVE_WITHOUT_ROOT(&tlo->tlg->group.go, go2) // FIXME: zrychlit
			if (go < go2) {
				count = go_intersections(go, go2, GEOM_INTERSECTION_ALL, &pass->intersections);
				for (i = 0; i < count; i++) {
					intersection = (struct geom_intersection *) garr_get(&pass->intersections, i);
					if (dist_pass_point_dist(pass, pass->temp, &intersection->point)) {
						DBG("inserting intersection at (%f,%f), dist=%f\n", intersection->point.x, intersection->point.y, pass->temp->dist);
						pass->temp->type = DIST_PASS_INTERSECTION;
						pass->temp->data.intersection.go1 = go;
						pass->temp->data.intersection.go2 = go2;
						pass->temp->data.intersection.geom = *intersection;
						dist_pass_insert_temp(pass);
					}
				}
			}

		GEOM_RTREE_RECT_INTERSECT_QUERY_END(walk1);
	}


	if (pass->flags & DIST_PASS_GOS) {
		switch (go->type) {
			case GOT_POINT:
				if (!dist_pass_point_dist(pass, pass->temp, &((struct go_point *)go)->geom.point))
					return;
			INSERT_GO_POINT:
				pass->temp->data.go.geom.point = pass->temp->point;
				pass->temp->data.go.geom.time = 0;
			INSERT_GO:
				pass->temp->type = DIST_PASS_GO;
				pass->temp->data.go.go = go;
				DBG("inserting GO, index=%d, dist=%f\n", pass->temp_index, pass->temp->dist);
				dist_pass_insert_temp(pass);
				return;

			case GOT_SEGMENT:
				if (!dist_pass_curve_dist(pass, pass->temp, &((struct go_segment*)go)->geom))
					return;
				goto INSERT_GO;

			case GOT_TEXT:
			case GOT_TEX_TEXT:
				// FIXME: nepocitat s RTree, ale ztransformovanym presnym bboxem kolem textu
				if (!dist_pass_box_dist(pass, pass->temp, bbox))
					return;
				goto INSERT_GO_POINT;

			case GOT_BEZIER:
			case GOT_DECORATOR_POINT:
			case GOT_ARROW:
			case GOT_ELARC:
				if (!dist_pass_box_dist(pass, pass->temp, bbox))
					return;
				pass->temp->type = DIST_PASS_GO_BBOX;
				pass->temp->data.go.go = go;
				DBG("inserting GO_BBOX, index=%d, dist=%f\n", pass->temp_index, pass->temp->dist);
				dist_pass_insert_temp(pass);
				return;

			default:
				return;
		}
	}
}


/* dist_pass_insert_tlo
 * - insert tlo to pass heap
 * - parameters:
 *     pass - pass pointer
 *     tlo - tlo pointer */

static void dist_pass_insert_tlo(struct dist_pass *pass, struct obj_tlo *tlo)
{
	struct geom_rtree_node *node;
	struct go *go;

	if (pass->flags & DIST_PASS_RTREE) {
		if (!tlo->rtree.root)
			return;
		node = tlo->rtree.root;
		if (!dist_pass_box_dist(pass, pass->temp, &node->bbox))
			return;
		pass->temp->type = DIST_PASS_NODE;
		pass->temp->data.node.node = node;
		DBG("inserting NODE, index=%d, dist=%.2f\n", pass->temp_index, pass->temp->dist);
		dist_pass_insert_temp(pass);
	}

	else
		WALK_GO_RECURSIVE_WITHOUT_ROOT(&tlo->tlg->group.go, go)
			dist_pass_insert_go(pass, go);
}


/* dist_pass_return */

static void dist_pass_return(struct dist_pass *pass, uns index)
{
#ifdef DIST_PASS_STAT
	pass->stat_returned++;
#endif

	switch (pass->result.type = pass->items[index].type) {
	case DIST_PASS_GO:
		pass->result.data.go.go = pass->items[index].data.go.go;
		pass->result.data.go.geom = pass->items[index].data.go.geom;
		DBG("Found GO %p\n", pass->result.data.go.go);
		break;
	case DIST_PASS_HANGER:
		pass->result.data.hanger.dist = pass->items[index].dist;
		pass->result.data.hanger.hanger = pass->items[index].data.hanger.hanger;
		DBG("Found hanger %p\n", pass->result.data.hanger.hanger);
		break;
	case DIST_PASS_ANCHOR:
		pass->result.data.anchor.dist = pass->items[index].dist;
		pass->result.data.anchor.anchor = pass->items[index].data.anchor.anchor;
		DBG("Found anchor %p\n", pass->result.data.anchor.anchor);
		break;
	case DIST_PASS_INTERSECTION:
		pass->result.data.intersection.dist = pass->items[index].dist;
		pass->result.data.intersection.go1 = pass->items[index].data.intersection.go1;
		pass->result.data.intersection.go2 = pass->items[index].data.intersection.go2;
		pass->result.data.intersection.geom = pass->items[index].data.intersection.geom;
		DBG("Found intersection of %p and %p\n", pass->result.data.intersection.go1, pass->result.data.intersection.go2);
		break;
	default:
		BUG();
	}
}


/* dist_pass_expand_node */

static void dist_pass_expand_node(struct dist_pass *pass, uns index)
{
	uns i;
	struct geom_rtree_node *node, *child;

	DBG("expanding NODE, index=%d, dist=%.2f\n", index, pass->items[index].dist);
	node = pass->items[index].data.node.node;
	dist_pass_delete(pass, index);
	dist_pass_free_item(pass, index);

	if (node->height)
		for (i = 0; i < node->count; i++) {
			child = node->child[i];
			if (dist_pass_box_dist(pass, pass->temp, &child->bbox)) {
				pass->temp->type = DIST_PASS_NODE;
				pass->temp->data.node.node = child;
				DBG("inserting NODE, index=%d, dist=%.2f\n", pass->temp_index, pass->temp->dist);
				dist_pass_insert_temp(pass);
			}
		}

	else
		for (i = 0; i < node->count; i++)
			dist_pass_insert_go(pass, SKIP_BACK(struct rtree_bbox, rtree, node->child[i])->go);
}


/* dist_pass_expand_go */

/*static void dist_pass_expand_go(struct dist_pass *pass, uns index)
{
	DBG("expanding GO, index=%d, dist=%.2f\n", index, pass->items[index].dist);
	dist_pass_delete(pass, index);
	dist_pass_return(pass, index);
}*/


/* dist_pass_expand_go_bbox */

static void dist_pass_expand_go_bbox(struct dist_pass *pass, uns index)
{
	struct dist_pass_item *item;
	struct go *go;
	void *geom;
	//struct expansion *expansion;

	DBG("expanding GO_BBOX, index=%d, dist=%0.2f\n", index, pass->items[index].dist);

	item = &pass->items[index];
	go = item->data.go.go;

	// SELECTION HACK

	geom = go2geom(go);

	if (dist_pass_curve_dist(pass, item, geom)) {
		item->type = DIST_PASS_GO;
		dist_pass_update_min(pass);
		return;
	}
	dist_pass_delete(pass, index);
	dist_pass_free_item(pass, index);

	// END OF SELECTION HACK
	
	
	/*if (pass->flags & DIST_PASS_BBOX_DIST) {
		dist_pass_delete(pass, index);
		dist_pass_return(pass, index);
		return;
	}*/

	//expansion = go_expand(go);

	/*if (!(pass->flags & DIST_PASS_EXPAND_PATH))*/ {
		/*if (dist_pass_path_dist(pass, item, &expansion->path)) {
			item->type = DIST_PASS_GO;
			DBG("inserting GO, index=%d, dist=%.2f\n", index, item->dist);
			dist_pass_update_min(pass);
		}
		else {
			dist_pass_delete(pass, index);
			dist_pass_free_item(pass, index);
		}
		expansion_unref(expansion);*/
/*		if (dist_pass_curve_dist(pass, item, go_to_curve_type(go), go_to_curve_ptr(go))) { TODO
			item->type = DIST_PASS_GO;
			DBG("inserting GO, index=%d, dist=%.2f\n", index, item->dist);
			dist_pass_update_min(pass);
		}
		else {
			dist_pass_delete(pass, index);
			dist_pass_free_item(pass, index);
		} */
	}

	/*else {
		BUG();
		TODO
		dist_pass_delete(pass, index);
		pindex = index;
		for (i = 0; i < expansion->path.obj_count; i++) {
			bezier = &GEOM_PATH_NTH_BEZIER(&expansion->path, i);
			if (geom_bezier_bounding_box(bezier, &bbox) >= 0) {
				dist = dist_pass_box_dist(pass, &bbox);
				if (isfinite(dist) && !(pass->flags & DIST_PASS_MAX_DIST && dist > pass->max_dist)) {
					nindex = dist_pass_alloc_item(pass);
					nitem = &pass->items[nindex];
					nitem->type = DIST_PASS_BEZIER_BBOX;
					nitem->data.bezier.bezier = bezier;
					nitem->data.bezier.next = pindex;
					nitem->dist = dist;
					DBG("inserting BEZIER_BBOX, index=%d, dist=%0.2f\n", index, dist);
					dist_pass_insert(pass, nindex);
					pindex = index;
				}
			}
		}
		if (pindex != index) {
		}
		else {
			dist_pass_free_item(pass, index);
			expansion_unref(expansion);
		}
	}*/
}


/* dist_pass_expand_bezier_bbox */

/*static void dist_pass_expand_bezier_bbox(struct dist_pass *pass UNUSED, uns index UNUSED)
{
TODO	struct dist_pass_item *item;

	DBG("expanding BEZIER_BBOX, index=%d, dist=%.2f\n", index, pass->items[index].dist);

	item = &pass->items[index];
	item->dist = dist_pass_bezier_dist(pass, item->data.bezier.bezier);
	if (isfinite(item->dist) && !(pass->flags & DIST_PASS_MAX_DIST && item->dist > pass->max_dist)) {
		item->type = DIST_PASS_BEZIER;
		DBG("inserting BEZIER, index=%d, dist=%0.2f\n", index, item->dist);
		dist_pass_increase(pass, index);
	}
	else {
		dist_pass_delete(pass, index);
		dist_pass_free_item(pass, index);
	}

}*/


/* dist_pass_expand_bezier */

/*static void dist_pass_expand_bezier(struct dist_pass *pass UNUSED, uns index UNUSED)
{
	uns go_index, i, j;

	DBG("expanding BEZIER, index=%d, dist=%0.2f\n", index, pass->items[index].dist);

	go_index = pass->items[index].data.bezier.go;
	pass->items[go_index].dist = pass->items[index].dist;

	i = index;
	do {
		j = pass->items[i].data.bezier.next;
		dist_pass_delete(pass, i);
		dist_pass_free_item(pass, i);
		i = j;
	} while (i != index);

	expansion_unref(pass->items[go_index].data.go.expansion);
	dist_pass_return(pass, go_index);
}*/


/* dist_pass_next */

int dist_pass_next(struct dist_pass *pass, struct dist_pass_result *result)
{
	uns index;
	struct dist_pass_item *item;

	do {
		if (!pass->count) {
			DBG("no more pass items\n");
			return 0;
		}

		index = pass->heap[0];
		item = &pass->items[index];

		ASSERT (!(pass->flags & DIST_PASS_MAX_DIST && item->dist > pass->max_dist));

		switch (item->type) {
			case DIST_PASS_NODE:
				dist_pass_expand_node(pass, index);
				break;

			case DIST_PASS_GO_BBOX:
				dist_pass_expand_go_bbox(pass, index);
				break;

			case DIST_PASS_GO:
			case DIST_PASS_HANGER:
			case DIST_PASS_ANCHOR:
			case DIST_PASS_INTERSECTION:
				dist_pass_delete(pass, index);
				dist_pass_return(pass, index);
				goto RETURN;

			/*case DIST_PASS_BEZIER_BBOX:
				dist_pass_expand_bezier_bbox(pass, index);
				break;

			case DIST_PASS_BEZIER:
				dist_pass_expand_bezier(pass, index);
				break;*/

			default:
				BUG();
		}
	} while (1);

RETURN:
/*#ifdef DEBUG
	if (pass->result.go)
		DBG("pass returns go %p, dist=%.2f, point=(%.2f,%.2f), time=%.2f\n", pass->result.go, pass->result.dist, pass->result.point.x, pass->result.point.y, pass->result.time);
#endif*/
	if (result)
		*result = pass->result;
	return 1;
}


/* dist_pass_init
 * - initialize dist_pass structure
 * - parameters:
 *     pass - pass pointer
 *     type - pass type
 *     flags - pass options
 *     filter - go types filter
 *     max_dist - max distance
 *     tlo - tlo to pass */

static void dist_pass_init(struct dist_pass *pass, enum dist_pass_type type, uns flags, struct geom_rectangle *box, double max_dist, struct obj_tlo *tlo)
{
	ASSERT(pass);
	ASSERT(tlo);

	DBG("Initializing dist pass\n");

	pass->type = type;
	pass->flags = flags;
	if (box)
		pass->box = *box;
	pass->max_dist = max_dist;
	pass->size = DIST_PASS_INITIAL_SIZE;
	pass->count = 0;
	pass->used = 1;
	pass->free = -1;
	pass->items = xmalloc(pass->size * sizeof(struct dist_pass_item));
	pass->heap = xmalloc(pass->size * sizeof(uns));
	pass->temp_index = 0;
	pass->temp = pass->items;
	if (pass->flags & DIST_PASS_INTERSECTIONS)
		garr_init(&pass->intersections, 32, sizeof(struct geom_intersection));

#ifdef DIST_PASS_STAT
	pass->stat_tlo = tlo;
	pass->stat_box_dist = 0;
	pass->stat_point_dist = 0;
	pass->stat_segment_dist = 0;
	pass->stat_bezier_dist = 0;
	pass->stat_returned = 0;
#endif

	dist_pass_insert_tlo(pass, tlo);
}


/* dist_pass_destroy
 * - free pass internal structures */

static void dist_pass_destroy(struct dist_pass *pass)
{
	ASSERT(pass);

	DBG("Destroying dist pass.\n");

#ifdef SHOW_STATS
	dist_pass_dump(pass);
#endif

	if (pass->flags & DIST_PASS_INTERSECTIONS)
		garr_cleanup(&pass->intersections);
	xfree(pass->items);
	xfree(pass->heap);
}


/* center_pass_init */

void center_pass_init(struct center_pass *cp, uns flags, struct geom_rectangle *box, struct geom_point *center, real radius, struct obj_tlo *tlo)
{
	ASSERT(cp);
	GEOM_ASSERT(point, center);
	ASSERT(isfinite(radius));

	cp->center = *center;
	dist_pass_init(&cp->pass, DIST_PASS_CENTER_PASS, flags, box, radius, tlo);
}


/* center_pass_destroy */

void center_pass_destroy(struct center_pass *cp)
{
	ASSERT(cp);

	dist_pass_destroy(&cp->pass);
}


/* center_pass_nearest */

int center_pass_nearest(uns flags, struct geom_rectangle *box, struct geom_point *center, real radius, struct obj_tlo *tlo, struct dist_pass_result *result)
{
	struct center_pass cp;
	int found;

	ASSERT(result);
	center_pass_init(&cp, flags, box, center, radius, tlo);
	found = center_pass_next(&cp, result);
	center_pass_destroy(&cp);
	return found;
}


/* direction_pass_init */

/*void direction_pass_init(struct direction_pass *dp, uns flags, uns filter, struct geom_rectangle *box, struct geom_vector *vector, struct obj_tlo *tlo)
{
	DBG("Initializing direction-pass\n");

	ASSERT(dp);
	ASSERT(!(filter & ~DIRECTION_PASS_ALL));
	ASSERT(!(flags & DIST_PASS_MAX_DIST));
	GEOM_VECTOR_ASSERT(vector);

	if (!filter)
		filter = DIRECTION_PASS_ALL;
	dp->vector = *vector;
	dist_pass_init(&dp->pass, DIST_PASS_DIRECTION_PASS, flags, filter, box, 0, tlo);
}*/


/* direction_pass_destroy */

/*void direction_pass_destroy(struct direction_pass *dp)
{
	DBG("Destroying direction-pass\n");

	ASSERT(dp);
	dist_pass_destroy(&dp->pass);
}*/


