#ifndef _VRR_LIB_CACHE_H_
#define _VRR_LIB_CACHE_H_

/** @file
 *
 * VRRLIB - Cache.
 *
 * Simple limited cache implementation with LRU freeing mechanism.
 * Cache may be shared between multiple users.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/clists.h"
#include <limits.h>

/* Unlimited cache constants. */
#define CACHE_UNLIMITED_MEMORY_LIMIT INT_MAX
#define CACHE_UNLIMITED_COUNT_MAX (~(uns)0)

/* Default cache limits. */
#define CACHE_DEFAULT_MEMORY_LIMIT (5 * 1024 * 1024)
#define CACHE_DEFAULT_COUNT_MIN 2
#define CACHE_DEFAULT_COUNT_MAX 100000

/* Maximum number of cache users. */
#define CACHE_MAX_USERS 0x100

/**
 * Cache node structure.
 * Direct use of CACHE_NODE macro should be only in structure header because of aligning.
 */
#define CACHE_NODE cnode cache_lru_node; u8 cache_user_id
struct cache_node {
	CACHE_NODE;
};

/**
 * Cache user strucure.
 * Each user represents callback to force remove its cached items).
 * Callback may free the node and call cache_used_sub(cache, size).
 * No cache_remove(cache, node) is needed.
 * Other calls to cache inside callback are supported but there is no cycle detection
 * while allocating and inserting nodes.
 */
struct cache_user {
	uns user_id;
	void (*callback)(struct cache_node *node);
};

/**
 * Cache main strurure - limits and internal varriables.
 */
struct cache {
	uns memory_limit;
	uns count_min;
	uns count_max;
	int memory_free;
	uns count;
	clist lru_list;
	struct cache_user *users[CACHE_MAX_USERS];
};

/**
 * Global cache structure.
 * Shortcuts are defined in gcache.h.
 */
extern struct cache global_cache;


/** Removes node from LRU list. */
static inline void cache_remove(struct cache *cache, struct cache_node *node)
{
	ASSERT(cache->users[node->cache_user_id] != NULL);
	clist_remove(&node->cache_lru_node);
	cache->count--;
#ifdef DEBUG
	node->cache_user_id = 0;
#endif
}

/** Moves node to head of LRU list. */
static inline void cache_touch(struct cache *cache, struct cache_node *node)
{
	ASSERT(cache->users[node->cache_user_id] != NULL);
	clist_remove(&node->cache_lru_node);
	clist_add_head(&cache->lru_list, &node->cache_lru_node);
}

/** Removes least recently used node and calls users callback to free it. Cache must not me empty. */
static inline void cache_remove_lru(struct cache *cache)
{
	struct cache_node *node = SKIP_BACK(struct cache_node, cache_lru_node, cache->lru_list.head.prev);
	struct cache_user *user = cache->users[node->cache_user_id];
	cache_remove(cache, node);
	user->callback(node);
}

/** Removes all nodes from cache and calls users callbacks to free them. Cache may be empty. */
static inline void cache_remove_all(struct cache *cache)
{
	while (!clist_empty(&cache->lru_list))
		cache_remove_lru(cache);
}

/** Internal function used when allocating on cache or decreasing cache limits. */
static inline void cache_used_update(struct cache *cache)
{
	while (cache->count > cache->count_min && cache->memory_free < 0)
		cache_remove_lru(cache);
}

/**
 * Adds a given users node to head of LRU list.
 * Node must be fully initialized to allow callback.
 * Remove callback may happen anytime between this function (possibly inside) and cache_remove.
 */
static inline void cache_add(struct cache *cache, struct cache_user *user, struct cache_node *node)
{
	ASSERT(cache->users[user->user_id] != NULL);
	clist_add_head(&cache->lru_list, &node->cache_lru_node);
	node->cache_user_id = user->user_id;
	cache->count++;
	if (cache->count == cache->count_max)
		cache_remove_lru(cache);
	cache_used_update(cache);
}

/** Increase allocated memory info hold by the cache. */
static inline void cache_used_inc(struct cache *cache, uns size)
{
	cache->memory_free -= size;
	cache_used_update(cache);
}

/** Decrease allocated memory info hold by the cache. */
static inline void cache_used_dec(struct cache *cache, uns size)
{
	cache->memory_free += size;
	ASSERT(cache->memory_free <= (int)cache->memory_limit);
}

/** Return used memory size. */
static inline uns cache_used_info(struct cache *cache)
{
	return cache->memory_limit - cache->memory_free;
}

/** Increase/decrease allocated memory info hold by the cache. */
static inline void cache_used_mod(struct cache *cache, int size)
{
	cache->memory_free -= size;
	ASSERT(cache->memory_free <= (int)cache->memory_limit);
	cache_used_update(cache);
}

/**
 * Initialize cache.
 * Set limits to cache defaults.
 */
void cache_init(struct cache *cache);

/**
 * Destroys cache.
 * All nodes are removed and users are destroyed automatically.
 */
void cache_destroy(struct cache *cache);

/**
 * Cache cache limits.
 */

void cache_set_limits(struct cache *cache, uns memory_limit, uns count_min, uns count_max);
#define CACHE_LEAVE_MEMORY_LIMIT 0
#define CACHE_LEAVE_COUNT_MIN (~(uns)0)
#define CACHE_LEAVE_COUNT_MAX 0

/**
 * Initialize cache user.
 */
void cache_user_init(struct cache_user *user, struct cache *cache, void (*callback)(struct cache_node *));

/**
 * Destroys cache user.
 * All user nodes must be removed from its cache before.
 */
void cache_user_destroy(struct cache_user *user, struct cache *cache);


#endif
