/*
 * VRRLIB - Small arrays
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include <stdio.h>
#include "lib/lib.h"
#include "lib/sarray.h"
#include "string.h"

#define MIN_NODE_SIZE 4
#define MIN_COUNT_NODE 4

// Invariant: s^3 * n = input (approximate)
static inline uns sarray_sqrt(uns n) { uns s = 4; while(n > 20) { n >>= 3; s <<= 1; } return s; }

static void sarray_copy(struct sarray_node* tn, int ti, struct sarray_node* sn, int si, int c)
{
	memcpy(&tn->array[ti], &sn->array[si], c * sizeof(element));
}

struct sarray* sarray_init(struct sarray* s)
{
	s->count = s->count_element = 0;
	s->size = MIN_COUNT_NODE;
	s->array = xmalloc(sizeof(void*) * MIN_COUNT_NODE);
	return s;
}

void sarray_finish(struct sarray* s)
{
	for(int i = 0; i < s->count; i++)
		xfree(s->array[i]);
	xfree(s->array);
}

int sarray_find_node(struct sarray* s, element e)
{
	int start = 0, end = s->count - 1, middle;
	do
	{
		middle = (start + end) / 2;
		if(e > s->array[middle]->array[0])
			start = middle + 1;
		else if(e < s->array[middle]->array[0])
			end = middle - 1;
		else
			return middle;
	} while(start <= end);
	return start > 0 ? start-1 : 0;
}

int sarray_find_element(struct sarray_node* n, element e)
{
	int start = 0, end = n->count - 1, middle;
	do		// Invariant e < array[end+1] and e > array[start-1] (suppose: array[-1] = -infinity and array[count] = infinity).
	{
		middle = (start + end) / 2;
		if(e > n->array[middle])
			start = middle + 1;
		else if(e < n->array[middle])
			end = middle - 1;
		else return middle;
	} while(start <= end);		// start = end + 1 on finish so array[start-1] < e < array[start]. It's what I want.
	return start;
}

int sarray_is_member(struct sarray* s, element e)
{
	if(s->count == 0)
		return 0;
	struct sarray_node *n = s->array[sarray_find_node(s, e)];
	int i = sarray_find_element(n, e);
	if(i < n->count && n->array[i] == e)
		return 1;
	return 0;
}

static void sarray_insert_node(struct sarray *s, int index, struct sarray_node* n)
{
	if(s->count == s->size)
	{
		struct sarray_node** sa2 = xmalloc(sizeof(void*) * 2 * s->size), **sa1 = s->array;
		int i;
		for(i = 0; i < index; i++)
			sa2[i] = sa1[i];
		sa2[index] = n;
		for(; i < s->size; i++)
			sa2[i+1] = sa1[i];
		s->array = sa2;
		s->size *= 2;
	}
	else
	{
		for(int i = s->count; i > index; i--)
			s->array[i] = s->array[i-1];
		s->array[index] = n;
	}
	s->count++;
}

element sarray_insert(struct sarray* s, element e)
{
	if(s->count == 0)
	{
		struct sarray_node* n = xmalloc(sizeof(struct sarray_node) + sizeof(element) * MIN_NODE_SIZE);
		n->size = MIN_NODE_SIZE;
		n->array[0] = e;
		s->array[0] = n;
		s->count = n->count = s->count_element = 1;
		return e;
	}
	int nindex = sarray_find_node(s, e);
	struct sarray_node* n = s->array[nindex];
	int index = sarray_find_element(n, e);
	if(n->count == n->size)
	{
		uns size = sarray_sqrt(s->count_element);
		if(size >= 2 * n->count)
		{
			struct sarray_node* n2 = xmalloc(sizeof(struct sarray_node) + sizeof(element) * size);
			s->array[nindex] = n2;
			n2->size = size;
			n2->count = n->count + 1;
			sarray_copy(n2, 0, n, 0, index);
			n2->array[index] = e;
			sarray_copy(n2, index + 1, n, index, n->count - index);
			xfree(n);
		}
		else
		{
			if(size < n->count)
				size = n->count;
			struct sarray_node* n2 = xmalloc(sizeof(struct sarray_node) + sizeof(element) * size);
			sarray_insert_node(s, nindex+1, n2);
			n2->size = size;
			n2->count = (n->count + 1) / 2;
			n->count = n->count - n2->count + 1;
			if(index < n->count)
			{
				sarray_copy(n2, 0, n, n->count - 1, n2->count);
				for(int i = n->count; i > index; i--)
					n->array[i] = n->array[i-1];
				n->array[index] = e;
			}
			else
			{
				sarray_copy(n2, 0, n, n->count, index - n->count);
				n2->array[index - n->count] = e;
				sarray_copy(n2, index - n->count + 1, n, index, n->size - index);
			}
		}
	}
	else
	{
		ASSERT(index == n->count || e != n->array[index]);
		for(int j = n->count; j > index; j--)
			n->array[j] = n->array[j-1];
		n->array[index] = e;
		n->count++;
	}
	s->count_element++;
	return e;
}

static void sarray_remove_node(struct sarray* s, int index)
{
	xfree(s->array[index]);
	for(int i = index + 1; i < s->count; i++)
		s->array[i-1] = s->array[i];
	s->count--;
}

element sarray_remove(struct sarray* s, element e)
{
	ASSERT(s->count);
	int nindex = sarray_find_node(s, e);
	struct sarray_node* n = s->array[nindex];
	int index = sarray_find_element(n, e);
	ASSERT(index < n->count && n->array[index] == e);
	if(n->count == 1)
		sarray_remove_node(s, nindex);
	else
	{
		struct sarray_node* p = s->array[nindex-1];
		if(nindex && (p->count + n->count) < p->size)
		{
			sarray_copy(p, p->count, n, 0, index);
			sarray_copy(p, p->count + index, n, index + 1, n->count - index - 1);
			p->count += n->count - 1;
			sarray_remove_node(s, nindex);
		}
		else
		{
			for(index++; index < n->count; index++)
				n->array[index-1] = n->array[index];
			n->count--;
		}
	}
	s->count_element--;
	return e;
}

void sarray_dump(struct sarray* s)
{
	struct sarray_node *n;
	msg("sarray: count = %u, size = %u, count_element = %u\n", s->count, s->size, s->count_element);
	for(int i = 0; i < s->count; i++)
	{
		n = s->array[i];
		msg("\tindex = %d, size = %u, count = %u, elements: ", i, n->size, n->count);
		for(uns i = 0; i < n->count; i++)
			msg("%d, ", n->array[i]);
		msg("\n");
	}
}

