/** @file
 * Transactions in GUILE.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <guile/gh.h>
#include "kernel/kernel.h"
#include "build/path.h"

#include "scheme/gl_trans.h"
#include "scheme/glt_common.h"
#include "scheme/glt_kernel.h"

SCM curry = SCM_BOOL_F;
SCM make_thunk = SCM_BOOL_F;
static SCM vrr_exception_handler = SCM_BOOL_F;
static int exception_happened;

// raise exception called 'transaction-failed'
// this should be called in TRANS_FAILED sections in Scheme binding code
// this is special case, in general case use propagate_trans_fail.
void throw_transaction_failed (const char * dsc)
{
  scm_error_scm (symbol_transaction_failed, SCM_BOOL_F, scm_makfrom0str ("Transaction failed with message ~s"), scm_list_1 (scm_makfrom0str (dsc)), SCM_BOOL_F);
}

// just re-throw exception
static void forward_exception (SCM exception)
{
  if (SCM_FALSEP (exception)) return;

  SCM_ASSERT (SCM_NFALSEP (scm_pair_p (exception)), exception, SCM_ARG1, "throw-1");
  scm_throw (SCM_CAR (exception), SCM_CDR (exception));
}



static SCM
format_exception_body (void *data)
{
  SCM *args = data;
  return (scm_simple_format (SCM_BOOL_F, args [0], args [1]));
}

static SCM
format_exception_handler (UNUSED void *data, UNUSED SCM tag, UNUSED SCM throw_args)
{ return (SCM_BOOL_F); }

static SCM
format_exception (SCM str, SCM args)
{
  SCM push [2] = {str, args};
  return (scm_internal_catch (SCM_BOOL_T, format_exception_body, push, format_exception_handler, NULL));
}

// if any exception happened, then call trans_fail with appropriate arg,
// this should be called after call_guile_thunk
void
propagate_exception (SCM exception)
{
  if (SCM_FALSEP (exception)) return;
  if (SCM_NFALSEP (scm_pair_p (exception)) && SCM_SYMBOLP (SCM_CAR (exception)))
   {
      if (SCM_EQ_P (SCM_CAR (exception), symbol_transaction_failed) &&
	  SCM_NFALSEP (scm_pair_p (SCM_CDR (exception))) && SCM_STRINGP (SCM_CADR (exception)))
	trans_fail (SCM_STRING_CHARS (SCM_CADR (exception)));
      else if ((scm_ilength (exception) >= 4) && SCM_STRINGP (SCM_CADDR (exception))
	       && SCM_NFALSEP (scm_list_p (SCM_CADDDR (exception))))
	{
	  SCM txt = format_exception (SCM_CADDR (exception), SCM_CADDDR (exception));
	  if (SCM_STRINGP (txt)) // decoding successful
	    trans_fail ("%s", SCM_STRING_CHARS (txt));
	  else  // failure in exception decoding
	    trans_fail ("Scheme exception (strange) %s", SCM_SYMBOL_CHARS (SCM_CAR (exception)));
	}
      else
	trans_fail ("Scheme exception %s", SCM_SYMBOL_CHARS (SCM_CAR (exception)));
    }
  else printf ("UNKNOWN EXCEPTION\n");
  // trans_fail ("Unknown - strange Scheme exception");
}

// Generate appropriate exception
// this should be called in TRANS_FAILED sections in Scheme binding code
// when trans_fail may be generated by exception in inner Scheme code.
// in special case (no inner scheme code) call to propagate_trans_fail
// can be replaced with call to throw_transaction_failed.
void
propagate_trans_fail (const char * dsc)
{
  throw_transaction_failed (dsc);
  /*
  if (SCM_FALSEP (stored_exception))

  else forward_exception (stored_exception);
  */
}


static SCM
gl_vrr_exception_handler(SCM exception)
{
   DBG("Default hander for transaction exception called\n");
   exception_happened = 1;
   return (exception);
}

// function for calling non-trivial scheme code.
// If exception happened during call, then it is stored in *exception_ptr,
// otherwise thee is SCM_BOOL_F
SCM
call_guile_thunk (SCM thunk, SCM *exception_ptr)
{
  exception_happened = 0;

  SCM retval = scm_call_with_dynamic_root (thunk, vrr_exception_handler);
  if (exception_happened)
    {
      *exception_ptr = retval;
      retval = SCM_UNSPECIFIED;
    }
  else *exception_ptr = SCM_BOOL_F;

  return (retval);
}

// function for calling non-trivial scheme code.
// Exceptions are propagated with trans_fail
SCM call_guile_fn_from_c (SCM fn, SCM args)
{
  SCM exception = SCM_BOOL_F;
  SCM thunk = scm_apply_1 (make_thunk, fn, args);
  SCM retval = call_guile_thunk (thunk, &exception);
  propagate_exception (exception);
  return (retval);
}

static struct obj_tlo*
get_tlo_or_universum(SCM proxy)
{
  if (is_specific_proxy(proxy, T_OBJ, OT_TLO, -1) ||
      is_specific_proxy(proxy, T_OBJ, OT_UNIVERSE, -1))
    return (struct obj_tlo *) SCM_SMOB_DATA (proxy);
  else
    return NULL;
}

/** \brief Make command in main trans.
 * \param proxy is a tlo where the command will be executed.
 * \param thunk Command to execute.
 * \param handler Exception handler. Optional, if unused then exception is propagated.
 * \param dsc Name of this command for the undo. Optional.
 */

static SCM gl_trans_main(SCM proxy, SCM thunk, SCM handler, SCM dsc)
{
  char buf[TRANS_ERR_SIZE], *c_dsc;
  struct obj_tlo * tlo;
  volatile SCM retval = SCM_UNSPECIFIED;
  SCM exception = SCM_BOOL_F;

  SCM_ASSERT(tlo = get_tlo_or_universum(proxy), proxy, SCM_ARG1, "trans-main");
  SCM_ASSERT(SCM_NFALSEP (scm_procedure_p(thunk)), thunk, SCM_ARG2, "trans-main");
  if(handler != SCM_UNDEFINED)
    SCM_ASSERT(SCM_NFALSEP (scm_procedure_p(handler)), handler, SCM_ARG3, "trans-main");
  if(dsc != SCM_UNDEFINED)
    {
      SCM_ASSERT(SCM_STRINGP(dsc), dsc, SCM_ARG4, "trans-main");
      c_dsc = SCM_STRING_CHARS (dsc);
    }
  else
    c_dsc = "Anonymous scheme transaction";

  // FIXME check for argcount (what)

  DBG ("ENTER %p\n", &tlo);

  TRANS_BEGIN_MAIN (tlo, c_dsc, buf)
    {
      retval = call_guile_thunk (thunk, &exception);
      DBG ("AFTER %p\n", &tlo);
      // scm_write (exception, SCM_UNDEFINED);
      propagate_exception (exception);
    }
  TRANS_FAILED
    {
      DBG ("FAILED %p\n", &tlo);
      if (handler == SCM_UNDEFINED) forward_exception (exception);
      else retval = scm_apply_0 (handler, exception);
    }
  TRANS_END;

  DBG ("FINISH %p\n", &tlo);

  return retval;
}

static SCM gl_trans(SCM thunk, SCM handler)
{
  char buf[TRANS_ERR_SIZE];
  volatile SCM retval = SCM_UNSPECIFIED;
  SCM exception = SCM_BOOL_F;

  SCM_ASSERT (SCM_NFALSEP (scm_procedure_p (thunk)), thunk, SCM_ARG1, "trans");
  if (handler != SCM_UNDEFINED)
    SCM_ASSERT (SCM_NFALSEP (scm_procedure_p (handler)), handler, SCM_ARG2, "trans");

  TRANS_BEGIN (buf)
    {
      retval = call_guile_thunk (thunk, &exception);
      propagate_exception (exception);
    }
  TRANS_FAILED
    {
      if (handler == SCM_UNDEFINED) forward_exception (exception);
      else retval = scm_apply_0 (handler, exception);
    }
  TRANS_END;

  return retval;
}

static SCM gl_trans_fail(SCM dsc)
{
  SCM_ASSERT (SCM_STRINGP(dsc), dsc, SCM_ARG1, "trans-fail");
  throw_transaction_failed (SCM_STRING_CHARS (dsc));
  return (SCM_UNSPECIFIED);
}
 
static SCM gl_undo(SCM proxy)
{
	struct obj_tlo* tlo;
	char buf[TRANS_ERR_SIZE];

	SCM_ASSERT(tlo = get_tlo_or_universum(proxy), proxy, SCM_ARG1, "undo");

	if(!trans_undo(tlo, buf))
		msg("Undo failed: %s\n", buf);
	return proxy; // buf should be returned
}

static SCM gl_redo(SCM proxy)
{
	struct obj_tlo* tlo;
	char buf[TRANS_ERR_SIZE];

	SCM_ASSERT(tlo = get_tlo_or_universum(proxy), proxy, SCM_ARG1, "redo");

	if(!trans_redo(tlo, buf))
		msg("Redo failed: %s\n", buf);
	return proxy; // buf should be returned
}

void
gl_trans_bindings_init(void)
{
  scm_c_define_gsubr("trans-main", 2, 2, 0, gl_trans_main);
  scm_c_define_gsubr("trans", 1, 1, 0, gl_trans);
  scm_c_define_gsubr("trans-fail", 1, 0, 0, gl_trans_fail);
  scm_c_define_gsubr("undo", 1, 0, 0, gl_undo);
  scm_c_define_gsubr("redo", 1, 0, 0, gl_redo);

  scm_c_export ("trans-main", "trans", "trans-fail", "undo", "redo", NULL);
}


void
gl_trans_init (void)
{
  vrr_exception_handler = scm_c_make_gsubr("vrr-exception-handler", 0, 0, 1, gl_vrr_exception_handler);
}
