;;; Load mechanism
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.

(define-module (vrr load))

(use-modules (vrr misc))
(use-modules (vrr low))
(use-modules (vrr support))
(use-modules (vrr property))

(export load-doc)

;; actions: just (read-enable 'positions)

;;; Tokenize routines for load (warning: extreme perversion)


(define (convert-to-tokens struct)
  (cond
   ((null? struct)
    (list (cons 'open-list #f) (cons 'close-list #f)))
   ((list? struct)
    (append! (list (cons 'open-list (source-properties struct)))
	     (apply append! (map convert-to-tokens struct))
	     (list (cons 'close-list #f))))
   ((pair? struct)
    (parse-error-int "Invalid token - improper list: ~s" #f (list (last-pair struct))))
   ((eof-object? struct)
    (list (cons 'end-of-file #f)))
   (else
    (list (cons 'data struct)))))

(define (make-token-stream port)
  (vector port '() #f))

(define (fill-token-buffer! stream)
  (if (null? (vector-ref stream 1))
      (vector-set! stream 1 (convert-to-tokens (read (vector-ref stream 0))))))

(define (read-next-token stream)
  (fill-token-buffer! stream)
  (let ((old-data (vector-ref stream 1)))
    (vector-set! stream 1 (cdr old-data))
    (if (and (eq? (caar old-data) 'open-list) (cdar old-data))
	(vector-set! stream 2 (cdar old-data)))
    (car old-data)))

(define (look-next-token stream)
  (fill-token-buffer! stream)
  (car (vector-ref stream 1)))

(define (get-last-source-properties stream) (vector-ref stream 2))


;;; Load

(read-enable 'positions)

(define (parse-error-int string sp args)
  (if sp
      (scm-error 'parse-error #f (string-append "Parse error after ~s:~s:~s - " string)
		 (cons* (assq-ref sp 'filename) (assq-ref sp 'line)
			(assq-ref sp 'column) args) #f)
      (scm-error 'parse-error #f (string-append "Parse error - " string) args #f)))


(define (load-internal token-stream)
  (define (parse-error string . args)
    (parse-error-int  string (get-last-source-properties token-stream) args))

  (define (want-specific type pred dsc)
    (let ((data (read-next-token token-stream)))
      (if (and (eq? (car data) type) (pred (cdr data)))
	  (if (eq? (car data) 'data)
	      (cdr data)
	      *unspecified*)
	  (parse-error "invalid token ~s (expected ~s)"
		       (if (eq? (car data) 'data)
			   (cdr data)
			   (car data)) dsc))))

  (define (look-specific type pred)
    (let ((data (look-next-token token-stream)))
      (and (eq? (car data) type) (pred (cdr data)))))

  (define (want-data) (want-specific 'data true 'data))
  (define (want-open-list) (want-specific 'open-list true 'open-list))
  (define (want-close-list) (want-specific 'close-list true 'close-list))
  (define (want-symbol) (want-specific 'data symbol? 'symbol))
  (define (want-number) (want-specific 'data number? 'number))

  (define (look-open-list) (look-specific 'open-list true))
  (define (look-close-list) (look-specific 'close-list true))

  (define (want-sexp)
    (if (look-open-list)
	(begin
	  (want-open-list)
	  (do ((acc '() (cons (want-sexp) acc)))
	      ((look-close-list)
	       (want-close-list)
	       (reverse acc))))
	(want-data)))

  (define (go-get-hanger-by-name-or-fail a b)
    (or (go-get-hanger-by-name a b)
	(parse-error-int "unsupported go hanger '~s' at '~s'" #f (list b (proxy-subtype a)))))

  (define (go-get-hanger-curve-or-fail a)
    (or (go-get-hanger-curve a)
	(parse-error-int "unsupported go curve hanger at '~s'" #f (list (proxy-subtype a)))))

  (define (read-properties obj)
    (define id #f)
    (want-open-list)
    (do () ((look-close-list))
      (want-open-list)
      (let* ((prop-name (want-symbol))
	     (prop-value (want-sexp)))
	(cond
	 ((and (proxy? obj 'go 'elliptic-arc)
	       (eq? prop-name 'conic))
	  (apply set-conic! (cons obj prop-value)))
	 ((not (eq? prop-name 'id))
	  (set-property! obj prop-name prop-value))
	 ((natural? prop-value)
	  (set! id prop-value))
	 ((string? prop-value)
	  (set! id prop-value)
	  (set-property! obj 'id prop-value))
	 (else
	  (parse-error "invalid object identifier definition"))))
      (want-close-list))
    (want-close-list)
    id)

  (define (read-tlo-content tlo)
    (define ht (make-hash-table 255)) ; larger value?

    (define (table-lookup id)
      (define item (hash-ref ht id))
      (if (proxy? item) item #f))

    (define (table-register-todo id todo)
      (hash-set! ht id (cons todo (hash-ref ht id '()))))

					;    (define (table-register-patch id h h-type)
					;    (hash-set! ht id (cons (cons h hname) (hash-ref ht id '()))))

    (define (table-register-go id go)
      (let ((patches (hash-ref ht id '())))
	(if (not (list? patches)) (parse-error "more objects with one ID: ~S ~S" patches go))
	(for-each (lambda (todo) (todo go)) patches)
	(hash-set! ht id go)))

    (define (read-go-link none)
      (if (look-open-list)
	  (read-go)
	  (let ((id (want-data)))
	    (if (not (or (string? id) (number? id) (and none (not id))))
		(parse-error "invalid referenced ID: ~S" id))
	    id)))

    (define (know-link go)
      (if (proxy? go) go (table-lookup go)))

    (define (make-rehang-todo fake-hanger hanger-type)
      (lambda (go)
	(anchor-rehang (hanger-get-first-anchor fake-hanger)
		       (go-get-hanger-by-name-or-fail go hanger-type))))

    (define (register-rehang id fake-hanger hanger-type)
      (table-register-todo id (make-rehang-todo fake-hanger hanger-type))
      fake-hanger)

    ;; read-hanger-inside doesn't return hanger, but promise of hanger
    (define (read-pos-hanger-in)
      (let ((hanger-type (want-symbol)))
	(if (eq? hanger-type 'coords)
	    (let* ((x (want-number)) (y (want-number))) (hanger-new-mc-xy x y))
	    (let ((link (read-go-link #t)))
	      (delay
		(let ((go (know-link link)))
		  (if go
		      (go-get-hanger-by-name-or-fail go hanger-type)
		      (register-rehang link (hanger-new-mc-xy 0 0) hanger-type))))))))

    (define (read-pos-hanger)
      (want-open-list)
      (let ((retval (read-pos-hanger-in)))
	(want-close-list)
	retval))


    (define (read-cur-hanger)
      (let ((link (read-go-link #t)))
	(if link
	    (delay
	      (let ((go (know-link link)))
		(if go
		    (go-get-hanger-curve-or-fail go)
		    (register-rehang link (hanger-new-mc-curve) 'curv))))
	    (hanger-new-mc-curve))))

    ;; warning - #t (not #f) in after slot to link at front.
    (define (make-link-package grp what after)
      (lambda (input-what input-after)
	(if input-what  (set! what  input-what))
	(if input-after (set! after input-after))
	(if (and what after)
	    (if (boolean? after)
		(go-relink-head what grp)
		(go-relink what grp after)))))

    (define (make-what-todo pkg)
      (lambda (go) (pkg go #f)))

    (define (make-after-todo pkg)
      (lambda (go) (pkg #f go)))

    (define (read-go-args type)
      (map safe-force ;; read-hanger return promise of hanger
	   (map (lambda (x)
		  (case x
		    ((hanger) (read-pos-hanger))
		    ((hanger-c) (read-cur-hanger))
		    ((data) (want-data))
		    ;; ((go) (read-hanger)) ((number) (want-number)) ((string) (want-string))
		    (else (error "BUG in read-go-header"))))
		(or (go-get-source-dsc type)
		    (parse-error "unsupported go subtype (~s)" type)))))

    (define (read-group-content-raw)
      (want-open-list)
      (do ((acc '() (cons (read-go-link #f) acc)))
	  ((look-close-list)
	   (want-close-list)
	   (reverse acc))))

    (define (read-group-content grp)
      (let ((list (read-group-content-raw)))
	(if (not (null? list))
	    (let rec ((act (car list))
		      (old #t)
		      (known-act (know-link (car list)))
		      (known-old #t)
		      (rest (cdr list)))
	      (let ((pkg (make-link-package grp known-act known-old)))
		(pkg #f #f) ; in case that everything is ok
		(if (not known-act) (table-register-todo act (make-what-todo pkg)))
		(if (not known-old) (table-register-todo old (make-after-todo pkg))))
	      (if (not (null? rest))
		  (rec (car rest) act (know-link (car rest)) known-act (cdr rest)))))))

    (define (read-go-header type)
      (if (eq? type 'top-level-group)
	  (tlo-get-tlg (get-trans-tlo))
	  (make-go type (read-go-args type))))

    (define (read-go)
      (want-open-list)
      (let* ((go (read-go-header (want-symbol)))
	     (id (read-properties go)))
	(if id (table-register-go id go))
	(if (proxy? go 'go 'group)
	    (read-group-content go))
	(want-close-list)
	go))

    ;; body of read-tlo-content
    (want-open-list)
    (do () ((look-close-list))
      (read-go))
    (want-close-list)
    tlo)

  (define (read-obj-children obj)
    (want-open-list)
    (do () ((look-close-list))
      (obj-link-tail (read-obj) obj))
    (want-close-list))

  (define (read-tlo-rest)
    (let ((tlo (tlo-new)))
      (read-properties tlo)
      (trans-main tlo
		  (lambda ()
		    (read-tlo-content tlo)))
      (want-close-list)
      tlo))

  (define (read-doc-rest)
    (let ((doc (doc-new)))
      (let ((version (want-number)))
	(if (not (eqv? version 0.6))
	    (parse-error "unsupported file version ~s" version)))
      (read-properties doc)
      (read-obj-children doc)
      (want-close-list)
      doc))

  (define (read-univ-rest)
    (let ((version (want-number)))
      (if (not (eqv? version 0.6))
	  (parse-error "unsupported file version ~s" version)))
    (read-properties universe)
    (want-close-list)
    #f)

  (define (read-obj)
    (want-open-list)
    (let ((type (want-symbol)))
      (case type
	((page tlo) (read-tlo-rest))
	((document) (read-doc-rest))
	((universe universum) (read-univ-rest))
	(else (parse-error "unsupported obj type (~s)" type)))))

  ;; body of load-internal
  (let ((doc (read-obj)))
    (want-specific 'end-of-file true 'end-of-file)
    (if doc (obj-link-tail doc universe))
    doc))

(define (load-doc filename)
  (trans-main universe
	      (lambda ()
		(call-with-input-file filename
		  (lambda (port) (load-internal (make-token-stream port)))))
	      (lambda x  (apply throw x))
	      "Document load"))

