;;; Support functions for common user functions
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.

(define-module (vrr support))

(use-modules (vrr misc))
(use-modules (vrr low))

;; actions: fills tables (idtables, src-table)

(define (typearg-error fn arg val)
  (scm-error 'wrong-type-arg fn "Wrong type argument in position ~S: ~S" (list arg val) #f))

(define (typearg2-error fn arg val wanted)
  (if wanted
      (scm-error 'wrong-type-arg fn "Wrong type argument in position ~S: ~S (wanted ~S)" (list arg val wanted) #f)
      (scm-error 'wrong-type-arg fn "Wrong type argument in position ~S: ~S" (list arg val) #f)))

;;; Support functions for common user functions

(define (make-idtable . items)
  (let* ((l (length items))
	 (t1 (make-vector l))
	 (t2 (make-hash-table (* l 2))))
    (for-each (lambda (key value)
		(vector-set! t1 value key)
		(hashq-set! t2 key value))
	      items (iota l))
    (cons t1 t2)))

(define (make-idtable-lookup-1 idt)
  (lambda (n) (vector-ref (car idt) n)))

(define (make-idtable-lookup-2 idt)
  (lambda (k) (hashq-ref (cdr idt) k)))

(define conic-table (make-idtable
		     'start-entire 'start-dif 'point-entire 'point-dif
		     'ccw 'cw 'smaller 'bigger 'middle 'opposite 'entire))

(define-public code-conic (make-idtable-lookup-2 conic-table))
(define-public decode-conic (make-idtable-lookup-1 conic-table))

(define cap-style-table (make-idtable 'butt 'round 'projecting))

(define-public code-cap-style (make-idtable-lookup-2 cap-style-table))
(define-public decode-cap-style (make-idtable-lookup-1 cap-style-table))

(define join-style-table (make-idtable 'miter 'round 'bevel))

(define-public code-join-style (make-idtable-lookup-2 join-style-table))
(define-public decode-join-style (make-idtable-lookup-1 join-style-table))

(define alignment-x-table (make-idtable 'ref-left 'ref-center 'ref-right 'ref-relative
					'bbox-left 'bbox-center 'bbox-right 'bbox-relative))

(define-public code-alignment-x (make-idtable-lookup-2 alignment-x-table))
(define-public decode-alignment-x (make-idtable-lookup-1 alignment-x-table))

(define alignment-y-table (make-idtable 'baseline 'bbox-left 'bbox-center 'bbox-right 'bbox-relative))

(define-public code-alignment-y (make-idtable-lookup-2 alignment-y-table))
(define-public decode-alignment-y (make-idtable-lookup-1 alignment-y-table))

(define arrow-alignment-table (make-idtable  'front 'back))

(define-public code-arrow-alignment (make-idtable-lookup-2 arrow-alignment-table))
(define-public decode-arrow-alignment (make-idtable-lookup-1 arrow-alignment-table))

(define arrow-front-table (make-idtable 'straight 'parabolic))

(define-public code-arrow-front (make-idtable-lookup-2 arrow-front-table))
(define-public decode-arrow-front (make-idtable-lookup-1 arrow-front-table))

(define arrow-back-table (make-idtable 'none 'straight 'poly 'parabolic))

(define-public code-arrow-back (make-idtable-lookup-2 arrow-back-table))
(define-public decode-arrow-back (make-idtable-lookup-1 arrow-back-table))

(define-public src-table (make-hash-table 31))
(define cons-table (make-hash-table 31))


(define (define-source subtype srclist)
  (hashq-set! src-table subtype srclist))

(define-public (define-constructor subtype make-fn)
  (hashq-set! cons-table subtype make-fn))

(define-public (make-go subtype args)
  (apply (hashq-ref cons-table subtype) args))

;;; Common user functions

(define-public (vrr-type-assert fn . args)
  (let loop ((args args) (order 1))
    (define (check-next value ok wanted rest)
      (if (not ok)
	  (typearg2-error fn order value wanted))
      (loop rest (+ order 1)))
    (cond 
     ((null? args) #t)
     ((boolean? (cadr args))
      (check-next (car args) (cadr args) (caddr args) (cdddr args)))
     ((eq? (cadr args) 'natural)
      (check-next (car args) (natural? (car args)) 'natural (cddr args)))
     ((eq? (cadr args) 'real)
      (check-next (car args) (real? (car args)) 'real (cddr args)))
     ((eq? (cadr args) 'symbol)
      (check-next (car args) (string? (car args)) 'string (cddr args)))
     ((eq? (cadr args) 'string)
      (check-next (car args) (string? (car args)) 'string (cddr args)))
     ((eq? (cadr args) 'obj)
      (check-next (car args) (proxy? (car args) 'obj) 'obj (cddr args)))
     ((eq? (cadr args) 'go)
      (check-next (car args) (proxy? (car args) 'go) 'go (cddr args)))
     ((eq? (cadr args) 'anchor)
      (check-next (car args) (proxy? (car args) 'anchor) 'anchor (cddr args)))
     ((eq? (cadr args) 'hanger)
      (check-next (car args) (proxy? (car args) 'hanger) 'hanger (cddr args)))
     (else
      (error "bug in type specification of fn ~S" fn)))))

;(define (property-get-subtype obj name)
; (let ((x (property-get-subtype-i obj name)))
;   (and x (property-type-symbol (car x) (cdr x)))))


(define-source 'point 1)
(define-source 'segment 2)

(define-source 'ellipse-by-center-2-radii-rotation-2-angles 
  (list 0 'a-radius 'b-radius 'rotation 'start 'dif 'conic)) ;; OBSOLETE

(define-source 'ellipse-by-center-2-radii-rotation (list 0 'a-radius 'b-radius 'rotation))
(define-source 'ellipse-by-2-foci-point 3)
(define-source 'ellipse-by-3-points-smallest 3)
(define-source 'ellipse-by-3-points-rotation-eccentricity (list 0 1 2 'rotation 'ecc))
(define-source 'ellipse-by-center-point-rotation-eccentricity (list 0 1 'rotation 'ecc))

(define-source 'quadratic-bezier (list 0 'weight-0 1 'weight-1 2 'weight-2))
(define-source 'cubic-bezier (list 0 'weight-0 1 'weight-1 2 'weight-2 3 'weight-3))

(define (text-get-font-family go) (font-family (car (property-get-int go 'font))))
(define (text-get-font-style go)  (font-style  (car (property-get-int go 'font))))

(define-source 'text (list 0 'text text-get-font-family text-get-font-style 'size))
(define-source 'tex-text (list 0 'text))

(define-source 'parametric-point (list -1 'parameter))
(define-source 'intersection (list -1 -2 'parameter))
(define-source 'decorator-point (list 0 'vertices 'radius 'rotation))
(define-source 'decorator-arrow (list -1 'parameter 'arrow-length 'rotation))

(define-source 'group '())
(define-source 'path '())


(define-public (lookup name) (o-lookup name))
(define-public (look name) (o-look name))

