/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>


int vcl_property_sizes [VCL_PROP_COUNT] =
  {[VCL_PROP_FILL_COLOR] = sizeof (u32),
   [VCL_PROP_STROKE_COLOR] = sizeof (u32),
   [VCL_PROP_STROKE_WIDTH] = sizeof (double),
   [VCL_PROP_STROKE_CAP_STYLE] = sizeof (u8),
   [VCL_PROP_STROKE_JOIN_STYLE] = sizeof (u8)};

int vcl_property_offsets [VCL_PROP_COUNT] =
  {[VCL_PROP_FILL_COLOR] = OFFSETOF(struct vcl_context, fill_color),
   [VCL_PROP_STROKE_COLOR] = OFFSETOF(struct vcl_context, stroke_color),
   [VCL_PROP_STROKE_WIDTH] = OFFSETOF(struct vcl_context, stroke_width),
   [VCL_PROP_STROKE_CAP_STYLE] = OFFSETOF(struct vcl_context, stroke_cap_style),
   [VCL_PROP_STROKE_JOIN_STYLE] = OFFSETOF(struct vcl_context, stroke_join_style)};


struct vcl_context *
vcl_context_new (void)
{
  struct vcl_context *ct = xmalloc (sizeof (struct vcl_context));
  vcl_ga_init (&(ct -> storage), sizeof (struct vcl_context), 1);
  // sizeof (struct vcl_context) in init_items arg is heuristic;
  geom_transform2_set_identity (&(ct -> t));
  ct -> dist_transform = 1;

  //  ct -> fill_color = get_color (255, 255, 128, 255);
  ct -> fill_color = get_color (255, 255, 255, 0);
  ct -> stroke_color = get_color (0, 0, 0, 255);
  ct -> stroke_width = 0;
  ct -> stroke_cap_style = VCL_CAP_ROUND;
  ct -> stroke_join_style = VCL_JOIN_ROUND;
  return (ct);
}

void
vcl_context_destroy (struct vcl_context *ct)
{
  vcl_ga_finish (&(ct -> storage));
  xfree (ct);
}

void
vcl_context_set_property (struct vcl_context *ct, int prop_id, void *data)
{
  int bytes = vcl_property_sizes [prop_id];
  u8 *slot = vcl_property_offsets [prop_id] + (u8 *) ct;
  u8 *buff = vcl_ga_allocate (&(ct -> storage), bytes);
  memcpy (buff, slot, bytes);
  memcpy (slot, data, bytes);
}

void
vcl_context_unset_property (struct vcl_context *ct, int prop_id)
{
  int bytes = vcl_property_sizes [prop_id];
  u8 *slot = vcl_property_offsets [prop_id] + (u8 *) ct;
  u8 *buff = vcl_ga_unallocate (&(ct -> storage), bytes);
  // I know that buffer is not freed so i can read.
  memcpy (slot, buff, bytes);
}

void
vcl_context_preply_transform (struct vcl_context *ct, void *trans)
{
  struct geom_transform2 *buff = vcl_ga_allocate (&(ct -> storage),
						 sizeof (struct geom_transform2));
  *buff = ct -> t;
  vcl_transformation_primary_preply_matrix  (trans, &(ct -> t . primary), &(ct -> t . primary));
  vcl_transformation_inverted_preply_matrix (trans, &(ct -> t . inverted), &(ct -> t . inverted));
  ct -> dist_transform = -1;
}

void
vcl_context_unpreply_transform (struct vcl_context *ct)
{
  struct geom_transform2 *buff = vcl_ga_unallocate (&(ct -> storage),
						    sizeof (struct geom_transform2));
  // I know that buffer is not freed so i can read.
  ct -> t = *buff;
  ct -> dist_transform = -1;
}

void
vcl_context_update_dist_transform (struct vcl_context *ct)
{
  struct geom_vector v = {1, 0};
  geom_vector_transform (&v, &(ct -> t . primary), &v);
  ct -> dist_transform = sqrt (v . dx * v . dx + v . dy * v  . dy);
}

static inline void
convert_rectangle (struct vcl_context *ct, const struct vcl_rectangle *src, struct geom_rectangle *dst)
{
  struct geom_rectangle tmp;
  tmp . ll . x = src -> lx;
  tmp . ll . y = src -> ly;
  tmp . ur . x = src -> hx;
  tmp . ur . y = src -> hy;
  geom_rectangle_transform (&tmp, &(ct -> t . inverted), dst);
}

void
vcl_context_walk_node (struct vcl_context *ct, void *painter,
		       void *node, const struct vcl_rectangle *bbox,
		       const struct geom_rectangle *bbox_local)
{
  struct geom_rectangle new_bbox_local;

  /*
  if (vcl_char_p (node))
    {
      struct geom_rectangle r;
      vcl_node_get_bbox (node, &r);
      draw_bbox (painter, &r);
    }
  */

  if (node == NULL) return;

  if (vcl_property_p (node))
    {
      int i;
      int j = vcl_property_get_count (node);
      for (i = 0; i < j; i ++)
	{
	  vcl_context_set_property (ct, vcl_property_get_nth_type (node, i),
				    vcl_property_get_nth_value (node, i));
	}
    }

  if (vcl_if_transformation_p (node))
    {
      /*      printf ("BEFORE %p:\n", node);
      geom_transform_dump (&(ct -> t . primary));
      geom_transform_dump (&(ct -> t . inverted));
      geom_rectangle_dump (bbox_local);
      */
      vcl_context_preply_transform (ct, node);
      convert_rectangle (ct, bbox, &new_bbox_local);
      bbox_local = &new_bbox_local;
      /*            printf ("AFTER %p:\n", node);
      geom_transform_dump (&(ct -> t . primary));
      geom_transform_dump (&(ct -> t . inverted));
      geom_rectangle_dump (bbox_local);
      */
    }

  if (vcl_if_composite_p (node))
    vcl_painter_draw_composite (painter, node, bbox, bbox_local);
  else if (vcl_if_enclosure_p (node))
    vcl_context_walk_node (ct, painter, vcl_enclosure_get_child (node), bbox, bbox_local);
  else
    vcl_painter_draw_leaf (painter, node, bbox, bbox_local);

  if (vcl_if_transformation_p (node))
    vcl_context_unpreply_transform (ct);

  if (vcl_property_p (node))
    {
      int i;
      int j = vcl_property_get_count (node);
      for (i = j - 1; i >= 0; i --)
	{
	  vcl_context_unset_property (ct, vcl_property_get_nth_type (node, i));
	}
    }
}

struct cb_data
{
  struct vcl_context *ct;
  void *painter;
  const struct vcl_rectangle *bbox;
  const struct geom_rectangle *bbox_local;
};

static void cb (void *child, void *data_ptr)
{
  const struct cb_data * data = data_ptr;
  vcl_context_walk_node (data -> ct, data -> painter, child, data -> bbox, data -> bbox_local);
}

void
vcl_context_walk_children (struct vcl_context *ct, void *painter,
		           void *node, const struct vcl_rectangle *bbox,
			   const struct geom_rectangle *bbox_local)
{
  struct cb_data cb_data = {ct, painter, bbox, bbox_local};
  vcl_composite_get_children_in_bbox_backwards (node, bbox_local, cb, &cb_data);
}
