/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_group;

static inline int
want_redraw (struct vcl_group * obj)
{ return  (! (obj -> parent == NULL)); }

static void
altered (struct vcl_group * obj, const struct geom_rectangle *where)
{
  if (want_redraw (obj) && (! geom_rectangle_is_empty (where)))
    vcl_container_child_altered (obj -> parent, obj, where);
}

static void
changed (struct vcl_group * obj, const struct geom_rectangle *old_bbox, const struct geom_rectangle *new_bbox)
{
  struct geom_rectangle temp;

  geom_rectangle_merge (old_bbox, new_bbox, &temp);

  // FIXME shrink bbox when new_bbox is smaller?
  if (geom_rectangle_is_in (&temp, &(obj -> bbox)))
    altered (obj, &temp);
  else /* new bbox is not empty, because old_bbox is always in obj's bbox */
    {
      struct geom_rectangle old_bbox = obj -> bbox;

      geom_rectangle_merge (&(obj -> bbox), new_bbox, &(obj -> bbox));
      if (want_redraw (obj))
	vcl_container_child_changed (obj -> parent, obj, &old_bbox, &(obj -> bbox));
    }
}

static void vcl_group_remove_child (struct vcl_group * obj, void * child);


// OBJECT interface

static void
vcl_group_destroy (struct vcl_group * obj)
{
  ASSERT (obj -> parent == NULL);

  for (int i = 0; i < obj -> nodes . used ; i++)
    {
      void * child = VCL_GA_REF (&(obj -> nodes), i, void *);

      vcl_node_set_parent (child, NULL);
      vcl_object_destroy (child);
    }

  vcl_ga_finish (&(obj -> nodes));
  xfree (obj);
}


// NODE interface

static void
vcl_group_set_data (struct vcl_group * obj, int data)
{ obj -> data = data; }

static int
vcl_group_get_data (struct vcl_group * obj)
{ return (obj -> data); }

static void
vcl_group_set_parent (struct vcl_group * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_group_get_parent (struct vcl_group * obj)
{ return (obj -> parent); }

static void
vcl_group_get_bbox (struct vcl_group * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> bbox; }


// CONTAINER interface

static void
vcl_group_child_altered (struct vcl_group * obj, void * child UNUSED,
			 const struct geom_rectangle *where)
{ altered (obj, where); }

static void
vcl_group_child_changed (struct vcl_group * obj, void * child UNUSED,
			 const struct geom_rectangle *old_bbox,
			 const struct geom_rectangle *new_bbox)
{ changed (obj, old_bbox, new_bbox); }


// COMPOSITE interface

static void
vcl_group_get_children (struct vcl_group * obj, void (* cb)(void *, void * ), void * cb_data)
{
  for (int i = obj -> nodes . used - 1; i >= 0; i--)
    {
	void * child = VCL_GA_REF (&(obj -> nodes), i, void *);
	cb (child, cb_data);
    }
}

static void
vcl_group_get_children_backwards (struct vcl_group * obj, void (* cb)(void *, void * ), void * cb_data)
{
  for (int i = 0; i < obj -> nodes . used; i++)
    {
	void * child = VCL_GA_REF (&(obj -> nodes), i, void *);
	cb (child, cb_data);
    }
}

static void
vcl_group_get_children_in_bbox (struct vcl_group * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void * ), void * cb_data)
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox))
    vcl_group_get_children (obj, cb, cb_data);
}

static void
vcl_group_get_children_in_bbox_backwards (struct vcl_group * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void * ), void * cb_data)
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox))
    vcl_group_get_children_backwards (obj, cb, cb_data);
}

/*
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox))
    {
      for (int i = 0; i < obj -> nodes . used; i++)
        {
	  void * child = VCL_GA_REF (&(obj -> nodes), i, void *);
	  struct geom_rectangle child_bbox;
	  vcl_node_get_bbox (child, &child_bbox);
	  if (geom_rectangle_have_intersection (&child_bbox, wanted_bbox))
	    cb (child, cb_data);
	}
    }
}
*/

// PLACEMENT interface

static void
vcl_group_add_child_on_top (struct vcl_group * obj, void * child)
{
  struct geom_rectangle child_bbox;

  int i = vcl_ga_new_item (&(obj -> nodes));
  VCL_GA_REF (&(obj -> nodes), i, void *) = child;

  vcl_node_set_parent (child, obj);
  vcl_node_set_data (child, i);

  vcl_node_get_bbox (child, &child_bbox);
  changed (obj, NULL, &child_bbox);
}

static void
vcl_group_remove_child (struct vcl_group * obj, void * child)
{
  struct geom_rectangle child_bbox;
  int i;

  for (i = vcl_node_get_data (child) + 1; i < obj -> nodes . used; i ++)
    {
      void * tmp = VCL_GA_REF (&(obj -> nodes), i, void *);
      vcl_node_set_data (tmp, i - 1);
      VCL_GA_REF (&(obj -> nodes), i - 1, void *) = tmp;
    }

  obj -> nodes . used --;
  vcl_node_set_parent (child, NULL);

  vcl_node_get_bbox (child, &child_bbox);
  changed (obj, &child_bbox, NULL);
}


// specific functions

struct vcl_group *
vcl_group_new (void)
{
  struct vcl_group *obj = xmalloc (sizeof (struct vcl_group));
  vcl_object_init (obj, vcl_vt_group);

  vcl_ga_init (&(obj -> nodes), 4, sizeof (void *));
  geom_rectangle_set_empty (&(obj -> bbox));

  obj -> parent = NULL;
  obj -> data = 0;

  return (obj);
}


void
vcl_group_main (void)
{
  vcl_vt_group = vcl_class_new ("group");

  vcl_class_add_interface (vcl_vt_group, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_group_destroy);

  vcl_class_add_interface (vcl_vt_group, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_group_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_group_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_group_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_group_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_group_get_bbox);

  vcl_class_add_interface (vcl_vt_group, vcl_if_container);
  vcl_class_add_method_li (vcl_mid_container_child_altered, vcl_group_child_altered);
  vcl_class_add_method_li (vcl_mid_container_child_changed, vcl_group_child_changed);

  vcl_class_add_interface (vcl_vt_group, vcl_if_composite);
  vcl_class_add_method_li (vcl_mid_composite_get_children, vcl_group_get_children);
  vcl_class_add_method_li (vcl_mid_composite_get_children_backwards, vcl_group_get_children_backwards);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox, vcl_group_get_children_in_bbox);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox_backwards, vcl_group_get_children_in_bbox_backwards);

  vcl_class_add_interface (vcl_vt_group, vcl_if_placement);
  vcl_class_add_method_li (vcl_mid_placement_add_child_on_top, vcl_group_add_child_on_top);
  vcl_class_add_method_li (vcl_mid_placement_remove_child, vcl_group_remove_child);
}

