/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

#include <gdk/gdkx.h>
#include <cairo-xlib.h>
vcl_vtable *vcl_vt_painter_cairo;

int vcl_to_cairo_cap_styles [VCL_CAP_COUNT] =
  {[VCL_CAP_BUTT] = CAIRO_LINE_CAP_BUTT,
   [VCL_CAP_ROUND] = CAIRO_LINE_CAP_ROUND,
   [VCL_CAP_PROJECTING] = CAIRO_LINE_CAP_SQUARE};

int vcl_to_cairo_join_styles [VCL_JOIN_COUNT] =
  {[VCL_JOIN_MITER] = CAIRO_LINE_JOIN_MITER,
   [VCL_JOIN_ROUND] = CAIRO_LINE_JOIN_ROUND,
   [VCL_JOIN_BEVEL] = CAIRO_LINE_JOIN_BEVEL};


// OBJECT interface

static void
vcl_painter_cairo_destroy (struct vcl_painter_cairo * obj)
{
  vcl_context_destroy (obj -> ct);
  if (obj -> cr != NULL) cairo_destroy (obj -> cr);
  if (obj -> sf != NULL) cairo_surface_destroy (obj -> sf);
  xfree (obj);
}


// PAINTER interface

static void
vcl_painter_cairo_draw_begin (struct vcl_painter_cairo * obj, UNUSED const struct vcl_rectangle *bbox)
{
  GdkDrawable *dst;
  gint dx, dy, w, h;

  gdk_window_get_internal_paint_info (obj -> w -> window, &dst, &dx, &dy);
  gdk_drawable_get_size (dst, &w, &h);
  obj -> sf = cairo_xlib_surface_create (GDK_DRAWABLE_XDISPLAY (dst),
							GDK_DRAWABLE_XID (dst), 
							GDK_VISUAL_XVISUAL(gdk_drawable_get_visual (dst)), w, h);
  cairo_surface_set_device_offset (obj -> sf, -dx, -dy);

  obj -> cr = cairo_create (obj -> sf);
  cairo_set_source_rgb (obj -> cr, 1.0, 1.0, 1.0);
  cairo_paint (obj -> cr);

  //  Should i do some clipping? I don't thing so ..
  // (because everything si drawn to pixmap of matching size)
  //  cairo_rectangle (obj -> cr, bbox -> lx, bbox -> ly, bbox -> hx, bbox -> hy);

  obj -> dx = dx;
  obj -> dy = dy;
}

static void
vcl_painter_cairo_draw_end (struct vcl_painter_cairo * obj)
{
  cairo_destroy (obj -> cr);
  cairo_surface_destroy (obj -> sf);
  obj -> cr = NULL;
  obj -> sf = NULL;
}

static void
vcl_painter_cairo_draw_node (struct vcl_painter_cairo * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{ vcl_context_walk_node (obj -> ct, obj, node, bbox, bbox_local); }

static void
draw_shape (struct vcl_painter_cairo * obj, void * node, UNUSED const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{
  struct geom_point * buff;
  void * it = NULL;
  uns closed = 0;
  uns points = 0;

  double w = obj -> ct -> stroke_width * vcl_context_get_dist_transform (obj -> ct);
  cairo_set_line_cap (obj -> cr, vcl_to_cairo_cap_styles [obj -> ct -> stroke_cap_style]);
  cairo_set_line_join (obj -> cr, vcl_to_cairo_join_styles [obj -> ct -> stroke_join_style]);
  cairo_set_line_width (obj -> cr, (w > 0) ? w : 1);

  do
    {
      it = vcl_shape_polygonize (node, it, 0.2, &(obj -> ct -> t . primary), bbox_local, &(obj -> buff), &(obj -> buff_len), &closed, &points, 0);

      if (points == 0) continue;
      buff = (struct geom_point *) obj -> buff;

      cairo_move_to (obj -> cr, buff [0] . x, buff [0] . y);
      for (uns i = 1; i < points; i ++)
	cairo_line_to (obj -> cr, buff [i] . x, buff [i] . y);

      if (closed) cairo_close_path (obj -> cr);

      u32 c = obj -> ct -> fill_color;
      cairo_set_source_rgba (obj -> cr, get_r (c) / 255.0,  get_g (c) / 255.0, get_b (c) / 255.0, get_a (c) / 255.0);
      cairo_fill_preserve (obj -> cr);

      c = obj -> ct -> stroke_color;
      cairo_set_source_rgba (obj -> cr, get_r (c) / 255.0,  get_g (c) / 255.0, get_b (c) / 255.0, get_a (c) / 255.0);
      cairo_stroke (obj -> cr);

    } while (it != NULL);
}

static void
draw_mask (struct vcl_painter_cairo * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{
  struct vcl_rectangle rb;
  cairo_surface_t *txt;

  vcl_mask_render (node, &(obj -> ct -> t . primary), bbox, bbox_local, &(obj -> buff), &(obj -> buff_len), &rb);
  txt = cairo_image_surface_create_for_data (obj -> buff, CAIRO_FORMAT_A8, rb . hx - rb . lx,
					     rb . hy - rb . ly, rb . hx - rb . lx);

  u32 c = obj -> ct -> fill_color;
  cairo_set_source_rgba (obj -> cr, get_r (c) / 255.0,  get_g (c) / 255.0, get_b (c) / 255.0, get_a (c) / 255.0);

  // After reading documentation i think that commented code is a correct way,
  //  but reality says something different. 
  //  cairo_mask_surface (obj -> cr, txt, rb . lx, rb . ly);
  cairo_mask_surface (obj -> cr, txt, rb . lx - obj -> dx, rb . ly - obj -> dy);

  cairo_surface_destroy (txt);
}

static void
vcl_painter_cairo_draw_leaf (struct vcl_painter_cairo * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{
  if (vcl_if_shape_p (node))
    draw_shape (obj, node, bbox, bbox_local);
  else if (vcl_if_mask_p (node))
    draw_mask (obj, node, bbox, bbox_local);
}

static void
vcl_painter_cairo_draw_composite (struct vcl_painter_cairo * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{ vcl_context_walk_children (obj -> ct, obj, node, bbox, bbox_local); }

// specific functions

struct vcl_painter_cairo *
vcl_painter_cairo_new (GtkWidget *wg)
{
  struct vcl_painter_cairo *obj = xmalloc (sizeof (struct vcl_painter_cairo));
  vcl_object_init (obj, vcl_vt_painter_cairo);

  obj -> w = wg;
  obj -> buff = NULL;
  obj -> buff_len = 0;
  obj -> cr = NULL;
  obj -> ct = vcl_context_new ();

  return (obj);
}

void
vcl_painter_cairo_main (void)
{
  vcl_vt_painter_cairo = vcl_class_new ("painter-cairo");

  vcl_class_add_interface (vcl_vt_painter_cairo, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_painter_cairo_destroy);

  vcl_class_add_interface (vcl_vt_painter_cairo, vcl_if_painter);
  vcl_class_add_method_li (vcl_mid_painter_draw_begin, vcl_painter_cairo_draw_begin);
  vcl_class_add_method_li (vcl_mid_painter_draw_end, vcl_painter_cairo_draw_end);
  vcl_class_add_method_li (vcl_mid_painter_draw_node, vcl_painter_cairo_draw_node);
  vcl_class_add_method_li (vcl_mid_painter_draw_leaf, vcl_painter_cairo_draw_leaf);
  vcl_class_add_method_li (vcl_mid_painter_draw_composite, vcl_painter_cairo_draw_composite);
}
